#
# Copyright (c) 2013 Duo Security
# All Rights Reserved
#
import base64
import hmac
import hashlib
import time


ENCRYPT_PLAIN = 0
ENCRYPT_AES256CFB = 1
ENCRYPT_TYPES = set((ENCRYPT_PLAIN, ENCRYPT_AES256CFB))
SIG_WINDOW = 300


def secure_compare(str1, str2):
    """
    Data-independent comparison for fixed-length secrets.

    WARNING: Do _NOT_ use for comparisons where length is unknown by the
    attacker (eg. for passwords), but only use for fixed-length comparisons
    like an HMAC.

    There's lots of subtleties here...Python 3.x will have a proper version
    written in C analogous to BSD's timingsafe_bcmp(3).
    """
    bytes1 = bytearray(str(str1))
    bytes2 = bytearray(str(str2))

    if len(bytes1) != len(bytes2):
        return False

    result = 0
    for x, y in zip(bytes1, bytes2):
        result |= x ^ y

    return result == 0


def sign_message(secret, data_bytes, sig_time):
    return _hmac_sha256(secret, data_bytes, sig_time)


def _hmac_sha256(key, data_bytes, sig_time):
    """
    Return the signature string for the given time and key.
    """
    canon_bytes = base64.b64encode(data_bytes) + '|' + str(sig_time)
    return hmac.HMAC(key, canon_bytes, hashlib.sha256).hexdigest()


def verify_sig(data_bytes, sig, drpc_creds, now=None, sig_window=SIG_WINDOW):
    """
    Return True iff the signature is correct.

    now: time (accounting for offset from Duo servers). None if you want system time
    """
    if not isinstance(sig, dict):
        return False

    if not now:
        now = int(time.time())

    sig_time = sig.get('time')
    sig_sig = sig.get('sig')
    if not (sig_time and sig_sig):
        return False

    if abs(sig_time - now) > sig_window:
        return False

    actual_sig = sign_message(drpc_creds.get_secret(), data_bytes, str(sig_time))
    return secure_compare(sig_sig, actual_sig)
