#
# Copyright (c) 2011 Duo Security
# All Rights Reserved
#

import base64
import datetime
import functools
import hashlib
import hmac
import json
import time
import urllib

from twisted.internet import defer
from twisted.web.http import HTTPClient as twistedHTTPClient

from duoauthproxy.lib.primary_only_manager import PrimaryOnlyManager
from . import http

FAILMODE_SAFE = 'safe'
FAILMODE_SECURE = 'secure'
FAILMODES = [FAILMODE_SAFE, FAILMODE_SECURE]

FAILMODE_SAFE_MSG = 'Failmode Safe - Allowed Duo login on preauth failure'
FAILMODE_SECURE_MSG = 'Failmode Secure - Denied Duo login on preauth failure'

PRIMARY_ONLY_FAILMODE_SAFE_MSG = 'Primary Only Mode - Secondary authentication bypassed'


class DuoAPIError(Exception):
    # Clients in FAILMODE_SAFE may fail open iff fail_open is True.
    fail_open = False


class DuoAPIFailClosedError(DuoAPIError):
    pass


class DuoAPIFailOpenError(DuoAPIError):
    fail_open = True


class DuoAPIProxyJoinError(DuoAPIError):
    fail_open = True


class DuoAPIMalformedResponseError(DuoAPIFailOpenError):
    def __init__(self, key):
        msg = ("Malformed API Response - invalid or missing '%s'"
               % key)
        super(DuoAPIMalformedResponseError, self).__init__(msg)


API_RESULT_AUTH = 'auth'
API_RESULT_ALLOW = 'allow'
API_RESULT_DENY = 'deny'
API_RESULT_ENROLL = 'enroll'


def should_server_fail_open(server_section_failmode, is_fail_open):
    """Checks the server section fail mode configuration, the fail mode of the exception and if the Authentication Proxy is running in primary authentication only mode.

    Args:
        server_section_failmode (str): the failmode configuration for a server section ('safe' or 'secure')
        is_fail_open (bool): the fail_open value of an DuoAPIError

    Returns:
        bool: True if the server section is configured to fail safe and either the AuthProxy is running in primary only mode or the exception type allows failing open. Otherwise False.
    """
    return server_section_failmode == FAILMODE_SAFE and (is_fail_open or PrimaryOnlyManager.is_primary_only_enabled())


def get_fail_open_msg():
    """Determines the fail open message depending on if the Authentication Proxy
    is running in primary mode or not.

    Returns:
        string: a fail open message
    """
    if PrimaryOnlyManager.is_primary_only_enabled():
        return PRIMARY_ONLY_FAILMODE_SAFE_MSG

    return FAILMODE_SAFE_MSG


class AsyncDuoClient(object):
    # overall default timeout: whatever the requestor decides
    TIMEOUT_DEFAULT = 0

    def __init__(self, host, duo_creds, port=443,
                 requestor=http.request, timeout_default=TIMEOUT_DEFAULT):
        """
        * requestor: Function API-compatible with http.request. Must
          return a Deferred.
        """
        self.duo_creds = duo_creds
        self.host = host
        self.port = port
        self.requestor = requestor
        self.time_offset = 0  # Between local clock and service.

        # per-instance default timeout
        self.timeout_default = timeout_default

    @staticmethod
    def encode_params(params):
        """Like urllib.urlencode, but sorted by name, and ~ values escaped."""
        result = []
        for k, v in sorted(params.iteritems()):
            if isinstance(k, unicode):
                k = k.encode('utf-8')
            if isinstance(v, unicode):
                v = v.encode('utf-8')
            result.append('%s=%s' % (urllib.quote(k, '~'), urllib.quote(v, '~')))
        return '&'.join(result)

    def _sign_request(self, method, uri, params, date=None):
        """Return request signature as HTTP basic authorization header."""
        url = self._get_canonical_url(method, uri, params, date)
        return self._get_signature(url)

    def _get_canonical_url(self, method, uri, params, date):
        if not isinstance(params, str):
            params = self.encode_params(params)

        if date is None:
            canon = []
        else:
            canon = [date]
        canon += [method.upper(), self.host.lower(), uri, params]
        canon = '\n'.join(canon)
        return canon

    def _get_signature(self, canon):
        ctx = hmac.new(self.duo_creds.get_secret(), canon, hashlib.sha1)
        auth = '%s:%s' % (self.duo_creds.get_identity(), ctx.hexdigest())
        return 'Basic ' + base64.b64encode(auth)

    def _format_request(self, method, uri, params):
        url = 'https://%s:%d%s' % (self.host, self.port, uri)
        qs = self.encode_params(params)
        headers = {'Authorization': self._sign_request(method, uri, qs)}
        if method == 'GET':
            if qs:
                url += '?' + qs
                qs = None
        else:
            headers['Content-Type'] = 'application/x-www-form-urlencoded'

        return (method, url, qs, headers)

    def _request(self, method, url, body, headers, timeout=None,
                 force_fail_closed=False, **kwargs):
        if PrimaryOnlyManager.is_primary_only_enabled():
            raise DuoAPIFailOpenError('The proxy is operating in primary-only mode')

        if timeout is None:
            timeout = self.timeout_default
        d = self.requestor(method, url, body, headers,
                           timeout=timeout,
                           **kwargs)

        def err_func(err):
            if force_fail_closed:
                raise DuoAPIFailClosedError(
                    'API Request Failed: %r' % err.value
                )
            else:
                raise DuoAPIFailOpenError(
                    'API Request Failed: %r' % err.value
                )
        d.addErrback(err_func)
        return d

    @defer.inlineCallbacks
    def call(self, method, uri, params, timeout=None,
             force_fail_closed=False, expect_type=dict):
        request_args = self._format_request(method, uri, params)
        code, data, headers = yield self._request(
            *request_args,
            force_fail_closed=force_fail_closed,
            timeout=timeout
        )
        res = self._parse_response(code, data, headers,
                                   expect_type=expect_type)
        defer.returnValue(res)

    @staticmethod
    def _parse_response(code, data, headers, expect_type=dict):
        res = None
        try:
            rsp = json.loads(data)
            if rsp['stat'] == 'OK':
                res = rsp['response']
            elif rsp['stat'] == 'FAIL':
                msg = ('%s: %s' % (rsp['code'], rsp['message']))
                if str(code).startswith('5'):
                    raise DuoAPIFailOpenError(msg)
                else:
                    raise DuoAPIFailClosedError(msg)
            else:
                raise DuoAPIMalformedResponseError('stat')
        except (ValueError, KeyError):
            # some sort of parse error. if we got a 200 response, this
            # is weird; otherwise, it's less weird
            if code != 200:
                msg = 'API request failed: HTTP Error %d' % code
            else:
                msg = 'Invalid Remote API Response: %r' % data
            # response?
            if str(code).startswith('5'):
                raise DuoAPIFailOpenError(msg)
            else:
                raise DuoAPIFailClosedError(msg)
        if not (res and isinstance(res, expect_type)):
            raise DuoAPIFailOpenError('Malformed API Response')
        return res

    @defer.inlineCallbacks
    def preauth(self, username, client_ip):
        if client_ip is None:
            client_ip = '0.0.0.0'

        params = dict(user=username, ipaddr=client_ip)
        preauth_res = yield self.call('POST', '/rest/v1/preauth', params)

        # validate preauth result format
        result = preauth_res.get('result', None)
        if result in (API_RESULT_ALLOW, API_RESULT_DENY, API_RESULT_ENROLL):
            if 'status' not in preauth_res:
                raise DuoAPIMalformedResponseError('status')
        elif result == API_RESULT_AUTH:
            if not isinstance(preauth_res.get('factors', None), dict):
                raise DuoAPIMalformedResponseError('factors')
            elif 'prompt' not in preauth_res:
                raise DuoAPIMalformedResponseError('prompt')
        else:
            raise DuoAPIMalformedResponseError('result')

        defer.returnValue(preauth_res)

    @defer.inlineCallbacks
    def auth(self, username, passcode, client_ip):
        if client_ip is None:
            client_ip = '0.0.0.0'

        params = dict(user=username, factor='auto', auto=passcode,
                      ipaddr=client_ip)

        auth_res = yield self.call(
            'POST', '/rest/v1/auth', params, force_fail_closed=True)

        # validate auth result format
        potential_auth_results = (API_RESULT_ALLOW, API_RESULT_DENY)
        if not auth_res.get('result', None) in potential_auth_results:
            raise DuoAPIMalformedResponseError('result')
        elif 'status' not in auth_res:
            raise DuoAPIMalformedResponseError('status')
        defer.returnValue(auth_res)

    @defer.inlineCallbacks
    def proxy_init(self, username):
        init_res = yield self.call('POST', '/rest/v1/tx/proxy_init',
                                   dict(user=username))

        # validate init result format
        if 'proxy_txid' not in init_res:
            raise DuoAPIMalformedResponseError('proxy_txid')
        try:
            init_res['proxy_txid'] = init_res['proxy_txid'].encode('ascii')
        except Exception:
            raise DuoAPIMalformedResponseError('proxy_txid')
        defer.returnValue(init_res)

    @defer.inlineCallbacks
    def proxy_finish(self, auth_cookie):
        finish_res = yield self.call('POST', '/rest/v1/tx/proxy_finish',
                                     dict(auth_cookie=auth_cookie))

        # validate init result format
        if 'valid_cookie' not in finish_res:
            raise DuoAPIMalformedResponseError('valid_cookie')
        elif finish_res['valid_cookie'] and 'user' not in finish_res:
            raise DuoAPIMalformedResponseError('user')
        defer.returnValue(finish_res)

    @defer.inlineCallbacks
    def ping(self):
        start = int(time.time())
        res = yield self.call('GET', '/auth/v2/ping', {})

        # Validate result format.
        server_time = res.get('time')
        if not server_time:
            raise DuoAPIMalformedResponseError('time')

        self.time_offset = res['time'] - start
        defer.returnValue(res)

    @defer.inlineCallbacks
    def proxy_join(self, drpc_server_factory_class, drpc_path):
        yield self.ping()       # Update self.time_offset.

        # Open a request normally (using configured HTTPS proxy,
        # etc.), but upgrade protocol to DRPC instead of calling API.
        method = 'POST'
        _status, rpc_protocol, _headers = yield self._request(
            method=method,
            # URL used only for scheme, host, and port.
            url=('https://{0}:{1:d}/'.format(self.host, self.port)),
            body='',            # overridden in protocol.
            headers={},         # overridden in protocol to sign with v2.
            timeout=self.timeout_default,
            protocol_class=functools.partial(
                HTTPClientDRPC,
                duo_client=self,
                drpc_server_factory_class=drpc_server_factory_class,
                drpc_path=drpc_path,
            ),
            disconnect=False,
        )
        defer.returnValue(rpc_protocol)


class HTTPClientDRPC(twistedHTTPClient, object):
    """
    HTTP client for upgrade to DRPC.
    """

    def __init__(self,
                 duo_client,
                 drpc_server_factory_class,
                 drpc_path,
                 *args, **kwargs):
        self.duo_client = duo_client
        self.drpc_server_factory_class = drpc_server_factory_class
        self.drpc_path = drpc_path
        self.rpc_server = None
        super(HTTPClientDRPC, self).__init__(*args, **kwargs)

    def lineReceived(self, line):
        if self.rpc_server:
            return self.rpc_server.dataReceived(line)
        else:
            return super(HTTPClientDRPC, self).lineReceived(line)

    def rawDataReceived(self, data):
        if self.rpc_server:
            return self.rpc_server.dataReceived(data)
        if self.length is not None:
            data, rest = data[:self.length], data[self.length:]
            self.length -= len(data)
        else:
            rest = b''
        self.handleResponsePart(data)
        if self.length == 0:
            self.handleResponseEnd()
            if self.rpc_server:
                self.rpc_server.dataReceived(rest)
            else:
                self.setLineMode(rest)

    def connectionMade(self):
        path = self.drpc_path
        method = 'POST'

        d = datetime.datetime.utcnow()
        d += datetime.timedelta(seconds=self.duo_client.time_offset)
        date = d.strftime('%a, %d %b %Y %H:%M:%S -0000')

        self.sendCommand(method, path)
        self.sendHeader(b'Host', self.factory.headers.get(b'host',
                                                          self.factory.host))
        self.sendHeader(b'User-Agent', self.factory.agent)
        auth = self.duo_client._sign_request(
            method=method,
            uri=path,
            params={},
            date=date,
        )
        self.sendHeader(b'Authorization', auth)
        self.sendHeader(b'Connection', 'upgrade')
        self.sendHeader(b'Content-Length', '0')
        self.sendHeader(b'Content-Type', 'application/x-www-form-urlencoded')
        self.sendHeader(b'Date', date)
        self.sendHeader(b'Upgrade', 'drpc')
        self.endHeaders()

    def sendCommand(self, command, path):
        self.transport.writeSequence([command,
                                      b' ',
                                      path,
                                      b' HTTP/1.1\r\n'])

    def handleStatus(self, version, status, message):
        self.factory.gotStatus(version, status, message)

    def handleResponse(self, response):
        def err(why):
            """
            Send error back to original caller and close the
            connection.
            """
            self.transport.abortConnection()
            self.factory.noPage(DuoAPIProxyJoinError(why))
            self.factory._disconnectedDeferred.callback(None)

        # Parse HTTP response.
        status = self.factory.status
        if str(status) not in ('101', '200'):
            err('Unexpected HTTP status: ' + str(status))
            return
        try:
            response = json.loads(response)
        except Exception:
            err('Malformed HTTP response')
            return
        if not isinstance(response, dict):
            err('Malformed HTTP response')
            return
        if response.get('stat') != 'OK':
            err('Failed HTTP response')
            return
        response = response.get('response')
        if not isinstance(response, dict):
            err('Malformed HTTP response')
            return
        if response.get('call_message_type') != 0:
            err('Unknown protocol version')
            return

        # Calculate time offset.
        server_time = response.get('time')
        if server_time:
            time_offset = server_time - int(time.time())
        else:
            time_offset = self.duo_client.time_offset

        # Build DRPC protocol to take over the connection.
        trans = self.transport
        factory = self.drpc_server_factory_class(
            drpc_creds=self.duo_client.duo_creds.create_drpc_credentials(),
            time_offset=time_offset,
            auto_zlib_min_length=response.get('auto_zlib_min_length'),
            enable_unzlib=True,
            idle_rpc_timeout=response.get('idle_rpc_timeout', 600),
        )
        self.rpc_server = factory.buildProtocol(trans.getPeer())
        self.rpc_server.makeConnection(trans)

        # Tell the transport to send data to the RPC server from now
        # on instead of the original protocol. If TLS-wrappered,
        # however, need to get plaintext, not the raw data.
        if trans.TLS:
            # trans.protocol is a TLSMemoryBIOProtocol initiated by
            # startTLS because the connection uses an HTTPS proxy.
            trans.protocol.wrappedProtocol = self.rpc_server
        elif hasattr(trans, 'wrappedProtocol'):
            # From connectSSL().
            trans.wrappedProtocol = self.rpc_server
        else:
            # Not TLS. Update raw, unwrapped transport.
            trans.protocol = self.rpc_server

        # Smuggle DRPC protocol (and protocol.factory) back to caller.
        self.factory.page(self.rpc_server)
        self.factory._disconnectedDeferred.callback(None)

    def connectionLost(self, reason):
        if self.rpc_server:
            return self.rpc_server.connectionLost(reason)
        else:
            self.factory.noPage(DuoAPIProxyJoinError('Connection lost'))
            if not self.factory._disconnectedDeferred.called:
                self.factory._disconnectedDeferred.callback(None)
            return super(HTTPClientDRPC, self).connectionLost(reason)
