#
# Copyright (c) 2011 Duo Security
# All Rights Reserved
#

"""
The enrollment package provides an interface to managing an enrollment code. An enrollment code is generated by Trustedpath, provided
to Gary who then inputs the enrollment code when registering a new Authentication Proxy.

An enrollment code is a base64 encoded json blob. It's attributes contains values such as the api_host and the short lived enrollment key.
"""

import base64
import json


class EnrollmentError(Exception):
    pass


class EnrollmentCode(object):
    @classmethod
    def create(cls, encoded_enrollment_code):
        """Takes in a base64 encoded JSON string and constructs an object with attributes needed to enroll an Authentication Proxy.

        Args:
           enrollment_code: (str) A base64 encoded json blob that contains attributes needed for enrolling an Authentication Proxy.

        Returns:
           enrollment: (EnrollmentCode) An object that contains attributes needed to enroll an Authentication Proxy
        """
        try:
            decoded_enrollment_code = json.loads(base64.b64decode(encoded_enrollment_code))

            enrollment = cls(api_host=decoded_enrollment_code['api_host'],
                             enrollment_key=decoded_enrollment_code['enrollment_key'])
        except StandardError:
            raise EnrollmentError

        return enrollment

    def __init__(self, api_host=None, enrollment_key=None):
        """
        Args:
            api_host: (str) the endpoint to perform Duo API calls
            enrollment_key: (str) The enrollment key generated by Trustedpath used to enroll an Auth Proxy
        """
        self.api_host = api_host
        self.enrollment_key = enrollment_key
