import struct

from twisted.internet import defer
from twisted.internet import protocol
from twisted.internet import reactor

from ldaptor import ldapfilter
from ldaptor.protocols import pureldap
from ldaptor.protocols.ldap import ldapclient
from ldaptor.protocols.ldap import ldaperrors
from ldaptor.protocols.ldap.distinguishedname import DistinguishedName
from ldaptor.protocols.ldap.ldapsyntax import LDAPEntry

from cryptography.hazmat.primitives import hashes
from cryptography.exceptions import UnsupportedAlgorithm

from ..ssl_verify import create_context_factory
from .. import log, ntlm

try:
    import sspi
    import sspicon
    import win32security
except ImportError:
    sspi = None
    sspicon = None
    win32security = None

# Not defined in sspicon, for some reason...
ISC_REQ_NO_INTEGRITY = 0x00800000L

AD_AUTH_TYPE_PLAIN = 'plain'
AD_AUTH_TYPE_NTLM_V1 = 'ntlm1'
AD_AUTH_TYPE_NTLM_V2 = 'ntlm2'
AD_AUTH_TYPE_SSPI = 'sspi'
AD_AUTH_TYPES = [
    AD_AUTH_TYPE_PLAIN,
    AD_AUTH_TYPE_NTLM_V1,
    AD_AUTH_TYPE_NTLM_V2,
    AD_AUTH_TYPE_SSPI,
]

AD_TRANSPORT_CLEAR = 'clear'
AD_TRANSPORT_LDAPS = 'ldaps'
AD_TRANSPORT_STARTTLS = 'starttls'
AD_TRANSPORTS_WITH_SSL = [AD_TRANSPORT_LDAPS, AD_TRANSPORT_STARTTLS]
AD_TRANSPORTS = AD_TRANSPORTS_WITH_SSL + [AD_TRANSPORT_CLEAR]

DEFAULT_SSL_VERIFY_DEPTH = 9

# active directory error code given when we search using an attribute range
# that exceeds the total number of elements in the multi-value attribute.
ERR_AD_CANT_RETRIEVE_ATTS = '00002121'


class ADClientError(Exception):
    pass


class SSPIError(ADClientError):
    pass


class ADClientProtocol(ldapclient.LDAPClient, object):
    """
    LDAPClient with utility methods needed for proxying.
    """

    def __init__(self):
        super(ADClientProtocol, self).__init__()
        self._tls_ensured = None
        self.unsolicited_notification_handler = None

    def connectionMade(self):
        super(ADClientProtocol, self).connectionMade()

        if self.factory.transport_type == AD_TRANSPORT_STARTTLS:
            self._tls_ensured = self.startTLS(self.factory.ssl_context_factory)
        else:
            self._tls_ensured = defer.succeed(True)
        self.factory.deferred.callback(self)

    def user_filter_object(self, username_match=None):
        """
        Return a filterObject for searching for a username attribute.

        If username_match is not None, add a filter requiring that the
        username attribute match that expression.
        """
        equality_matches = {}

        if username_match is not None:
            equality_matches.update(username_match)

        # Members of the given security group, if specified.
        if self.factory.security_group:
            equality_matches['memberOf'] = self.factory.security_group

        # Build BER representations of all this, starting with the
        # built-in filter to match only objects that are users.
        filter_objs = [self.factory.user_filter]
        for key, value in equality_matches.iteritems():
            attr_obj = pureldap.LDAPAttributeDescription(key)
            value_obj = pureldap.LDAPAssertionValue(value)
            filter_objs.append(
                pureldap.LDAPFilter_equalityMatch(attr_obj, value_obj))

        # Config can supply an optional extra filter expression to
        # be AND'd with the above.
        if self.factory.ldap_filter is not None:
            filter_objs.append(self.factory.ldap_filter)

        return pureldap.LDAPFilter_and(filter_objs)

    @defer.inlineCallbacks
    def dn_to_username(self, dn, client_factory):
        """
        Return a username usable for Duo auth.

        The return value may include a domain. If necessary, that must
        be handled with a domain username normalization policy on the
        Duo integration.
        """
        try:
            dn = DistinguishedName(dn)
            entry = LDAPEntry(self, dn)
        except Exception:
            base_entry = LDAPEntry(self, client_factory.search_dn)
            username_filter = self._create_username_filter(dn)
            if username_filter is None:
                raise

            if 'userPrincipalName' in username_filter:
                # We think it's a UPN formatted bind 'username@domain'
                res = yield base_entry.search(
                    filterObject=self.user_filter_object(username_filter)
                )
            elif 'sAMAccountName' in username_filter:
                # We think its a NetBIOS formmated bind 'domain\username'
                res = yield base_entry.search(
                    filterObject=self.user_filter_object(username_filter),
                    attributes=(
                        'msDS-PrincipalName',
                    ),
                )
                # If result has multiple entries with the same sAMAccountName
                # we can use msDS-Principal to check all domains. Only AD will
                # support this. All others will return no user
                if len(res) > 1 and client_factory.domain_discovery:
                    for match in res:
                        if list(match.get('msDS-PrincipalName', [None]))[0] == dn:
                            defer.returnValue(dn)

            if len(res) != 1:
                defer.returnValue(None)
            else:
                defer.returnValue(dn)

        result = yield entry.search(
            filterObject=self.user_filter_object(),
            attributes=(
                self.factory.username_attribute,
            ),
        )
        if len(result) != 1:
            defer.returnValue(None)
        attr_set = result[0].get(self.factory.username_attribute)
        if not attr_set:
            defer.returnValue(None)
        defer.returnValue(list(attr_set)[0])

    @staticmethod
    def _create_username_filter(dn):
        if ',' not in dn and dn.count('@') == 1:
            username_attribute = {'userPrincipalName': dn}
        elif ',' not in dn and dn.count('\\') == 1:
            _, samaccountname = dn.split('\\', 1)
            username_attribute = {'sAMAccountName': samaccountname}
        else:
            username_attribute = None

        return username_attribute

    @defer.inlineCallbacks
    def perform_unbind(self):
        op = pureldap.LDAPUnbindRequest()
        yield self.send_noResponse(op)

    @defer.inlineCallbacks
    def perform_bind_plain(self, dn, password):
        op = pureldap.LDAPBindRequest(dn=dn, auth=password, sasl=False)
        response = yield self.send(op)
        if response.resultCode != ldaperrors.Success.resultCode:
            raise ldaperrors.get(response.resultCode, response.errorMessage)

    @defer.inlineCallbacks
    def perform_bind_ntlm(self, dn, username, password, domain,
                          workstation,
                          ntlm_version):
        ntlm_negotiate_msg = ntlm.create_negotiate_msg()
        op = pureldap.LDAPBindRequest(
            dn=dn, auth=('GSS-SPNEGO', ntlm_negotiate_msg), sasl=True)
        response = yield self.send(op)
        if response.resultCode != ldaperrors.LDAPSaslBindInProgress.resultCode:
            raise ldaperrors.get(response.resultCode, response.errorMessage)

        ntlm_challenge_msg = response.serverSaslCreds.value
        ntlm_authenticate_msg = ntlm.create_authenticate_msg(
            ntlm_negotiate_msg, ntlm_challenge_msg, username, password,
            domain, workstation,
            ntlm_version)
        op = pureldap.LDAPBindRequest(
            dn=dn, auth=('GSS-SPNEGO', ntlm_authenticate_msg), sasl=True)
        response = yield self.send(op)
        if response.resultCode != ldaperrors.Success.resultCode:
            raise ldaperrors.get(response.resultCode, response.errorMessage)

    @defer.inlineCallbacks
    def perform_bind_sspi(self, dn, username, password, domain,
                          permit_implicit, targetspn=None):
        """Perform bind using native windows SSPI mechanism. If no
        username, password, or domain is provided, then we'll attempt
        to use the authproxy's existing process credentials to perform
        the bind.

        If targetspn is provided (and valid), then theoretically we
        might use kerberos instead of NTLM

        Returns:
            No return value from this function. Finishing without raising an exception means the
            bind succeeded
        Raises:
            LDAPUnwillingToPerform: If SSPI auth is not supported
            SSPIError: If the SSPI negotiation fails
        """

        if sspi is None:
            msg = 'The SSPI bind type is only supported on Windows.'
            log.err(msg)
            raise ldaperrors.LDAPUnwillingToPerform(msg)

        auth_info = (username, domain, password)
        # omitting 'domain' from this if statement; that way we can
        # specify ntlm_domain in config and have it apply to user
        # auth, but not automatically trip us into using configured,
        # rather than implicit, service account creds
        if not (username or password):
            if permit_implicit:
                auth_info = None
            else:
                # even passing a tuple of empty values still appears
                # to trigger the implicit-auth mechanism, so we need
                # to be explicit
                msg = 'Implicit authentication forbidden for this request.'
                log.err(msg)
                raise ldaperrors.LDAPUnwillingToPerform(msg)

        # SSPI negotiation can request some application-level
        # encryption/authentication, but AD apparently throws a fit if
        # you leave this on when using SSL/TLS (and expects you to
        # actually somehow sign all your LDAP requests otherwise)
        scflags = ISC_REQ_NO_INTEGRITY

        # these return codes mean we need to continue the handshake
        sspi_continue = set([
            sspicon.SEC_I_CONTINUE_NEEDED,
            sspicon.SEC_I_COMPLETE_AND_CONTINUE])
        # any return code not in this set should be considered an error
        sspi_ok = sspi_continue.union([
            sspicon.SEC_E_OK,
            sspicon.SEC_I_COMPLETE_NEEDED])

        # explicitly specifying 'Kerberos' or 'NTLM' will also work
        # here, but technically if we say the SASL type is GSS-SPNEGO,
        # we're supposed to use 'Negotiate'
        ca = sspi.ClientAuth(
            'Negotiate', auth_info=auth_info, targetspn=targetspn,
            scflags=scflags)
        data = None
        # This negotiation is made up of challenge and response messages.
        # We will continue responding to challenges until a success or
        # failure case is hit
        while True:
            # get the next step in the handshake
            err, out_buf = ca.authorize(data)
            if err not in sspi_ok:
                raise SSPIError('SSPI negotiation failed', err)

            # format request and send it
            op = pureldap.LDAPBindRequest(
                dn=dn,
                auth=('GSS-SPNEGO', out_buf[0].Buffer),
                sasl='True',
            )
            response = yield self.send(op)
            if response.resultCode \
                    == ldaperrors.LDAPSaslBindInProgress.resultCode:
                peercert = self._get_peercert()
                data = self._create_buffer_array(ca.pkg_info['MaxToken'], response.serverSaslCreds.value, peercert)
            elif response.resultCode == ldaperrors.Success.resultCode:
                break
            else:
                raise ldaperrors.get(
                    response.resultCode, response.errorMessage)

            # if SSPI said we're done, but ldap response doesn't
            # agree, that's weird
            if err not in sspi_continue:
                raise SSPIError(
                    'SSPI negotiation should\'ve finished by now', err)

    def perform_bind(self,
                     auth_type,
                     dn, username, password, domain, workstation,
                     permit_implicit=False):
        if auth_type == AD_AUTH_TYPE_PLAIN:
            return self.perform_bind_plain(dn, password)
        elif auth_type == AD_AUTH_TYPE_NTLM_V1:
            return self.perform_bind_ntlm(dn, username, password, domain,
                                          workstation,
                                          ntlm_version=1)
        elif auth_type == AD_AUTH_TYPE_NTLM_V2:
            return self.perform_bind_ntlm(dn, username, password, domain,
                                          workstation,
                                          ntlm_version=2)
        elif auth_type == AD_AUTH_TYPE_SSPI:
            return self.perform_bind_sspi(dn, username, password, domain,
                                          permit_implicit)

    @defer.inlineCallbacks
    def send(self, op, controls=None, handler=None, handle_msg=False):
        yield self._tls_ensured
        send_result = yield (
            super(ADClientProtocol, self).send(op, controls=controls, handler=handler, handle_msg=handle_msg)
        )
        defer.returnValue(send_result)

    @defer.inlineCallbacks
    def send_multiResponse(self, op, handler, *args, **kwargs):
        yield self._tls_ensured
        send_result = yield (
            super(ADClientProtocol, self).send_multiResponse(op, handler, *args, **kwargs)
        )
        defer.returnValue(send_result)

    @defer.inlineCallbacks
    def send_noResponse(self, op):
        yield self._tls_ensured
        super(ADClientProtocol, self).send_noResponse(op)

    def unsolicitedNotification(self, msg):
        """Passes the message back the authproxy server section to determine
        what action should be taken to inform the downstream client
        """
        super(ADClientProtocol, self).unsolicitedNotification(msg)
        if self.unsolicited_notification_handler is not None:
            self.unsolicited_notification_handler(msg)

    def _startTLS(self, ctx):
        if not self.connected:
            raise ldapclient.LDAPClientConnectionLostException
        elif self.onwire:
            raise ldapclient.LDAPStartTLSBusyError, self.onwire
        else:
            op = pureldap.LDAPStartTLSRequest()
            # call super's send to avoid a circular dependency on yield self._tls_ensured
            d = super(ADClientProtocol, self).send(op)
            d.addCallback(self._cbStartTLS, ctx)
            return d

    def _create_buffer_array(self, max_token_size, challenge_response, peercert):
        """ Given the challenge response from our BindRequest build up an array of the buffers to be used as a
        response to that challenge. This will most often just be the challenge itself as well as the channel binding
        token if using a transport with SSL
        Args:
            max_token_size (int): As defined by ClientAuth this value is the maximum size of a token for the handshake
            challenge_response (str) This arg contains the raw bytes of the challenge from the server
            peercert: Peer SSL certificate taken off of a transport
        Returns:
            PySecBufferDescType: The array of PySecBufferTypes
        """
        buffer_array = win32security.PySecBufferDescType()

        challenge_buffer = win32security.PySecBufferType(max_token_size, sspicon.SECBUFFER_TOKEN)
        challenge_buffer.Buffer = challenge_response
        buffer_array.append(challenge_buffer)

        # To support servers that have turned on LdapEnforceChannelBinding we add this token
        if peercert:
            try:
                appdata = self._create_appdata_from_peercert(peercert)
            except UnsupportedAlgorithm:
                log.msg("Skipping the creation of the CBT due to unsupported hash algorithm")
            else:
                cbt_buffer = self._create_channel_binding_token(max_token_size, appdata)
                buffer_array.append(cbt_buffer)

        return buffer_array

    def _get_peercert(self):
        if self.factory.transport_type in AD_TRANSPORTS_WITH_SSL:
            peercert = self.transport.getPeerCertificate()
            if not peercert and self.debug:
                log.msg("SSL transport was specified but we are unable to get peercertificate. CBT will not be created.")
        else:
            peercert = None
        return peercert

    @staticmethod
    def _create_appdata_from_peercert(peercert):
        """Pull the peercert from the connection and then massage it into the proper appdata

        Returns:
            appdata: Combination of the hashed cert and some metadata
        Raises:
            UnsupportedAlgorithm: If the cert's signature hash algorithm is not a single hash
        """
        # We convert to the cryptography library's object representation of a cert so that we have more functionality.
        # Specifically we want the signature_hash_algorithm
        crypto_peercert = peercert.to_cryptography()
        try:
            hash_algo = crypto_peercert.signature_hash_algorithm
        except UnsupportedAlgorithm as e:
            log.err(str(e))
            raise e
        if isinstance(hash_algo, (hashes.MD5, hashes.SHA1)):
            # https://tools.ietf.org/html/rfc5929#section-4.1
            hash_algo = hashes.SHA256()

        hashed_cert = crypto_peercert.fingerprint(hash_algo)
        return 'tls-server-end-point:'.encode('ASCII') + hashed_cert

    @staticmethod
    def _create_channel_binding_token(max_token_size, appdata):
        """ Create a channel binding token to bind the LDAP layer to the underlying TLS layer.
        We do this by hashing the peercert and packing it into a buffer to be sent with an LDAPBindRequest.
        See T34588 for more detail. """

        # This struct.pack creates the SEC_CHANNEL_BINDINGS structure followed by the actual appdata
        # https://docs.microsoft.com/en-us/windows/desktop/api/sspi/ns-sspi-_sec_channel_bindings
        application_data_length = len(appdata)
        application_data_offset = 32
        struct_plus_data = struct.pack('<LLLLLLLL{}s'.format(len(appdata)), 0, 0, 0, 0, 0, 0, application_data_length, application_data_offset, appdata)

        cbtbuf = win32security.PySecBufferType(max_token_size, sspicon.SECBUFFER_CHANNEL_BINDINGS)
        cbtbuf.Buffer = struct_plus_data
        return cbtbuf


class ADClientFactory(protocol.ClientFactory, object):
    protocol = ADClientProtocol

    user_filter = ldapfilter.parseFilter(
        '(|'
        # AD: match only users.
        # <http://social.technet.microsoft.com/wiki/contents/articles/5392.active-directory-ldap-syntax-filters.aspx>
        '(&(objectClass=user)(objectCategory=person))'
        # RFC 2798 schemas:
        '(objectClass=inetOrgPerson)'
        # OpenLDAP Core schema
        '(objectClass=organizationalPerson)'
        ')')

    def __init__(self,
                 timeout,
                 transport_type,
                 ssl_ca_certs,
                 ssl_verify_depth,
                 ssl_verify_hostname,
                 domain_discovery=False,
                 username_attribute='sAMAccountName',
                 at_attribute='userPrincipalName',
                 security_group=None,
                 ldap_filter=None,
                 debug=False,
                 is_logging_insecure=False):
        self.timeout = timeout
        self.domain_discovery = domain_discovery
        self.username_attribute = username_attribute
        self.at_attribute = at_attribute
        self.transport_type = transport_type
        self.ssl_ca_certs = ssl_ca_certs
        self.ssl_context_factory = None
        self.ssl_verify_depth = ssl_verify_depth
        self.ssl_verify_hostname = ssl_verify_hostname
        self.security_group = security_group
        self.ldap_filter = ldap_filter
        self.debug = debug
        self.is_logging_insecure = is_logging_insecure
        self.deferred = defer.Deferred()

    def buildProtocol(self, addr):
        p = super(ADClientFactory, self).buildProtocol(addr)
        p.debug = self.debug
        p.is_logging_insecure = self.is_logging_insecure
        return p

    def clientConnectionFailed(self, connector, reason):
        log.msg('AD Connection failed: %r' % reason)
        self.deferred.errback(
            ADClientError('AD Connection failed: %s' % reason))

    def stopFactory(self):
        super(ADClientFactory, self).stopFactory()
        try:
            self.deferred.errback(
                ADClientError('AD Connection closed prematurely'))
        except defer.AlreadyCalledError:
            pass

    def connect_ldap(self, host, port):
        if self.transport_type in (AD_TRANSPORT_STARTTLS, AD_TRANSPORT_LDAPS):
            ssl_hostname = (host if self.ssl_verify_hostname else None)
            self.ssl_context_factory = create_context_factory(
                hostnames=ssl_hostname, caCerts=self.ssl_ca_certs,
                verifyDepth=self.ssl_verify_depth)

        if self.transport_type == AD_TRANSPORT_LDAPS:
            return reactor.connectSSL(
                host, port, self, self.ssl_context_factory,
                timeout=self.timeout)
        else:
            return reactor.connectTCP(
                host, port, self, timeout=self.timeout)
