#
# Copyright (c) 2013 Duo Security
# All Rights Reserved
#
import functools
import hashlib
import hmac
import time

from twisted.internet import defer
from twisted.internet import task

from duoauthproxy import __version__
from duoauthproxy.lib import (
    base,
    duo_async,
    log,
)
from duoauthproxy.lib.duo_creds import DuoCreds
import drpc.v2 as drpc


class CloudApiFactory(drpc.ServerFactory):

    def __init__(self, module, **kwargs):
        super(CloudApiFactory, self).__init__(**kwargs)
        self.module = module
        module.register_drpc_call_provider(self)

    def get_func_for_call(self, call_name):
        return self.module.get_func_for_drpc_call(call_name)

    def get_drpc_calls(self):
        return {
            'ping': self.do_ping,
            'identify': self.do_identify
        }

    def log_msg(self, summary, **kwargs):
        try:
            log.msg(summary)
        except Exception:
            pass

    def log_err(self, e, summary):
        try:
            log.err(e, summary)
        except Exception:
            pass

    def do_ping(self):
        info = super(CloudApiFactory, self).do_ping()
        info['version'] = __version__
        return info

    def do_identify(self, challenge):
        return {
            'identities': [
                {
                    'ikey': creds.get_identity(),
                    'sig': hmac.new(creds.get_secret().encode('utf-8'),
                                    challenge.encode('utf-8'),
                                    hashlib.sha1).hexdigest(),
                }
                for creds in self.module.identities.values()
            ],
        }


class Module(base.ServerModule):
    no_client = True

    reconnect_interval = 60

    def __init__(self, config, _primary_ator=None):
        log.msg('Cloud Connector Module Configuration:')
        log.config(config, (lambda k: k in (
            'skey',
            'skey_protected',
            'service_account_password',
            'service_account_password_protected',
        )))
        self.debug = config.get_bool('debug', False)
        self.duo_creds = DuoCreds(
            config.get('ikey'),
            config.get_protected_str('skey_protected', 'skey'),
        )
        self.duo_client = self.make_duo_client(config, duo_creds=self.duo_creds)
        self._check_connection_lc = None
        self.rpc_server = None

        self.identities = {}
        self.identities[self.duo_creds.get_identity()] = self.duo_creds

        self.drpc_server_factory_class = functools.partial(
            CloudApiFactory,
            module=self,
        )
        self.drpc_path = '/auth/v2/proxy_join'

        self.drpc_calls = {}

    def get_func_for_drpc_call(self, call_name):
        if call_name in self.drpc_calls:
            return self.drpc_calls[call_name]

        log.err(None, 'Unknown DRPC call \'{0}\''.format(call_name))
        return None

    def register_drpc_call_provider(self, call_provider):
        self.drpc_calls.update(call_provider.get_drpc_calls())

    def startService(self):
        super(Module, self).startService()
        self._check_connection_lc = task.LoopingCall(self._check_connection)
        self._check_connection_lc.start(self.reconnect_interval,
                                        True)

    def stopService(self):
        super(Module, self).stopService()
        if self._check_connection_lc:
            self._check_connection_lc.stop()
        if self.rpc_server and self.rpc_server:
            self.rpc_server.transport.abortConnection()
            self.rpc_server = None

    @defer.inlineCallbacks
    def _check_connection(self):
        disconnected = False
        if not self.rpc_server:
            discon_msg = "no rpc server"
            disconnected = True
        elif self.rpc_server.transport.disconnected:
            discon_msg = "transport disconnected"
            disconnected = True
        elif not self.rpc_server.transport.connected:
            discon_msg = "transport not connected"
            disconnected = True
        elif (self.rpc_server.factory.idle_rpc_timeout is not None and
                self.rpc_server.last_rpc is not None):
            last_rpc = int(time.time() - self.rpc_server.last_rpc)
            if last_rpc > self.rpc_server.factory.idle_rpc_timeout:
                discon_msg = "Missed pings for {0} seconds, maximum {1} seconds allowed.".format(
                    last_rpc, self.rpc_server.factory.idle_rpc_timeout)
                disconnected = True

        if disconnected:
            # No connection to service found! Fix that.
            log.msg("DRPC Disconnected: {0}".format(discon_msg))
            log.msg('(Re)connecting to service...')
            self.rpc_server = None
            try:
                self.rpc_server = yield self.duo_client.proxy_join(
                    self.drpc_server_factory_class,
                    self.drpc_path)
            except duo_async.DuoAPIError as e:
                log.msg('Error connecting to service: {0}'.format(e.message))
