#
# Copyright (c) 2019 Duo Security
# All Rights Reserved
#
import functools
import six

from duoauthproxy.lib import ldap, log
from duoauthproxy.lib.ssl_verify import load_ca_bundle
import drpc.v2 as drpc

from ldaptor import ldapfilter
from ldaptor.protocols import pureber, pureldap
from ldaptor.protocols.ldap import ldaperrors, distinguishedname
from ldaptor.protocols.ldap.ldaperrors import LDAPOperationsError
from ldaptor.protocols.ldap.ldapsyntax import LDAPEntry
from twisted.internet import defer, reactor

# ERR_LDAP* errors are returned by the proxy to our service over DDRPC.
ERR_LDAP_BIND_FAILED = 'LDAP bind failed'
ERR_LDAP_CONFIGURATION_FAILED = 'LDAP configuration failed'
ERR_LDAP_CONNECTION_FAILED = 'LDAP connection failed'
ERR_LDAP_PW_ENCRYPT_FAILED = 'password decryption failed'
ERR_LDAP_SEARCH_FAILED = 'LDAP search failed'
ERR_LDAP_TIMEOUT = 'LDAP timeout'
ERR_LDAP_BAD_AD_CONFIGURATION = 'Bad AD client configuration value'
ERR_LDAP_AD_SEARCH_FAILED = 'LDAP search on AD service failed'

# AUTH* messages are human readable values returned in combination with
# whether the authentication was successful or not
AUTH_INVALID_USER = 'Invalid User'
AUTH_FAILED = 'Active Directory authentication failed'
AUTH_SUCCEEDED = 'Active Directory authentication succeeded'

CONFIG_BIND_USER = 'service_account_username'
CONFIG_BIND_PASSWORD = 'service_account_password'

RFC_2696_CONTROL_TYPE = '1.2.840.113556.1.4.319'

PAGINATION_TYPE_CRITICAL = 'force'
PAGINATION_TYPE_ENABLE = 'True'
PAGINATION_TYPE_DISABLE = 'False'

DEFAULT_PAGE_SIZE = 5000
# ldap multi-value attributes can only be returned in segments of 1500 values
# per request. actually, it's controlled by a server-side setting but 1500 is
# the default. it's incredibly unlikely that anyone would make this SMALLER
# than 1500. if they make it larger then it shouldn't make too much difference.
LDAP_ATTR_DEFAULT_RANGE_LEN = 1500

PRINTABLE = set('0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ!"#$%&\'()*+,-./:;<=>?@[]^_`{|}~ ')


class LdapSyncClientProtocol(ldap.client.ADClientProtocol):
    """ This class exists soley to improve the logging during directory sync calls. Twisted
    logs class names as a prefix """
    pass


class LdapSyncClientFactory(ldap.client.ADClientFactory):
    """ This class exists soley to improve the logging during directory sync calls. Twisted
    logs class names as a prefix """
    protocol = LdapSyncClientProtocol


class LdapDrpcPlugin(object):
    def __init__(self, config):
        self.debug = config.get_bool('debug', False)
        self.is_logging_insecure = config.get_bool('is_logging_insecure', False)

        self.service_account_username = config.get_str(CONFIG_BIND_USER, '')
        self.service_account_password = config.get_protected_str(
            '{0}_protected'.format(CONFIG_BIND_PASSWORD),
            CONFIG_BIND_PASSWORD,
            '')

    def get_drpc_calls(self):
        return {
            'ldap_search': self.do_ldap_search,
            'ldap_authentication': self.do_ldap_authentication
        }

    @staticmethod
    @drpc.inlineCallbacks
    def _get_client(host,
                    port,
                    transport_type,
                    ssl_verify_depth,
                    ssl_verify_hostname,
                    ssl_ca_certs,
                    timeout,
                    debug,
                    is_logging_insecure):
        if ssl_ca_certs:
            ssl_ca_certs = load_ca_bundle(ssl_ca_certs)
            if not ssl_ca_certs:
                # Didn't parse out any PEM certificates.
                raise drpc.CallBadArgError(['ssl_ca_certs'])
        else:
            # Ensure ssl_ca_certs is a list.
            ssl_ca_certs = []

        is_ssl = transport_type != ldap.client.AD_TRANSPORT_CLEAR
        if is_ssl:
            if ssl_verify_hostname and not ssl_ca_certs:
                log.msg('Missing required configuration item: '
                        "'SSL verify hostname' requires that "
                        "'SSL CA certs' also be specified "
                        '(and non-empty).')
                raise drpc.CallError(ERR_LDAP_CONFIGURATION_FAILED)

        try:
            factory = LdapSyncClientFactory(
                timeout=timeout,
                transport_type=transport_type,
                ssl_verify_depth=ssl_verify_depth,
                ssl_verify_hostname=ssl_verify_hostname,
                ssl_ca_certs=ssl_ca_certs,
                debug=debug,
                is_logging_insecure=is_logging_insecure,
            )
        except Exception as e:
            log.err(e, ERR_LDAP_CONFIGURATION_FAILED)
            raise drpc.CallError(ERR_LDAP_CONFIGURATION_FAILED, {
                'error': str(e),
            })

        try:
            factory.connect_ldap(host, port)
            client = yield factory.deferred
        except Exception as e:
            log.err(e, ERR_LDAP_CONNECTION_FAILED)
            raise drpc.CallError(ERR_LDAP_CONNECTION_FAILED, {
                'error': str(e),
            })

        def timeout_cb():
            log.msg('LDAP operation timed out')
            client.transport.abortConnection()
        timeout_dc = reactor.callLater(timeout, timeout_cb)

        defer.returnValue((client, timeout_dc))

    def _verify_ldap_config_args(self,
                                 bind_dn,
                                 bind_pw,
                                 auth_type,
                                 transport_type):
        if auth_type not in ldap.client.AD_AUTH_TYPES:
            raise drpc.CallBadArgError(['auth_type'])
        if transport_type not in ldap.client.AD_TRANSPORTS:
            raise drpc.CallBadArgError(['transport_type'])
        if auth_type != ldap.client.AD_AUTH_TYPE_SSPI \
                and (bind_dn == '' or bind_pw == ''):
            e = drpc.CallError(ERR_LDAP_CONFIGURATION_FAILED, {
                'authproxy_configuration_error': 'Missing {0} or {1}'.format(
                    CONFIG_BIND_USER, CONFIG_BIND_PASSWORD)
            })
            log.err(e, ERR_LDAP_CONFIGURATION_FAILED)
            raise e

    @drpc.inlineCallbacks
    def do_ldap_authentication(self,
                               host,
                               port,
                               username,
                               password,
                               base_dn='',
                               ntlm_domain=None,
                               ntlm_workstation=None,
                               auth_type=ldap.client.AD_AUTH_TYPE_NTLM_V2,
                               transport_type=ldap.client.AD_TRANSPORT_STARTTLS,
                               ssl_verify_depth=ldap.client.DEFAULT_SSL_VERIFY_DEPTH,
                               ssl_verify_hostname=True,
                               ssl_ca_certs=None,
                               timeout=60,
                               username_attribute='samaccountname'):
        """
        * username_attribute: attribute within AD that we will compare the
          username against.

        See do_ldap_search for further argument documentation.
        """
        bind_dn = self.service_account_username
        bind_pw = self.service_account_password

        log.msg(("Performing LDAP authentication: "
                 "host={host} port={port} base_dn={base_dn} "
                 "auth_type={auth_type} transport_type={transport_type} "
                 "ssl_verify_depth={ssl_verify_depth} ssl_verify_hostname={ssl_verify_hostname} "
                 "ssl_ca_certs={ssl_ca_certs} username_attribute={username_attribute}"
                 ).format(host=host, port=port, base_dn=base_dn, auth_type=auth_type,
                          transport_type=transport_type, ssl_verify_depth=ssl_verify_depth,
                          ssl_verify_hostname=ssl_verify_hostname, username_attribute=username_attribute,
                          ssl_ca_certs=ssl_ca_certs is not None))

        self._verify_ldap_config_args(
            bind_dn,
            bind_pw,
            auth_type,
            transport_type
        )

        cl, timeout_dc = yield self._get_client(
            host,
            port,
            transport_type,
            ssl_verify_depth,
            ssl_verify_hostname,
            ssl_ca_certs,
            timeout,
            self.debug,
            self.is_logging_insecure
        )

        principal_name = ''
        if username_attribute.lower() == 'samaccountname' and '\\' in username:
            principal_name = username
            ntlm_domain, username = username.split('\\', 1)

        try:
            # Primary bind. This authenticates us to AD and allows us to
            # make search queries.
            try:
                yield cl.perform_bind(
                    auth_type=auth_type,
                    dn=bind_dn,
                    username=bind_dn,
                    password=bind_pw,
                    domain=ntlm_domain,
                    workstation=ntlm_workstation,
                    permit_implicit=True
                )
            except Exception as e:
                if timeout_dc.active():
                    log.err(e, ERR_LDAP_BIND_FAILED)
                    raise drpc.CallError(ERR_LDAP_BIND_FAILED, {
                        'error': str(e),
                    })
                else:
                    log.err(e, ERR_LDAP_TIMEOUT)
                    raise drpc.CallError(ERR_LDAP_TIMEOUT, {
                        'during': 'bind',
                        'error': str(e),
                    })

            username_match = {cl.factory.username_attribute: username}
            filterObject = cl.user_filter_object(username_match=username_match)

            # Search for the user. With AD, the user's cn is his/her full
            # name, not username so we need to search for a specific
            # attribute value instead (typically, the 'sAMAccountName'
            # attribute)
            try:
                entry = LDAPEntry(cl, base_dn)
                result = yield entry.search(
                    filterObject=filterObject,
                    attributes=(
                        username_attribute,
                        'msDS-PrincipalName',
                    ),
                )

                if cl.factory.domain_discovery:
                    # Use the domain returned in the user's principal name
                    result_pnames = [
                        u['msDS-PrincipalName'].__iter__().next()
                        for u in result
                        if 'msDS-PrincipalName' in u and u['msDS-PrincipalName']
                    ]
                    if len(result_pnames) > 1:
                        log.msg("Domain discovery failed due to multiple user results.")
                    elif len(result_pnames) == 1 and '\\' in result_pnames[0]:
                        ntlm_domain = result_pnames[0].split('\\')[0]

                # If we have more than one result, narrow it down by msDS-PN
                if len(result) != 1:
                    for user in result:
                        if ('msDS-PrincipalName' in user and principal_name and
                                user['msDS-PrincipalName'].pop() == principal_name):
                            result = [user]
                            break
                    else:
                        defer.returnValue({
                            'success': False,
                            'msg': AUTH_INVALID_USER,
                        })
            except distinguishedname.InvalidRelativeDistinguishedName as e:
                log.err(e, ERR_LDAP_BAD_AD_CONFIGURATION)
                raise drpc.CallError(ERR_LDAP_BAD_AD_CONFIGURATION, {
                    'error': str(e),
                })
            except Exception as e:
                log.err(e, ERR_LDAP_AD_SEARCH_FAILED)
                raise drpc.CallError(ERR_LDAP_AD_SEARCH_FAILED, {
                    'error': str(e),
                })

            # Secondary bind. Assuming the user we queried exists, attempt
            # to bind as them (this is essentially performing an authentication).
            try:
                yield cl.perform_bind(
                    auth_type=auth_type,
                    dn=str(result[0].dn),
                    username=username,
                    password=password,
                    domain=ntlm_domain,
                    workstation=ntlm_workstation,
                    permit_implicit=False
                )
            except Exception:
                defer.returnValue({
                    'success': False,
                    'msg': AUTH_FAILED,
                })
        finally:
            if timeout_dc.active():
                timeout_dc.cancel()
            try:
                cl.transport.abortConnection()
            except Exception:
                pass

        defer.returnValue({
            'success': True,
            'msg': AUTH_SUCCEEDED,
        })

    @drpc.inlineCallbacks
    def do_ldap_search(self,
                       host, port, base_dn,
                       filter_text=None,
                       attributes=None,
                       ntlm_domain=None,
                       ntlm_workstation=None,
                       auth_type=ldap.client.AD_AUTH_TYPE_NTLM_V2,
                       transport_type=ldap.client.AD_TRANSPORT_STARTTLS,
                       ssl_verify_depth=ldap.client.DEFAULT_SSL_VERIFY_DEPTH,
                       ssl_verify_hostname=True,
                       ssl_ca_certs=None,
                       pagination=PAGINATION_TYPE_CRITICAL,
                       page_size=DEFAULT_PAGE_SIZE,
                       max_result_size=None,
                       timeout=60):
        """
        * host: LDAP IP address we will perform actions against.
        * port: LDAP server port.
        * filter_text: Filter user search.
        * attributes: Retrieve only the listed attributes. If None,
          retrieve all attributes.
        * ntlm_{domain, workstation}: Windows authentication mechanism.
          Allows for bypassing primary bind if already within the domain.
        * auth_type: Bind method to use, e.g. NTLM, Plain, SSPI, etc.
        * transport_type: Method of transportation to use, e.g. Clear, TLS,
          LDAPS, etc.
        * timeout: If either establishing the connection or binding
          and searching take longer than this number of seconds
          the response will be an error.
        * page_size: As in RFC 2696.
        * max_result_size: If paging, stop requesting results after
          when at least this number of results have been received.
        """

        if attributes is None:
            attributes = set()

        try:
            attributes = set(k.lower() for k in attributes)
            if 'userpassword' in attributes:
                attributes.remove('userpassword')
        except Exception as e:
            log.err(e)
            raise drpc.CallBadArgError(['attributes'])

        page_size = parse_positive_int(page_size, 'page_size')
        if max_result_size is not None:
            max_result_size = parse_positive_int(
                max_result_size, 'max_result_size')

        if not (filter_text is None
                or isinstance(filter_text, six.string_types)):
            raise drpc.CallBadArgError(['filter_text'])

        bind_dn = self.service_account_username
        bind_pw = self.service_account_password

        log.msg(("Performing LDAP search: "
                 "host={host} port={port} base_dn={base_dn} "
                 "auth_type={auth_type} transport_type={transport_type} "
                 "ssl_verify_depth={ssl_verify_depth} ssl_verify_hostname={ssl_verify_hostname} "
                 "ssl_ca_certs={ssl_ca_certs} attributes={attributes}"
                 ).format(host=host, port=port, base_dn=base_dn, auth_type=auth_type,
                          transport_type=transport_type, ssl_verify_depth=ssl_verify_depth,
                          ssl_verify_hostname=ssl_verify_hostname, attributes=attributes,
                          ssl_ca_certs=ssl_ca_certs is not None)
                )

        self._verify_ldap_config_args(
            bind_dn,
            bind_pw,
            auth_type,
            transport_type
        )

        cl, timeout_dc = yield self._get_client(
            host,
            port,
            transport_type,
            ssl_verify_depth,
            ssl_verify_hostname,
            ssl_ca_certs,
            timeout,
            self.debug,
            self.is_logging_insecure
        )

        try:
            try:
                yield cl.perform_bind(
                    auth_type=auth_type,
                    dn=bind_dn,
                    username=bind_dn,
                    password=bind_pw,
                    domain=ntlm_domain,
                    workstation=ntlm_workstation,
                    permit_implicit=True
                )
            except Exception:
                if timeout_dc.active():
                    log.err(None, ERR_LDAP_BIND_FAILED)
                    raise drpc.CallError(ERR_LDAP_BIND_FAILED)
                else:
                    log.err(None, ERR_LDAP_TIMEOUT)
                    raise drpc.CallError(ERR_LDAP_TIMEOUT, {
                        'during': 'bind',
                    })

            try:
                if filter_text is not None:
                    filter_text = filter_text.encode('utf-8')

                result = yield self._paged_search(
                    client=cl,
                    base_dn=base_dn,
                    filter_text=filter_text,
                    attributes=[a.encode('utf-8') for a in attributes],
                    pagination=pagination,
                    page_size=page_size,
                    max_result_size=max_result_size,
                )

                # any object that has a 'member' attribute is a potential
                # group object. if that attribute is an empty list then the
                # group is either empty or contains more values than the
                # server's configured maximum attribute length (usually 1500)
                # results.
                # try to check for members using range before giving up.
                try_ranged = [obj for obj in result if obj.get('member') == []]
                if len(try_ranged):
                    yield self.try_ranged_search(cl, base_dn, try_ranged)

            except Exception:
                if timeout_dc.active():
                    log.err(None, ERR_LDAP_SEARCH_FAILED)
                    raise drpc.CallError(ERR_LDAP_SEARCH_FAILED)
                else:
                    log.err(None, ERR_LDAP_TIMEOUT)
                    raise drpc.CallError(ERR_LDAP_TIMEOUT, {
                        'during': 'search',
                    })
        finally:
            if timeout_dc.active():
                timeout_dc.cancel()
            try:
                cl.transport.abortConnection()
            except Exception:
                pass

        defer.returnValue({
            'results': result,
        })

    @defer.inlineCallbacks
    def _paged_search(self, client, base_dn, filter_text, attributes,
                      pagination=PAGINATION_TYPE_CRITICAL,
                      max_result_size=None,
                      page_size=DEFAULT_PAGE_SIZE):
        """
        Given a bound client, search exhaustively using RFC 2696.
        Return a list of dictionaries containing the attributes
        of the resulting entries.

        * attributes: Set of lower-case byte-strings.
        """
        res = []

        def handle_msg(msg, d):
            try:
                if isinstance(msg.value, pureldap.LDAPSearchResultDone):
                    e = ldaperrors.get(
                        msg.value.resultCode, msg.value.errorMessage)
                    if isinstance(e, (ldaperrors.Success, ldaperrors.LDAPSizeLimitExceeded)):
                        cookie = get_cookie(msg.controls)
                        d.callback((None, cookie))
                    else:
                        d.callback((e, None))
                elif isinstance(msg.value, pureldap.LDAPSearchResultEntry):
                    obj = {}

                    # Always send DN. Overwrite DN from attribute set, if any.
                    obj['distinguishedname'] = [
                        escape_bytes(str(msg.value.objectName))]

                    for k, vs in msg.value.attributes:
                        # Smash attribute name case.
                        k = str(k.lower())

                        # Server may not honor attributes (e.g.
                        # SearchByTreeWalkingMixin).
                        if attributes and k.lower() not in attributes:
                            continue

                        # Covert value to list and encode for JSON.
                        vs = [escape_bytes(str(v)) for v in vs]

                        obj[k] = vs

                    # Refuse to return certain attributes even if all
                    # attributes were requested.
                    obj.pop('userpassword', None)

                    res.append(obj)
            except Exception as e:
                log.err()
            finally:
                return isinstance(msg.value, (
                    pureldap.LDAPBindResponse,
                    pureldap.LDAPSearchResultDone,
                ))

        if filter_text:
            filter_obj = ldapfilter.parseFilter(filter_text)
        else:
            filter_obj = None
        op = pureldap.LDAPSearchRequest(
            baseObject=base_dn,
            scope=pureldap.LDAP_SCOPE_wholeSubtree,
            derefAliases=0,
            sizeLimit=0,
            timeLimit=0,
            typesOnly=0,
            filter=filter_obj,
            attributes=attributes,
        )

        if pagination == PAGINATION_TYPE_CRITICAL:
            # AD may ignore the RFC 2696 control if it is not
            # critical. The caller can override this default if the
            # control should be present but not critical or absent.
            criticality = True
        else:
            criticality = False

        cookie = pureber.BEROctetString('')
        while True:
            if pagination == PAGINATION_TYPE_DISABLE:
                controls = None
            else:
                controls = [
                    (
                        RFC_2696_CONTROL_TYPE,
                        criticality,
                        pureber.BERSequence([
                            pureber.BERInteger(page_size),
                            cookie,
                        ]),
                    ),
                ]

            d = defer.Deferred()
            yield client.send(
                op=op,
                controls=controls,
                handler=functools.partial(handle_msg, d=d),
                handle_msg=True,
            )

            # handle_msg() is synchronous so d should be called by the
            # time it returns to LDAPClient.handle().
            if d.called:
                e, cookie = yield d
            else:
                log.err('Paging cookie not found!')
                break
            if e is not None:
                # So the RPC caller can distinguish between problems
                # with the search (e.g. bad configuration) and
                # searches that return no results.
                raise e
            if not cookie.value:
                break
            if max_result_size is not None and len(res) > max_result_size:
                break
        defer.returnValue(res)

    def next_range(self, previous_range, previous_result_empty,
                   range_len=LDAP_ATTR_DEFAULT_RANGE_LEN):
        """ Calculate the next range to use for requesting attribute values
            that exceed the servers maximum attr length setting.
            If the previous search caused an error because the range segment
            did not include values, try again with a '*' as the upper range
            to get any remaining members.
        """
        if previous_result_empty:
            return (0 if previous_range is None else previous_range[0], '*')
        if previous_range is None:
            return (0, range_len - 1)
        else:
            return (previous_range[1] + 1, previous_range[1] + range_len)

    @defer.inlineCallbacks
    def try_ranged_search(self, client, base_dn, objs):
        """ So, you are an empty group object or a group with too many members
            to return in one shot.
            Let's try to grab those members a few at a time until we determine
            that the group is empty or we've built up the full list of members.
        """
        entry = LDAPEntry(client, base_dn)
        for obj in objs:
            member_range = None
            previous_range = None
            previous_result_empty = False
            # loop until we've exhausted all range segments
            while True:
                # can't help you without a dn. sorry.
                if not obj.get('distinguishedname'):
                    break
                previous_range = member_range
                member_range = self.next_range(
                    previous_range, previous_result_empty)

                # create an ldap filter object to search with.
                # pureldap will escape the distinguishedname, making it safe.
                # Since we're reading the distinguishedname from search results,
                # we'll need to unescape it before it gets re-escaped.
                attr_obj = pureldap.LDAPAttributeDescription(
                    'distinguishedname')
                value_obj = pureldap.LDAPAssertionValue(
                    distinguishedname.unescape(obj['distinguishedname'][0]))
                filter_object = pureldap.LDAPFilter_equalityMatch(
                    attr_obj, value_obj)

                attributes = [
                    'member;range={0}-{1}'.format(
                        member_range[0], member_range[1])]
                try:
                    ranged_res = yield entry.search(
                        filterObject=filter_object, attributes=attributes)
                except LDAPOperationsError as e:
                    # searched for a range that exceeds the total number of
                    # elements in the attribute.
                    # if this is the first instance of this error, we may
                    # just need to try again with a '*' as the upper bound
                    # to get the last few members.
                    # if this is the second time (we are using the '*'), then
                    # we already have all of the members and can break.
                    if ldap.client.ERR_AD_CANT_RETRIEVE_ATTS in e.message:
                        if previous_result_empty:
                            break
                        previous_result_empty = True
                    # something bad happened... give up
                    else:
                        log.err(e)
                        break
                # no results; we're done here
                if len(ranged_res) == 0:
                    break
                else:
                    # update the result object with the newly discovered
                    # members
                    for k, vs in ranged_res[0].items():
                        k = str(k.lower())
                        vs = [escape_bytes(str(v)) for v in vs]
                        if k.startswith('member;'):
                            obj['member'].extend(vs)


def escape_bytes(s):
    return ''.join([
        c if c in PRINTABLE else '\\{0:02x}'.format(ord(c))
        for c in s
    ])


def get_cookie(controls):
    for control in controls:
        # control was parsed by pureldap.LDAPControl.fromBER but is
        # now [type, criticality, value], not an LDAPControl object.
        if control[0] == RFC_2696_CONTROL_TYPE:
            value = pureber.BERSequence.fromBER(
                tag=None,
                content=control[2],  # The control's value.
                berdecoder=pureber.BERDecoderContext())
            _size, cookie = value[0]
            return cookie
    return None


def parse_positive_int(v, arg_name):
    try:
        v = int(v)
    except Exception:
        v = None
    if v is None or v <= 0:
        raise drpc.CallBadArgError([arg_name])
    else:
        return v
