#!/usr/bin/env python
#
# grep for any use of the python "random" module EXCEPT SystemRandom.
#
# By default python uses a Mersenne Twister implementation
# (http://en.wikipedia.org/wiki/Mersenne_twister) to generate random
# values, and this is not suitable for cryptographic
# purposes. SystemRandom uses os.urandom to get random data, which is
# generally a much better choice.
#
# ask akgood@ if you have any questions.
#

import ast
import os
import subprocess
import sys


class RandomVisitor(ast.NodeVisitor):
    def __init__(self, filename):
        self.filename = filename
        self.bad_nodes = []

    def visit_Attribute(self, node):
        if (isinstance(node.value, ast.Name)
                and node.value.id == 'random'
                and node.attr != 'SystemRandom'):

            self.bad_nodes.append(node)

    def visit_ImportFrom(self, node):
        if (node.module == 'random'
                and any(alias.name != 'SystemRandom' for alias in node.names)):

            self.bad_nodes.append(node)


def find_files(root, exclude_dirs=None, exclude_files=None):
    if exclude_dirs is None:
        exclude_dirs = []

    if exclude_files is None:
        exclude_files = []

    exclude_args = (['--exclude-dir={}'.format(s) for s in exclude_dirs] +
                    ['--exclude={}'.format(s) for s in exclude_files])
    cmd = (['grep', '-l', '--null', '-m', '1', '-R', '--include=*.py'] +
           exclude_args + ['random', root])

    p = subprocess.Popen(
        cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    output, _ignored = p.communicate()
    p.wait()

    files = output.strip('\0').split('\0')
    return files


def check_tree(root, exclude_dirs=None, exclude_files=None):
    files = find_files(root, exclude_dirs, exclude_files)
    bad_files = {}

    for name in files:
        with open(name, 'r') as fp:
            contents = fp.read()
            try:
                m = ast.parse(contents)
            except SyntaxError as e:
                bad_files[name] = e
            else:
                visitor = RandomVisitor(name)
                visitor.visit(m)
                if visitor.bad_nodes:
                    bad_files[name] = visitor.bad_nodes
    return bad_files


def selftest():
    TEST_ROOT = './06-testcases'
    all_files = [s for s in os.listdir('./06-testcases')
                 if s.endswith('.py')]
    bad_files = check_tree('./06-testcases')

    success = True

    for filename in all_files:
        path = '{}/{}'.format(TEST_ROOT, filename)
        if filename.startswith('pass') and path in bad_files:
            err('failed self-test on {}!'.format(filename))
            success = False
        elif filename.startswith('fail') and path not in bad_files:
            err('failed self-test on {}!'.format(filename))
            success = False
    return success


def main():
    ROOT = os.environ.get('SRC', '.')
    EXCLUDE_DIRS = ['dist', 'third-party', 'duoauthproxy-build', 'test']
    EXCLUDE_FILES = ['test_fipscustomize.py', '_fipscustomize.py']

    print 'checking for improper "random" use...'

    #  if not selftest():
    #    sys.exit(1)

    bad_files = check_tree(ROOT, EXCLUDE_DIRS, EXCLUDE_FILES)

    if bad_files:
        for name, failure in bad_files.iteritems():
            if isinstance(failure, SyntaxError):
                err('{}: syntax error at line {}'.format(
                    name, failure.lineno))
            else:
                for node in failure:
                    err('{}: bad "random" module use at line {} - {}'.format(
                        name, node.lineno, ast.dump(node)))
        err('detected an improper use of the "random" module (or syntax error)')
        err('you should use only random.SystemRandom!')
        sys.exit(1)

    print 'OK!'


def err(msg):
    print >>sys.stderr, msg


if __name__ == '__main__':
    main()
