#!/usr/bin/env python
#
# grep for any use of the python "xml" module for parsing EXCEPT defusedxml. Etree and Minidom are BAD.
# Using TreeBuilder is okay because it does not parse xml
#
# ask vikask@ or security@ if you have any questions.
#

import ast
import os
import subprocess
import sys
import argparse


class OnlyDefusedVisitor(ast.NodeVisitor):
    def __init__(self, filename):
        self.filename = filename
        self.bad_nodes = []

    def visit_Import(self, node):
        if node.names:
            for alias in node.names:
                fields = alias.name.split('.')
                # xml, lxml, xmlrpclib are blacklisted except xml.sax.saxutils
                if ((fields[0] == 'xmlrpclib' or fields[0] == 'lxml') or
                        (fields[0] == 'xml' and
                         alias.name != 'xml.sax.saxutils')):
                    self.bad_nodes.append(node)

    def visit_ImportFrom(self, node):
        if node.module:
            # its OK to import the exception type, but nothing more
            if (node.module == 'xml.etree.ElementTree' and
                    len(node.names) == 1 and
                    node.names[0].name == 'ParseError'):
                return
            # xml, lxml, xmlrpclib are blacklisted except xml.sax.saxutils
            if (node.module.split('.')[0] in ('xml', 'lxml', 'xmlrpclib') and
                    node.module != 'xml.sax.saxutils' and
                    (node.module != 'xml.sax' or
                     any(a.name != 'saxutils' for a in node.names))):
                self.bad_nodes.append(node)


class LxmlWithUnsafeVisitor(OnlyDefusedVisitor):

    def visit_Import(self, node):
        if not node.names:
            return
        names = node.names
        # process any imports that are just lxml as OnlyDefusedVisitor
        # the remaining imports are lxml, if all named "unsafe_" can skip them
        if all(a.name == 'lxml' and
               a.asname is not None and
               a.asname.startswith('unsafe_') for a in names):
            return
        # strange import combinations mixing multiple modules aren't skipped
        else:
            super(LxmlWithUnsafeVisitor, self).visit_Import(node)

    def visit_ImportFrom(self, node):
        if not node.module:
            return
        names = node.names
        module = node.module.split('.')
        # process any imports that are just lxml as OnlyDefusedVisitor
        # the remaining imports are lxml, if all named "unsafe_" can skip them
        if all(module[0] == 'lxml' and
               a.asname is not None and
               a.asname.startswith('unsafe_') for a in names):
            return
        # any non-usafe aliases means the code should fail audit
        else:
            super(LxmlWithUnsafeVisitor, self).visit_ImportFrom(node)


def find_files(root, exclude_list):
    exclude_dir_args = ['--exclude-dir={}'.format(s) for s in exclude_list]
    cmd = (['grep', '-l', '--null', '-m', '1', '-R', '--include=*.py']
           + exclude_dir_args + ['xml', root])

    p = subprocess.Popen(
        cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    output, _ignored = p.communicate()
    p.wait()
    return output.strip('\0').split('\0')


def check_tree(root, exclude_list, Visitor):
    files = find_files(root, exclude_list)
    bad_files = {}
    if not files:
        raise ValueError('no files matched "{}"'.format(root))
    for name in files:
        with open(name, 'r') as fp:
            contents = fp.read()
            try:
                m = ast.parse(contents)
            except SyntaxError as e:
                bad_files[name] = e
            else:
                visitor = Visitor(name)
                visitor.visit(m)
                if visitor.bad_nodes:
                    bad_files[name] = visitor.bad_nodes
    return bad_files


def main():
    ROOT = os.environ.get('SRC', '.')
    EXCLUDE = ['dist', 'third-party', 'duoauthproxy-build', ]

    message_lines = """detected at least one improper use of an xml parser (or syntax error)\n')
you should use only defusedxml for parsing xml from untrusted input!
""".format()
    print 'checking for improper "xml" use...'
    total_bad_files = 0
    bad_files = check_tree(
        ROOT,
        EXCLUDE,
        OnlyDefusedVisitor)
    if bad_files:
        for name, failure in bad_files.iteritems():
            if isinstance(failure, SyntaxError):
                sys.stderr.write('{} syntax error at line {}\n'.format(
                    name, failure.lineno))
            else:
                for node in failure:
                    sys.stderr.write(
                        '{}: bad "xml" module use at line {}: {}\n'.format(
                            name, node.lineno, ast.dump(node)))
        for m in message_lines:
            sys.stderr.write(m)
        total_bad_files += len(bad_files.keys())

    if total_bad_files:
        sys.exit(1)

    print 'OK!'


def self_test():
    ROOT = '15-testcases'
    UNSAFE = 'unsafe'

    print 'checking for improper "xml" use...'

    general_bad_files = check_tree(ROOT, (UNSAFE,), OnlyDefusedVisitor)
    subdir = os.path.join(ROOT, UNSAFE)
    unsafe_bad_files = check_tree(subdir, [], LxmlWithUnsafeVisitor)
    number_of_fail_cases = (8, 4)  # normal failures, unsafe failures
    result = (len(general_bad_files), len(unsafe_bad_files))
    if result != number_of_fail_cases:
        msg = "{} cases should have bad modules but {} bad modules detected\n"
        sys.stderr.write(
            msg.format(str(number_of_fail_cases), str(result)))

    bad_files = dict(general_bad_files, **unsafe_bad_files)
    if len(bad_files) > 0:
        for (name, failure) in [(name, failure) for name, failure in bad_files.iteritems() if 'pass' in name]:
            for node in failure:
                sys.stderr.write(
                    "{}: identified bad module use at line {}: {}\n".format(
                        name, node.lineno, ast.dump(node)))

    if result != number_of_fail_cases:
        sys.stderr.write('detected at least one failed self-test\n')
        sys.exit(1)

    print 'OK!'


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--test', action='store_true', default=False)

    args = parser.parse_args()

    # always run the self test!
    #  self_test()

    if not args.test:
        main()
