#!/usr/bin/env python
#
# grep for any use of unsafe functions in python.

import ast
import os
import subprocess
import sys


class IllegalFunctionVisitor(ast.NodeVisitor):
    def __init__(self, filename):
        self.filename = filename
        self.bad_nodes = []
        self.safe_input_import = False

    def visit_Call(self, node):
        # Check for usage of eval
        if (isinstance(node, ast.Call) and
                isinstance(node.func, ast.Name) and
                node.func.id == 'eval'):
            self.bad_nodes.append(node)

        # Check for bad usage of input, as long as it's not from six.moves
        # (where you can call input in a safe way)
        if (isinstance(node, ast.Call) and
                isinstance(node.func, ast.Name) and
                node.func.id == 'input' and
                not self.safe_input_import):
            self.bad_nodes.append(node)

        # Check for bad usage of os.system
        if (isinstance(node, ast.Call) and
                isinstance(node.func, ast.Attribute) and
                isinstance(node.func.value, ast.Name) and
                node.func.value.id == 'os' and
                node.func.attr == 'system'):
            self.bad_nodes.append(node)

    def visit_Exec(self, node):
        # Check for usage of exec
        self.bad_nodes.append(node)

    def visit_ImportFrom(self, node):
        # input from six.moves is valid, so if input is imported in
        # a safe way, allow input to be used for the rest of the file
        if (node.module == 'six.moves' and
                any(alias.name == 'input' for alias in node.names)):
            self.safe_input_import = True

        # Check for use of pdb
        if node.module == 'pdb':
            self.bad_nodes.append(node)

        # Check for use of os.system
        if (node.module == 'os' and
                any(alias.name == 'system' for alias in node.names)):
            self.bad_nodes.append(node)

    def visit_Import(self, node):
        # Check for usage of pdb
        if any(alias.name == 'pdb' for alias in node.names):
            self.bad_nodes.append(node)


def find_files(root, exclude_list=None):
    if exclude_list is None:
        exclude_list = []

    exclude_args = ['--exclude-dir={}'.format(s) for s in exclude_list]
    cmd = (['grep', '-l', '--null', '-m', '1', '-R', '--include=*.py']
           + exclude_args
           + ['-E', 'eval|input|pdb|exec|system', root])

    p = subprocess.Popen(
        cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    output, _ignored = p.communicate()
    p.wait()
    return output.strip('\0').split('\0')


def check_tree(root, exclude_list=None):
    files = find_files(root, exclude_list)
    bad_files = {}

    for name in files:
        with open(name, 'r') as fp:
            contents = fp.read()
            try:
                m = ast.parse(contents)
            except SyntaxError as e:
                bad_files[name] = e
            else:
                visitor = IllegalFunctionVisitor(name)
                visitor.visit(m)
                if visitor.bad_nodes:
                    bad_files[name] = visitor.bad_nodes
    return bad_files


def selftest():
    TEST_ROOT = './18-testcases'
    all_files = [s for s in os.listdir('./18-testcases')
                 if s.endswith('.py')]
    bad_files = check_tree('./18-testcases')

    success = True
    for filename in all_files:
        path = '{}/{}'.format(TEST_ROOT, filename)
        if filename.startswith('pass') and path in bad_files:
            err('failed self-test on {}!'.format(filename))
            success = False
        elif filename.startswith('fail') and path not in bad_files:
            err('failed self-test on {}!'.format(filename))
            success = False
    return success


def main():
    ROOT = os.environ.get('SRC', '.')
    EXCLUDE = ['duoauthproxy-build', 'third-party', 'dist']

    print 'checking for use of illegal functions...'

    #  if not selftest():
    #      sys.exit(1)

    bad_files = check_tree(ROOT, EXCLUDE)

    if bad_files:
        for name, failure in bad_files.iteritems():
            if isinstance(failure, SyntaxError):
                err('{}: syntax error at line {}'.format(
                    name, failure.lineno))
            else:
                for node in failure:
                    err('{}: Detected illegal function use at {} - {}'.format(
                        name, node.lineno, ast.dump(node)))
        err("Detected use of an illegal function. Don't do that!")
        sys.exit(1)

    print 'OK!'


def err(msg):
    print >>sys.stderr, msg


if __name__ == '__main__':
    main()
