# This module defines code to run before and after certain events. See: https://behave.readthedocs.io/en/stable/tutorial.html?highlight=before_all#environmental-controls
from os import getcwd, chdir, kill
from platform import system

from test.features.fixtures.install import base
from test.features.fixtures.install.nix import LinuxInstaller
from test.features.fixtures.install.windows import WindowsInstaller
from test.features.fixtures.servers.servers import backing_ldap_server, backing_fake_duo_server

import signal
import psutil

from behave import use_fixture


def before_scenario(context, _scenario):
    context.starting_dir = getcwd()
    context.installer.clear_log_file()  # start with a clean slate after each scenario

    # `config_settings` will store config values that will be used when constructing an
    # authproxy config. This let's us chain configuration steps before constructing
    # an AP config and starting the Auth Proxy.
    context.config_settings = {}


def after_scenario(context, scenario):
    if scenario.feature.name == u'Primary Only Mode':
        kill(context.pid, 9)
    chdir(context.starting_dir)
    context.config_settings = {}  # reset the config settings for the next scenario to use
    context.installer.save_log(scenario)


def before_all(context):
    context.installer = _get_installer()
    use_fixture(backing_fake_duo_server, context)
    use_fixture(backing_ldap_server, context)
    base.setup(context)


def after_all(context):
    context.installer = _get_installer()
    base.teardown(context)

    _cleanup_fake_server_processes(context.backing_ldap_server_process.pid)
    _cleanup_fake_server_processes(context.backing_fake_duo_server_process.pid)


def before_tag(context, tag):
    pass


def after_tag(context, tag):
    if tag == "remove_auth_events_log":
        # The authproxy has to be stopped in order to remove the auth_events
        # log on windows. Only stop and start the service if it's already
        # running, though, since we want the state to stay the same outside
        # of removing the log file.
        system_is_windows = _system_is_windows()
        if system_is_windows and context.installer.is_running():
            context.installer.stop()
            context.installer.remove_auth_events_log()
            context.installer.start()
        else:
            context.installer.remove_auth_events_log()


def _get_installer():
    if _system_is_windows():
        return WindowsInstaller()
    else:
        return LinuxInstaller()


def _system_is_windows():
    return system() == 'Windows'


def _cleanup_fake_server_processes(pid):
    parent = psutil.Process(pid)
    children = parent.children()

    # We need to explicitly iterate over the child processes and terminate them
    # otherwise they will stick around and cause gitlab to hang on Windows
    for process in children:
        process.send_signal(signal.SIGTERM)
    parent.send_signal(signal.SIGTERM)
