import os
import subprocess
import shutil

from behave import fixture
import duoauthproxy

from test.features.fixtures import configs


@fixture
def setup(context, **kwargs):
    context.installer.install()
    context.installer.write_config(configs.get())
    context.installer.copy_fakeduo_ca()
    context.installer.clear_log_file()
    context.installer.start()


@fixture
def teardown(context, **kwargs):
    """Stop the duo auth proxy and remove it"""
    context.installer.stop()
    context.installer.uninstall()


class Installer(object):
    """Abstract base class for both Windows and Linux installers"""
    def __init__(self):
        """Inheriting classes must define install_dir"""
        log_dir = os.path.join(self.install_dir, 'log')
        self.log_file = os.path.join(log_dir, 'authproxy.log')
        self.auth_events_log = os.path.join(log_dir, 'authevents.log')
        self.tool_log_file = os.path.join(log_dir, 'connectivity_tool.log')
        self.conf_file = os.path.join(self.install_dir, 'conf', 'authproxy.cfg')

    def clear_log_file(self):
        with open(self.log_file, "w"):
            pass

    # save the log files after each scenario just in case we want to manually inspect them
    def save_log(self, scenario):
        filename = scenario.feature.name.replace(" ", "_").lower() + "-" + scenario.name.replace(" ", "_").lower() + ".log"
        source = self.log_file
        target = os.path.join(os.path.join(os.getcwd()), 'test', 'features', 'artifacts', filename)
        shutil.copyfile(source, target)

    def remove_auth_events_log(self):
        if os.path.isfile(self.auth_events_log):
            os.remove(self.auth_events_log)

    def write_config(self, config_data):
        with open(self.conf_file, "w") as config_handle:
            config_handle.write(config_data)

    def copy_fakeduo_ca(self):
        source = os.path.join('test', 'features', 'fixtures', 'servers', 'fakeduo-rootca.pem')
        target = os.path.join(self.install_dir, 'conf', 'fakeduo-rootca.pem')

        shutil.copyfile(source, target)

    # We don't care to verify the cert for tests
    def clear_ca_bundle(self):
        with open(os.path.join(self.install_dir, 'conf', 'ca-bundle.crt'), "w"):
            pass

    def install(self):
        pass

    def uninstall(self):
        pass

    def is_running(self):
        pass

    def start(self):
        pass

    def stop(self):
        pass

    def restart(self):
        pass

    @staticmethod
    def _run(runner, args, **kwargs):
        """Wrapper used to make calls with subprocesses
        Args:
            runner: One of subprocess functions such as call or pOpen
            args: args for the runner
            kwargs: kwargs for the runner
        Returns:
            The result of the runner
        """
        print "{0}".format(' '.join(args))
        return runner(args, **kwargs)


def get_authproxy_version():
    return duoauthproxy.__version__


def get_commit_hash():
    """ Call out to a subprocess to get the latest git commit hash.
    Returns:
        str: Commit hash or '' if we failed to get the commit
    """
    try:
        commit_hash = subprocess.check_output(["git", "log", "--pretty=format:%h", "-n", "1"])
    except subprocess.CalledProcessError as e:
        print "Error getting commit hash:"
        print e.message
        print "Continuing without one."
        return ''

    return commit_hash
