#
# Copyright (c) 2012 Duo Security
# All Rights Reserved
#
import contextlib
import copy
import itertools
import platform
import re
import tempfile
import traceback
import unittest

from pyrad import packet
import mox
import mock
import netaddr
import six

from twisted.internet import defer
from twisted.internet import reactor
import twisted.test.proto_helpers
import twisted.trial.unittest

from ldaptor.protocols.ldap import ldaperrors

import duoauthproxy.modules.mock_client
from duoauthproxy.lib import duo_async, base, forward_serv, log, config_provider
from duoauthproxy.lib.radius.base import radius_dictionary
from duoauthproxy.lib.radius.challenge import _ChallengeState
from duoauthproxy.lib.radius.server import parse_radius_secrets
from duoauthproxy.lib.radius.server import _ProxyRequest
from duoauthproxy.lib.validation.config.config_toolbox import ConfigTestToolbox
from duoauthproxy.modules.ad_client import _ADServiceClientProtocol

import mock_objects

from duoauthproxy.lib.validation.connectivity.connectivity_toolbox import ConnectivityTestToolbox
from duoauthproxy.lib.validation.test_resolver import BaseTestResolver
from duoauthproxy.lib.validation.connectivity import connectivity_results


def sleep(seconds):
    d = defer.Deferred()
    reactor.callLater(seconds, d.callback, None)
    return d


class ProtectedValidatingConfigDict(config_provider.ConfigDict):
    """ConfigDict with "protected" keys that can't be accessed with get_str().

    Sublass of ConfigDict that requires keys beginning with specific "protected"
    prefixes to be accessed with get_protected_str() instead of get_str().
    """
    prefixes_accessed_with_protection = (
        'service_account_password',
        'skey',
        'secret',
        'radius_secret_'
    )
    protected_key_prefixes = (
        'service_account_password_protected',
        'skey_protected',
        'secret_protected',
        'radius_secret_protected_'
    )

    def called_from(self):
        stack = traceback.extract_stack()
        # [-1] called_from
        # [-2] get_str
        # [-3] get_protected_str (hopefully)
        return stack[-3][2]

    def get_str(self, key, default=None):

        # some keys aren't allowed to get_str at all!
        if key.startswith(self.protected_key_prefixes):
            raise Exception(
                "%s should only be accessed with get_protected_str" % key
            )

        # some keys can only be get_str'd from get_protected_str
        # to ensure that we'll try to get the protected variant first
        if key.startswith(self.prefixes_accessed_with_protection):
            if self.called_from() != 'get_protected_str':
                raise Exception(
                    "%s should be accessed using get_protected_str"
                    " rather than directly with get_str."
                    " get_str was called by %s " % (key, self.called_from())
                )

        return config_provider.ConfigDict.get_str(self, key, default)


class TestCase(twisted.trial.unittest.TestCase):
    """Basic test case class inherited by all other test case classes.

    Defines various assert functions not defined in Python 2.6.
    """
    if hasattr(unittest.TestCase, 'assertRaisesRegexp'):
        assertRaises = unittest.TestCase.assertRaises
        assertRaisesRegexp = unittest.TestCase.assertRaisesRegexp
    else:
        # Python 2.6 assertRaises isn't a context manager;
        # assertRaisesRegexp isn't present.
        @contextlib.contextmanager
        def assertRaises(self, exception):
            try:
                yield
            except exception:
                pass
            else:
                self.fail("Didn't raise " + exception.__name__)

        @contextlib.contextmanager
        def assertRaisesRegexp(self, exception, regexp):
            try:
                yield
            except exception as e:
                if not re.search(regexp, str(e)):
                    self.fail("Exception %s didn't match regexp %s"
                              % (str(e), regexp))
            else:
                self.fail("Didn't raise " + exception.__name__)

    if not hasattr(unittest.TestCase, 'assertItemsEqual'):
        # Python 2.6 lacks assertItemsEqual. Unlike the upstream
        # version this does not generate diffs or handle unhashable
        # items.
        def assertItemsEqual(self, expected_seq, actual_seq):
            return self.assertEqual(set(expected_seq),
                                    set(actual_seq))

    if not hasattr(unittest.TestCase, 'assertIs'):
        # Python 2.6 lacks assertIs.
        def assertIs(self, a, b):
            return self.assertTrue(a is b)

    if not hasattr(unittest.TestCase, 'assertIsNot'):
        # Python 2.6 lacks assertIsNot.
        def assertIsNot(self, a, b):
            return self.assertTrue(a is not b)

    if not hasattr(unittest.TestCase, 'assertIn'):
        # Python 2.6 lacks assertIn.
        def assertIn(self, a, b, msg=None):
            return self.assertTrue(a in b, msg=msg)

    if not hasattr(unittest.TestCase, 'assertNotIn'):
        # Python 2.6 lacks assertNotIn.
        def assertNotIn(self, a, b, msg=None):
            return self.assertTrue(a not in b, msg=msg)

    if not hasattr(unittest.TestCase, 'assertIsInstance'):
        # Python 2.6 lacks assertIsInstance.
        def assertIsInstance(self, a, b, msg=None):
            return self.assertTrue(isinstance(a, b), msg=msg)

    if not hasattr(unittest.TestCase, 'assertIsNone'):
        # Python 2.6 lacks assertIsNone.
        def assertIsNone(self, a, msg=None):
            return self.assertTrue(a is None, msg=msg)

    if not hasattr(unittest.TestCase, 'assertIsNotNone'):
        # Python 2.6 lacks assertIsNotNone.
        def assertIsNotNone(self, a, msg=None):
            return self.assertTrue(a is not None, msg=msg)

    def wrap_do_connect(self, clock, seconds=0, to_override_module=twisted.internet.tcp.BaseClient, to_override_name='doConnect'):
        """
        This is a utility function for wrapping the "callLater" method in twisted
        for the scope of a single function call with the goal of speeding up time
        from the perspective of the network connection. Typically used in the testing
        of timeouts.

        Args:
            clock (twisted.internet.task.Clock): A clock which will be used to manipulate time
            seconds (int): Amount of seconds to have pass directly after the overriden function is called
            to_override_module (module): Module containing callable in which reactor.callLater will be replaced
            to_override_name (string): Name of callable in to_override_module in which reactor.callLater will be replaced
        """

        oldDoConnect = getattr(to_override_module, to_override_name)

        def wrapped_do_connect(*args, **kwargs):
            oldCallLater = twisted.web.client.reactor.callLater
            try:
                twisted.web.client.reactor.callLater = clock.callLater
                result = oldDoConnect(*args, **kwargs)

                if seconds:
                    clock.advance(seconds)
            finally:
                twisted.web.client.reactor.callLater = oldCallLater
            return result
        return self.patch(to_override_module, to_override_name, wrapped_do_connect)

    def assert_call_count_with_args(self, mocked_method, expected_call_count, *args, **kwargs):
        """
        Asserts that the mocked method was called exactly the expected number
        of times with the specified arguments.

        Note that, aside from allowing for specifying the number of expected
        calls, this differs from Mock.assert_called_once_with.
        assert_called_once_with asserts that the total method call count is 1
        and that the method was called with the specified arguments. This method
        asserts that the mocked method was called with the specified arguments
        expected_call_count times but does not care about other calls of the
        method using different arguments.

        Args:
            mocked_method (Mock): The mocked method to assert against
            expected_call_count (int): The expected call count for the method
            *args: Positional arguments the method should have been called with
            **kwargs: Keyword arguments the method should have been called with

        """
        call_to_check = mock.call(*args, **kwargs)
        matching_calls = [call for call in mocked_method.call_args_list if call == call_to_check]
        times_called = len(matching_calls)
        error_message = 'Expected method \'{method}\' to be called with specified ' \
                        'args {expected_call_count} times. ' \
                        'Actual call count: {times_called}. args: {args}. kwargs: {kwargs}'
        self.assertEqual(expected_call_count, times_called, error_message.format(
            method=mocked_method, expected_call_count=expected_call_count,
            times_called=times_called, args=args, kwargs=kwargs))

    def assertUnorderedUnhashableListsEqual(self, list1, list2, msg=None):
        temp1 = list(list1)
        temp2 = list(list2)
        for item in list1:
            try:
                temp2.remove(item)
            except ValueError:
                pass
        for item in list2:
            try:
                temp1.remove(item)
            except ValueError:
                pass
        err_messages = []
        if temp1:
            err_messages.append("list1 contains the following items not in list2: {}. ".format(', '.join(str(x) for x in temp1)))
        if temp2:
            err_messages.append("list1 contains the following items not in list2: {}. ".format(', '.join(str(x) for x in temp2)))
        if err_messages:
            if msg:
                err_messages.append(msg)
            self.fail("\n".join(err_messages))


class StringTransport(twisted.test.proto_helpers.StringTransport):
    """Custom Twisted Transport mock object used to emulate a network transport.

    Used as a fake network transport by protocols like DuoAsyncClient or
    CloudApiFactory so test requests and responses can be sent and received
    properly during testing.
    """
    aborted = False
    disconnected = False
    TLS = False

    def abortConnection(self):
        self.aborted = True
        self.disconnected = True


def ReopenableTemporaryFile(data):
    """Stores keys/certs in a temp file, often for Windows-based tests.

    On Windows, unlike POSIX, a NamedTemporaryFile must be closed before it can
    be re-opened. Even within a single process. In this case tearDown must
    delete it manually.

    Can't subclass NamedTemporaryFile because it's not a class!
    """
    close_before_reopen = (
        platform.system() == 'Windows' or
        platform.system().startswith('CYGWIN')
    )
    fh = tempfile.NamedTemporaryFile(delete=(not close_before_reopen))
    fh.write(data)
    fh.flush()
    if close_before_reopen:
        fh.close()
    return fh


class ServerModuleTest(object):
    """Class used to test RADIUS/LDAP server modules.

    Sets basic vars and defines helper functions used for testing server module
    classes (among others) and their various (pre)authentication flows.

    Because of the way python's unittest module works, setUp() and tearDown()
    are run before/after every test function beginning with 'test'.

    Attributes:
        self.instance: Instance of the server type defined by server_section_name
            (eg. radius_server_concat, radius_server_auto)
        self.client_instance: Mocked-out instance of the client, which normally
            does primary authentication on behalf of self.instance
        self.f_instance: Instance of forward_serv, which receives packets and
            forwards them to self.instance
        self.instance.protocol.client: Mocked-out instance of AsyncDuoClient,
            which normally handles communications with the Duo API
        RADIUS attribute dicts: When passed to _test_auth(), emulates various
            attributes of a packet sent by a client's response packet

    Mocks out:
        self.client_instance: mocked RadiusClient
        self.instance.protocol.client: mocked AsyncDuoClient
        self.f_instance.transport: mocked TransportQueue
    """
    def setUp(self, extra_server_config, server_module, server_section_name):
        self.radius_packet_id = itertools.count(2)
        self.username = 'me'
        self.password = 'word2pass'
        self.default_factor = 'i am the default factor'

        self.mox = mox.Mox()

        self.pass_through_radius_attrs = {}
        self.client_config = config_provider.ConfigDict({
            'username': self.username,
            'password': self.password,
            'chap_challenge': 'You killed my father',
            'chap2_response': 'Prepare to die',
            'chap2_expired': 'Expired',
            'chap2_cpw': 'New Password Please!'
        })

        self.chap2_request_attrs = {
            'MS-CHAP-Challenge': 'You killed my father',
            'MS-CHAP2-Response': 'Prepare to die'
        }

        self.chap2_bad_request_attrs = {
            'MS-CHAP-Challenge': 'You killed my father',
            'MS-CHAP2-Response': 'Fail me.'
        }

        self.chap2_cpw_attrs = {
            'MS-CHAP2-CPW': 'New Password Please!'
        }

        self.chap2_bad_cpw_attrs = {
            'MS-CHAP2-CPW': 'Fail Me'
        }

        self.chap2_expired_request_attrs = {
            'MS-CHAP-Challenge': 'You killed my father',
            'MS-CHAP2-Response': 'Expired'
        }

        # 691 = Auth Failed
        self.chap2_error_attr = {
            'MS-CHAP-Error': [u'\x00E=691 R=0 V=3']
        }

        # 648 = Password Expired
        self.chap2_ps_expired_attr = {
            'MS-CHAP-Error': [u'\x00E=648 R=0 V=3']
        }

        self.client_instance = duoauthproxy.modules.mock_client.Module(
            self.client_config,
        )

        self.radius_client_port = 9563
        self.radius_ip = '1.2.3.4'
        self.secret = 's3cr3t'
        self.radius_ip2 = '10.2.68.19'
        self.secret2 = 'woohoo'
        server_config = {
            'radius_ip_1': self.radius_ip,
            'radius_secret_1': self.secret,
            'radius_ip_2': self.radius_ip2,
            'radius_secret_2': self.secret2,
            'ikey': 'nope',
            'skey': 'nuh-uh',
        }
        server_config.update(extra_server_config)
        server_config = config_provider.ConfigDict(server_config)
        self.instance = server_module.Module(
            server_config,
            self.client_instance,
            server_section_name,
        )

        server_port = 12345
        server_networks = {self.instance: parse_radius_secrets(server_config).keys()}
        server_names = {self.instance: 'test-server'}
        self.f_instance = forward_serv.Module(server_port, server_networks, server_names, '', False)

        server = self.f_instance.protocol
        server.transport = mock_objects.BasicMockTransport()
        self.instance.protocol.client = self.mox.CreateMock(
            duo_async.AsyncDuoClient)
        setattr(self.instance.protocol.client, 'host', 'api.test')
        setattr(self.instance.protocol.client, 'port', 444)

    def tearDown(self):
        self.f_instance.stopService()
        self.instance.protocol.cleanup_all()
        self.client_instance.stopService()
        self.mox.UnsetStubs()

    def create_cpw_req(self, bad=False):
        """Helper function for MS-CHAP2 test functions.

        Builds a Change-Password request packet to be sent in challenge_state
        when using _test_auth().

        Returns:
            _ProxyRequest of the auth request
        """
        request = packet.AuthPacket(
            code=packet.AccessRequest,
            secret='secret',
            dict=radius_dictionary()
        )
        request.source = 'Bwa ha ha'
        if bad:
            cpw_attrs = self.chap2_bad_cpw_attrs
        else:
            cpw_attrs = self.chap2_cpw_attrs
        for attr, value in cpw_attrs.iteritems():
            request[attr] = value

        request['User-Name'] = self.username
        return _ProxyRequest(request)

    # Call preauth functions first if you want _test_auth() to also preauth

    def enqueue_duo_preauth(self, result, status=None, groups=None):
        server = self.instance.protocol
        duo_client = server.client
        if status is None:
            status = str(result)
        method_call = duo_client.preauth(mox.IsA(six.string_types),
                                         mox.Or(mox.IsA(type(None)), mox.IsA((six.string_types))))
        preauth_ret = {
            'result': result,
            'status': status,
            'factors': {
                '1': 'push1',
                '2': 'phone1',
                '3': 'sms1',
                'default': self.default_factor,
            },
            'prompt': '?',
        }
        if groups is not None:
            preauth_ret['groups'] = groups
        method_call.AndReturn(preauth_ret)
        return method_call

    def enqueue_duo_preauth_error_fail_open(self):
        e = duo_async.DuoAPIFailOpenError()
        return self.enqueue_duo_preauth(None).AndRaise(e)

    def enqueue_duo_preauth_error_fail_closed(self):
        e = duo_async.DuoAPIFailClosedError()
        return self.enqueue_duo_preauth(None).AndRaise(e)

    # Call these first when you want _test_auth to also do secondary auth

    def enqueue_duo_auth(self, result, status=None,
                         factor=mox.IsA(six.string_types)):
        server = self.instance.protocol
        duo_client = server.client
        if status is None:
            status = str(result)
        method_call = duo_client.auth(mox.IsA(six.string_types),
                                      factor,
                                      # Or() can't actually compare to
                                      # a specific value.
                                      mox.Or(mox.IsA(type(None)),
                                             mox.IsA(six.string_types)))
        method_call.AndReturn({
            'result': result,
            'status': status,
        })
        return method_call

    def enqueue_duo_auth_error_fail_open(self):
        e = duo_async.DuoAPIFailOpenError()
        return self.enqueue_duo_auth(None).AndRaise(e)

    def enqueue_duo_auth_error_fail_closed(self):
        e = duo_async.DuoAPIFailClosedError()
        return self.enqueue_duo_auth(None).AndRaise(e)

    # Call duo_proxy_init functions before calling _test_auth() if you want
    # DuoAsyncClient to do a mock proxy_init() during _test_auth()
    # Useful mainly for iframe server tests that require a transaction id
    # before doing secondary auth

    def enqueue_duo_proxy_init(self, proxy_txid='lalala'):
        server = self.instance.protocol
        duo_client = server.client
        method_call = duo_client.proxy_init(mox.IsA(six.string_types))
        method_call.AndReturn({
            'proxy_txid': proxy_txid,
        })
        return method_call

    def enqueue_duo_proxy_init_error_fail_open(self):
        e = duo_async.DuoAPIFailOpenError()
        return self.enqueue_duo_proxy_init().AndRaise(e)

    def enqueue_duo_proxy_init_error_fail_closed(self):
        e = duo_async.DuoAPIFailClosedError()
        return self.enqueue_duo_proxy_init().AndRaise(e)

    def enqueue_duo_proxy_finish(self, valid_cookie, username):
        server = self.instance.protocol
        duo_client = server.client
        method_call = duo_client.proxy_finish(mox.IsA(six.string_types))
        method_call.AndReturn({
            'valid_cookie': valid_cookie,
            'user': username,
        })
        return method_call

    def enqueue_duo_proxy_finish_error_fail_open(self):
        e = duo_async.DuoAPIFailOpenError()
        return self.enqueue_duo_proxy_finish(None, None).AndRaise(e)

    def enqueue_duo_proxy_finish_error_fail_closed(self):
        e = duo_async.DuoAPIFailClosedError()
        return self.enqueue_duo_proxy_finish(None, None).AndRaise(e)

    @defer.inlineCallbacks
    def _test_auth(self, expect_code,
                   username,
                   password=None,
                   pw_codec='utf-8',
                   failmode=duo_async.FAILMODE_SECURE,
                   request_attrs=None,
                   reject_attrs=None,
                   challenge_response=None,
                   challenge_send_user_password=True,
                   challenge_state=None,
                   expect_msg_contains=None):
        """Performs a mock authentication an end user would make against the server.

        A request packet is made, with attributes such as username, password,
        RADIUS attrs, and challenge-response being set.

        _test_auth will cause the server instance to begin processing the packet,
        but it will NOT do any actual primary/secondary auth or preauthing on its
        own - use enqueue_preauth(), enqueue_duo_preauth(), enqueue_duo_proxy_init(),
        or their variants first before calling _test_auth()!

        Args:
            failmode: if base.FAILMODE_SAFE, server will fail open; if FAILMODE_SECURE,
                server will fail closed
            request_attrs: dict of attributes and their values to be added to
                the authentication request being sent
            reject_attrs: dict of attributes and their values expected in response
                packets that are of type AccessReject
            challenge_response: string of pretend user's reply to the challenge
                presented by the server. eg. 'push', 'anything', ''
            challenge_send_user_password: if True, encrypts challenge_response
                and includes it in User-Password field of request packet
            challenge_state: dict containing the challenge state, if any.
                Stored in server's list of pending challenges.
                Only needed if the server type you're working with keeps track
                of current challenges, such as in radius_server_auto or
                radius_server_challenge.
                Example challenge_state:
                {
                    'challenge_handler': self.instance.protocol.handle_auth_challenge ||
                                         self.instance.protocol.handle_enroll_challenge,
                    'is_chap2': True || False,
                    'factors': {'1': 'sms'}, (optional)
                    'prompt': '?' (optional)
                }
            expect_msg_contains (string): Expected string in the Reply-Message
                field of response packet yielded by _test_process_packet

        Yields:
            An AuthPacket containing the response packet.

        Raises:
            AssertionError: if proper attributes aren't in the response (as in
                the case of pass-through attributes or an AccessReject packet),
                unexpected RADIUS response type, non-matching reply/response IDs,
                wrong Reply-Message field
        """
        if request_attrs is None:
            request_attrs = {}
        if challenge_state is None:
            challenge_state = {}
        server = self.instance.protocol
        server.failmode = failmode

        radius_packet_id = self.radius_packet_id.next()
        request = packet.AuthPacket(code=packet.AccessRequest,
                                    id=radius_packet_id,
                                    secret=self.secret,
                                    dict=radius_dictionary())

        if username is not None:
            request['User-Name'] = username

        for attr_name, value in request_attrs.items():
            if isinstance(value, six.string_types):
                request.AddAttribute(attr_name, value)
            else:
                for v in value:
                    request.AddAttribute(attr_name, v)

        if password is not None:
            request['User-Password'] = request.PwCrypt(password)

        if challenge_response is not None:
            if challenge_send_user_password:
                # Use-Password gets replaced with the pretend user's response to
                # the challenge and encrypted as per the RADIUS RFC spec
                request['User-Password'] = request.PwCrypt(challenge_response)
            challenge_id = server._create_challenge_id()
            request['State'] = challenge_id

            # all comparisons with "original request" pass! amazing!
            orig_request = copy.deepcopy(request)
            orig_request.source = (self.radius_ip, self.radius_client_port)
            orig_request.username = self.username

            challenge_state = copy.copy(challenge_state)
            challenge_state['primary_res'] = base.AuthResult(
                success=True,
                msg='moo',
                radius_attrs=self.pass_through_radius_attrs,
            )

            server.challenges = {
                challenge_id: _ChallengeState(
                    initial_request=orig_request,
                    challenge_id=challenge_id,
                    state=challenge_state,
                ),
            }

        res_packet = yield self._test_process_packet(request)
        self.assertEqual(res_packet.id,
                         radius_packet_id,
                         msg="Response ID {0} does not match request ID {1}".format(res_packet.id, radius_packet_id))
        self.assertEqual(res_packet.code,
                         expect_code,
                         msg="Response code {0} does not match expected code {1}".format(res_packet.code, expect_code))
        if expect_msg_contains is not None:
            self.assertIn(expect_msg_contains,
                          res_packet['Reply-Message'],
                          msg="Response packet does not contain expected message '{0}'".format(expect_msg_contains))

        # Verify Access-Reject attributes
        if res_packet.code == packet.AccessReject and reject_attrs is not None:
            for attr_name, attr_value in reject_attrs.iteritems():
                self.assertIn(
                    attr_name,
                    res_packet,
                    msg="Expected attribute '{0}' not found in response packet".format(attr_name)
                )
                self.assertEqual(
                    res_packet[attr_name],
                    attr_value,
                    msg="Expected attribute '{0}' value '{1}', but response has '{2}'".format(
                        attr_name,
                        attr_value,
                        res_packet[attr_name])
                )

        # Verify pass-through RADIUS attributes we sent were actually passed through
        if res_packet.code == packet.AccessAccept:
            for attr_name, attr_value in self.pass_through_radius_attrs.items():
                self.assertIn(
                    attr_name,
                    res_packet,
                    "Expected attribute '{0}' not found in response packet".format(attr_name)
                )
                self.assertEqual(
                    res_packet[attr_name],
                    attr_value,
                    "Expected attribute '{0}' value '{1}', but response has '{2}'".format(
                        attr_name,
                        attr_value,
                        res_packet[attr_name])
                )

        server.cleanup_all()

    @defer.inlineCallbacks
    def _test_process_packet(self, auth_packet):
        """Helper function to _test_auth().

        Sends auth_packet to self.f_instance, which forwards the packet to
        self.instance's datagramReceived(). self.instance will then handle the
        request accordingly and send back a response packet.

        Mox will verify that all functions leading up to authentication occur
        and in their proper order.

        Since it's used in _test_auth(), don't use this on its own unless
        you need to make a custom authentication request packet, such as in the
        case of EAP.

        Yields:
            A response AuthPacket.
        """
        server = self.f_instance.protocol
        req_data = auth_packet.RequestPacket()
        client_addr = (self.radius_ip,
                       self.radius_client_port)

        self.mox.ReplayAll()
        yield server.datagramReceived(req_data, client_addr)
        self.mox.VerifyAll()
        self.mox.ResetAll()

        self.assertEqual(
            len(server.transport.queue),
            1,
            "Transport queue length mismatch, expected 1 but found {0}".format(len(server.transport.queue))
        )
        res_data, res_dest = server.transport.queue.pop(0)
        self.assertEqual(res_dest,
                         client_addr,
                         "Response destination IP mismatch: expected {0} but got {1}".format(res_dest, client_addr))
        response_packet = packet.AuthPacket(packet=res_data,
                                            dict=radius_dictionary())
        defer.returnValue(response_packet)

    def test_parse_radius_secrets(self):
        server = self.instance.protocol
        self.assertEqual(server.secrets, {
            netaddr.IPNetwork(self.radius_ip).cidr: self.secret,
            netaddr.IPNetwork(self.radius_ip2).cidr: self.secret2,
        })


class ServerModuleTestPassThrough(ServerModuleTest):
    """ServerModuleTest with pass-through values.

    Variant on ServerModuleTest where self.pass_through_radius_attrs is sent with
    every packet the server instance sends and should still be present when you
    get back a response packet.
    """
    def setUp(self, *args, **kwargs):
        super(ServerModuleTestPassThrough, self).setUp(*args, **kwargs)
        self.pass_through_radius_attrs = {
            'Filter-Id': ['61'],
        }
        self.client_instance.pass_through_radius_attrs = {
            'Filter-Id': '61',
        }


class ServerModuleTestPassThroughMultipleValues(ServerModuleTest):
    """ServerModuleTestPassThrough but with more than one pass-through value.
    """
    def setUp(self, *args, **kwargs):
        super(ServerModuleTestPassThroughMultipleValues,
              self).setUp(*args, **kwargs)
        self.pass_through_radius_attrs = {
            'Filter-Id': ['61', 'dogcow'],
        }
        self.client_instance.pass_through_radius_attrs = {
            'Filter-Id': ['61', 'dogcow'],
        }


class ClientProtectedConfigParsingTestCase(TestCase):
    """Class for testing client modules with protected configs.

    Client module test classes can inherit from this class to test that
    sensitive config items are accessed with get_protected_str() rather
    than get_str() (and that non-sensitive keys without protected prefixes
    don't require get_protected_str).
    """
    def _make_instance(self, more_config_items, client_module):
        """Creates a client instance that uses a protected config.

        Use on a per-test basis rather than using this once in eg. a setUp()
        call to make sure no extra keys from more_config_items are accidentally
        reused across tests.

        Args:
            more_config_items: Dict containing additional config attributes to
                add to the client config. Config keys containing a protected
                prefix as well as their values should be passed here.

        Returns:
            A client Module instance of type client_module with a config
            containing more_config_items
        """
        config_items = {
            'host_1': '1.2.3.4',
            'service_account_username': 'adminer',
            'search_dn': 'CN=Users,DC=bongos,DC=test',
            # MoreADLikeLDAPServer isn't AD-like enough for <ROOT>.
            'bind_dn': 'CN=adminer adminer,CN=Users,DC=bongos,DC=test',
            # Nor NTLM.
            'auth_type': 'plain',
        }
        config_items.update(more_config_items)
        config = ProtectedValidatingConfigDict(config_items)
        return client_module.Module(config)


class ServerProtectedConfigParsingTestCase(TestCase):
    """Class for testing server modules with protected configs.

    Server module test classes can inherit from this class to test that sensitive
    client and server config items are accessed with get_protected_str() rather
    than get_str() (and that non-sensitive keys without protected prefixes
    don't require get_protected_str).
    """
    def _make_instance(self, more_config_items, server_module, server_section_name):
        """Creates a server instance with a protected client and server config.

        Other test classes inheriting from this class will need to ensure that
        any test cases they run use get_protected_str() when interacting with its
        own server or client config.

        Args:
            more_config_items: Dict containing key-value pairs to be added to the
                protected server config. Pass in your protected prefix-including
                attributes here.

        Returns:
            A server Module instance of type server_section_name with a protected
            client and server config.
        """
        client_config = {
            'host_1': '1.2.3.4',
            'port_1': '1812',
            'secret': 'secret!',
        }
        config = ProtectedValidatingConfigDict(client_config)
        client = duoauthproxy.modules.radius_client.Module(config)

        server_config = {
            'radius_ip_1': '127.0.0.1',
            'ikey': 'nope',
        }
        server_config.update(more_config_items)
        return server_module.Module(
            ProtectedValidatingConfigDict(server_config),
            client,
            server_section_name,
        )


class AuthLoggingTest(object):
    """
    Test case mixin that allows for testing SIEM logging done by calls to
    log.auth_standard().

    log.auth() is mocked out, with _mock_auth() simply appending dict-ified SIEM
    logs to self.auth_logged_events as they happen.

    The logged events can then be tested with the assert_logged() method, by
    providing the expected result.
    """
    def setUp(self):
        super(AuthLoggingTest, self).setUp()
        self.patch(log._instance, 'auth', self._mock_auth)
        self.patch(log, 'auth', log._instance.auth)

        self.auth_logged_events = []

    def _mock_auth(self, **kwargs):
        self.auth_logged_events.append(kwargs)

    def assert_logged(self, expected):
        """
        After running a function or doing an in-test auth flow, compares the
        logged events in self.auth_logged_events to those of expected.
        Compares # of events, property types, and property values.

        Args:
            expected (list of dicts): list of expected log events in order of occurrence.
        """
        comparison_message = "Expected {0}; Actual {1}".format(expected, self.auth_logged_events)
        self.assertEqual(len(expected),
                         len(self.auth_logged_events),
                         "Number of events mismatch: {0}".format(comparison_message))
        for index, expected_event in enumerate(expected):
            actual_event = self.auth_logged_events[index]
            event_comparison_message = "Entry {0}; Expected {1}; Actual {2}".format(index, expected_event, actual_event)
            for expected_property in expected_event:
                self.assertIn(expected_property,
                              actual_event,
                              "Property {0} not found in actual result: {1}".format(expected_property,
                                                                                    event_comparison_message))
                self.assertEqual(expected_event[expected_property],
                                 actual_event[expected_property],
                                 "Results for property {0} do not match: {1}".format(expected_property,
                                                                                     event_comparison_message))


class MockConfigTestToolbox(ConfigTestToolbox):
    """
    Toolbox used to mock away any config tests that rely on external resources
    """

    def __init__(self, fail_test_file=False, fail_test_directory=False):
        self.fail_tests_file = fail_test_file
        self.fail_test_directory = fail_test_directory

    def test_file_readable(self, config, file_path_key):
        return not self.fail_tests_file

    def test_is_valid_directory(self, config, dir_path_key):
        return not self.fail_test_directory


class MockConnectivityTestToolbox(ConnectivityTestToolbox):
    def __init__(self,
                 fail_listen_udp=False,
                 fail_listen_tcp=False,
                 fail_validate_api_creds=False,
                 fail_ping=False,
                 fail_time_drift=False,
                 fail_connect_tcp=False,
                 fail_connect_radius=False,
                 fail_connect_with_http_proxy=False,
                 fail_ssl_certs=False,
                 fail_bind_service_account=False,
                 fail_search_has_results=False,
                 fail_valid_filter=False,
                 fail_listen_ssl=False,
                 fail_test_ssl_creds=False):
        self.fail_listen_udp = fail_listen_udp
        self.fail_listen_tcp = fail_listen_tcp
        self.fail_validate_api_creds = fail_validate_api_creds
        self.fail_ping = fail_ping
        self.fail_time_drift = fail_time_drift
        self.fail_connect_tcp = fail_connect_tcp
        self.fail_connect_radius = fail_connect_radius
        self.fail_connect_with_http_proxy = fail_connect_with_http_proxy
        self.fail_ssl_certs = fail_ssl_certs
        self.fail_bind_service_account = fail_bind_service_account
        self.fail_search_has_results = fail_search_has_results
        self.fail_valid_filter = fail_valid_filter
        self.fail_listen_ssl = fail_listen_ssl
        self.fail_test_ssl_creds = fail_test_ssl_creds

    def test_listen_udp(self, *args, **kwargs):
        if self.fail_listen_udp:
            return ModuleConnectivityTest.BAD_LISTEN_UDP_RESULT
        else:
            return ModuleConnectivityTest.GOOD_LISTEN_UDP_RESULT

    def test_listen_tcp(self, *args, **kwargs):
        if self.fail_listen_tcp:
            return ModuleConnectivityTest.BAD_LISTEN_TCP_RESULT
        else:
            return ModuleConnectivityTest.GOOD_LISTEN_TCP_RESULT

    def test_ping_duo(self, *args, **kwargs):
        if self.fail_ping:
            return ModuleConnectivityTest.BAD_PING_RESULT
        else:
            return ModuleConnectivityTest.GOOD_PING_RESULT

    def test_time_drift(self, *args, **kwargs):
        if self.fail_time_drift:
            return ModuleConnectivityTest.BAD_TIME_DRIFT_RESULT
        else:
            return ModuleConnectivityTest.GOOD_TIME_DRIFT_RESULT

    def test_validate_api_credentials(self, *args, **kwargs):
        if self.fail_validate_api_creds:
            return ModuleConnectivityTest.BAD_VALIDATE_RESULT
        else:
            return ModuleConnectivityTest.GOOD_VALIDATE_RESULT

    def test_connect_tcp(self, *args, **kwargs):
        if self.fail_connect_tcp:
            return ModuleConnectivityTest.BAD_CONNECT_TCP_RESULT
        else:
            return ModuleConnectivityTest.GOOD_CONNECT_TCP_RESULT

    def test_connect_radius(self, *args, **kwargs):
        if self.fail_connect_radius:
            return ModuleConnectivityTest.BAD_CONNECT_RADIUS_RESULT
        else:
            return ModuleConnectivityTest.GOOD_CONNECT_RADIUS_RESULT

    def test_connect_with_http_proxy(self, *args, **kwargs):
        if self.fail_connect_with_http_proxy:
            return ModuleConnectivityTest.BAD_CONNECT_WITH_HTTP_PROXY_RESULT
        else:
            return ModuleConnectivityTest.GOOD_CONNECT_WITH_HTTP_PROXY_RESULT

    def test_ssl_certs(self, cert_file_path):
        if self.fail_ssl_certs:
            return ModuleConnectivityTest.BAD_CERTS_RESULT
        else:
            return ModuleConnectivityTest.GOOD_CERTS_RESULT

    def test_bind_service_account(self, *args, **kwargs):
        if self.fail_bind_service_account:
            return ModuleConnectivityTest.BAD_BIND_SERVICE_ACCOUNT
        else:
            return ModuleConnectivityTest.GOOD_BIND_SERVICE_ACCOUNT

    def test_ldap_search_has_results(self, *args, **kwargs):
        if self.fail_search_has_results:
            return ModuleConnectivityTest.BAD_SEARCH_HAS_RESULTS
        else:
            return ModuleConnectivityTest.GOOD_SEARCH_HAS_RESULTS

    def test_valid_filter(self, *args, **kwargs):
        if self.fail_valid_filter:
            return ModuleConnectivityTest.BAD_VALID_FILTER
        else:
            return ModuleConnectivityTest.GOOD_VALID_FILTER

    def test_ssl_credentials(self, *args, **kwargs):
        if self.fail_test_ssl_creds:
            return ModuleConnectivityTest.BAD_SSL_CREDS_RESULT
        else:
            return ModuleConnectivityTest.GOOD_SSL_CREDS_RESULT

    def test_listen_ssl(self, *args, **kwargs):
        if self.fail_listen_ssl:
            return ModuleConnectivityTest.BAD_LISTEN_SSL_RESULT
        else:
            return ModuleConnectivityTest.GOOD_LISTEN_SSL_RESULT


class MockTesterResolver(BaseTestResolver):
    MOCK_RESULT = connectivity_results.SKIPPED_SECTION_RESULT

    def find_tester(self, section_name):
        return self.mock_tester

    def mock_tester(self, *args, **kwargs):
        return self.MOCK_RESULT


class MockProtocol(_ADServiceClientProtocol):
    """A mock LDAPClientProtocol for faking binds and unbinds"""
    def __init__(self, bind_success):
        class DummyFactory(object):
            pass
        factory = DummyFactory()
        factory.service_account_username = ''
        factory.search_dn = ''
        self.factory = factory
        self.bind_success = bind_success
        self.bound = False

    def primary_bind(self):
        if self.bind_success:
            self.bound = True
            return defer.succeed("Great success")
        else:
            return defer.fail(ldaperrors.LDAPException("Failed bind"))

    def perform_unbind(self):
        self.bound = False
        return defer.succeed("Great success")

    def user_filter_object(self):
        return "mock_filter"


class ConnectivityMockException(Exception):
    def __eq__(self, other):
        return isinstance(other, ConnectivityMockException)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return "String of ConnectivityMockException"

    def __repr__(self):
        return self.__str__()


class ModuleConnectivityTest(TestCase):
    TEST_EXCEPTION = ConnectivityMockException()
    TEST_SSL_EXCEPTION = ConnectivityMockException()
    TEST_SSL_EXCEPTION.message = [('openssl', 'style', 'error')]

    GOOD_CONFIG_CHECK_RESULT = connectivity_results.ConfigCheckResult([])

    GOOD_TIME_DRIFT_RESULT = connectivity_results.TimeDriftResult(
        True,
        10
    )

    BAD_TIME_DRIFT_RESULT = connectivity_results.TimeDriftResult(
        False,
        200
    )

    GOOD_PING_RESULT = connectivity_results.DuoPingResult(
        True,
        'good_api_host'
    )

    BAD_PING_RESULT = connectivity_results.DuoPingResult(
        False,
        'bad_api_host',
        exception=TEST_EXCEPTION
    )

    GOOD_VALIDATE_RESULT = connectivity_results.ValidateApiCredentialsResult(
        True,
        'good_ikey',
        'good_host'
    )

    BAD_VALIDATE_RESULT = connectivity_results.ValidateApiCredentialsResult(
        False,
        'bad_ikey',
        'bad_host',
        exception=TEST_EXCEPTION
    )

    GOOD_LISTEN_UDP_RESULT = connectivity_results.ListenResult(
        True,
        1812,
        '0.0.0.0'
    )

    BAD_LISTEN_UDP_RESULT = connectivity_results.ListenResult(
        False,
        1812,
        '0.0.0.0',
        exception=TEST_EXCEPTION
    )

    GOOD_LISTEN_TCP_RESULT = connectivity_results.ListenResult(
        True,
        389,
        '0.0.0.0'
    )

    BAD_LISTEN_TCP_RESULT = connectivity_results.ListenResult(
        True,
        389,
        '0.0.0.0',
        exception=TEST_EXCEPTION
    )

    GOOD_CONNECT_TCP_RESULT = connectivity_results.ConnectResult(
        True,
        'good_host',
        389
    )

    BAD_CONNECT_TCP_RESULT = connectivity_results.ConnectResult(
        False,
        'bad_host',
        389,
        exception=TEST_EXCEPTION
    )

    GOOD_CONNECT_RADIUS_RESULT = connectivity_results.RadiusConnectResult(
        True,
        'good_host',
        1812
    )

    BAD_CONNECT_RADIUS_RESULT = connectivity_results.RadiusConnectResult(
        False,
        'bad_host',
        1812,
    )

    GOOD_CONNECT_WITH_HTTP_PROXY_RESULT = connectivity_results.HttpWebProxyResult(
        True,
        'good_host',
        80
    )

    BAD_CONNECT_WITH_HTTP_PROXY_RESULT = connectivity_results.HttpWebProxyResult(
        False,
        'bad_host',
        80,
        exception=TEST_EXCEPTION
    )

    BAD_BIND_SERVICE_ACCOUNT = connectivity_results.LdapBindResult(
        False,
        'bad_username',
        exception=TEST_EXCEPTION
    )

    GOOD_BIND_SERVICE_ACCOUNT = connectivity_results.LdapBindResult(
        True,
        'good_username'
    )

    ERROR_SEARCH_HAS_RESULTS = connectivity_results.LdapSearchResult(
        False,
        'error_dn',
        'error_filter',
        exception=TEST_EXCEPTION
    )

    BAD_SEARCH_HAS_RESULTS = connectivity_results.LdapSearchResult(
        False,
        'bad_dn',
        'bad_filter'
    )

    GOOD_SEARCH_HAS_RESULTS = connectivity_results.LdapSearchResult(
        True,
        'good_dn',
        'good_filter'
    )

    GOOD_LISTEN_SSL_RESULT = connectivity_results.ListenResult(
        True,
        636,
        '0.0.0.0'
    )

    BAD_LISTEN_SSL_RESULT = connectivity_results.ListenResult(
        False,
        636,
        '0.0.0.0'
    )

    GOOD_KEY_RESULT = connectivity_results.SslKeyFileResult(
        True,
        'good_key_path'
    )

    BAD_KEY_RESULT = connectivity_results.SslKeyFileResult(
        False,
        'bad_key_path',
        exception=TEST_EXCEPTION
    )

    GOOD_CERTS_RESULT = connectivity_results.SslCertFileResult(
        True,
        'good_cert_path'
    )

    BAD_CERTS_RESULT = connectivity_results.SslCertFileResult(
        False,
        'bad_cert_path',
        exception=TEST_EXCEPTION
    )

    GOOD_CONTEXT_RESULT = connectivity_results.SslContextResult(
        True,
        None,
        'tls1.2',
    )

    BAD_CONTEXT_RESULT = connectivity_results.SslContextResult(
        False,
        None,
        'ssl2',
        exception=TEST_SSL_EXCEPTION,
    )

    GOOD_SSL_CREDS_RESULT = connectivity_results.SslResult(
        GOOD_KEY_RESULT,
        GOOD_CERTS_RESULT,
        GOOD_CONTEXT_RESULT,
    )

    BAD_SSL_CREDS_RESULT = connectivity_results.SslResult(
        BAD_KEY_RESULT,
        BAD_CERTS_RESULT,
        BAD_CONTEXT_RESULT,
    )

    SUCCESS_TOOLBOX = MockConnectivityTestToolbox()
