#
# Copyright (c) 2013 Duo Security
# All Rights Reserved
#
import io
import os
import sys
import re
import textwrap
import mock

import duoauthproxy.proxy
from duoauthproxy.lib.config_provider import ConfigProvider
from duoauthproxy.lib.config_error import ConfigError
from duoauthproxy.lib import const

from base import TestCase

LINUX_LINE_ENDING = '\n'
WINDOWS_LINE_ENDING = '\r\n'


class testOutput:
    def __init__(self):
        self.fservs = {}
        self.configs = {}
        self.log_auth_events = False


outinst = testOutput()

# Mock classes/functions


def mockFunction(*args, **kwargs):
    pass


def mockParse(*args):
    return {}


class mockLogObserver(object):
    def __init__(self, *args, **kwargs):
        pass

    def __call__(self, *args, **kwargs):
        pass


class mockFileLogObserver(object):
    def __init__(self, outFile, formatEvent):
        outinst.log_auth_events = (outFile.name == 'authevents.log')

    def __call__(self, *args, **kwargs):
        pass


class mockLogFile(object):
    def __init__(self, name, directory, rotateLength, maxRotatedFiles, defaultMode):
        self.name = name


class mockApplication:
    def __init__(self, *args, **kwargs):
        pass

    def setComponent(self, *args):
        pass

    def addService(self, *args):
        pass


class mockModule(object):
    def __init__(self, *args, **kwargs):
        self.emit = None
    setServiceParent = mockFunction


class mockServerModule(mockModule):
    pass


class mockClientModule(mockModule):
    pass


class mockServer(mockServerModule):
    class Module(mockServerModule):
        pass


class mockServerConcat(mockServerModule):
    class Module(mockServerModule):
        def __init__(self, *args, **kwargs):
            outinst.configs['radius_server_concat'] = args[0]


class mockCloudModule(mockServerModule):
    class Module(mockServerModule):
        no_client = True

        def register_drpc_call_provider(self, _provider):
            pass


class mockldapServer(mockServerModule):
    class Module(mockServerModule):
        def __init__(self, config, client, server_section_name):
            global outinst
            outinst.fservs.setdefault('ldap', 0)
            outinst.fservs['ldap'] += 1
            outinst.configs['ldap_server_auto'] = config


class mockFServer(mockModule):
    class Module(mockModule):
        def __init__(self, port, servers, server_names, interface, **kwargs):
            global outinst
            outinst.fservs[(port, interface)] = server_names.values()


class mockClient(mockClientModule):
    class Module(mockClientModule):
        pass


class mockRadiusClient(mockClientModule):
    class Module(mockClientModule):
        def __init__(self, *args, **kwargs):
            outinst.configs['radius_client'] = args[0]


class mockFipsManager(mockModule):
    def run(self):
        pass

    def status(self):
        pass


def alwaystrue(x, y):
    return True


class TestConfigParser(TestCase):
    def setUp(self):
        config_f = io.BytesIO()
        self.config_f = config_f

        self.crt_bundle_f = io.BytesIO()
        self.crt_bundle_f.write('bla')
        self.crt_bundle_f.seek(0, os.SEEK_SET)

        def mockOpen(filename, readmode):
            if "ca-bundle.crt" in filename:
                return self.crt_bundle_f
            else:
                config_f.seek(0, os.SEEK_SET)
                return self.config_f

        # Override every global it is possible to define in the Python
        # language. Remember the originals and restore them before the
        # next TestCase instance runs.
        self.orig = []
        for namespace, ident, mocker in (
            (duoauthproxy.proxy.base, 'ServerModule', mockServerModule),
            (duoauthproxy.proxy.base, 'ClientModule', mockClientModule),
            (duoauthproxy.proxy.modules, 'cloud', mockCloudModule),
            (duoauthproxy.proxy.modules, 'http_proxy', mockCloudModule),
            (duoauthproxy.proxy.modules, 'radius_client', mockRadiusClient),
            (duoauthproxy.proxy.modules, 'ad_client', mockClient),
            (duoauthproxy.proxy.modules, 'radius_server_concat', mockServerConcat),
            (duoauthproxy.proxy.modules, 'radius_server_auto', mockServer),
            (duoauthproxy.proxy.modules, 'radius_server_duo_only', mockServer),
            (duoauthproxy.proxy.modules, 'radius_server_challenge', mockServer),
            (duoauthproxy.proxy.modules, 'radius_server_iframe', mockServer),
            (duoauthproxy.proxy.modules, 'ldap_server_auto', mockldapServer),
            (duoauthproxy.proxy, 'FileLogObserver', mockFileLogObserver),
            (duoauthproxy.proxy, 'textFileLogObserver', mockLogObserver),
            (duoauthproxy.proxy, 'FilteringLogObserver', mockLogObserver),
            (duoauthproxy.proxy, 'LegacyLogObserverWrapper', mockLogObserver),
            (duoauthproxy.proxy, 'isinstance', alwaystrue),
            (duoauthproxy.proxy, 'parse_radius_secrets', mockParse),
            (duoauthproxy.proxy, 'forward_serv', mockFServer),
            (duoauthproxy.proxy, 'open', mockOpen),
            (duoauthproxy.proxy, 'Application', mockApplication),
            (duoauthproxy.proxy, 'LogFile', mockLogFile),
            (duoauthproxy.lib.config_provider, 'open', mockOpen),
        ):
            orig = getattr(namespace, ident, None)
            self.orig.append((namespace, ident, orig))
            setattr(namespace, ident, mocker)

    def tearDown(self):
        # Can't reload() modules here; everything that imports a copy
        # of AuthResult will be unable to construct it because there
        # will be two different AuthResult classes. :(
        for namespace, ident, orig in self.orig:
            if orig is None:
                delattr(namespace, ident)
            else:
                setattr(namespace, ident, orig)

    def assert_good_config_parse(self, config, expected_fservs, expected_configs=None,
                                 create_app_args=None):
        """Checks if config passed in is parseable by creating an AuthProxy instance.

        Args:
            config (dict): Example AP configuration.
            expected_fservs (dict): Expected instances of DuoForwardServ mapping
                (port, ip) to a list of server section headers in config associated
                with that interface. If ldap_server_auto sections are present in
                config, this will contain { 'ldap': # }, where # is the number of
                ldap_server_auto sections.

                Example:
                {
                    (12345, '127.0.0.1'): [
                        'radius_server_concat',
                        'radius_server_concat2',
                    ],
                }
            expected_configs (dict): Optional. Every client or server section
                name in config (no main) mapped to its own section configuration.
            create_app_args (bool): if True, parse arguments in sys.argv during
                call to create_application()

        Note:
            If more than one server is listed for a given interface in config or
            expected_fservs, a CannotListenError will not be raised during the
            call to create_application() since all server Module constructors
            have been mocked out.
        """
        self.config_f.seek(0, os.SEEK_SET)
        self.config_f.write(config)
        self.config_f.flush()

        outinst.fservs = {}
        outinst.configs = {}
        duoauthproxy.proxy.create_application(args=create_app_args)

        for y in expected_fservs.keys():
            self.assertIn(y, outinst.fservs)
            if y == 'ldap':
                self.assertEqual(expected_fservs[y], outinst.fservs[y])
            else:
                self.assertEqual(len(expected_fservs[y]),
                                 len(outinst.fservs[y]))
                for z in expected_fservs[y]:
                    self.assertIn(z, outinst.fservs[y])

        if expected_configs is not None:
            for y in expected_configs.keys():
                self.assertIn(y, outinst.configs)
                self.assertEqual(expected_configs[y], outinst.configs[y])

    def assert_bad_config_parse(self, config):
        self.config_f.seek(0, os.SEEK_SET)
        self.config_f.write(config)
        self.config_f.flush()

        outinst.fservs = {}
        outinst.configs = {}
        with self.assertRaises(ConfigError):
            duoauthproxy.proxy.create_application()

    def assert_auth_log_flag_set(self, config, expected_value):
        self.config_f.seek(0, os.SEEK_SET)
        self.config_f.write(config)
        self.config_f.flush()

        outinst.log_auth_events = False
        duoauthproxy.proxy.create_application()
        self.assertEqual(outinst.log_auth_events, expected_value)

    def test_main_basic(self):
        config = textwrap.dedent("""
            [main]
            client=ad_client
            server=radius_server_concat

            [ad_client]
            host=srgs
            service_account_username=gsrg
            service_account_password=srgs
            search_dn=gsr

            [radius_server_concat]
            ikey=oreghu
            skey=srge
            api_host=srgs
            radius_ip_1=grs
            radius_secret_1=sgrs
            failmode=safe
        """)
        expected_fservs = {
            (const.DEFAULT_RADIUS_PORT, ''): ['radius_server_concat'],
        }
        self.assert_good_config_parse(config, expected_fservs)

    def test_main_override(self):
        config = textwrap.dedent("""
            [main]
            client=ad_client
            server=radius_server_concat

            [radius_client]
            host=1.2.3.4
            secret=thisisaradiussecret

            [ad_client]
            host=srsegs
            service_account_username=gsrg
            service_account_password=srgs
            search_dn=gsr

            [radius_server_concat]
            ikey=oreghu
            skey=srge
            api_host=srgs
            radius_ip_1=grs
            radius_secret_1=sgrs
            failmode=safe
            client=radius_client
        """)
        expected_fservs = {
            (const.DEFAULT_RADIUS_PORT, ''): ['radius_server_concat'],
        }
        self.assert_good_config_parse(config, expected_fservs)

    def test_no_main_or_client(self):
        config = textwrap.dedent("""
            [radius_client]
            host=1.2.3.4
            secret=thisisaradiussecret

            [radius_server_concat]
            ikey=oreghu
            skey=srge
            api_host=srgs
            radius_ip_1=grs
            radius_secret_1=sgrs
            failmode=safe
        """)
        expected_fservs = {
            (const.DEFAULT_RADIUS_PORT, ''): ['radius_server_concat'],
        }
        self.assert_good_config_parse(config, expected_fservs)

    def test_invalid_client(self):
        config = textwrap.dedent("""
            [herpderp_client]
            host=1.2.3.4
            secret=thisisaradiussecret

            [radius_server_concat]
            ikey=oreghu
            skey=srge
            api_host=srgs
            radius_ip_1=grs
            radius_secret_1=sgrs
            failmode=safe
            client=herpderp_client
        """)
        self.assert_bad_config_parse(config)

    def test_forward_serv(self):
        config = textwrap.dedent("""
            [radius_client]
            host=1.2.3.4
            secret=thisisaradiussecret

            [forward_serv]
            ikey=oreghu
            skey=srge
            api_host=srgs
            radius_ip_1=grs
            radius_secret_1=sgrs
            failmode=safe
        """)
        self.assert_bad_config_parse(config)

    def test_invalid_server(self):
        config = textwrap.dedent("""
            [radius_client]
            host=srgs
            service_account_username=gsrg
            service_account_password=srgs
            search_dn=gsr

            [foobar_server]
            ikey=oreghu
            skey=srge
            api_host=srgs
            radius_ip_1=grs
            radius_secret_1=sgrs
            failmode=safe
        """)
        self.assert_bad_config_parse(config)

    def test_two_servers_one_client(self):
        config = textwrap.dedent("""
            [radius_client]
            host=srgs
            service_account_username=gsrg
            service_account_password=srgs
            search_dn=gsr

            [radius_server_concat]
            ikey=oreghu
            skey=srge
            api_host=srgs
            radius_ip_1=grs
            radius_secret_1=sgrs
            failmode=safe

            [radius_server_concat2]
            ikey=oreghu
            skey=srge
            api_host=srgs
            radius_ip_1=grs
            radius_secret_1=sgrs
            failmode=secure
        """)
        expected_fservs = {
            (const.DEFAULT_RADIUS_PORT, ''): [
                'radius_server_concat',
                'radius_server_concat2',
            ],
        }
        self.assert_good_config_parse(config, expected_fservs)

    def test_two_servers_two_clients(self):
        config = textwrap.dedent("""
            [radius_client]
            host=1.2.3.4
            secret=thisisaradiussecret

            [ad_client]
            host=srgs
            service_account_username=gsrg
            service_account_password=srgs
            search_dn=gsr

            [radius_server_concat]
            ikey=oreghu
            skey=srge
            api_host=srgs
            radius_ip_1=grs
            radius_secret_1=sgrs
            failmode=safe
            client=radius_client

            [radius_server_concat2]
            ikey=oreghu
            skey=srge
            api_host=srgs
            radius_ip_1=grs
            radius_secret_1=sgrs
            failmode=secure
            client=ad_client
        """)
        expected_fservs = {
            (const.DEFAULT_RADIUS_PORT, ''): [
                'radius_server_concat',
                'radius_server_concat2',
            ],
        }
        self.assert_good_config_parse(config, expected_fservs)

    def test_lots_of_servers(self):
        config = textwrap.dedent("""
            [radius_client]
            host=1.2.3.4
            secret=thisisaradiussecret

            [radius_server_concat]
            ikey=oreghu
            skey=srge
            api_host=srgs
            radius_ip_1=grs
            radius_secret_1=sgrs
            failmode=safe

            [radius_server_concat2]
            ikey=oreghu
            skey=srge
            api_host=srgs
            radius_ip_1=grs
            radius_secret_1=sgrs
            failmode=secure

            [radius_server_concat3]
            ikey=oreghu
            skey=srge
            api_host=srgs
            radius_ip_1=grs
            radius_secret_1=sgrs
            failmode=secure
            port=12345

            [radius_server_concat4]
            ikey=oreghu
            skey=srge
            api_host=srgs
            radius_ip_1=grs
            radius_secret_1=sgrs
            failmode=safe
            port=12345

            [radius_server_concat43597438]
            ikey=oreghu
            skey=srge
            api_host=srgs
            radius_ip_1=grs
            radius_secret_1=sgrs
            failmode=secure
            port=54321
        """)
        expected_fservs = {
            (12345, ''): [
                'radius_server_concat3',
                'radius_server_concat4',
            ],
            (54321, ''): [
                'radius_server_concat43597438',
            ],
            (const.DEFAULT_RADIUS_PORT, ''): [
                'radius_server_concat',
                'radius_server_concat2',
            ],
        }
        self.assert_good_config_parse(config, expected_fservs)

    def test_ldap_ssl_and_not_ssl(self):
        config = textwrap.dedent("""
            [ad_client]
            host=srgs
            service_account_username=gsrg
            service_account_password=srgs
            search_dn=gsr

            [ldap_server_auto]
            ikey=oreghu
            skey=srge
            api_host=srgs
            radius_ip_1=grs
            radius_secret_1=sgrs
            failmode=safe

            [ldap_server_auto2]
            ikey=oreghu
            skey=srge
            api_host=srgs
            radius_ip_1=grs
            radius_secret_1=sgrs
            failmode=secure
            ssl_key_path=sslstuff.bla
            ssl_cert_path=sslpath/pathypath/file.file
        """)
        expected_fservs = {'ldap': 2}
        self.assert_good_config_parse(config, expected_fservs)

    def test_two_servers_different_interfaces(self):
        config = textwrap.dedent("""
            [radius_client]
            host=1.2.3.4
            secret=thisisaradiussecret

            [radius_server_concat]
            ikey=oreghu
            skey=srge
            api_host=srgs
            radius_ip_1=grs
            radius_secret_1=sgrs
            failmode=safe
            interface=eth0

            [radius_server_concat2]
            ikey=oreghu
            skey=srge
            api_host=srgs
            radius_ip_1=grs
            radius_secret_1=sgrs
            failmode=secure
            interface=eth1
        """)
        expected_fservs = {
            (const.DEFAULT_RADIUS_PORT, 'eth0'): ['radius_server_concat'],
            (const.DEFAULT_RADIUS_PORT, 'eth1'): ['radius_server_concat2'],
        }
        self.assert_good_config_parse(config, expected_fservs)

    def test_one_cloud_module_no_client_or_server(self):
        config = textwrap.dedent("""
            [cloud]
            ikey=foo
            skey=bar
            host=baz
        """)
        self.assert_good_config_parse(config, {})

    def test_two_cloud_modules(self):
        config = textwrap.dedent("""
            [cloud]
            ikey=foo
            skey=bar
            host=baz

            [cloud2]
            ikey=totally
            skey=legit
            host=config
        """)
        self.assert_bad_config_parse(config)

    def test_only_irrelevant_modules(self):
        config = textwrap.dedent("""
            [whatever]
        """)
        self.assert_bad_config_parse(config)

    def test_no_modules(self):
        config = textwrap.dedent("""
        """)
        self.assert_bad_config_parse(config)

    def test_debug(self):
        config = textwrap.dedent("""
            [main]
            debug = 1
            [radius_client]
            [radius_server_concat]
        """)
        expected_fservs = {
            (const.DEFAULT_RADIUS_PORT, ''): ['radius_server_concat'],
        }
        expected_configs = {
            'radius_client': {
                'debug': 'True',
            },
            'radius_server_concat': {
                'debug': 'True',
            }
        }
        self.assert_good_config_parse(config, expected_fservs, expected_configs)

    def test_logging_insecure(self):
        config = textwrap.dedent("""
            [main]
            [radius_client]
            [radius_server_concat]
        """)
        expected_fservs = {
            (const.DEFAULT_RADIUS_PORT, ''): ['radius_server_concat'],
        }
        expected_configs = {
            'radius_client': {
                'debug': 'True',
                'is_logging_insecure': 'True',
            },
            'radius_server_concat': {
                'debug': 'True',
                'is_logging_insecure': 'True',
            },
        }
        orig_args = sys.argv
        sys.argv = ["duoauthproxy", "--logging-insecure"]
        try:
            self.assert_good_config_parse(
                config,
                expected_fservs,
                expected_configs,
                create_app_args=True
            )
        finally:
            sys.argv = orig_args

    def test_line_endings_mixed(self):
        linux_line_endings_config = LINUX_LINE_ENDING.join([
            "[radius_client]",
            "host=1.2.3.4",
            "secret=thisisaradiussecret",
            "service_account_username=gsrg",
            "service_account_password=srgs",
            "search_dn=gsr",
        ])

        windows_line_endings_config = WINDOWS_LINE_ENDING.join([
            "[radius_server_concat]",
            "ikey=oreghu",
            "skey=srge",
            "failmode=safe",
            "port=32312",
            "interface=interface",
        ])

        config = LINUX_LINE_ENDING.join([
            linux_line_endings_config,
            windows_line_endings_config
        ])

        expected_fservs = {(32312, 'interface'): ['radius_server_concat']}
        expected_configs = {
            'radius_client': {
                'host': '1.2.3.4',
                'secret': 'thisisaradiussecret',
                'service_account_username': 'gsrg',
                'service_account_password': 'srgs',
                'search_dn': 'gsr',
            },
            'radius_server_concat': {
                'ikey': 'oreghu',
                'skey': 'srge',
                'failmode': 'safe',
                'port': '32312',
                'interface': 'interface',
            },
        }
        self.assert_good_config_parse(config, expected_fservs, expected_configs)

    def test_line_endings_windows_only(self):
        config = WINDOWS_LINE_ENDING.join([
            "[radius_client]",
            "host=1.2.3.4",
            "secret=thisisaradiussecret",
            "service_account_username=gsrg",
            "service_account_password=srgs",
            "search_dn=gsr",
            "",
            "[radius_server_concat]",
            "ikey=oreghu",
            "skey=srge",
            "failmode=safe",
            "port=32312",
            "interface=interface",
        ])
        expected_fservs = {(32312, 'interface'): ['radius_server_concat']}
        expected_configs = {
            'radius_client': {
                'host': '1.2.3.4',
                'secret': 'thisisaradiussecret',
                'service_account_username': 'gsrg',
                'service_account_password': 'srgs',
                'search_dn': 'gsr',
            },
            'radius_server_concat': {
                'ikey': 'oreghu',
                'skey': 'srge',
                'failmode': 'safe',
                'port': '32312',
                'interface': 'interface',
            },
        }
        # re.sub should not be executed for this config since it does not have mixed line endings.
        # Raise an exception if re.sub is executed.

        def raiser(*args, **kwargs):
            raise Exception("uh oh!")
        self.patch(re, "sub", raiser)
        self.assert_good_config_parse(config, expected_fservs, expected_configs)

    def test_line_endings_unix_only(self):
        config = textwrap.dedent("""
            [radius_client]
            host=1.2.3.4
            secret=thisisaradiussecret
            service_account_username=gsrg
            service_account_password=srgs
            search_dn=gsr

            [radius_server_concat]
            ikey=oreghu
            skey=srge
            failmode=safe
            port=32312
            interface=interface
        """)
        expected_fservs = {(32312, 'interface'): ['radius_server_concat']}
        expected_configs = {
            'radius_client': {
                'host': '1.2.3.4',
                'secret': 'thisisaradiussecret',
                'service_account_username': 'gsrg',
                'service_account_password': 'srgs',
                'search_dn': 'gsr',
            },
            'radius_server_concat': {
                'ikey': 'oreghu',
                'skey': 'srge',
                'failmode': 'safe',
                'port': '32312',
                'interface': 'interface',
            },
        }

        def raiser(*args, **kwargs):
            raise Exception("uh oh!")
        self.patch(re, "sub", raiser)
        self.assert_good_config_parse(config, expected_fservs, expected_configs)

    def test_same_section_names_unix_line_endings(self):
        config = textwrap.dedent("""
            [ad_client]
            host=srgs2
            service_account_username=gsrg2
            service_account_password=srgs2
            search_dn=srgs

            [ad_client]
            host=srgs3
            service_account_username=gsrg3
            service_account_password=srgs3
            search_dn=srgs

            [ad_client]
            host=srgs3
            service_account_username=gsrg3
            service_account_password=srgs3
            search_dn=srgs

            [ldap_server_auto]
            ikey=oreghu
            skey=srge
            api_host=srgs
            radius_ip_1=grs
            radius_secret_1=sgrs
            failmode=secure
            ssl_key_path=sslstuff.bla
            ssl_cert_path=sslpath/pathypath/file.file

            [ldap_server_auto]
            ikey=oreghu
            skey=srge
            api_host=srgs
            radius_ip_1=grs
            radius_secret_1=sgrs
            failmode=safe
            client=ad_client
        """)
        self.assert_bad_config_parse(config)

    def test_same_section_names_windows_line_endings(self):
        config = WINDOWS_LINE_ENDING.join([
            "[radius_client]",
            "host=1.2.3.4",
            "secret=thisisaradiussecret",
            "service_account_username=gsrg",
            "service_account_password=srgs",
            "search_dn=gsr",
            "",
            "[radius_server_concat]",
            "ikey=oreghu",
            "skey=srge",
            "failmode=safe",
            "port=32312",
            "interface=interface",
            "",
            "[radius_server_concat]",
            "ikey=oreghu",
            "skey=srge",
            "failmode=safe",
            "port=32312",
            "interface=interface",
        ])
        self.assert_bad_config_parse(config)

    def test_special_characters_in_entries(self):
        config = textwrap.dedent("""
            [radius_client]
            host=1.2.3.4
            secret=specialchars%(c#)$)!(

            [radius_server_concat]
            ikey=oreghu
            skey=srge
            api_host=srgs
            radius_ip_1=grs
            radius_secret_1=s%(g&#(k!rs
            failmode=safe
        """)
        expected_fservs = {
            (const.DEFAULT_RADIUS_PORT, ''): ['radius_server_concat'],
        }
        self.assert_good_config_parse(config, expected_fservs)

    def test_lowercase_sections_uppercase_keys_success(self):
        config = textwrap.dedent("""
            [ad_client]
            HOST=1.2.3.4
            SERVICE_ACCOUNT_USERNAME=duoservice
            SERVICE_ACCOUNT_PASSWORD=password1
            SEARCH_DN=DC=example,DC=com
            SECURITY_GROUP_DN=CN=DuoVPNUsers,OU=Groups,DC=example,DC=com

            [radius_server_iframe]
            IKEY=DIXXXXXXXXXXXXXXXXXX
            SKEY=XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
            API_HOST=api-XXXXXXXX.duosecurity.com
            TYPE=citrix_netscaler
            FAILMODE=safe
            RADIUS_IP_1=5.6.7.8
            RADIUS_SECRET_1=thisisalsoaradiussecret
            CLIENT=ad_client
        """)
        expected_fservs = {
            (const.DEFAULT_RADIUS_PORT, ''): ['radius_server_iframe'],
        }
        self.assert_good_config_parse(config, expected_fservs)

    def test_uppercase_sections_lowercase_keys_success(self):
        config = textwrap.dedent("""
            [AD_CLIENT]
            host=1.2.3.4
            service_account_username=duoservice
            service_account_password=password1
            search_dn=DC=example,DC=com
            security_group_dn=CN=DuoVPNUsers,OU=Groups,DC=example,DC=com

            [RADIUS_SERVER_IFRAME]
            ikey=DIXXXXXXXXXXXXXXXXXX
            skey=XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
            api_host=api-XXXXXXXX.duosecurity.com
            type=citrix_netscaler
            failmode=SAFE
            radius_ip_1=5.6.7.8
            radius_secret_1=thisisalsoaradiussecret
            client=ad_client
        """)
        expected_fservs = {
            (const.DEFAULT_RADIUS_PORT, ''): ['radius_server_iframe'],
        }
        self.assert_good_config_parse(config, expected_fservs)

    def test_all_upper_main_success(self):
        config = textwrap.dedent("""
            [MAIN]
            debug = 1
            [radius_client]
            [radius_server_concat]
        """)
        expected_fservs = {
            (const.DEFAULT_RADIUS_PORT, ''): ['radius_server_concat'],
        }
        expected_configs = {
            'radius_client': {
                'debug': 'True',
            },
            'radius_server_concat': {
                'debug': 'True',
            },
        }
        self.assert_good_config_parse(config, expected_fservs, expected_configs)

    def test_mixed_case_main_success(self):
        config = textwrap.dedent("""
            [MaIn]
            debug = 1
            [radius_client]
            [radius_server_concat]
        """)
        expected_fservs = {
            (const.DEFAULT_RADIUS_PORT, ''): ['radius_server_concat'],
        }
        expected_configs = {
            'radius_client': {
                'debug': 'True',
            },
            'radius_server_concat': {
                'debug': 'True',
            },
        }
        self.assert_good_config_parse(config, expected_fservs, expected_configs)

    @mock.patch.object(ConfigProvider, 'from_ini_file')
    def test_config_provider_error(self, mock_from_ini_file):
        mock_from_ini_file.side_effect = ConfigError('config error')

        self.assert_bad_config_parse('config')

    def test_link_casing_success(self):
        config = textwrap.dedent("""
            [AD_CLIENT]
            host=1.2.3.4
            service_account_username=duoservice
            service_account_password=password1
            search_dn=DC=example,DC=com
            security_group_dn=CN=DuoVPNUsers,OU=Groups,DC=example,DC=com

            [RADIUS_SERVER_IFRAME]
            ikey=DIXXXXXXXXXXXXXXXXXX
            skey=XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
            api_host=api-XXXXXXXX.duosecurity.com
            type=citrix_netscaler
            failmode=SAFE
            radius_ip_1=5.6.7.8
            radius_secret_1=thisisalsoaradiussecret
            client=AD_CLIENT
        """)
        expected_fservs = {
            (const.DEFAULT_RADIUS_PORT, ''): ['radius_server_iframe'],
        }
        self.assert_good_config_parse(config, expected_fservs)

    def test_authlogs_enabled(self):
        config = textwrap.dedent("""
            [main]
            log_auth_events = true
            [radius_client]
            [radius_server_concat]
        """)

        self.assert_auth_log_flag_set(config, True)

    def test_authlogs_disabled(self):
        config = textwrap.dedent("""
            [main]
            log_auth_events = false
            [radius_client]
            [radius_server_concat]
        """)

        self.assert_auth_log_flag_set(config, False)

    def test_authlogs_default(self):
        config = textwrap.dedent("""
            [main]
            [radius_client]
            [radius_server_concat]
        """)

        self.assert_auth_log_flag_set(config, False)

    def test_ldap_server_auto_main_interface_inherited(self):
        config = textwrap.dedent("""
            [main]
            interface = interface
            [ad_client]
            [ldap_server_auto]
        """)

        expected_fservs = {'ldap': 1}
        expected_configs = {
            'ldap_server_auto': {
                'interface': 'interface',
            },
        }

        self.assert_good_config_parse(config, expected_fservs, expected_configs)

    def test_ldap_server_auto_partial_interface_inheritance(self):
        config = textwrap.dedent("""
            [main]
            interface = interface1
            [ad_client]
            [ldap_server_auto]
            interface = interface2
        """)

        expected_fservs = {'ldap': 1}
        expected_configs = {
            'ldap_server_auto': {
                'interface': 'interface2',
            },
        }
        self.assert_good_config_parse(config, expected_fservs, expected_configs)

    def test_REM_as_comment_parse(self):
        config = textwrap.dedent("""
            REM this is a comment
            [cloud]
            ikey=foo
            skey=bar
            host=baz
        """)
        self.assert_good_config_parse(config, {})

    def test_bad_REM_comment_parse(self):
        config = textwrap.dedent("""
            REMthis this is a comment
            [cloud]
            ikey=foo
            skey=bar
            host=baz
        """)
        self.assert_bad_config_parse(config)

    def test_weird_REM_comment_parse(self):
        config = textwrap.dedent("""
            ReM this this is a comment
            [cloud]
            ikey=foo
            skey=bar
            host=baz
        """)
        self.assert_good_config_parse(config, {})


if __name__ == '__main__':
    import unittest
    unittest.main()
