import textwrap
import unittest
import re
import StringIO
import mock

from duoauthproxy.lib import config_provider
from duoauthproxy.lib.config_error import ConfigError
from duoauthproxy.lib.config_provider import ConfigDict
from base import TestCase


class TestConfigProvider(TestCase):
    GOOD_CONFIG = textwrap.dedent("""
        [main]
        key1=value1

        [cloud]
        key1=value1
        key2=value2

        [http_proxy]
        key1=value1
        key2=value2

        [ad_client]
        key1=value1
        key2=value2

        [ldap_server_auto]
        key1=value1
        key2=value2

        [radius_client]
        key1=value1
        key2=value2

        [radius_server_auto1]
        key1=value1
        key2=value2

        [radius_server_auto2]
        key1=value1
        key2=value2

        [radius_server_auto3]
        key1=value1
        key2=value2
        key3=value3
    """)

    UNIX_LINE_ENDING = '\n'
    UNIX_LINE_ENDING_REGEX = '(?<!\r)\n'
    WIN_LINE_ENDING = '\r\n'
    WIN_LINE_ENDING_REGEX = '\r\n'

    def assert_only_unix_line_endings(self, test_string):
        self.assertIsNone(re.search(self.WIN_LINE_ENDING_REGEX, test_string))
        self.assertIsNotNone(re.search(self.UNIX_LINE_ENDING_REGEX, test_string))

    def assert_only_win_line_endings(self, test_string):
        self.assertIsNone(re.search(self.UNIX_LINE_ENDING_REGEX, test_string))
        self.assertIsNotNone(re.search(self.WIN_LINE_ENDING_REGEX, test_string))

    def test_good_config_section_list_success(self):
        expected_result = ['main',
                           'cloud',
                           'http_proxy',
                           'ad_client',
                           'ldap_server_auto',
                           'radius_client',
                           'radius_server_auto1',
                           'radius_server_auto2',
                           'radius_server_auto3',
                           ]

        config_string = self.GOOD_CONFIG
        cp = config_provider.ConfigProvider(config_string)
        actual_result = cp.list_sections()

        self.assertItemsEqual(expected_result, actual_result)

    def test_good_config_one_section_success(self):
        expected_result = ConfigDict({'key1': 'value1', 'key2': 'value2'})

        config_string = self.GOOD_CONFIG
        cp = config_provider.ConfigProvider(config_string)
        actual_result = cp.get_section_config('ldap_server_auto')

        self.assertItemsEqual(expected_result, actual_result)

    def test_find_duplicate_sections_when_no_duplicates_exist(self):
        expected_result = []

        config_string = self.GOOD_CONFIG
        actual_result = config_provider.find_duplicate_sections(config_string)

        self.assertItemsEqual(expected_result, actual_result)

    def test_find_duplicate_sections_when_duplicates_exist(self):
        expected_result = ['main', 'radius_server_auto', 'ldap_server_auto2']

        config_string = textwrap.dedent("""
            [main]
            key=value

            [MaIn]
            key=value

            [mAiN]
            key=value

            [radius_server_auto]
            key1=value1
            key2=value2

            [RADIUS_SERVER_AUTO]
            key1=value1
            key2=value2

            [ldap_server_auto1]
            key1=value1
            key2=value2

            [ldap_server_auto2]
            key1=value1
            key2=value2

            [ldap_server_auto2]
            key1=value1
            key2=value2
        """)

        actual_result = config_provider.find_duplicate_sections(config_string)
        self.assertItemsEqual(expected_result, actual_result)

    def test_duplication_with_whitespace(self):
        expected_result = ['main']

        config_string = textwrap.dedent("""
            [main]
            key1=value1

            [main]{0}
            key1=value1
        """.format('\t'))
        actual_result = config_provider.find_duplicate_sections(config_string)
        self.assertItemsEqual(expected_result, actual_result)

    def test_duplication_with_inline_comments(self):
        expected_result = ['main']

        config_string = textwrap.dedent("""
            [main]
            key1=value1

            [main] ; this is a comment
            key1=value1
        """)

        actual_result = config_provider.find_duplicate_sections(config_string)
        self.assertItemsEqual(expected_result, actual_result)

    def test_duplication_with_commented_out_section(self):
        expected_result = []

        config_string = textwrap.dedent("""
            [main]
            key1=value1

            ;[main]  this is a comment
            ;key1=value1
        """)

        actual_result = config_provider.find_duplicate_sections(config_string)
        self.assertItemsEqual(expected_result, actual_result)

    @mock.patch.object(config_provider, 'find_duplicate_sections')
    def test_duplication_raises_error(self, mock_find_duplicates):
        mock_find_duplicates.return_value = ['main']

        with self.assertRaises(ConfigError):
            config_provider.ConfigProvider('config string')

    def test_sectionless_config_errors(self):
        config_string = textwrap.dedent("""
        ;[commented]
        not=a_section

        still_not=a_section
        """)

        with self.assertRaises(ConfigError):
            config_provider.ConfigProvider(config_string)

    def test_config_contains_parsing_error(self):
        config_string = textwrap.dedent("""
            [main]
            key=value

            REthis is a bad comment
        """)
        with self.assertRaises(ConfigError):
            config_provider.ConfigProvider(config_string)

    def test_uppercase_main_found(self):
        expected_result = 'MAIN'

        section_list = ['MAIN', 'ldap_server_auto', 'HTTP_PROXY']
        actual_result = config_provider.find_main_section_name(section_list)

        self.assertEqual(expected_result, actual_result)

    def test_mixed_case_main_found(self):
        expected_result = 'MaIn'

        section_list = ['MaIn', 'LDAP_SERVER_AUTO', 'hTtP_PROxy']
        actual_result = config_provider.find_main_section_name(section_list)

        self.assertEqual(expected_result, actual_result)

    def test_get_main_section_success(self):
        expected_result = ConfigDict({'key1': 'value1'})

        config_string = self.GOOD_CONFIG
        cp = config_provider.ConfigProvider(config_string)
        actual_result = cp.get_main_section_config()

        self.assertItemsEqual(expected_result, actual_result)

    def test_get_main_section_mixed_case_success(self):
        expected_result = ConfigDict({'key1': 'value1'})

        config_string = textwrap.dedent("""
            [mAIn]
            key1=value1
        """)
        cp = config_provider.ConfigProvider(config_string)
        actual_result = cp.get_main_section_config()

        self.assertItemsEqual(expected_result, actual_result)

    def test_get_main_section_missing(self):
        expected_result = ConfigDict({})

        config_string = textwrap.dedent("""
            [cloud]
            key1=value1

            [ldap_server_auto]
            key1=value1
        """)
        cp = config_provider.ConfigProvider(config_string)
        actual_result = cp.get_main_section_config()

        self.assertItemsEqual(expected_result, actual_result)

    def test_get_all_sections(self):
        expected_result = {
            'main': ConfigDict({'key1': 'value1'}),
            'cloud': ConfigDict({'key1': 'value1', 'key2': 'value2'}),
            'http_proxy': ConfigDict({'key1': 'value1', 'key2': 'value2'}),
            'ad_client': ConfigDict({'key1': 'value1', 'key2': 'value2'}),
            'ldap_server_auto': ConfigDict({'key1': 'value1', 'key2': 'value2'}),
            'radius_client': ConfigDict({'key1': 'value1', 'key2': 'value2'}),
            'radius_server_auto1': ConfigDict({'key1': 'value1', 'key2': 'value2'}),
            'radius_server_auto2': ConfigDict({'key1': 'value1', 'key2': 'value2'}),
            'radius_server_auto3': ConfigDict({'key1': 'value1', 'key2': 'value2', 'key3': 'value3'}),
        }
        cp = config_provider.ConfigProvider(self.GOOD_CONFIG)
        actual_result = cp.get_all_sections()

        self.assertDictEqual(expected_result, actual_result)

    def test_get_section_missing(self):
        expected_result = ConfigDict({})

        config_string = self.GOOD_CONFIG
        cp = config_provider.ConfigProvider(config_string)
        actual_result = cp.get_section_config("Nope")

        self.assertItemsEqual(expected_result, actual_result)

    def test_normalize_line_endings_win_only(self):
        test_string = textwrap.dedent("""
           [main]{0}
           key1=value1{1}
        """.format(self.WIN_LINE_ENDING, self.WIN_LINE_ENDING))

        test_string = config_provider.normalize_line_endings(test_string)
        self.assert_only_win_line_endings(test_string)

    def test_normalize_line_endings_unix_only(self):
        test_string = textwrap.dedent("""
           [main]{0}
           key1=value1{1}
        """.format(self.UNIX_LINE_ENDING, self.UNIX_LINE_ENDING))

        test_string = config_provider.normalize_line_endings(test_string)
        self.assert_only_unix_line_endings(test_string)

    def test_normalize_line_endings_mixed(self):
        test_string = textwrap.dedent("""
           [main]{0}
           key1=value1{1}
        """.format(self.UNIX_LINE_ENDING, self.WIN_LINE_ENDING))

        test_string = config_provider.normalize_line_endings(test_string)
        self.assert_only_win_line_endings(test_string)

    def test_from_readable_success(self):
        mock_config_file = StringIO.StringIO(self.GOOD_CONFIG)

        cp = config_provider.ConfigProvider.from_ini_file(mock_config_file, "mock_file")

        expected_sections = ['main',
                             'cloud',
                             'http_proxy',
                             'ad_client',
                             'ldap_server_auto',
                             'radius_client',
                             'radius_server_auto1',
                             'radius_server_auto2',
                             'radius_server_auto3',
                             ]
        actual_sections = cp.list_sections()
        self.assertItemsEqual(expected_sections, actual_sections)

        expected_config = ConfigDict({'key1': 'value1', 'key2': 'value2', 'key3': 'value3'})
        actual_config = cp.get_section_config('radius_server_auto3')
        self.assertItemsEqual(expected_config, actual_config)


class TestMainSectionOverride(TestCase):
    def test_client_server_override(self):
        expected_result = ConfigDict({'client': 'ad_client1'})

        config_string = textwrap.dedent("""
            [main]
            client=ad_client1
            server=ldap_server_auto

            [ldap_server_auto]
            client=ad_client2
        """)
        cp = config_provider.ConfigProvider(config_string)
        actual_result = cp.get_section_config('ldap_server_auto')

        self.assertEqual(expected_result.get_str('client'), actual_result.get_str('client'))

    def test_client_server_override_unnecessary(self):
        expected_result = ConfigDict({'client': 'radius_client'})

        config_string = textwrap.dedent("""
            [main]
            client=ad_client1
            server=ldap_server_auto

            [radius_server_auto]
            client=radius_client
        """)
        cp = config_provider.ConfigProvider(config_string)
        actual_result = cp.get_section_config('radius_server_auto')

        self.assertEqual(expected_result.get_str('client'), actual_result.get_str('client'))

    def test_debug_override(self):
        expected_result = ConfigDict({'ikey': 'ikey', 'debug': 'True'})

        config_string = textwrap.dedent("""
            [main]
            DeBug=TruE

            [radius_server_auto]
            ikey=ikey
        """)
        cp = config_provider.ConfigProvider(config_string)
        actual_result = cp.get_section_config('radius_server_auto')

        self.assertDictEqual(expected_result, actual_result)

    def test_debug_override_unnecessary(self):
        expected_result = ConfigDict({'ikey': 'ikey'})

        config_string = textwrap.dedent("""
            [main]
            debug=false

            [radius_server_auto]
            ikey=ikey
        """)
        cp = config_provider.ConfigProvider(config_string)
        actual_result = cp.get_section_config('radius_server_auto')

        self.assertDictEqual(expected_result, actual_result)

    def test_interface_override(self):
        expected_result = ConfigDict({'ikey': 'ikey', 'interface': 'main_interface'})

        config_string = textwrap.dedent("""
            [main]
            interface=main_interface

            [radius_server_auto]
            ikey=ikey
        """)
        cp = config_provider.ConfigProvider(config_string)
        actual_result = cp.get_section_config('radius_server_auto')

        self.assertDictEqual(expected_result, actual_result)

    def test_interface_override_http_proxy(self):
        expected_result = ConfigDict({'api_host': 'api_host.test', 'interface': 'main_interface'})

        config_string = textwrap.dedent("""
            [main]
            interface=main_interface

            [http_proxy]
            api_host=api_host.test
        """)
        cp = config_provider.ConfigProvider(config_string)
        actual_result = cp.get_section_config('http_proxy')

        self.assertDictEqual(expected_result, actual_result)

    def test_interface_already_defined(self):
        expected_result = ConfigDict({'ikey': 'ikey', 'interface': 'radius_interface'})

        config_string = textwrap.dedent("""
            [main]
            interface=main_interface

            [radius_server_auto]
            ikey=ikey
            interface=radius_interface
        """)
        cp = config_provider.ConfigProvider(config_string)
        actual_result = cp.get_section_config('radius_server_auto')

        self.assertDictEqual(expected_result, actual_result)

    def test_mix_interface_defined(self):
        expected_radius_result = ConfigDict({'ikey': 'ikey', 'interface': 'radius_interface'})
        expected_ldap_result = ConfigDict({'ikey': 'ikey', 'interface': 'main_interface'})

        config_string = textwrap.dedent("""
            [main]
            interface=main_interface

            [radius_server_auto]
            ikey=ikey
            interface=radius_interface

            [ldap_server_auto]
            ikey=ikey
        """)
        cp = config_provider.ConfigProvider(config_string)
        actual_radius_result = cp.get_section_config('radius_server_auto')
        actual_ldap_result = cp.get_section_config('ldap_server_auto')

        self.assertDictEqual(expected_radius_result, actual_radius_result)
        self.assertDictEqual(expected_ldap_result, actual_ldap_result)

    def test_interface_only_server_sections(self):
        expected_server_result = ConfigDict({'ikey': 'ikey', 'interface': 'main_interface'})
        expected_client_result = ConfigDict({'host': 'host'})

        config_string = textwrap.dedent("""
            [main]
            interface=main_interface

            [ldap_server_auto3]
            ikey=ikey

            [ad_client2]
            host=host
        """)
        cp = config_provider.ConfigProvider(config_string)
        actual_server_result = cp.get_section_config('ldap_server_auto3')
        actual_client_result = cp.get_section_config('ad_client2')

        self.assertDictEqual(expected_server_result, actual_server_result)
        self.assertDictEqual(expected_client_result, actual_client_result)

    def test_no_main_interface(self):
        expected_radius_result = ConfigDict({'ikey': 'ikey'})
        expected_ldap_result = ConfigDict({'ikey': 'ikey', 'interface': 'ldap_interface'})

        config_string = textwrap.dedent("""
            [main]
            ignore=ignore

            [radius_server_auto]
            ikey=ikey

            [ldap_server_auto]
            ikey=ikey
            interface=ldap_interface
        """)
        cp = config_provider.ConfigProvider(config_string)
        actual_radius_result = cp.get_section_config('radius_server_auto')
        actual_ldap_result = cp.get_section_config('ldap_server_auto')

        self.assertDictEqual(expected_radius_result, actual_radius_result)
        self.assertDictEqual(expected_ldap_result, actual_ldap_result)

    def test_http_ca_certs_override(self):
        expected_result = ConfigDict({'ikey': 'ikey', 'http_ca_certs_file': '/tmp/mycerts.pem'})
        config_string = textwrap.dedent("""
            [main]
            http_ca_certs_file=/tmp/mycerts.pem

            [ldap_server_auto]
            ikey=ikey
        """)
        cp = config_provider.ConfigProvider(config_string)
        actual_result = cp.get_section_config('ldap_server_auto')

        self.assertDictEqual(expected_result, actual_result)

    def test_http_proxy_override(self):
        expected_ldap_section = ConfigDict({'ikey': 'ikey', 'http_proxy_host': 'proxy.host', 'http_proxy_port': '80'})
        expected_cloud_section = ConfigDict({'ikey': 'ikey', 'http_proxy_host': 'proxy.host', 'http_proxy_port': '80'})
        config_string = textwrap.dedent("""
            [main]
            http_proxy_host=proxy.host

            [ldap_server_auto]
            ikey=ikey

            [cloud]
            ikey=ikey
        """)
        cp = config_provider.ConfigProvider(config_string)
        actual_ldap_section = cp.get_section_config('ldap_server_auto')
        actual_cloud_section = cp.get_section_config('cloud')

        self.assertDictEqual(expected_ldap_section, actual_ldap_section)
        self.assertDictEqual(expected_cloud_section, actual_cloud_section)

    def test_http_proxy_override_with_port(self):
        expected_result = ConfigDict({'ikey': 'ikey', 'http_proxy_host': 'proxy.host', 'http_proxy_port': '8080'})
        config_string = textwrap.dedent("""
            [main]
            http_proxy_host=proxy.host
            http_proxy_port=8080

            [ldap_server_auto]
            ikey=ikey
        """)
        cp = config_provider.ConfigProvider(config_string)
        actual_result = cp.get_section_config('ldap_server_auto')

        self.assertDictEqual(expected_result, actual_result)

    def test_override_off(self):
        expected_result = ConfigDict({'ikey': 'ikey'})

        config_string = textwrap.dedent("""
            [main]
            debug=true
            interface=main_interface
            http_ca_certs_file=/tmp/mycerts.pem

            [radius_server_auto]
            ikey=ikey
        """)
        cp = config_provider.ConfigProvider(config_string, apply_overrides=False)
        actual_result = cp.get_section_config('radius_server_auto')

        self.assertDictEqual(expected_result, actual_result)


class TestGetSectionClient(TestCase):
    def test_section_client_ldap(self):
        config_string = textwrap.dedent("""
            [ldap_server_auto]
            ikey=foo

            [ad_client]
            host=blah

            [ad_client2]
            host=bar
        """)
        cp = config_provider.ConfigProvider(config_string, apply_overrides=False)
        self.assertEqual(cp.get_section_client('ldap_server_auto'), None)

    def test_section_client_only_one(self):
        config_string = textwrap.dedent("""
            [radius_server_auto]
            ikey=foo

            [ad_client]
            host=blah
        """)
        cp = config_provider.ConfigProvider(config_string, apply_overrides=False)
        self.assertEqual(cp.get_section_client('radius_server_auto'), 'ad_client')

    def test_section_client_mismtached_ad(self):
        '''AD Clients will pick up a radius client if it's the only section
        available.
        '''
        config_string = textwrap.dedent("""
            [ldap_server_auto]
            ikey=foo

            [radius_client]
            host=blah
        """)
        cp = config_provider.ConfigProvider(config_string, apply_overrides=False)
        self.assertEqual(cp.get_section_client('ldap_server_auto'), 'radius_client')

    def test_section_client_capitalization(self):
        config_string = textwrap.dedent("""
            [LDAP_SERVER_AUTo]
            ikey=foo
            client=AD_cliENT1

            [ad_client1]
            host=blah
        """)
        cp = config_provider.ConfigProvider(config_string, apply_overrides=False)

        # Flagging this as an error is bussiness logic not part of determining
        # the client name
        self.assertEqual(cp.get_section_client('ldap_server_auto'), 'ad_client1')

    def test_section_client_ad_over_main(self):
        ''' An LDAP Server Auto section looking for a client should select any
        ad_client section before falling back to the main default client.
        '''
        config_string = textwrap.dedent("""
            [main]
            client=radius_client

            [ldap_server_auto]
            ikey=foo

            [radius_client]
            host=blah

            [ad_client2]
            host=blah2
        """)
        cp = config_provider.ConfigProvider(config_string, apply_overrides=False)
        self.assertEqual(cp.get_section_client('ldap_server_auto'), 'ad_client2')

    def test_section_client_radius_fallback_to_main(self):
        ''' A Radius Server Auto section looking for a client should select the
        client specified in main if none is provided
        '''
        config_string = textwrap.dedent("""
            [main]
            client=radius_client2

            [radius_server_auto]
            ikey=foo

            [radius_client]
            host=blah

            [radius_client2]
            host=blah2

            [ad_client2]
            host=blah2
        """)
        cp = config_provider.ConfigProvider(config_string, apply_overrides=False)
        self.assertEqual(cp.get_section_client('radius_server_auto'), 'radius_client2')


class TestGetModuleType(TestCase):
    def test_server(self):
        expected_result = 'server'
        actual_result = config_provider.get_module_type('radius_server_auto3')
        self.assertEqual(expected_result, actual_result)

    def test_client(self):
        expected_result = 'client'
        actual_result = config_provider.get_module_type('ad_client4')
        self.assertEqual(expected_result, actual_result)

    def test_cloud(self):
        expected_result = 'cloud'
        actual_result = config_provider.get_module_type('cloud')
        self.assertEqual(expected_result, actual_result)

    def test_http_proxy(self):
        expected_result = 'http_proxy'
        actual_result = config_provider.get_module_type('http_proxy2')
        self.assertEqual(expected_result, actual_result)

    def test_no_match(self):
        expected_result = ''
        actual_result = config_provider.get_module_type('nothing7')
        self.assertEqual(expected_result, actual_result)


class TestIsKeyProtectable(TestCase):
    def test_not_protectable(self):
        result = config_provider._is_key_protectable('ikey')
        self.assertFalse(result)

    def test_protectable(self):
        for option in ['skey', 'service_account_password', 'secret']:
            result = config_provider._is_key_protectable(option)
            self.assertTrue(result, "option not protectable: {}".format(option))

    def test_radius_secret(self):
        """Radius secret is the one tricky key because it has numbers appended."""
        for option in ['radius_secret_1', 'radius_secret_2', 'radius_secret_23456']:
            result = config_provider._is_key_protectable(option)
            self.assertTrue(result, "option not protectable: {}".format(option))

    def test_already_protected(self):
        for option in ['radius_secret_protected_1', 'skey_protected', 'service_account_password_protected']:
            result = config_provider._is_key_protectable(option)
            self.assertFalse(result, "option not protectable: {}".format(option))


class TestGenerateProtectedOption(TestCase):
    def test_skey(self):
        expected = 'skey_protected'
        actual = config_provider._generate_protected_option('skey')
        self.assertEqual(expected, actual)

    def test_radius_secret(self):
        expected = 'radius_secret_protected_22'
        actual = config_provider._generate_protected_option('radius_secret_22')
        self.assertEqual(expected, actual)


class TestKeyDuplication(TestCase):
    def test_last_write_wins(self):
        expected = 'second_value'
        config_string = textwrap.dedent("""
            [radius_server_auto]
            skey=skey
            skey=second_value
        """)
        config = config_provider.ConfigProvider(config_string)
        actual = config.get_section_config('radius_server_auto')['skey']
        self.assertEqual(actual, expected)

    def test_multi_section_last_write_wins(self):
        expected = 'second_value'
        config_string = textwrap.dedent("""
            [radius_server_auto]
            skey=first_value
            skey=second_value

            [radius_server_auto2]
            skey=third_value
        """)
        config = config_provider.ConfigProvider(config_string)
        actual = config.get_section_config('radius_server_auto')['skey']
        self.assertEqual(actual, expected)


class TestKeyStandardization(TestCase):
    def test_getting_an_unlisted_key(self):
        config_string = textwrap.dedent("""
            [radius_server_auto]
            key_with_no_standard_form=value
            skey=skey
        """)
        config = config_provider.ConfigProvider(config_string)
        self.assertTrue('key_with_no_standard_form' in config.get_section_config('radius_server_auto'))

    def test_getting_a_listed_key(self):
        config_string = textwrap.dedent("""
            [radius_server_eap]
            ikey=ikey
            skey=skey
            concat_mode=test_value
        """)
        config = config_provider.ConfigProvider(config_string)
        self.assertEqual(config.get_section_config('radius_server_eap')['allow_concat'], 'test_value')

    def test_get_standard_key(self):
        for key in [None, 'a', 1]:
            self.assertEqual(key, config_provider.get_standard_key(key))


class TestAddEncryptedValue(TestCase):
    def setUp(self):
        self.MOCK_ENCRYPT_VALUE = 'AAA'

        config_string = textwrap.dedent("""
            [radius_server_auto]
            skey=skey
        """)
        self.cp = config_provider.ConfigProvider(config_string, apply_overrides=False)

    @mock.patch('duoauthproxy.lib.protect.PROTECT_ENABLED', False)
    def test_protect_not_enabled(self):
        with self.assertRaises(ConfigError):
            self.cp._add_encrypted_value('radius_server_auto', 'skey')

    @mock.patch('duoauthproxy.lib.protect.PROTECT_ENABLED', True)
    @mock.patch('duoauthproxy.lib.protect.protect')
    def test_protect_skey(self, mock_protect):
        mock_protect.return_value = self.MOCK_ENCRYPT_VALUE
        expected_result = ConfigDict({'skey': 'skey', 'skey_protected': self.MOCK_ENCRYPT_VALUE})

        self.cp._add_encrypted_value('radius_server_auto', 'skey')

        actual_result = self.cp.get_section_config('radius_server_auto')
        self.assertDictEqual(expected_result, actual_result)


class TestRemoveUnencryptedValue(TestCase):
    def setUp(self):
        config_string = textwrap.dedent("""
            [radius_server_auto]
            ikey=ikey
            skey=skey
        """)
        self.cp = config_provider.ConfigProvider(config_string, apply_overrides=False)

    def test_bad_section(self):
        with self.assertRaises(ConfigError):
            self.cp._remove_unencrypted_value('WRONGIUS_SERVER_AUTO', 'skey')

    def test_option_doesnt_exist(self):
        with self.assertRaises(ConfigError):
            self.cp._remove_unencrypted_value('radius_server_auto', 'WRONGkey')

    def test_good_removal(self):
        expected_result = ConfigDict({'ikey': 'ikey'})

        self.cp._remove_unencrypted_value('radius_server_auto', 'skey')
        actual_result = self.cp.get_section_config('radius_server_auto')

        self.assertDictEqual(expected_result, actual_result)


class TestEncryptAllValues(TestCase):
    def setUp(self):
        self.MOCK_ENCRYPT_VALUE = 'AAA'

    @mock.patch('duoauthproxy.lib.protect.PROTECT_ENABLED', True)
    @mock.patch('duoauthproxy.lib.protect.protect')
    def test_full_config(self, mock_protect):
        mock_protect.return_value = self.MOCK_ENCRYPT_VALUE
        expected_result = {
            'main': ConfigDict({'debug': 'True'}),
            'ad_client': ConfigDict({'service_account_username': 'user', 'service_account_password_protected': self.MOCK_ENCRYPT_VALUE}),
            'radius_client': ConfigDict({'host': '127.0.0.1', 'secret_protected': self.MOCK_ENCRYPT_VALUE}),
            'radius_server_auto2': ConfigDict({'ikey': 'ikey', 'skey_protected': self.MOCK_ENCRYPT_VALUE, 'radius_ip_1': '127.0.0.1', 'radius_secret_protected_1': self.MOCK_ENCRYPT_VALUE}),
        }

        config_string = textwrap.dedent("""
            [main]
            debug=True

            [radius_client]
            host=127.0.0.1
            secret=secret

            [ad_client]
            service_account_username=user
            service_account_password=pass

            [radius_server_auto2]
            ikey=ikey
            skey=skey
            radius_ip_1=127.0.0.1
            radius_secret_1=shh_secret
        """)
        cp = config_provider.ConfigProvider(config_string, apply_overrides=False)

        cp.encrypt_all_values()
        actual_result = cp.get_all_sections()
        self.assertDictEqual(expected_result, actual_result)


if __name__ == '__main__':
    unittest.main()
