#
# Copyright (c) 2013 Duo Security
# All Rights Reserved
#
import itertools
import os
import struct

import OpenSSL
from pyrad import packet
from twisted.internet import defer
from base import TestCase, ServerModuleTest, ReopenableTemporaryFile, AuthLoggingTest
from duoauthproxy.lib import eap, duo_async, log
from duoauthproxy.lib.duo_async import FAILMODE_SAFE, FAILMODE_SECURE
from duoauthproxy.lib.config_provider import ConfigDict
from duoauthproxy.lib.radius.base import radius_dictionary
from duoauthproxy.modules import radius_server_eap, mock_client


class mock_server(object):
    def __init__(self):
        self.server_section_name = 'Mock server'


class mock_request(object):
    def __init__(self):
        self.username = 'Nobody'
        self.client_ip = '1.2.3.4'


class mock_eap(eap.EAPSession):
    def __init__(self, test, **kwargs):
        self.test = test
        self.server = mock_server()
        self.current_request = mock_request()
        super(mock_eap, self).__init__(**kwargs)

    def success(self, session):
        self.test.success = True

    def errback(self, session, reason):
        self.test.errback.append(reason)
        return reason

    def gtc_received(self, session, passcode, prompt):
        self.test.passcodes.append(passcode)


def create_keys():
    cert = OpenSSL.crypto.X509()
    cert.get_subject().CN = "ca.example.com"
    cert.gmtime_adj_notBefore(0)
    cert.gmtime_adj_notAfter(24 * 60 * 60)
    cert.set_issuer(cert.get_subject())
    key = OpenSSL.crypto.PKey()
    key.generate_key(OpenSSL.crypto.TYPE_RSA, 2048)
    pkey = OpenSSL.crypto.dump_privatekey(OpenSSL.crypto.FILETYPE_PEM, key)
    cert.sign(key, "sha1")
    cert.set_pubkey(key)
    cert = OpenSSL.crypto.dump_certificate(OpenSSL.crypto.FILETYPE_PEM,
                                           cert)
    temp_key_file = ReopenableTemporaryFile(pkey)
    temp_cert_file = ReopenableTemporaryFile(cert)

    pkey_name = temp_key_file.name
    cert_name = temp_cert_file.name
    return (pkey_name, cert_name, (temp_key_file, temp_cert_file))


class mock_preauth(object):
    def __init__(self, result, status, default_factor, api_excp=None):
        self.preauth_res = {
            'result': result,
            'status': status,
            'factors': {
                '1': 'push1',
                '2': 'phone1',
                '3': 'sms1',
            },
            'prompt': '?'}
        if default_factor:
            self.preauth_res['factors']['default'] = default_factor
        self.api_excp = api_excp
        self.called = 0

    def __call__(self, username, client_ip):
        self.called += 1
        if self.api_excp:
            raise self.api_excp
        return self.preauth_res


class mock_duo_auth(object):
    def __init__(self, expected_factor='push'):
        self.expected_factor = expected_factor
        self.called = 0

    def __call__(self, request, primary_res, factor):
        self.called += 1

        if self.expected_factor is None:
            success = (factor is None)
        else:
            success = (factor == self.expected_factor)

        if success:
            response_packet = request.packet.CreateReply()
            response_packet.code = packet.AccessAccept
        else:
            response_packet = request.packet.CreateReply()
            response_packet.code = packet.AccessReject
        return response_packet


class BaseEapTest(TestCase):
    def get_id(self, packet):
        return ord(packet[1])

    def get_code(self, packet):
        return ord(packet[0])

    def get_type(self, packet):
        return ord(packet[4])

    def get_data(self, packet):
        return packet[5:]

    def get_peap_data(self, packet):
        return packet[6:]

    def get_len(self, packet):
        return struct.unpack('>H', packet[2:4])[0]

    def remove_header(self, packet):
        return packet[4:]

    def verify_packet(self, response, code=None, id=None,
                      type=None, length=None, data=None):
        if code:
            self.assertEqual(self.get_code(response), code)
        if id:
            self.assertEqual(self.get_id(response), id)
        if type:
            self.assertEqual(self.get_type(response), type)
        if length:
            self.assertEqual(self.get_len(response), length)
        if data:
            self.assertEqual(self.get_data(response), data)

    def tls_decrypt(self, con, data):
        con.bio_write(data)
        return con.recv(4096)

    def tls_encrypt(self, con, data):
        con.send(data)
        return con.bio_read(4096)

    @defer.inlineCallbacks
    def start_peap(self, session, id):
        nak_pack = eap.EAPPacket(eap.EAP_CODE_RESPONSE, id, eap.EAP_TYPE_NAK, 25)
        yield session.add_message(nak_pack.render())
        response = yield session.next_message()
        self.verify_packet(response, eap.EAP_CODE_REQUEST, id + 1, eap.EAP_TYPE_PEAP, data=chr(32))

        ctx = OpenSSL.SSL.Context(OpenSSL.SSL.TLSv1_2_METHOD)
        ctx.use_privatekey_file(self.pkey_name)
        ctx.use_certificate_file(self.cert_name)
        self.ssl_con = OpenSSL.SSL.Connection(ctx, None)
        self.ssl_con.set_connect_state()
        try:
            self.ssl_con.do_handshake()
        except OpenSSL.SSL.WantReadError:
            pass

        client_hello = self.ssl_con.bio_read(4096)
        peap_packet = eap.EAPPacket(eap.EAP_CODE_RESPONSE, id + 1, eap.EAP_TYPE_PEAP, client_hello)
        session.add_message(peap_packet.render())
        response = yield session.next_message()
        self.verify_packet(response, eap.EAP_CODE_REQUEST, id + 2, eap.EAP_TYPE_PEAP)
        data = self.get_peap_data(response)

        try:
            self.tls_decrypt(self.ssl_con, data)
        except OpenSSL.SSL.WantReadError:
            pass

        client_key_ex = self.ssl_con.bio_read(4096)
        peap_packet = eap.EAPPacket(eap.EAP_CODE_RESPONSE, id + 2, eap.EAP_TYPE_PEAP, client_key_ex)
        session.add_message(peap_packet.render())
        response = yield session.next_message()
        self.verify_packet(response, eap.EAP_CODE_REQUEST, id + 3, eap.EAP_TYPE_PEAP)
        data = self.get_peap_data(response)
        self.ssl_con.bio_write(data)


class TestEap(BaseEapTest):
    def setUp(self):
        self.pkey_name, self.cert_name, self.tmp_fhs = create_keys()
        self.session = mock_eap(self, pkey=self.pkey_name, certs=self.cert_name)
        self.errback = []
        self.success = False
        self.passcodes = []

    def tearDown(self):
        for fh in self.tmp_fhs:
            if fh.close_called and os.path.isfile(fh.name):
                os.remove(fh.name)
                pass
            pass

    def peap_send(self, con, data, id):
        encrypted = self.tls_encrypt(con, data)
        peap_packet = eap.EAPPacket(eap.EAP_CODE_RESPONSE, id, eap.EAP_TYPE_PEAP, encrypted)
        self.session.add_message(peap_packet.render())

    @defer.inlineCallbacks
    def test_gtc(self):
        ''' Test whether a GTC packet inside PEAP triggers gtc_received callback '''
        yield self.start_peap(self.session, 0)
        gtc_packet = eap.EAPPacket(eap.EAP_CODE_RESPONSE, 0, eap.EAP_TYPE_GTC, 'codes here!')
        self.peap_send(self.ssl_con, self.remove_header(gtc_packet.render()), 3)

        self.assertEqual(len(self.passcodes), 1)
        self.assertEqual(self.passcodes[0], 'codes here!')

    @defer.inlineCallbacks
    def test_success(self):
        ''' Test that a session in PEAP accept state sends and confirms a
        MS-Result TLV and calls success '''
        self.start_peap(self.session, 0)
        self.session.state = eap.EAP_SESSION_PEAP_ACCEPT
        response = yield self.session.next_message()
        self.verify_packet(response, eap.EAP_CODE_REQUEST, 3, eap.EAP_TYPE_PEAP)
        data = self.get_peap_data(response)
        decrypted_packet = self.tls_decrypt(self.ssl_con, data)
        success_tlv = eap.EAPPacket(eap.EAP_CODE_REQUEST, 1, eap.EAP_TYPE_TLV, True)
        self.assertEqual(self.remove_header(success_tlv.render()), decrypted_packet)

        success_tlv_resp = eap.EAPPacket(eap.EAP_CODE_RESPONSE, 1, eap.EAP_TYPE_TLV, True)
        self.peap_send(self.ssl_con, success_tlv_resp.render(), 3)

        response = yield self.session.next_message()

        self.assertEqual(self.success, True)

    @defer.inlineCallbacks
    def test_id(self):
        ''' Test that upon receiving an ID packet, the session sends a GTC prompt '''
        id_packet = eap.EAPPacket(eap.EAP_CODE_RESPONSE, 0, eap.EAP_TYPE_IDENTITY, 'test')
        self.session.add_message(id_packet.render())
        response = yield self.session.next_message()
        self.verify_packet(
            response,
            eap.EAP_CODE_REQUEST,
            1,
            eap.EAP_TYPE_GTC,
            5 + len(self.session.gtc_message),
            self.session.gtc_message
        )

    def test_invalid_nak(self):
        ''' Test that a request for EAP-TLS triggers errback '''
        nak_pack = eap.EAPPacket(eap.EAP_CODE_RESPONSE, 0, eap.EAP_TYPE_NAK, 13)  # Eap-TLS
        self.session.add_message(nak_pack.render())

        self.assertEqual(len(self.errback), 1)
        self.assertEqual(self.errback[0], "Unknown Nak request 13.")


class TestEapModule(BaseEapTest, ServerModuleTest):
    def setUp(self):
        self.prompt = 'Stuff goes here'
        self.pkey_name, self.cert_name, self.tmp_fhs = create_keys()
        extra_config = self.get_server_module_config()
        ServerModuleTest.setUp(self, extra_config, radius_server_eap, 'radius_server_eap')
        self.radius_packet_id = itertools.count(0)
        self.addr = (self.radius_ip, 12345)
        self.state = ''
        self.pkey_name = self.pkey_name
        self.cert_name = self.cert_name
        self.client_addr = (self.radius_ip,
                            self.radius_client_port)
        self.config_items = {
            'username': self.username,
            'password': self.password,
        }
        config = ConfigDict(self.config_items)
        self.instance.protocol.primary_ator = mock_client.Module(config)

        # Disable cleanup_request calls to prevent reactor unclean errors
        self.instance.protocol._send_response = lambda x: x

    def get_server_module_config(self):
        extra_config = {
            'prompt': self.prompt,
            'pkey': self.pkey_name,
            'certs': self.cert_name
        }
        return extra_config

    def tearDown(self):
        for fh in self.tmp_fhs:
            if fh.close_called and os.path.isfile(fh.name):
                os.remove(fh.name)
                pass
            pass
        ServerModuleTest.tearDown(self)

    @defer.inlineCallbacks
    def get_challenge(self, eap_packet, send_only=False):
        radius_packet_id = self.radius_packet_id.next()
        request = packet.AuthPacket(code=packet.AccessRequest,
                                    id=radius_packet_id,
                                    secret=self.secret,
                                    dict=radius_dictionary())
        request.AddAttribute('User-Name', self.username)
        request.AddAttribute('EAP-Message', eap_packet.render())
        if self.state:
            request.AddAttribute('State', self.state)
        yield self.f_instance.protocol.datagramReceived(request.RequestPacket(), self.client_addr)

        if send_only:
            defer.returnValue(None)

        self.assertEqual(len(self.f_instance.protocol.transport.queue), 1, "No response packet was generated when expected")
        res_data, res_dest = self.f_instance.protocol.transport.queue.pop(0)
        response = packet.AuthPacket(
            packet=res_data,
            dict=radius_dictionary()
        )

        self.assertEqual(response.code, packet.AccessChallenge, 'Response code {0} does not match expected {1}'.format(response.code, packet.AccessChallenge))

        eap_message = ''.join(response['EAP-Message'])
        self.state = response['State'][0]
        defer.returnValue(eap_message)

    @defer.inlineCallbacks
    def get_peap_response(self, eap_packet, send_only=False):
        self.peap_id += 1
        encrypted_packet = self.tls_encrypt(self.ssl_con, self.remove_header(eap_packet.render()))
        peap_packet = eap.EAPPacket(
            eap.EAP_CODE_RESPONSE,
            self.peap_id,
            eap.EAP_TYPE_PEAP,
            encrypted_packet
        )

        eap_response = yield self.get_challenge(peap_packet, send_only)

        if send_only:
            defer.returnValue(None)

        self.verify_packet(eap_response, eap.EAP_CODE_REQUEST, self.peap_id + 1, eap.EAP_TYPE_PEAP)
        encrypted_response = self.get_peap_data(eap_response)
        decrypted_response = self.tls_decrypt(self.ssl_con, encrypted_response)
        decrypted_response = 'xxxx' + decrypted_response  # Add junk header
        defer.returnValue(decrypted_response)

    @defer.inlineCallbacks
    def basic_case_id(self):
        # Send ID and verify we get a GTC request back
        self.peap_id = 0
        id_packet = eap.EAPPacket(
            eap.EAP_CODE_RESPONSE,
            self.peap_id,
            eap.EAP_TYPE_IDENTITY,
            'test'
        )
        eap_response = yield self.get_challenge(id_packet)
        self.verify_packet(
            eap_response,
            eap.EAP_CODE_REQUEST,
            self.peap_id + 1,
            eap.EAP_TYPE_GTC,
            None,
            self.prompt
        )

    @defer.inlineCallbacks
    def basic_case_tls(self):
        # Cheat and start TLS by directly modifying session
        self.peap_id += 1
        session = self.instance.protocol.challenges.itervalues().next().state
        yield self.start_peap(session, self.peap_id)

    @defer.inlineCallbacks
    def basic_case_peap_id(self):
        # Send ID again through PEAP and verify we get a GTC request
        self.peap_id += 2
        id_packet = eap.EAPPacket(
            eap.EAP_CODE_RESPONSE,
            0,
            eap.EAP_TYPE_IDENTITY,
            'test'
        )
        eap_response = yield self.get_peap_response(id_packet)
        self.verify_packet(eap_response, None, None, eap.EAP_TYPE_GTC, None, self.prompt)

    @defer.inlineCallbacks
    def basic_case_pw(self):
        # Enqueue auth result, send password and verify we get a factor GTC request
        self.instance.protocol.client.preauth = mock_preauth(duo_async.API_RESULT_AUTH, '', 'push1')
        gtc_packet = eap.EAPPacket(eap.EAP_CODE_RESPONSE, 0,
                                   eap.EAP_TYPE_GTC, self.password)
        eap_response = yield self.get_peap_response(gtc_packet)
        self.verify_packet(eap_response, None, None, eap.EAP_TYPE_GTC, None)

    @defer.inlineCallbacks
    def basic_case_factor(self):
        # Send a factor and verify we get a MS success TLV
        self.instance.protocol.duo_auth = mock_duo_auth()
        gtc_packet = eap.EAPPacket(eap.EAP_CODE_RESPONSE, 0,
                                   eap.EAP_TYPE_GTC, 'push')
        eap_response = yield self.get_peap_response(gtc_packet)
        success_tlv = '!\x80\x03\x00\x02\x00\x01'
        self.assertEqual(self.remove_header(eap_response), success_tlv)

    @defer.inlineCallbacks
    def basic_case_tlv(self):
        # Send a success TLV back
        self.peap_id += 1
        tlv_packet = eap.EAPPacket(eap.EAP_CODE_RESPONSE, 0, eap.EAP_TYPE_TLV, True)
        encrypted_packet = self.tls_encrypt(self.ssl_con, self.remove_header(tlv_packet.render()))
        peap_packet = eap.EAPPacket(eap.EAP_CODE_RESPONSE, self.peap_id,
                                    eap.EAP_TYPE_PEAP, encrypted_packet)
        radius_packet_id = self.radius_packet_id.next()
        request = packet.AuthPacket(code=packet.AccessRequest,
                                    id=radius_packet_id,
                                    secret=self.secret,
                                    dict=radius_dictionary())
        request.AddAttribute('User-Name', self.username)
        request.AddAttribute('EAP-Message', peap_packet.render())
        if self.state:
            request.AddAttribute('State', self.state)
        yield self.f_instance.protocol.datagramReceived(request.RequestPacket(), self.client_addr)

    def basic_case_verify(self):
        # Verify we get an Access-Accept w/ MPPE keys
        self.assertEqual(len(self.f_instance.protocol.transport.queue), 1)
        res_data, res_dest = self.f_instance.protocol.transport.queue.pop(0)
        response = packet.AuthPacket(
            packet=res_data,
            dict=radius_dictionary()
        )
        self.assertEqual(response.code, packet.AccessAccept)
        self.assertIn('MS-MPPE-Send-Key', response.keys())
        self.assertIn('MS-MPPE-Recv-Key', response.keys())

    @defer.inlineCallbacks
    def setup_peap(self):
        yield self.basic_case_id()
        yield self.basic_case_tls()
        yield self.basic_case_peap_id()

    @defer.inlineCallbacks
    def assert_gtc_packet_is_rejected(self, gtc_packet):
        yield self.get_peap_response(gtc_packet, True)

        # Verify we get a reject packet
        self.assertEqual(len(self.f_instance.protocol.transport.queue), 1)
        res_data, res_dest = self.f_instance.protocol.transport.queue.pop(0)
        response = packet.AuthPacket(
            packet=res_data,
            dict=radius_dictionary()
        )
        self.assertEqual(response.code, packet.AccessReject)

    @defer.inlineCallbacks
    def assert_gtc_packet_is_auth_success(self, gtc_packet):
        # Verify we get a success packet
        eap_response = yield self.get_peap_response(gtc_packet)

        success_tlv = '!\x80\x03\x00\x02\x00\x01'
        self.assertEqual(self.remove_header(eap_response), success_tlv)

        yield self.basic_case_tlv()
        yield self.basic_case_verify()


class TestEapModuleChallenge(TestEapModule):

    @defer.inlineCallbacks
    def test_basic_case(self):
        ''' Performs a full session authentication, including password and factor prompts '''
        yield self.setup_peap()
        yield self.basic_case_pw()
        yield self.basic_case_factor()
        yield self.basic_case_tlv()
        yield self.basic_case_verify()

    @defer.inlineCallbacks
    def test_preauth_allow(self):
        yield self.setup_peap()

        # Set preauth to return 'allow'
        self.instance.protocol.client.preauth = mock_preauth(
            duo_async.API_RESULT_ALLOW, 'no duo for you', '')
        # Send primary password; make sure we get a success
        gtc_packet = eap.EAPPacket(
            eap.EAP_CODE_RESPONSE, 0, eap.EAP_TYPE_GTC, self.password)
        yield self.assert_gtc_packet_is_auth_success(gtc_packet)

    @defer.inlineCallbacks
    def test_second_preauth_allow(self):
        yield self.setup_peap()
        # Primary auth will call preauth, which is mocked to return AUTH
        yield self.basic_case_pw()

        # Sending the factor will cause a second preauth, mock that to return 'allow'
        self.instance.protocol.client.preauth = mock_preauth(
            duo_async.API_RESULT_ALLOW, 'no duo for you', '')
        # Send factor choice; make sure we get a success
        gtc_packet = eap.EAPPacket(
            eap.EAP_CODE_RESPONSE, 0, eap.EAP_TYPE_GTC, 'push')
        yield self.assert_gtc_packet_is_auth_success(gtc_packet)

    @defer.inlineCallbacks
    def test_preauth_enroll(self):
        yield self.setup_peap()

        # Set preauth to return 'enroll'
        self.instance.protocol.client.preauth = mock_preauth(
            duo_async.API_RESULT_ENROLL, 'you gotta enroll', '')

        # send primary password
        gtc_packet = eap.EAPPacket(
            eap.EAP_CODE_RESPONSE, 0, eap.EAP_TYPE_GTC, self.password)
        eap_response = yield self.get_peap_response(gtc_packet)
        self.verify_packet(eap_response, None, None, eap.EAP_TYPE_GTC, None)

        # make sure whatever we send next is unconditionally rejected
        self.instance.protocol.duo_auth = mock_duo_auth()
        gtc_packet = eap.EAPPacket(
            eap.EAP_CODE_RESPONSE, 0, eap.EAP_TYPE_GTC, 'push')
        yield self.assert_gtc_packet_is_rejected(gtc_packet)

    @defer.inlineCallbacks
    def test_preauth_deny(self):
        yield self.setup_peap()

        # Set preauth to return 'enroll'
        self.instance.protocol.client.preauth = mock_preauth(
            duo_async.API_RESULT_DENY, 'DE-NIED', '')

        # send primary password, make sure we get a reject
        gtc_packet = eap.EAPPacket(
            eap.EAP_CODE_RESPONSE, 0, eap.EAP_TYPE_GTC, self.password)
        yield self.assert_gtc_packet_is_rejected(gtc_packet)

    @defer.inlineCallbacks
    def test_incorrect_password(self):
        yield self.setup_peap()

        # make our mock preauth function
        m = mock_preauth(duo_async.API_RESULT_AUTH, '', 'push1')
        self.instance.protocol.client.preauth = m

        # Enqueue auth result, send wrong password and verify we get rejection
        gtc_packet = eap.EAPPacket(eap.EAP_CODE_RESPONSE, 0,
                                   eap.EAP_TYPE_GTC, self.password + 'DEADBEEF')
        yield self.assert_gtc_packet_is_rejected(gtc_packet)

        # ensure we did not call preauth!
        self.assertEqual(0, m.called)

    @defer.inlineCallbacks
    def test_incorrect_duo(self):
        yield self.setup_peap()

        # send password as normal
        yield self.basic_case_pw()

        # apply our mock duo_auth function
        self.instance.protocol.duo_auth = mock_duo_auth()

        # send wrong duo factor and verify we get a rejection
        gtc_packet = eap.EAPPacket(eap.EAP_CODE_RESPONSE, 0,
                                   eap.EAP_TYPE_GTC, 'XXXXXX')
        yield self.assert_gtc_packet_is_rejected(gtc_packet)

    @defer.inlineCallbacks
    def test_empty_credentials(self):
        ''' In challenge/response mode, if we send an empty password
        at any point, we should just get a repeat of the last
        challenge'''
        yield self.setup_peap()

        # do it 5 times in a row just for kicks
        for i in range(5):
            # send empty primary password
            gtc_packet = eap.EAPPacket(eap.EAP_CODE_RESPONSE, 0,
                                       eap.EAP_TYPE_GTC, '')
            eap_response = yield self.get_peap_response(gtc_packet)
            self.verify_packet(eap_response, type=eap.EAP_TYPE_GTC, data=self.prompt)

        yield self.basic_case_pw()

        # send empty duo factor
        # do it 5 times in a row just for kicks
        for i in range(5):
            gtc_packet = eap.EAPPacket(eap.EAP_CODE_RESPONSE, 0,
                                       eap.EAP_TYPE_GTC, '')
            eap_response = yield self.get_peap_response(gtc_packet)
            self.verify_packet(
                eap_response, type=eap.EAP_TYPE_GTC)
            self.assertTrue(
                self.get_data(eap_response).startswith('Choose a secondary'))

        # make sure everything else finishes successfully
        yield self.basic_case_factor()
        yield self.basic_case_tlv()
        yield self.basic_case_verify()

    @defer.inlineCallbacks
    def test_fail_open(self):
        ''' Test results of an API error when failmode=safe '''
        yield self.setup_peap()

        # Set preauth to raise API error
        self.instance.protocol.failmode = FAILMODE_SAFE
        error = duo_async.DuoAPIFailOpenError()
        self.instance.protocol.client.preauth = mock_preauth(
            None, None, None, error)

        # Send GTC
        gtc_packet = eap.EAPPacket(eap.EAP_CODE_RESPONSE, 0,
                                   eap.EAP_TYPE_GTC, self.password)
        yield self.assert_gtc_packet_is_auth_success(gtc_packet)

    @defer.inlineCallbacks
    def test_fail_open_with_wrong_password(self):
        ''' failmode=safe should not allow bypass of primary auth! '''
        yield self.setup_peap()

        # Set preauth to raise API error
        self.instance.protocol.failmode = FAILMODE_SAFE
        error = duo_async.DuoAPIFailOpenError()
        self.instance.protocol.client.preauth = mock_preauth(
            None, None, None, error)

        # Send GTC
        gtc_packet = eap.EAPPacket(eap.EAP_CODE_RESPONSE, 0,
                                   eap.EAP_TYPE_GTC, self.password + 'DEADBEEF')
        yield self.assert_gtc_packet_is_rejected(gtc_packet)

    @defer.inlineCallbacks
    def test_fail_close(self):
        ''' Test results of an API error when failmode=secure '''
        yield self.setup_peap()

        # Set preauth to raise API error
        self.instance.protocol.failmode = FAILMODE_SECURE
        error = duo_async.DuoAPIFailOpenError()
        self.instance.protocol.client.preauth = mock_preauth(
            None, None, None, error)

        # Send GTC
        gtc_packet = eap.EAPPacket(eap.EAP_CODE_RESPONSE, 0,
                                   eap.EAP_TYPE_GTC, self.password)
        yield self.assert_gtc_packet_is_rejected(gtc_packet)


class TestEapModuleConcat(TestEapModule):
    def get_server_module_config(self):
        extra_config = super(
            TestEapModuleConcat, self).get_server_module_config()
        extra_config['allow_concat'] = 'True'
        return extra_config

    @defer.inlineCallbacks
    def test_basic_case_with_concat(self):
        '''Performs a full successful authentication with concat'''
        yield self.setup_peap()

        # setup mock duo_auth
        self.instance.protocol.duo_auth = mock_duo_auth()

        # send primary password and 'push'; make sure we get success
        gtc_packet = eap.EAPPacket(eap.EAP_CODE_RESPONSE, 0,
                                   eap.EAP_TYPE_GTC, self.password + ',push')

        # assert successful auth
        yield self.assert_gtc_packet_is_auth_success(gtc_packet)

    @defer.inlineCallbacks
    def test_preauth_allow_with_concat(self):
        yield self.setup_peap()

        self.instance.protocol.client.preauth = mock_preauth(
            duo_async.API_RESULT_ALLOW, 'no duo for you', '')

        # send primary password and 'push'; make sure we get success
        gtc_packet = eap.EAPPacket(eap.EAP_CODE_RESPONSE, 0,
                                   eap.EAP_TYPE_GTC, self.password + ',push')

        # assert successful auth
        yield self.assert_gtc_packet_is_auth_success(gtc_packet)

    @defer.inlineCallbacks
    def test_basic_case(self):
        '''Performs a full successful authentication with auto factor'''
        yield self.setup_peap()

        # setup mock duo_auth
        self.instance.protocol.duo_auth = mock_duo_auth(expected_factor=None)

        # send primary password only; make sure we get success
        gtc_packet = eap.EAPPacket(eap.EAP_CODE_RESPONSE, 0,
                                   eap.EAP_TYPE_GTC, self.password)

        # assert successful auth
        yield self.assert_gtc_packet_is_auth_success(gtc_packet)

    @defer.inlineCallbacks
    def test_wrong_password(self):
        """Make sure incorrect password is rejected"""
        yield self.setup_peap()

        # setup mock duo_auth
        m = mock_duo_auth(expected_factor=None)
        self.instance.protocol.duo_auth = m

        # wrong primary password (only); make sure we get rejected
        gtc_packet = eap.EAPPacket(eap.EAP_CODE_RESPONSE, 0,
                                   eap.EAP_TYPE_GTC, self.password + 'DEADBEEF')

        yield self.assert_gtc_packet_is_rejected(gtc_packet)

        # and also, make sure we didn't even attempt to call duo_auth
        self.assertEqual(m.called, 0)


class TestRadiusEapAuthLogging(AuthLoggingTest, TestEapModule):
    @defer.inlineCallbacks
    def test_non_eap_missing_password_failure_logged(self):
        request_packet = packet.AuthPacket(code=packet.AccessRequest,
                                           id=1,
                                           secret=self.secret,
                                           dict=radius_dictionary(),
                                           )
        request_packet['User-Name'] = self.username

        yield self._test_process_packet(request_packet)

        self.assert_logged(
            [
                {'auth_stage': log.AUTH_PRIMARY, 'status': log.AUTH_ERROR}
            ]
        )

    @defer.inlineCallbacks
    def test_rejected_auth_failure_logged(self):
        yield self.setup_peap()
        yield self.basic_case_pw()

        self.instance.protocol.duo_auth = mock_duo_auth()

        gtc_packet = eap.EAPPacket(eap.EAP_CODE_RESPONSE, 0,
                                   eap.EAP_TYPE_GTC, 'sms')

        yield self.assert_gtc_packet_is_rejected(gtc_packet)

        self.assert_logged(
            [
                {'auth_stage': log.AUTH_PRIMARY, 'status': log.AUTH_ALLOW},
                {'auth_stage': log.AUTH_SECONDARY, 'status': log.AUTH_REJECT}
            ]
        )

    @defer.inlineCallbacks
    def test_invalid_passcode_failure_logged(self):
        yield self.setup_peap()
        yield self.basic_case_pw()

        self.instance.protocol.duo_auth = mock_duo_auth()

        gtc_packet = eap.EAPPacket(eap.EAP_CODE_RESPONSE, 0,
                                   eap.EAP_TYPE_GTC, 'XXXXX')

        yield self.assert_gtc_packet_is_rejected(gtc_packet)

        self.assert_logged(
            [
                {'auth_stage': log.AUTH_PRIMARY, 'status': log.AUTH_ALLOW},
                {'auth_stage': log.AUTH_SECONDARY, 'status': log.AUTH_REJECT}
            ]
        )


class TestEapModuleConcatLogging(AuthLoggingTest, TestEapModule):
    def get_server_module_config(self):
        extra_config = super(TestEapModuleConcatLogging, self).get_server_module_config()
        extra_config['allow_concat'] = 'True'
        return extra_config

    @defer.inlineCallbacks
    def test_no_factors_failure_logged(self):
        yield self.setup_peap()

        # Mock preauth to return no factors
        self.instance.protocol.client.preauth = mock_preauth(duo_async.API_RESULT_AUTH, "User must authenticate", None)

        gtc_packet = eap.EAPPacket(eap.EAP_CODE_RESPONSE, 1,
                                   eap.EAP_TYPE_GTC, self.password + ",")
        yield self.assert_gtc_packet_is_rejected(gtc_packet)

        self.assert_logged(
            [
                {'username': self.username, 'auth_stage': log.AUTH_PRIMARY, 'status': log.AUTH_ALLOW},
                {'username': self.username, 'auth_stage': log.AUTH_SECONDARY, 'status': log.AUTH_ERROR}
            ]
        )


class TestEapProtocolLogging(AuthLoggingTest, TestEap):
    @defer.inlineCallbacks
    def test_ssl_error_failure_logged(self):
        def bad_ssl_recv(length):
            raise OpenSSL.SSL.Error()

        yield self.start_peap(self.session, 0)
        self.patch(self.session.ssl_con, 'recv', bad_ssl_recv)
        gtc_packet = eap.EAPPacket(eap.EAP_CODE_RESPONSE, 0, eap.EAP_TYPE_GTC, 'codes here!')
        self.peap_send(self.ssl_con, self.remove_header(gtc_packet.render()), 3)

        self.assert_logged(
            [
                {'auth_stage': 'Unknown', 'status': log.AUTH_ERROR}
            ]
        )

    def test_malformed_eap_packet_failure_logged(self):
        def bad_render(self):
            msg = chr(self.code)
            msg += chr(self.ID)
            if self.type:
                msg += struct.pack('>H', 6 + len(self.data))
                msg += chr(self.type)
            else:
                msg += struct.pack('>H', 4)
            msg += str(self.data)
            return msg

        self.patch(eap.EAPPacket, 'render', bad_render)

        eap_packet = eap.EAPPacket(eap.EAP_CODE_RESPONSE, 0, eap.EAP_TYPE_IDENTITY, 'test')
        self.session.add_message(eap_packet.render())

        self.assert_logged(
            [
                {'auth_stage': log.AUTH_UNKNOWN, 'status': log.AUTH_ERROR}
            ]
        )

    def test_eap_code_request_failure_logged(self):
        eap_packet = eap.EAPPacket(eap.EAP_CODE_REQUEST, 0, eap.EAP_TYPE_GTC, 'gtc code')
        self.session.add_message(eap_packet.render())

        self.assert_logged(
            [
                {'auth_stage': log.AUTH_UNKNOWN, 'status': log.AUTH_ERROR}
            ]
        )

    def test_unknown_nak_request_failure_logged(self):
        nak_pack = eap.EAPPacket(eap.EAP_CODE_RESPONSE, 0, eap.EAP_TYPE_NAK, 13)  # Eap-TLS
        self.session.add_message(nak_pack.render())

        self.assert_logged(
            [
                {'auth_stage': log.AUTH_UNKNOWN, 'status': log.AUTH_ERROR}
            ]
        )

    def test_unsupported_tlv_failure_logged(self):
        tlv_pack = eap.EAPPacket(eap.EAP_CODE_RESPONSE, 0, eap.EAP_TYPE_TLV, True)
        tlv_pack.data = struct.pack('BBBBBB', 128, 12, 0, 2, 0, 1)
        self.session.add_message(tlv_pack.render())

        self.assert_logged(
            [
                {'auth_stage': log.AUTH_UNKNOWN, 'status': log.AUTH_ERROR}
            ]
        )


if __name__ == '__main__':
    import unittest
    unittest.main()
