#
# Copyright (c) 2012 Duo Security
# All Rights Reserved
#
# pylint: disable=protected-access
"""
Test suite for FIPS related code

Running in FIPS mode is determined by passing in a FIPS_MODE environment variable.

Please note that the order of tests in this suite does matter (determinted by alphabetical order,
not location). If we enable FIPS, (intentionally) invoke usage of an unapproved cipher and disable fips
then for some reason we can't re-enable FIPS in subsequent tests. It seems unlikely we would
use this flow anywhere but this test suite.
"""

import unittest
import os
import platform
import random
import mock
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from duoauthproxy.lib import fips_manager
from duoauthproxy.lib.md5_wrapper import Depends


class mockConfig(object):
    def __init__(self, fips_mode):
        self.fips_mode = fips_mode

    def get_main_section_config(self):
        return self

    def get_bool(self, key, default):
        if key == 'fips_mode':
            return self.fips_mode

        return default


class DefaultFipsManagerTestCase(unittest.TestCase):
    """
    The superclass which initializes FipsManagerTestCase.
    Before we execute the test suite for this class, we store
    the FIPS status and then restore it when the test is complete.
    This let's the rest of the code base run in the appropriate
    FIPS mode.
    """
    @classmethod
    def setUpClass(cls):
        cls.original_mode = fips_manager.status()

    @classmethod
    def tearDownClass(cls):
        if cls.original_mode:
            fips_manager.enable()
        else:
            fips_manager._disable()

    def setUp(self):
        fips_manager._disable()

    def tearDown(self):
        fips_manager._disable()


class FipsManagerTestCase(DefaultFipsManagerTestCase):
    def test_get_openssl_version(self):
        self.assertTrue(fips_manager.get_openssl_version().startswith('OpenSSL 1.0.2o-fips'))

    @mock.patch.object(fips_manager.Binding.lib, 'FIPS_mode')
    def test_run_with_fips_flag_enabled(self, fips_mode):
        """
        When the FIPS flag is set, validate that we can successfully enable FIPS mode
        """
        fips_mode.return_value = True
        self.assertEqual(fips_manager.status(), 1)

    @mock.patch.object(fips_manager.Binding.lib, 'FIPS_mode_set')
    @mock.patch.object(fips_manager.Binding.lib, 'ERR_get_error')
    def test_run_with_fips_flag_enabled_failure(self, mockFipsGetError, mockFipsModeSet):
        """
        If we want to enable FIPS but can't for some reason, make sure we raise
        an exception with an openssl error code.
        """
        mockFipsModeSet.return_value = 0
        mockFipsGetError.return_value = 252104805L

        with self.assertRaises(Exception):
            fips_manager.enable()

    @mock.patch.object(fips_manager.Binding.lib, 'FIPS_mode')
    def test_run_with_fips_flag_disabled(self, fips_mode):
        """
        If the fips flag is disabled, then we do not enter FIPS mode.
        This should run irrespective of the value of the FIPS flag.
        """
        fips_mode.return_value = False
        self.assertEqual(fips_manager.status(), 0)

    @mock.patch.object(fips_manager.Binding.lib, 'FIPS_mode_set')
    def test_successfully_disabling_fips(self, fips_mode):
        """
        Validate that invoking the disable FIPS option works.
        This should run irrespective of the value of the FIPS flag.

        We don't usually test private functions but since this function is used by
        this test suite as a teardown step, it's probably good to test it.
        """
        fips_mode.return_value = 1
        self.assertEqual(fips_manager._disable(), 1)
        self.assertEqual(fips_manager.status(), 0)

    def test_run_with_fips_using_unapproved_cipher(self):
        """
        If we are in FIPS mode and we try to load an unapproved cipher than an
        exception should be raised saying the cipher has been disabled for FIPS.
        """
        hashes.Hash(hashes.MD5(), backend=default_backend())

        fips_manager.enable()

        with self.assertRaisesRegexp(BaseException, "disabled for fips"):
            hashes.Hash(hashes.MD5(), backend=default_backend())

    def test_md5_implementations_should_be_wrapped(self):
        import hashlib
        self.assertEqual(hashlib.new.__name__, "md5_with_safe_guard")
        self.assertEqual(hashlib.md5.__name__, "md5_with_safe_guard")

        import _hashlib
        self.assertEqual(_hashlib.new.__name__, "md5_with_safe_guard")
        self.assertEqual(_hashlib.openssl_md5.__name__, "md5_with_safe_guard")

    def test_dont_allow_md5_when_missing_security_guard_and_in_fips_mode(self):
        fips_manager.enable()

        import hashlib
        with self.assertRaisesRegexp(ValueError, "Must pass used_for_security"):
            hashlib.new('md5', 'hello').digest()

        with self.assertRaisesRegexp(ValueError, "Must pass used_for_security"):
            hashlib.md5('hello').digest()

        import _hashlib
        with self.assertRaisesRegexp(ValueError, "Must pass used_for_security"):
            _hashlib.new('md5', 'hello').digest()

        with self.assertRaisesRegexp(ValueError, "Must pass used_for_security"):
            _hashlib.openssl_md5('hello').digest()

    def test_dont_allow_md5_when_used_for_security_and_in_fips_mode(self):
        fips_manager.enable()

        import hashlib
        with self.assertRaisesRegexp(ValueError, "Must pass used_for_security"):
            hashlib.md5('hello', used_for_security=True).digest()

        with self.assertRaisesRegexp(ValueError, "Must pass used_for_security"):
            hashlib.new('md5', 'hello', used_for_security=True).digest()

        import _hashlib
        with self.assertRaisesRegexp(ValueError, "Must pass used_for_security"):
            _hashlib.new('md5', 'hello', used_for_security=True).digest()

        with self.assertRaisesRegexp(ValueError, "Must pass used_for_security"):
            _hashlib.openssl_md5('hello', used_for_security=True).digest()

    def test_allow_md5_when_not_in_fips_mode(self):
        import hashlib
        self.assertEqual(hashlib.md5('hello').hexdigest(), "5d41402abc4b2a76b9719d911017c592")
        self.assertEqual(hashlib.new('md5', 'hello').hexdigest(), "5d41402abc4b2a76b9719d911017c592")

        import _hashlib
        self.assertEqual(_hashlib.new('md5', 'hello').hexdigest(), "5d41402abc4b2a76b9719d911017c592")
        self.assertEqual(_hashlib.openssl_md5('hello').hexdigest(), "5d41402abc4b2a76b9719d911017c592")

    def test_allow_md5_used_for_security_false_and_in_fips_mode(self):
        fips_manager.enable()

        import hashlib

        self.assertEqual(hashlib.md5('hello', used_for_security=False).hexdigest(), "5d41402abc4b2a76b9719d911017c592")
        self.assertEqual(hashlib.new('md5', 'hello', used_for_security=False).hexdigest(), "5d41402abc4b2a76b9719d911017c592")

        import _hashlib
        self.assertEqual(_hashlib.new('md5', 'hello', used_for_security=False).hexdigest(), "5d41402abc4b2a76b9719d911017c592")
        self.assertEqual(_hashlib.openssl_md5('hello', used_for_security=False).hexdigest(), "5d41402abc4b2a76b9719d911017c592")

    def test_allow_md5_used_for_security_depends_and_in_fips_mode(self):
        fips_manager.enable()

        import hashlib

        self.assertEqual(hashlib.md5('hello', used_for_security=Depends).hexdigest(), "5d41402abc4b2a76b9719d911017c592")
        self.assertEqual(hashlib.new('md5', 'hello', used_for_security=Depends).hexdigest(), "5d41402abc4b2a76b9719d911017c592")

        import _hashlib
        self.assertEqual(_hashlib.new('md5', 'hello', used_for_security=Depends).hexdigest(), "5d41402abc4b2a76b9719d911017c592")
        self.assertEqual(_hashlib.openssl_md5('hello', used_for_security=Depends).hexdigest(), "5d41402abc4b2a76b9719d911017c592")

    def test_no_safe_guard_for_non_md5_digests(self):
        """
        We wrap `hashlib.new` and `_hashlib.new` but only require `used_for_security` if the digest being
        passed in is `md5`. This test ensures that other digests are not required to use the
        `used_for_security` keyword.
        """
        fips_manager.enable()

        import hashlib
        self.assertEqual(hashlib.new('sha1', 'hello').hexdigest(), "aaf4c61ddcc5e8a2dabede0f3b482cd9aea9434d")

    def test_md5_c_module_cannot_be_imported(self):
        """
        We rely on OpenSSL's MD5 and don't build the built-in MD5 implementation
        """
        with self.assertRaises(ImportError):
            import _md5  # noqa: F401

    @unittest.skipIf(platform.system() == 'Windows', "Windows can use OS RNG")
    def test_random_engine_is_openssl_in_fips_mode(self):
        self.assertTrue(fips_manager.OPENSSL_LIB.ENGINE_get_default_RAND() != fips_manager.OPENSSL_FFI.NULL)
        fips_manager.enable()
        self.assertTrue(fips_manager.OPENSSL_LIB.ENGINE_get_default_RAND() == fips_manager.OPENSSL_FFI.NULL)


class WrappingRandom(unittest.TestCase):
    @unittest.skipIf(platform.system() != 'Windows', "Windows only check")
    def test_os_urandom_not_wrapped_on_windows(self):
        self.assertNotEquals(os.urandom.__name__, '_duo_urandom_new_wrapper')

    @unittest.skipIf(platform.system() != 'Windows', "Windows only check")
    def test_random_module_not_wrapped_on_windows(self):
        self.assertNotEquals(random._urandom.__name__, '_duo_urandom_new_wrapper')

    @unittest.skipIf(platform.system() == 'Windows', "Windows doesn't use this wrapper")
    def test_os_urandom_references_our_urandom(self):
        """
        os.urandom becomes the base of all other PNRGs after our patches, so we
        have to make sure it is using our FIPS approved method
        """
        self.assertEquals(os.urandom.__name__, '_duo_urandom_new_wrapper')
        self.assertTrue(os.urandom(1))

    @unittest.skipIf(platform.system() == 'Windows', "Windows doesn't use this wrapper")
    def test_random_module_references_our_urandom(self):
        """
        Make sure the random module imported our implementation of urandom.
        random.SystemRandom uses this so make sure we can invoke it.
        """
        self.assertEquals(random._urandom.__name__, '_duo_urandom_new_wrapper')
        self.assertTrue(random.SystemRandom().random())

    @unittest.skipIf(platform.system() == 'Windows', "Windows doesn't use this wrapper")
    def test_random_classes_overwritten_with_SystemRandom(self):
        """Classes in the random module should be aliases of SystemRandom"""
        classes = [(name, value) for name, value in random.__dict__.items()
                   # Ignore private things and non-classes
                   if isinstance(value, type) and not name.startswith('_')]

        for name, klass in classes:
            self.assertEquals(
                klass,
                random.SystemRandom,
                "random.{} not overwritten".format(name))

    @unittest.skipIf(platform.system() == 'Windows', "Windows doesn't use this wrapper")
    def test_functions_are_overwritten(self):
        """Functions in the random module should come from SystemRandom"""
        functions = []
        for name, value in random.__dict__.items():
            if not name.startswith('_') and callable(value) and not isinstance(value, type):
                functions.append((name, value))

        for name, function in functions:
            source = function.__self__.__class__
            self.assertEquals(
                source,
                random.SystemRandom,
                "random.{} not overwritten".format(name))


if __name__ == '__main__':
    unittest.main()
