import unittest

import os
import platform
from multiprocessing import Pipe
from cryptography.hazmat.bindings.openssl import binding

_openssl_binding = binding.Binding()


def _rand_bytes():
    buf = _openssl_binding.ffi.new("unsigned char[]", [0] * 8)
    _openssl_binding.lib.RAND_bytes(buf, 8)
    return _openssl_binding.ffi.buffer(buf)[:]


class ForkSafetyTests(unittest.TestCase):
    # OpenSSL by default uses a userland PRNG whose seed is copied to the child process
    # after a fork. These tests ensure that the same random bytes are not generated
    # by both a parent and a child.

    @unittest.skipIf(platform.system() == 'Windows', "Windows doesn't use this wrapper")
    def test_forksafety_simple(self):
        """Compares random bytes generated by a parent and its first child"""
        # Note that this successful output from this test does not indicate that our
        # fork safety mitigation is working correctly. This tests a much simpler case
        # and should pass with any sane unpatched OpenSSL.
        # Make sure to run test_forksafety_pidloop if you are concerned about the mitigation.

        # Initialize RNG
        _rand_bytes()

        parent_conn, child_conn = Pipe()

        pid = os.fork()
        if pid == 0:  # first child
            parent_conn.close()
            data = _rand_bytes()
            child_conn.send(data)
            child_conn.close()
            os._exit(0)

        parent_data = _rand_bytes()
        client_data = parent_conn.recv()
        child_conn.close()
        parent_conn.close()

        self.assertNotEqual(parent_data, client_data)

    @unittest.skip("Takes a looong time and forkbombs the system")
    def test_forksafety_pidloop(self):
        """Compares random bytes generated by two different processes that share the same PID"""

        # A popular mitigation for fork-unsafety is to mix in the pid.
        # This test creates processes until we get one with the
        # same PID as the parent process.
        # This mitigation is used as of OpenSSL 1.0.2o non-FIPS and this
        # test will fail without our own additional mitigations.

        # This test should pass for a couple of reasons:
        #  - When not in FIPS mode, we use the OS random number generation facilities
        #  - FIPS mode OpenSSL's RNG mixes in a timestamp
        #  - In FIPS mode, we use a wrapper around os.fork() that feeds in /dev/urandom
        # NOTE: Unmitigated FIPS mode OpenSSL WILL PASS this test, so test w/ FIPS mode off

        # Initialize the RNG
        _rand_bytes()

        parent_conn, child_conn = Pipe()

        xpid = 0
        pid = os.fork()
        # We have to do the bulk of the test from within a child because Linux
        # won't give a child process the PID of the process group leader
        # (which the first parent process probably is)
        if pid == 0:  # first child
            parent_conn.close()
            data = _rand_bytes()
            child_conn.send(data)
            child_conn.close()
            os._exit(0)
        elif pid > 0:
            os.wait()
            while pid != xpid:
                xpid = os.fork()
                if xpid == 0:
                    parent_conn.close()
                    if os.getpid() == pid:
                        # We've looped around pids and have ended
                        # up w/ the same pid as the first child!
                        data = _rand_bytes()
                        child_conn.send(data)
                        child_conn.close()
                    os._exit(0)
                elif xpid > 0:  # Parent
                    os.wait()

            parent_bytes = parent_conn.recv()
            child_bytes = parent_conn.recv()
            child_conn.close()
            parent_conn.close()

            self.assertNotEqual(parent_bytes, child_bytes)


if __name__ == "__main__":
    unittest.main()
