#
# Copyright (c) 2018 Duo Security
# All Rights Reserved
#

import netaddr

from duoauthproxy.lib import ip_util

from base import TestCase


class TestValidIP(TestCase):
    def test_valid_ipv4(self):
        for x in ("12.0.40.1",
                  "10.0.0.0"):
            self.assertTrue(ip_util.is_valid_ip(x), x)

    def test_invalid_ipv4(self):
        for x in ("12.",
                  "10.10",
                  "0000005e5537"):
            self.assertFalse(ip_util.is_valid_ip(x), x)

    def test_valid_ipv6(self):
        for x in ("fe80::fc54:ff:fe2e:d505",
                  "::ffff:192.1.56.12",
                  "1050:0:0:0:5:600:300c:326"):
            self.assertTrue(ip_util.is_valid_ip(x), x)

    def test_valid_range_ip(self):
        for x in ("3.3.3.3-3.3.3.6",
                  "127.0.0.1-128.0.0.1"):
            self.assertTrue(ip_util.is_valid_ip(x), x)

    def test_invalid_range_ip(self):
        for x in ("3.3.3.20-3.3.3.6",
                  "260.0.0.0-127.0.0.1"):
            self.assertFalse(ip_util.is_valid_ip(x), x)

    def test_valid_cidr_ip(self):
        for x in ("127.0.0.1/32",
                  "127.0.0.1/255.255.255.0"):
            self.assertTrue(ip_util.is_valid_ip(x), x)

    def test_invalid_cidr_ip(self):
        for x in ("127.0.0.1/128",
                  "260.0.0.0/16"):
            self.assertFalse(ip_util.is_valid_ip(x), x)

    def test_empty_ip_string(self):
        self.assertFalse(ip_util.is_valid_ip(''))


class TestIpInNetwork(TestCase):
    def setUp(self):
        self.ip_in = "1.1.1.1"
        self.ip_out = "4.0.0.1"

        self.networks = [
            netaddr.IPNetwork("1.1.1.0/24"),
            netaddr.IPNetwork("2.2.2.2/32"),
            netaddr.IPNetwork("3.3.0.0/16")
        ]

    def test_empty_networks_failure(self):
        networks = []

        result = ip_util.is_ip_in_networks(self.ip_in, networks)

        self.assertFalse(result, 'IP should not be in an empty list of networks')

    def test_ip_in_networks_success(self):
        result = ip_util.is_ip_in_networks(self.ip_in, self.networks)

        self.assertTrue(result, "{0} should be in the networks".format(self.ip_in))

    def test_ip_in_networks_failure(self):
        result = ip_util.is_ip_in_networks(self.ip_out, self.networks)

        self.assertFalse(result, "{0} should not be in the networks".format(self.ip_out))

    def test_ip_address_type_success(self):
        ip = netaddr.IPAddress(self.ip_in)

        result = ip_util.is_ip_in_networks(ip, self.networks)

        self.assertTrue(result, "{0} should be in the networks.".format(ip))


if __name__ == '__main__':
    import unittest
    unittest.main()
