#
# Copyright (c) 2019 Duo Security
# All Rights Reserved
#
import operator

from base import TestCase
from duoauthproxy.lib import config_provider, ldap
from duoauthproxy.modules import ldap_drpc_plugin
from mock_ldap_server import ca_cert_pem, make_server_factory, MEGA_GROUP_MEMBERS
import drpc.v2 as drpc

from ldaptor.protocols import pureldap
from ldaptor.protocols.ldap import ldaperrors
from twisted.internet import defer
from twisted.python import log


class LdapMethodTestBase(TestCase):
    is_logging_insecure = 'False'
    listen_ldaps = False

    @defer.inlineCallbacks
    def setUp(self):
        # Set up a LDAP server to talk to
        self.server_port = 10389
        (self.server_factory,
         self.server_listener) = yield make_server_factory(
            self.server_port, ssl=self.listen_ldaps)

        self.password = 'password!'

        config_settings = {
            'debug': 'True',
            'is_logging_insecure': self.is_logging_insecure,
            'service_account_username': 'CN=adminer adminer,CN=Users,DC=bongos,DC=test',
            'service_account_password': self.password,
        }
        config = config_provider.ConfigDict(config_settings)

        self.ldap_plugin = ldap_drpc_plugin.LdapDrpcPlugin(config)

    @defer.inlineCallbacks
    def tearDown(self):
        yield self.server_listener.stopListening()

    def canonicalize_ldap_obj(self, obj):
        for k in ('objectclass', 'objectcategory', 'member'):
            if k in obj:
                obj[k].sort()

    def assert_search_results_equal(self, expected, actual):
        expected = list(sorted(expected,
                               key=operator.itemgetter('distinguishedname')))
        actual = list(sorted(actual,
                             key=operator.itemgetter('distinguishedname')))
        for i, obj in enumerate(actual):
            self.canonicalize_ldap_obj(expected[i])
            self.canonicalize_ldap_obj(obj)
            self.assertEqual(expected[i], obj)
        self.assertEqual(len(expected), len(actual))

    def do_ldap_search(self, **kwargs):
        """
        Call the ldap_search RPC endpoint.

        This is not wrapped by ServerProtocol.recv_msg (because
        there's no protocol ;) so it will either return what would
        become the value of 'result' in the DRPC result message or
        raise an exception. It will never return a full result
        message.

        MoreADLikeLDAPServer doesn't support the pagination controls,
        so most tests need pagination=cloud.PAGINATION_TYPE_DISABLE
        for now. Only omit if overriding handle_LDAPSearchRequest.
        """
        default_kwargs = {
            'host': 'localhost',
            'port': self.server_port,
            'base_dn': 'DC=bongos,DC=test',
            # MoreADLikeLDAPServer isn't AD-like enough for NTLM.
            'auth_type': 'plain',
            'ssl_verify_hostname': False,
            'transport_type': 'clear',
        }
        default_kwargs.update(kwargs)
        return self.ldap_plugin.do_ldap_search(**default_kwargs)

    def do_ldap_authentication(self, **kwargs):
        """
        Call the ldap_authentication RPC endpoint.
        """
        default_kwargs = {
            'host': 'localhost',
            'port': self.server_port,
            'auth_type': 'plain',
            'ssl_verify_hostname': False,
            'transport_type': 'clear',
        }
        default_kwargs.update(kwargs)

        return self.ldap_plugin.do_ldap_authentication(**default_kwargs)


class TestLdapDrpcPlugin(LdapMethodTestBase):
    @defer.inlineCallbacks
    def test_ldap_authenticate_right_password(self):
        res = yield self.do_ldap_authentication(
            username='jane',
            password='password',
            base_dn='CN=Users,DC=bongos,DC=test',
        )
        self.assertTrue(res['success'])
        self.assertEqual(res['msg'], ldap_drpc_plugin.AUTH_SUCCEEDED)

    @defer.inlineCallbacks
    def test_ldap_authenticate_wrong_password(self):
        res = yield self.do_ldap_authentication(
            username='jane',
            password='password!!',
            base_dn='CN=Users,DC=bongos,DC=test',
        )
        self.assertFalse(res['success'])
        self.assertEqual(res['msg'], ldap_drpc_plugin.AUTH_FAILED)

    @defer.inlineCallbacks
    def test_ldap_authenticate_invalid_user(self):
        res = yield self.do_ldap_authentication(
            username='jane!!',
            password='password',
            base_dn='CN=Users,DC=bongos,DC=test',
        )
        self.assertFalse(res['success'])
        self.assertEqual(res['msg'], ldap_drpc_plugin.AUTH_INVALID_USER)

    @defer.inlineCallbacks
    def test_ldap_search_simple(self):
        res = yield self.do_ldap_search(
            pagination=ldap_drpc_plugin.PAGINATION_TYPE_DISABLE,
        )
        expected = [
            {'cn': ['Users'],
             'distinguishedname': ['DC=bongos,DC=test'],
             'instancetype': ['4'],
             'iscriticalsystemobject': ['TRUE'],
             'name': ['Users'],
             'objectcategory': ['CN=Container,CN=Schema,CN=Configuration,DC=bongos,DC=test'],
             'objectclass': ['top', 'container'],
             'objectguid': ['\\82h\\94\\e1\\d1(\\aaD\\b5\\0b:\\e0~\\cc\\a1\\b8'],
             'systemflags': ['-1946157056']},
            {'cn': ['Mega Users,+"\\5c<>;'],
             'distinguishedname': ['CN=Mega Users\\5c,\\5c+\\5c"\\5c\\5c<\\5c>\\5c;,DC=bongos,DC=test'],
             'instancetype': ['4'],
             'iscriticalsystemobject': ['TRUE'],
             'member': sorted(MEGA_GROUP_MEMBERS),
             'name': ['Mega Users,+"\\5c<>;'],
             'objectcategory': ['CN=Container,CN=Schema,CN=Configuration,DC=bongos,DC=test'],
             'objectclass': ['top', 'container'],
             'objectguid': ['\\9a(8Q\\da*\\12MBK:\\df\\82\\c8r\\02'],
             'systemflags': ['-1946157056']},
            {'cn': ['Users'],
             'distinguishedname': ['CN=Users,DC=bongos,DC=test'],
             'instancetype': ['4'],
             'iscriticalsystemobject': ['TRUE'],
             'name': ['Users'],
             'objectcategory': ['CN=Container,CN=Schema,CN=Configuration,DC=bongos,DC=test'],
             'objectclass': ['top', 'container'],
             'objectguid': ['\\82h\\94\\e1\\d1(\\aaD\\b5\\0b:\\e0~\\cc\\a1\\b8'],
             'systemflags': ['-1946157056']},
            {'cn': ['Users'],
             'distinguishedname': ['DC=subdomain,DC=bongos,DC=test'],
             'instancetype': ['4'],
             'iscriticalsystemobject': ['TRUE'],
             'name': ['Users'],
             'objectcategory': ['CN=Container,CN=Schema,CN=Configuration,DC=bongos,DC=test'],
             'objectclass': ['container', 'top'],
             'objectguid': ['\\82h\\94\\e1\\d1(\\aaD\\b5\\0b:\\e0~\\cc\\a1\\b9'],
             'systemflags': ['-1946157056']},
            {'cn': ['Users'],
             'distinguishedname': ['CN=Users,DC=subdomain,DC=bongos,DC=test'],
             'instancetype': ['4'],
             'iscriticalsystemobject': ['TRUE'],
             'name': ['Users'],
             'objectcategory': ['CN=Container,CN=Schema,CN=Configuration,DC=bongos,DC=test'],
             'objectclass': ['container', 'top'],
             'objectguid': ['\\82h\\94\\e1\\d1(\\aaD\\b5\\0b:\\e0~\\cc\\a1\\b9'],
             'systemflags': ['-1946157056']},
            {'cn': ['adminer adminer'],
             'codepage': ['0'],
             'countrycode': ['0'],
             'displayname': ['adminer adminer'],
             'distinguishedname': ['CN=adminer adminer,CN=Users,DC=bongos,DC=test'],
             'givenname': ['adminer'],
             'instancetype': ['4'],
             'msds-principalname': ['bongos\\5cadminer'],
             'name': ['adminer adminer'],
             'objectcategory': ['CN=Person,CN=Schema,CN=Configuration,DC=bongos,DC=test', 'person'],
             'objectclass': ['person', 'top', 'user', 'organizationalPerson'],
             'objectguid': ['Kv\\cf\\f1t\\eb\\ddK\\b6\\ae\\90#\\c3\\eddq'],
             'objectsid': ['\\01\\05\\00\\00\\00\\00\\00\\05\\15\\00\\00\\00\\b2\\b44\\ec+\\da\\90\\8d}\\11\\97@W\\04\\00\\00'],
             'primarygroupid': ['513'],
             'samaccountname': ['adminer'],
             'sn': ['adminer'],
             'useraccountcontrol': ['66048'],
             'userprincipalname': ['adminer@bongos.test']},
            {'cn': ['Jane Blackwood'],
             'codepage': ['0'],
             'countrycode': ['0'],
             'displayname': ['Jane Blackwood'],
             'distinguishedname': ['CN=Jane Blackwood,CN=Users,DC=bongos,DC=test'],
             'givenname': ['Jane'],
             'instancetype': ['4'],
             'msds-principalname': ['bongos\\5cjane'],
             'name': ['Jane Blackwood'],
             'memberof':['CN=Users,DC=bongos,DC=test'],
             'objectcategory': ['user', 'person'],
             'objectclass': ['person', 'top', 'user', 'organizationalPerson'],
             'objectguid': ['\\91\\b0\\b7{Ds\\a2E\\99\\f7\\1a[\\17!_\\89'],
             'objectsid': ['\\01\\05\\00\\00\\00\\00\\00\\05\\15\\00\\00\\00\\b2\\b44\\ec+\\da\\90\\8d}\\11\\97@U\\04\\00\\00'],
             'primarygroupid': ['513'],
             'samaccountname': ['jane'],
             'sn': ['Blackwood'],
             'useraccountcontrol': ['66048'],
             'userprincipalname': ['jane@bongos.test']},
            {'cn': ['Fred Person'],
             'codepage': ['0'],
             'countrycode': ['0'],
             'displayname': ['Fred Person'],
             'distinguishedname': ['CN=Fred Person,CN=Users,DC=bongos,DC=test'],
             'givenname': ['Fred'],
             'instancetype': ['4'],
             'msds-principalname': ['bongos\\5cfred'],
             'name': ['Fred Person'],
             'objectcategory': ['CN=Person,CN=Schema,CN=Configuration,DC=bongos,DC=test', 'person'],
             'objectclass': ['person', 'top', 'user', 'organizationalPerson'],
             'objectguid': ['\\89\\ce\\02\\c7_YgL\\adtZ\\84\\16\\0b\\be/'],
             'objectsid': ['\\01\\05\\00\\00\\00\\00\\00\\05\\15\\00\\00\\00\\b2\\b44\\ec+\\da\\90\\8d}\\11\\97@T\\04\\00\\00'],
             'primarygroupid': ['513'],
             'samaccountname': ['fred'],
             'sn': ['Person'],
             'useraccountcontrol': ['66048'],
             'userprincipalname': ['fred@bongos.test']},
            {'cn': ['Dupe User 1'],
             'codepage': ['0'],
             'countrycode': ['0'],
             'displayname': ['Dupe User 1'],
             'distinguishedname': ['CN=Dupe User 1,CN=Users,DC=subdomain,DC=bongos,DC=test'],
             'givenname': ['Dupe'],
             'instancetype': ['4'],
             'msds-principalname': ['subdomain\\5cdupe'],
             'name': ['Dupe User 1'],
             'objectcategory': ['person', 'user'],
             'objectclass': ['organizationalPerson', 'person', 'top', 'user'],
             'objectguid': ['\\91\\b0\\b7{Ds\\a2E\\99\\f7\\1a[\\17!_\\89'],
             'objectsid': ['\\01\\05\\00\\00\\00\\00\\00\\05\\15\\00\\00\\00\\b2\\b44\\ec+\\da\\90\\8d}\\11w@U\\04\\00\\00'],
             'primarygroupid': ['513'],
             'samaccountname': ['dupe'],
             'sn': ['User'],
             'useraccountcontrol': ['66048'],
             'userprincipalname': ['dupe@subdomain.bongos.test']},
            {'cn': ['Dupe User 1'],
             'codepage': ['0'],
             'countrycode': ['0'],
             'displayname': ['Dupe User 2'],
             'distinguishedname': ['CN=Dupe User 2,CN=Users,DC=bongos,DC=test'],
             'givenname': ['Dupe'],
             'instancetype': ['4'],
             'msds-principalname': ['bongos\\5cdupe'],
             'name': ['Dupe User 2'],
             'objectcategory': ['person', 'user'],
             'objectclass': ['organizationalPerson', 'person', 'top', 'user'],
             'objectguid': ['\\91\\b0\\b7{Ds\\a2E\\99\\f7\\1a[\\17!_\\89'],
             'objectsid': ['\\01\\05\\00\\00\\00\\00\\00\\05\\15\\00\\00\\00\\b2\\b44\\ec+\\da\\90\\8d}\\11w@\\15\\04\\00\\00'],
             'primarygroupid': ['513'],
             'samaccountname': ['dupe'],
             'sn': ['User'],
             'useraccountcontrol': ['66048'],
             'userprincipalname': ['dupe@bongos.test']},
            {'cn': ['Principal Skinner'],
             'codepage': ['0'],
             'countrycode': ['0'],
             'displayname': ['Principal Skinner'],
             'distinguishedname': ['CN=Principal Skinner,CN=Users,DC=bongos,DC=test'],
             'givenname': ['Principal'],
             'instancetype': ['4'],
             'msds-principalname': ['S-1-336566957-3618751888-1237994806-1297237956-58068105-1020576660'],
             'name': ['Principal Skinner'],
             'objectcategory': ['CN=Person,CN=Schema,CN=Configuration,DC=bongos,DC=test', 'person'],
             'objectclass': ['person', 'top', 'user', 'organizationalPerson'],
             'primarygroupid': ['513'],
             'samaccountname': ['pskinner'],
             'sn': ['pskinner'],
             'useraccountcontrol': ['66048'],
             'userprincipalname': ['pskinner@bongos.test']},
            {'cn': ['Non AD'],
             'distinguishedname': ['CN=Non AD,CN=Users,DC=bongos,DC=test'],
             'givenname': ['Non'],
             'objectcategory': ['CN=Person,CN=Schema,CN=Configuration,DC=bongos,DC=test', 'person'],
             'objectclass': ['person', 'top', 'user', 'organizationalPerson'],
             'sn': ['AD']},
            {'cn': ['OpenLDAPer'],
             'distinguishedname': ['CN=OpenLDAPer,CN=Users,DC=bongos,DC=test'],
             'objectclass': ['person', 'top', 'organizationalPerson'],
             'sn': ['OpenLDAPer']},
        ]
        self.assert_search_results_equal(expected, res['results'])

    @defer.inlineCallbacks
    def test_ldap_search_dn_escapesies(self):
        res = yield self.do_ldap_search(
            filter_text='(cn=Mega Users,+"\\5c<>;)',
            attributes=['member'],
            pagination=ldap_drpc_plugin.PAGINATION_TYPE_DISABLE,
        )

        expected = [
            {
                'distinguishedname': [
                    'CN=Mega Users\\5c,\\5c+\\5c"\\5c\\5c\\5c<\\5c>\\5c;,DC=bongos,DC=test'
                ],
                'member': sorted(MEGA_GROUP_MEMBERS)
            }
        ]
        self.assert_search_results_equal(expected, res['results'])

    @defer.inlineCallbacks
    def test_ldap_search_filter_text_and_attributes(self):
        res = yield self.do_ldap_search(
            filter_text='(objectClass=person)',
            attributes=['objectclass', 'sN'],
            pagination=ldap_drpc_plugin.PAGINATION_TYPE_DISABLE,
        )

        expected = sorted([
            {'distinguishedname': ['CN=adminer adminer,CN=Users,DC=bongos,DC=test'],
             'objectclass': ['person', 'top', 'user', 'organizationalPerson'],
             'sn': ['adminer']},
            {'distinguishedname': ['CN=Principal Skinner,CN=Users,DC=bongos,DC=test'],
             'objectclass': ['person', 'top', 'user', 'organizationalPerson'],
             'sn': ['pskinner']},
            {'distinguishedname': ['CN=Non AD,CN=Users,DC=bongos,DC=test'],
             'objectclass': ['person', 'top', 'user', 'organizationalPerson'],
             'sn': ['AD']},
            {'distinguishedname': ['CN=OpenLDAPer,CN=Users,DC=bongos,DC=test'],
             'objectclass': ['person', 'top', 'organizationalPerson'],
             'sn': ['OpenLDAPer']},
            {'distinguishedname': ['CN=Dupe User 2,CN=Users,DC=bongos,DC=test'],
             'objectclass': ['person', 'top', 'user', 'organizationalPerson'],
             'sn': ['User']},
            {'distinguishedname': ['CN=Dupe User 1,CN=Users,DC=subdomain,DC=bongos,DC=test'],
             'objectclass': ['person', 'top', 'user', 'organizationalPerson'],
             'sn': ['User']},
            {'distinguishedname': ['CN=Jane Blackwood,CN=Users,DC=bongos,DC=test'],
             'objectclass': ['person', 'top', 'user', 'organizationalPerson'],
             'sn': ['Blackwood']},
            {'distinguishedname': ['CN=Fred Person,CN=Users,DC=bongos,DC=test'],
             'objectclass': ['person', 'top', 'user', 'organizationalPerson'],
             'sn': ['Person']},
        ], key=operator.itemgetter('distinguishedname'))
        self.assert_search_results_equal(expected, res['results'])

    @defer.inlineCallbacks
    def test_ldap_search_server_error(self):
        """
        If the LDAP server responds to the search with the error, the RPC
        caller should get an error instead of an empty result list.
        """
        def handle_LDAPSearchRequest(request, controls, reply):
            return pureldap.LDAPSearchResultDone(
                resultCode=ldaperrors.LDAPOperationsError.resultCode)
        self.server_factory.handle_LDAPSearchRequest = handle_LDAPSearchRequest

        with self.assertRaisesRegexp(
                drpc.CallError, ldap_drpc_plugin.ERR_LDAP_SEARCH_FAILED):
            yield self.do_ldap_search()

    @defer.inlineCallbacks
    def test_ldap_search_attribute_escaping(self):
        """
        Attribute values in search results must be escaped with escape_bytes().
        """
        res = yield self.do_ldap_search(
            pagination=ldap_drpc_plugin.PAGINATION_TYPE_DISABLE,
            filter_text=('(cn=Mega Users*)'),
        )
        mu = res['results'][0]

        # All attributes from the search result entry must be escaped.
        self.assertEqual('Mega Users,+"\\5c<>;', mu['cn'][0])

        # DN is a special case (copied out of the search result entry
        # even if it wasn't in the requested attribute list) so be
        # sure to test it, too.
        self.assertEqual(
            'CN=Mega Users\\5c,\\5c+\\5c"\\5c\\5c<\\5c>\\5c;,DC=bongos,DC=test',
            mu['distinguishedname'][0])

    @defer.inlineCallbacks
    def test_backing_response_timeout(self):
        self.server_factory.response_delay = 2
        with self.assertRaisesRegexp(drpc.CallError, ldap_drpc_plugin.ERR_LDAP_TIMEOUT):
            yield self.do_ldap_search(timeout=1)
        # Wait for cleanup of artificially delayed requests.
        yield self.server_factory.delayed

    @defer.inlineCallbacks
    def test_null_ssl_ca_certs(self):
        res = yield self.do_ldap_search(
            ssl_ca_certs=None,
            ssl_verify_hostname=False,
            transport_type=ldap.client.AD_TRANSPORT_CLEAR,
            pagination=ldap_drpc_plugin.PAGINATION_TYPE_DISABLE,
        )
        self.assertIsNot(None, res['results'])

    @defer.inlineCallbacks
    def test_empty_ssl_ca_certs(self):
        res = yield self.do_ldap_search(
            ssl_ca_certs='',
            ssl_verify_hostname=False,
            transport_type=ldap.client.AD_TRANSPORT_CLEAR,
            pagination=ldap_drpc_plugin.PAGINATION_TYPE_DISABLE,
        )
        self.assertIsNot(None, res['results'])

    @defer.inlineCallbacks
    def test_invalid_page_size(self):
        for v in (-1, 0, None):
            with self.assertRaisesRegexp(drpc.CallBadArgError,
                                         'page_size'):
                yield self.do_ldap_search(
                    page_size=v,
                )

    @defer.inlineCallbacks
    def test_invalid_max_result_size(self):
        for v in (-1, 0):
            with self.assertRaisesRegexp(drpc.CallBadArgError,
                                         'max_result_size'):
                yield self.do_ldap_search(
                    max_result_size=v,
                )

    @defer.inlineCallbacks
    def test_max_result_size(self):
        # so this doesn't test the case where paged_search() bails
        # early when a cookie is still active.
        res = yield self.do_ldap_search(
            pagination=ldap_drpc_plugin.PAGINATION_TYPE_DISABLE,
            max_result_size=1,
            attributes=('cn',),
        )
        expected = [
            {'distinguishedname': ['DC=bongos,DC=test'], 'cn': ['Users']},
            {'distinguishedname': ['CN=Mega Users\\5c,\\5c+\\5c"\\5c\\5c\\5c<\\5c>\\5c;,DC=bongos,DC=test'], 'cn': ['Mega Users,+"\\5c<>;']},
            {'distinguishedname': ['CN=Users,DC=bongos,DC=test'], 'cn': ['Users']},
            {'distinguishedname': ['CN=adminer adminer,CN=Users,DC=bongos,DC=test'], 'cn': ['adminer adminer']},
            {'distinguishedname': ['CN=Dupe User 2,CN=Users,DC=bongos,DC=test'], 'cn': ['Dupe User 1']},
            {'distinguishedname': ['CN=Jane Blackwood,CN=Users,DC=bongos,DC=test'], 'cn': ['Jane Blackwood']},
            {'distinguishedname': ['CN=Fred Person,CN=Users,DC=bongos,DC=test'], 'cn': ['Fred Person']},
            {'distinguishedname': ['DC=subdomain,DC=bongos,DC=test'], 'cn': ['Users']},
            {'distinguishedname': ['CN=Users,DC=subdomain,DC=bongos,DC=test'], 'cn': ['Users']},
            {'distinguishedname': ['CN=Dupe User 1,CN=Users,DC=subdomain,DC=bongos,DC=test'], 'cn': ['Dupe User 1']},
            {'distinguishedname': ['CN=Principal Skinner,CN=Users,DC=bongos,DC=test'], 'cn': ['Principal Skinner']},
            {'distinguishedname': ['CN=Non AD,CN=Users,DC=bongos,DC=test'], 'cn': ['Non AD']},
            {'distinguishedname': ['CN=OpenLDAPer,CN=Users,DC=bongos,DC=test'], 'cn': ['OpenLDAPer']},
        ]
        self.assert_search_results_equal(expected, res['results'])

    def test_next_range(self):
        # if previous result empty, but previous range is None, 0-*
        self.assertEqual((0, '*'), self.ldap_plugin.next_range(None, True))
        # if previous result empty, and previous range not None, (previous-start+1)-*
        self.assertEqual(
            (1500, '*'),
            self.ldap_plugin.next_range((1500, 2999), True, range_len=1500))
        # if previous result NOT empty and previous range is None: 0-(len-1)
        self.assertEqual(
            (0, 1499),
            self.ldap_plugin.next_range(None, False, range_len=1500))
        # if previous result NOT empty and previous range not None:
        #   (previous_end+1)-(previous_end+len)
        self.assertEqual(
            (4500, 5999),
            self.ldap_plugin.next_range((3000, 4499), False, range_len=1500))

    def test_verify_ldap_config_args_bad_auth_type(self):
        bind_dn = 'CN=Im a bind DN,DC=test'
        bind_pw = 'Passw0rd'
        auth_type = 'NOTREALAUTHTYPE'
        transport_type = 'clear'

        with self.assertRaises(drpc.CallBadArgError):
            self.ldap_plugin._verify_ldap_config_args(
                bind_dn,
                bind_pw,
                auth_type,
                transport_type)

    def test_verify_ldap_config_args_bad_transport_type(self):
        bind_dn = 'CN=Im a bind DN,DC=test'
        bind_pw = 'Passw0rd'
        auth_type = 'ntlm1'
        transport_type = 'NOTREALTRANSPORTTYPE'

        with self.assertRaises(drpc.CallBadArgError):
            self.ldap_plugin._verify_ldap_config_args(
                bind_dn,
                bind_pw,
                auth_type,
                transport_type)

    def test_verify_ldap_config_args_missing_creds(self):
        bind_dn = ''
        bind_pw = ''
        auth_type = 'plain'
        transport_type = 'clear'

        with self.assertRaises(drpc.CallError):
            self.ldap_plugin._verify_ldap_config_args(
                bind_dn,
                bind_pw,
                auth_type,
                transport_type)


class TestLdapDrpcPluginWithLdaps(LdapMethodTestBase):
    listen_ldaps = True

    @defer.inlineCallbacks
    def test_ssl_verify_hostname_enabled(self):
        res = yield self.do_ldap_search(
            ssl_ca_certs=ca_cert_pem,
            ssl_verify_hostname=True,
            transport_type=ldap.client.AD_TRANSPORT_LDAPS,
            pagination=ldap_drpc_plugin.PAGINATION_TYPE_DISABLE,
        )
        self.assertIsNot(None, res['results'])

    @defer.inlineCallbacks
    def test_ssl_verify_hostname_disabled(self):
        res = yield self.do_ldap_search(
            ssl_ca_certs=ca_cert_pem,
            ssl_verify_hostname=False,
            transport_type=ldap.client.AD_TRANSPORT_LDAPS,
            pagination=ldap_drpc_plugin.PAGINATION_TYPE_DISABLE,
        )
        self.assertIsNot(None, res['results'])

    @defer.inlineCallbacks
    def test_invalid_ssl_ca_certs(self):
        with self.assertRaisesRegexp(drpc.CallBadArgError,
                                     'args.+:.+ssl_ca_certs'):
            yield self.do_ldap_search(
                ssl_ca_certs='WTF',
                ssl_verify_hostname=True,
                transport_type=ldap.client.AD_TRANSPORT_LDAPS,
            )

    @defer.inlineCallbacks
    def test_empty_ssl_ca_certs(self):
        with self.assertRaisesRegexp(drpc.CallError,
                                     ldap_drpc_plugin.ERR_LDAP_CONFIGURATION_FAILED):
            yield self.do_ldap_search(
                ssl_ca_certs='',
                ssl_verify_hostname=True,
                transport_type=ldap.client.AD_TRANSPORT_LDAPS,
            )


class TestLdapDrpcPluginConfigFlagging(LdapMethodTestBase):

    @defer.inlineCallbacks
    def setUp(self):
        yield super(TestLdapDrpcPluginConfigFlagging, self).setUp()

        self.password_found = False
        self.original_log = log.msg

    @defer.inlineCallbacks
    def tearDown(self):
        yield super(TestLdapDrpcPluginConfigFlagging, self).tearDown()

        log.msg = self.original_log

    def _log_check_for_password(self, msg):
        if self.password in msg:
            self.password_found = True


class TestLdapDrpcPluginSecureLogging(TestLdapDrpcPluginConfigFlagging):
    is_logging_insecure = 'False'

    @defer.inlineCallbacks
    def test_plaintext_password_not_logged(self):
        log.msg = self._log_check_for_password
        yield self.do_ldap_search(pagination=ldap_drpc_plugin.PAGINATION_TYPE_DISABLE)
        self.assertFalse(self.password_found)


class TestLdapDrpcPluginInsecureLogging(TestLdapDrpcPluginConfigFlagging):
    is_logging_insecure = 'True'

    @defer.inlineCallbacks
    def test_plaintext_password_logged(self):
        log.msg = self._log_check_for_password
        yield self.do_ldap_search(pagination=ldap_drpc_plugin.PAGINATION_TYPE_DISABLE)
        self.assertTrue(self.password_found)
