#
# Copyright (c) 2013 Duo Security
# All Rights Reserved
#
from twisted.internet import defer
from twisted.internet import reactor

from duoauthproxy.lib import config_provider
from drpc.v2.credentials import DrpcCredentials
from duoauthproxy.modules import cloud, ldap_drpc_plugin

from base import StringTransport, TestCase, sleep

IKEY = 'fooey ikey'
SKEY = 'fooey skey'
APIHOST = 'fooey apihost'
DRPC_CREDS = DrpcCredentials(IKEY, SKEY)


class MockModule(object):
    debug = False
    is_logging_insecure = False


class MockPlugin(object):
    def get_drpc_calls(self):
        return {'mock': self.do_mock}

    def do_mock(self):
        return True


class TestModuleCloud(TestCase):
    def setUp(self):
        self.instance = self.make_instance()
        # Make some tests faster with an aggressive reconnect interval.
        # (Be sure to set this before startService() is called.)
        self.instance.reconnect_interval = 2
        self.instance.startService()

    def tearDown(self):
        self.instance.stopService()

    def make_instance(self):
        self.config_items = {
            'ikey': 'nope',
            'skey': 'nuh-uh',
        }
        config = config_provider.ConfigDict(self.config_items)
        instance = cloud.Module(config)

        def mock_proxy_join(drpc_server_factory_class, drpc_path):
            """
            Simulate upgrading an HTTP connection to the service.
            """
            factory = drpc_server_factory_class(
                drpc_creds=DRPC_CREDS,
            )
            protocol = factory.buildProtocol('/Test/Foo')
            trans = StringTransport()
            protocol.makeConnection(trans)
            return protocol
        instance.duo_client.proxy_join = mock_proxy_join
        return instance

    def test_start_stop_service(self):
        # Only run setUp and tearDown.
        pass

    @defer.inlineCallbacks
    def test_auto_reconnect(self):
        orig_server = self.instance.rpc_server
        orig_server.transport.abortConnection()
        d = defer.Deferred()
        reactor.callLater((self.instance.reconnect_interval + 1),
                          d.callback, None)
        yield d
        self.assertIsNot(orig_server, self.instance.rpc_server)
        self.assertFalse(
            self.instance.rpc_server.transport.disconnected)

    @defer.inlineCallbacks
    def test_idle_reconnect(self):
        # Force reconnection parameters
        self.instance.drpc_server_factory_class.idle_rpc_timeout = 1
        self.instance.rpc_server.last_rpc = 1

        orig_server = self.instance.rpc_server
        yield sleep(self.instance.reconnect_interval + 1)
        self.assertIsNot(orig_server, self.instance.rpc_server)
        self.assertFalse(
            self.instance.rpc_server.transport.disconnected)


class TestDrpcFunctionLookup(TestCase):

    def setUp(self):
        self.module = cloud.Module(config_provider.ConfigDict({'ikey': 'ikey', 'skey': 'skey'}))
        self.cloudApiFactory = cloud.CloudApiFactory(module=self.module, drpc_creds=DRPC_CREDS)

    def test_ping(self):
        ping_func = self.cloudApiFactory.get_func_for_call('ping')
        self.assertIsNotNone(ping_func, 'Ping should be found')

    def test_module_register_existing_success(self):
        call_name = 'mock'
        func_1 = self.module.get_func_for_drpc_call(call_name)
        self.assertIsNone(func_1, 'Function {0} should not exist yet'.format(call_name))

        self.module.register_drpc_call_provider(MockPlugin())

        func_2 = self.module.get_func_for_drpc_call(call_name)
        self.assertIsNotNone(func_2, 'Function {0} should exist after registering'.format(call_name))

    def test_module_register_nonexisting_success(self):
        call_name = 'never'
        plugin_instance = MockPlugin()

        func_1 = self.module.get_func_for_drpc_call(call_name)
        self.assertIsNone(func_1, 'Function {0} should not exist'.format(call_name))

        self.module.register_drpc_call_provider(plugin_instance)

        func_2 = self.module.get_func_for_drpc_call(call_name)
        self.assertIsNone(func_2, 'Function {0} should still not exist'.format(call_name))

    def test_empty_module_register_success(self):
        call_name = 'mock'

        class MockBlankPlugin(object):
            def get_drpc_calls(self):
                return {}
        self.module.register_drpc_call_provider(MockBlankPlugin())

        func_1 = self.module.get_func_for_drpc_call(call_name)
        self.assertIsNone(func_1, 'Function {0} should not exist'.format(call_name))

    def test_factory_lookup_success(self):
        call_name = 'mock'
        self.module.register_drpc_call_provider(MockPlugin())

        func = self.cloudApiFactory.get_func_for_call(call_name)
        self.assertIsNotNone(func, 'Should find a function for {0}'.format(call_name))

    def test_factory_lookup_failure(self):
        call_name = 'no_such_function'
        self.module.register_drpc_call_provider(MockPlugin())

        func = self.cloudApiFactory.get_func_for_call(call_name)
        self.assertIsNone(func, 'Should not find a function for {0}'.format(call_name))

    def test_ldap_plugin_lookup_registered_function(self):
        fake_config = config_provider.ConfigDict({})

        call_name = 'ldap_search'
        ldap_plugin = ldap_drpc_plugin.LdapDrpcPlugin(fake_config)

        self.module.register_drpc_call_provider(ldap_plugin)

        func = self.module.get_func_for_drpc_call(call_name)
        self.assertIsNotNone(func, 'Should find a function for {0}'.format(call_name))

    def test_ldap_plugin_lookup_unregisted_function(self):
        fake_config = config_provider.ConfigDict({})

        call_name = '_paged_search'
        ldap_plugin = ldap_drpc_plugin.LdapDrpcPlugin(fake_config)

        self.module.register_drpc_call_provider(ldap_plugin)

        func = self.module.get_func_for_drpc_call(call_name)
        self.assertIsNone(func, 'Should not find a function for {0}'.format(call_name))


if __name__ == '__main__':
    import unittest
    unittest.main()
