#
# Copyright (c) 2012 Duo Security
# All Rights Reserved
#
from twisted.internet._sslverify import TLSVersion
from twisted.internet.ssl import CertificateOptions, AcceptableCiphers
import mox
import mock
import six
import socket
import unittest

import OpenSSL

from twisted.internet import defer
from twisted.internet import error
from twisted.internet import reactor
from twisted.python import log
from twisted.python.failure import Failure

from ldaptor.protocols import pureldap
from ldaptor.protocols.ldap import ldapclient
from ldaptor.protocols.ldap import ldapconnector
from ldaptor.protocols.ldap import ldaperrors
import ldaptor.protocols.ldap.ldapsyntax

from duoauthproxy.modules.ldap_server_auto import (
    DEFAULT_DELIM,
    Module,
)
import duoauthproxy.modules.ad_client
import duoauthproxy.modules.ssl_server
from duoauthproxy.lib import duo_async, ldap
from duoauthproxy.lib.primary_only_manager import PrimaryOnlyManager
from duoauthproxy.lib import log as duo_log
from duoauthproxy.lib import config_provider

from base import sleep, TestCase, AuthLoggingTest
from mock_ldap_server import make_server_factory

# Use the localhost IP address Avoid twisted hostname lookup, because
# it has a race condition in which it may leave a reactor in a dirty
# state when a connection is made but no data is sent over it.
# http://phab.duo/D3924
LOCALHOST = "127.0.0.1"


class BaseDuoAutoLdapServer(TestCase):
    default_factor = 'i am the default factor'
    fred_password = 'pass2word'

    def setUp(self):
        self.backing_port = 10389
        self.backing_listener = None
        self.fred_dn = 'CN=Fred Person,CN=Users,DC=bongos,DC=test'
        self.jane_dn = 'CN=Jane Blackwood,CN=Users,DC=bongos,DC=test'
        self.concat_factor = 'phone9'
        self.mox = mox.Mox()
        self.duo_client = self.mox.CreateMock(
            duo_async.AsyncDuoClient,
        )
        self.client_config = config_provider.ConfigDict({
            'host': LOCALHOST,
            'port': str(self.backing_port),
            'auth_type': ldap.client.AD_AUTH_TYPE_PLAIN,
            'bind_dn': self.fred_dn,
            'service_account_username': 'fred',
            'service_account_password': self.fred_password,
            'search_dn': 'DC=bongos,DC=test',
            'domain_discovery': 'True',
        })
        self.primary_ator = duoauthproxy.modules.ad_client.Module(self.client_config)
        self.primary_ator.startService()

        self.server_module = None
        self.client = None
        self.server_config = {
            'port': '11389',
            'factors': 'auto',
            'ikey': 'ikey',
            'skey': 'skey',
            'api_host': 'api_host',
            'failmode': 'safe',
        }

    @defer.inlineCallbacks
    def _make_backing_factory(self, ssl=False, minimum_tls_version=None, cipher_list=''):
        (self.backing_factory,
         self.backing_listener) = yield make_server_factory(
             self.backing_port,
             ssl=ssl,
             minimum_tls_version=minimum_tls_version,
             cipher_list=cipher_list,
        )

    @defer.inlineCallbacks
    def _make_proxy(self,
                    failmode='safe',
                    exempt_primary_bind=None,
                    exempt_ous=None,
                    delim=None,
                    allow_concat=None,
                    network_timeout=None,
                    allow_searches_after_bind=None,
                    allow_unlimited_binds=None,
                    idle_timeout=None,
                    server_section_name='ldap_server_auto'):
        if exempt_ous is None:
            exempt_ous = []
        self.server_config['failmode'] = failmode
        if exempt_primary_bind is not None:
            self.server_config['exempt_primary_bind'] = exempt_primary_bind
        for i, exempt_ou in enumerate(exempt_ous, start=1):
            self.server_config['exempt_ou_%d' % i] = exempt_ou
        if delim is not None:
            self.server_config['delimiter'] = delim
        if allow_concat is not None:
            self.server_config['allow_concat'] = allow_concat
        if network_timeout:
            self.server_config['network_timeout'] = str(network_timeout)
        if idle_timeout:
            self.server_config['idle_timeout'] = str(idle_timeout)
        if allow_searches_after_bind is not None:
            self.server_config['allow_searches_after_bind'] = str(allow_searches_after_bind)
        if allow_unlimited_binds is not None:
            self.server_config['allow_unlimited_binds'] = str(allow_unlimited_binds)
        self.server_config = config_provider.ConfigDict(self.server_config)
        self.server_module = Module(
            self.server_config,
            self.primary_ator,
            server_section_name,
        )
        self.server_module.factory.client = self.duo_client
        self.server_module.startService()

        # This override links the DN '' to a tuple of the host:port letting the connections
        # use these values intead of a lookup.
        cfg = ldaptor.config.LDAPConfig(serviceLocationOverrides={
            '': (LOCALHOST, 11389),
        })
        self.client_creator = ldapconnector.LDAPClientCreator(
            reactor=reactor,
            protocolClass=ldapclient.LDAPClient,
        )
        self.client = yield self.client_creator.connect(
            dn='',
            overrides=cfg.getServiceLocationOverrides(),
        )

    @defer.inlineCallbacks
    def tearDown(self):
        if self.client is not None:
            self.client.transport.loseConnection()
            self.client = None
        self.primary_ator.stopService()
        if self.backing_listener:
            yield self.backing_listener.stopListening()
        if self.server_module is not None:
            yield self.server_module.stopService()
            self.server_module = None
        self.mox.UnsetStubs()

    @defer.inlineCallbacks
    def _bind(self, dn, password):
        entry = ldaptor.protocols.ldap.ldapsyntax.LDAPEntry(self.client,
                                                            dn)
        bound_entry = yield entry.bind(password)
        defer.returnValue(bound_entry)

    @defer.inlineCallbacks
    def assert_bind(self, dn, password):
        bound_entry = yield self._bind(dn, password)
        self.assertEqual(dn, bound_entry.dn)

    @defer.inlineCallbacks
    def _bind_and_search(self, **proxy_kwargs):
        self._make_backing_factory()
        yield self._make_proxy(**proxy_kwargs)

        yield self.assert_bind('CN=adminer adminer,CN=Users,DC=bongos,DC=test',
                               'password!')

        entry = ldaptor.protocols.ldap.ldapsyntax.LDAPEntry(self.client,
                                                            self.fred_dn)
        self.assertEqual(self.fred_dn,
                         entry.dn)

        result = yield entry.search(attributes=('sAMAccountName',))
        result = result[0]
        self.assertEqual('fred', list(result.get('sAMAccountName'))[0])

    @defer.inlineCallbacks
    def _bind_search_bind(self, succeed,
                          password=fred_password,
                          bind_requests_expected=3,
                          **proxy_kwargs):
        """
        Test the bind-search-bind pattern for one configuration.

        By default, expect three binds:
        * Primary bind from the client of the proxy.
        * Secondary bind from that client.
        * Primary bind by proxy itself before searching for username attribute.

        But the proxy will only send the third bind will only happen
        if second-factor authentication is performed.
        """
        self.mox.ReplayAll()
        yield self._bind_and_search(**proxy_kwargs)
        if succeed:
            yield self.assert_bind(self.fred_dn,
                                   password)
        else:
            with self.assertRaises(ldaperrors.LDAPInvalidCredentials):
                yield self.assert_bind(self.fred_dn,
                                       password)
        self.mox.VerifyAll()
        self.mox.ResetAll()
        self.assertEqual(
            bind_requests_expected,
            self.backing_factory.bind_requests_received,
        )

    def enqueue_duo_preauth(self, result, status=None,
                            default_factor=default_factor):
        if status is None:
            status = str(result)
        method_call = self.duo_client.preauth(mox.IsA(six.string_types),
                                              mox.Or(mox.IsA(type(None)), mox.IsA((six.string_types))))
        method_call.AndReturn({
            'result': result,
            'status': status,
            'factors': {
                '1': 'push1',
                '2': 'phone1',
                '3': 'sms1',
                'default': default_factor,
            },
            'prompt': '?',
        })
        return method_call

    def enqueue_duo_preauth_error_fail_open(self):
        e = duo_async.DuoAPIFailOpenError()
        return self.enqueue_duo_preauth(None).AndRaise(e)

    def enqueue_duo_preauth_error_fail_closed(self):
        e = duo_async.DuoAPIFailClosedError()
        return self.enqueue_duo_preauth(None).AndRaise(e)

    def enqueue_duo_auth(self, result, status=None,
                         factor=mox.IsA(six.string_types)):
        if status is None:
            status = str(result)
        method_call = self.duo_client.auth(
            mox.IsA(six.string_types),
            factor,
            client_ip=None,
        )
        method_call.AndReturn({
            'result': result,
            'status': status,
        })
        return method_call

    def enqueue_duo_auth_error_fail_open(self):
        e = duo_async.DuoAPIFailOpenError()
        return self.enqueue_duo_auth(None).AndRaise(e)

    def enqueue_duo_auth_error_fail_closed(self):
        e = duo_async.DuoAPIFailClosedError()
        return self.enqueue_duo_auth(None).AndRaise(e)


class TestDuoAutoLdapServer(BaseDuoAutoLdapServer, TestCase):

    def setUp(self):
        super(TestDuoAutoLdapServer, self).setUp()
        self.primary_ator = duoauthproxy.modules.ad_client.Module(self.client_config)
        self.primary_ator.startService()

    @defer.inlineCallbacks
    def tearDown(self):
        yield super(TestDuoAutoLdapServer, self).tearDown()
        self.primary_ator.stopService()

    @defer.inlineCallbacks
    def test_proxy_connection_made_and_connection_lost_race_condition(self):
        """
        Simulate a race condition where Proxy's connectionMade is started but
        yields before completion. connectionLost then starts while connectionMade
        yields. We want to ensure that connectionLost yields until the Proxy has obtained
        a client reference, thereby letting connectionMade complete. Not yielding in connectionLost
        allows for the possibility of hanging TCP connections from the client to the domain controller.

        For those following along at home here is the call chain for this test:

        connectionMade starts
        connectionMade calls make_connected_client
        connectionMade yields on make_connected_client, awaiting assignment for the client
        connectionLost starts and adds a callLater for make_connected_client's yielding deferred
        connectionLost yields until client is yielded via make_connected_client
        make_connected_client's defer fires and connectionMade receives its client
        connectionMade completes
        connectionLost completes since make_connected_client has returned
        """

        # mock AD client's transport
        mockTransport = mox.Mox().CreateMockAnything()
        mockTransport.setTcpKeepAlive(True)
        mockTransport.abortConnection()
        mox.Replay(mockTransport)

        # mock Proxy's AD client connection
        mockClient = mox.Mox().CreateMockAnything()
        mockClient.transport = mockTransport

        class ProxyTester(ldap.proxy.Proxy):

            def __init__(self, testinstance=None):
                super(ProxyTester, self).__init__()
                self.queue_on_lost_start = defer.Deferred()
                self.testinstance = testinstance

            @defer.inlineCallbacks
            def make_connected_client(self):
                # to guarantee that connectionLost is called before connectionMade has run to completion,
                # don't return the connection until queue_on_lost_start is fired by connectionLost
                client_to_return = yield self.queue_on_lost_start
                defer.returnValue(client_to_return)

            @defer.inlineCallbacks
            def connectionMade(self):
                yield super(ProxyTester, self).connectionMade()
                # verify that connectionMade completed and client is properly assigned
                # i.e. connectionLost didn't set client to none before completion
                self.testinstance.assertIsNotNone(self._Proxy__client_connected)
                self.testinstance.assertIsNotNone(self._Proxy__client)

            @defer.inlineCallbacks
            def connectionLost(self, reason):
                # fire queue_on_lost_start's callback on the next mainloop iteration
                # allowing connectionMade to complete. we don't fire the callback directly
                # because we want connectionLost to start but yield on self.__client
                reactor.callLater(0, self.queue_on_lost_start.callback, mockClient)
                # verify that connectionMade started and but hasn't received a client
                self.testinstance.assertIsNotNone(self._Proxy__client_connected)
                self.testinstance.assertIsNone(self._Proxy__client)
                yield super(ProxyTester, self).connectionLost(reason)

        proxy = ProxyTester(testinstance=self)
        connection_made = proxy.connectionMade()
        connection_lost = proxy.connectionLost(Failure(Exception('tcp connection dropped')))

        yield connection_made
        yield connection_lost

    @defer.inlineCallbacks
    def test_bind_search_connection(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_ALLOW)
        self.mox.ReplayAll()
        self._make_backing_factory()
        yield self._make_proxy()
        # Admin bind
        yield self.assert_bind('CN=adminer adminer,CN=Users,DC=bongos,DC=test',
                               'password!')

        entry = ldaptor.protocols.ldap.ldapsyntax.LDAPEntry(self.client,
                                                            self.fred_dn)
        self.assertEqual(self.fred_dn,
                         entry.dn)

        # User search
        result = yield entry.search(attributes=('sAMAccountName',))
        result = result[0]
        self.assertEqual('fred', list(result.get('sAMAccountName'))[0])
        # User bind
        yield self.assert_bind(self.fred_dn,
                               self.fred_password)

        # Additional search - ensure that the ldap server does not close
        # the connection after the second bind.
        result = yield entry.search(attributes=('sAMAccountName',))
        result = result[0]
        self.assertEqual('fred', list(result.get('sAMAccountName'))[0])

    @defer.inlineCallbacks
    def test_bind_search_disconnect(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_ALLOW)
        self.mox.ReplayAll()
        self._make_backing_factory()
        yield self._make_proxy()
        # Admin bind
        yield self.assert_bind('CN=adminer adminer,CN=Users,DC=bongos,DC=test',
                               'password!')

        # User bind
        yield self.assert_bind(self.fred_dn,
                               self.fred_password)

        # Ensure the protocol doesn't permit another bind.
        with self.assertRaises(error.ConnectionDone):
            yield self._bind(
                self.fred_dn,
                self.fred_password
            )

    @defer.inlineCallbacks
    def test_unlimited_binds(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_ALLOW)
        self.mox.ReplayAll()
        self._make_backing_factory()
        yield self._make_proxy(
            allow_unlimited_binds='true',
            exempt_primary_bind='false',
            exempt_ous=['CN=adminer adminer,CN=Users,DC=bongos,DC=test'])
        # Admin bind
        yield self.assert_bind('CN=adminer adminer,CN=Users,DC=bongos,DC=test',
                               'password!')

        # User bind
        yield self.assert_bind(self.fred_dn,
                               self.fred_password)

        # Ensure the protocol will allow additional binds with the flag enabled.
        yield self.assert_bind('CN=adminer adminer,CN=Users,DC=bongos,DC=test',
                               'password!')

    @defer.inlineCallbacks
    def test_disconnect_after_bind(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_ALLOW)
        self.mox.ReplayAll()
        self._make_backing_factory()
        yield self._make_proxy(allow_searches_after_bind=False)
        # Admin bind
        yield self.assert_bind('CN=adminer adminer,CN=Users,DC=bongos,DC=test',
                               'password!')

        entry = ldaptor.protocols.ldap.ldapsyntax.LDAPEntry(self.client,
                                                            self.fred_dn)
        self.assertEqual(self.fred_dn,
                         entry.dn)

        # User search
        result = yield entry.search(attributes=('sAMAccountName',))
        result = result[0]
        self.assertEqual('fred', list(result.get('sAMAccountName'))[0])

        # User bind
        yield self.assert_bind(self.fred_dn,
                               self.fred_password)

        # Ensure the protocol doesn't permit another search.
        with self.assertRaises(error.ConnectionDone):
            result = yield entry.search(attributes=('sAMAccountName',))

    @defer.inlineCallbacks
    def test_wrong_password(self):
        yield self._bind_search_bind(
            succeed=False,
            password=('not' + self.fred_password),
            bind_requests_expected=2,
        )

    @defer.inlineCallbacks
    def test_preauth_allow(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_ALLOW)
        yield self._bind_search_bind(succeed=True, failmode='safe')

    @defer.inlineCallbacks
    def test_preauth_deny(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_DENY)
        yield self._bind_search_bind(succeed=False, failmode='safe')

    @defer.inlineCallbacks
    def test_preauth_enroll(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_ENROLL)
        yield self._bind_search_bind(succeed=False, failmode='safe')

    @defer.inlineCallbacks
    def test_preauth_fail_open_failmode_safe(self):
        self.enqueue_duo_preauth_error_fail_open()
        yield self._bind_search_bind(succeed=True, failmode='safe')

    @defer.inlineCallbacks
    def test_preauth_fail_closed_failmode_safe(self):
        self.enqueue_duo_preauth_error_fail_closed()
        yield self._bind_search_bind(succeed=False, failmode='safe')

    @defer.inlineCallbacks
    def test_preauth_fail_open_failmode_secure(self):
        self.enqueue_duo_preauth_error_fail_open()
        yield self._bind_search_bind(succeed=False, failmode='secure')

    @defer.inlineCallbacks
    def test_preauth_fail_closed_failmode_secure(self):
        self.enqueue_duo_preauth_error_fail_closed()
        yield self._bind_search_bind(succeed=False, failmode='secure')

    @defer.inlineCallbacks
    def test_preauth_override_failmode_primary_only(self):
        with mock.patch.object(PrimaryOnlyManager, 'is_primary_only_enabled', return_value=True):
            self.enqueue_duo_preauth(duo_async.API_RESULT_ALLOW)
            yield self._bind_search_bind(succeed=True, failmode='secure')

    @defer.inlineCallbacks
    def test_auth_allow(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth(duo_async.API_RESULT_ALLOW)
        yield self._bind_search_bind(succeed=True, failmode='safe')

    @defer.inlineCallbacks
    def test_auth_deny(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth(duo_async.API_RESULT_DENY)
        yield self._bind_search_bind(succeed=False, failmode='safe')

    @defer.inlineCallbacks
    def test_auth_fail_open_failmode_safe(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth_error_fail_open()
        yield self._bind_search_bind(succeed=True, failmode='safe')

    @defer.inlineCallbacks
    def test_auth_fail_closed_failmode_safe(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth_error_fail_closed()
        yield self._bind_search_bind(succeed=False, failmode='safe')

    @defer.inlineCallbacks
    def test_auth_fail_open_failmode_secure(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth_error_fail_open()
        yield self._bind_search_bind(succeed=False, failmode='secure')

    @defer.inlineCallbacks
    def test_auth_fail_closed_failmode_secure(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth_error_fail_closed()
        yield self._bind_search_bind(succeed=False, failmode='secure')

    @defer.inlineCallbacks
    def test_backing_connect_timeout(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        s.bind((LOCALHOST, self.backing_port))
        try:
            yield self._make_proxy('safe', network_timeout=1)
            yield sleep(2)
            with self.assertRaises(ldaperrors.LDAPUnavailable):
                yield self.assert_bind(self.fred_dn,
                                       'password')
        finally:
            s.close()

    @defer.inlineCallbacks
    def test_backing_response_timeout(self):
        self._make_backing_factory()
        self.backing_factory.response_delay = 2
        yield self._make_proxy('safe', network_timeout=1)
        with self.assertRaises(error.ConnectionLost):
            yield self.assert_bind(self.fred_dn,
                                   'password')
        # wait for cleanup of artificially delayed requests
        yield self.backing_factory.delayed

    @defer.inlineCallbacks
    def test_idle_timeout(self):
        self._make_backing_factory()
        yield self._make_proxy('safe', idle_timeout=2)
        yield sleep(4)
        with self.assertRaises(ldapclient.LDAPClientConnectionLostException):
            yield self.assert_bind(self.fred_dn,
                                   'password')

    @defer.inlineCallbacks
    def test_exempt_primary_bind_true(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_ALLOW)
        yield self._bind_search_bind(succeed=True, failmode='safe')

    @defer.inlineCallbacks
    def test_exempt_primary_bind_false(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_ALLOW)
        self._make_backing_factory()
        yield self._make_proxy('safe', exempt_primary_bind='0')
        self.mox.ReplayAll()
        yield self.assert_bind(self.fred_dn,
                               self.fred_password)
        self.mox.VerifyAll()
        self.mox.ResetAll()

    @defer.inlineCallbacks
    def test_exempt_ous(self):
        yield self._bind_search_bind(
            succeed=True,
            exempt_primary_bind='false',
            failmode='safe',
            # one irrelevant, one that case-sensitively matches
            # self.fred_dn
            exempt_ous=['dc=foo', 'DC=test'],
            bind_requests_expected=2,
        )

    @defer.inlineCallbacks
    def test_exempt_ous_mixed_case(self):
        yield self._bind_search_bind(
            succeed=True,
            failmode='safe',
            # one irrelevant, one that case-INsensitively matches
            # self.fred_dn
            exempt_ous=['dc=foo', 'dc=tEst'],
            bind_requests_expected=2,
        )

    # Concat tests! Don't bother testing every variation of preauth,
    # auth, and failmode, only concat detection and functionality.

    @defer.inlineCallbacks
    def test_concat_preauth_allow(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_ALLOW)
        yield self._bind_search_bind(
            succeed=True,
            password=(self.fred_password
                      + DEFAULT_DELIM
                      + self.concat_factor),
        )

    @defer.inlineCallbacks
    def test_concat_preauth_auth_auth_allow(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth(duo_async.API_RESULT_ALLOW,
                              factor=self.concat_factor)
        yield self._bind_search_bind(
            succeed=True,
            password=(self.fred_password
                      + DEFAULT_DELIM
                      + self.concat_factor),
        )

    @defer.inlineCallbacks
    def test_concat_wrong_delim(self):
        wrong_delim = chr(ord(DEFAULT_DELIM) + 1)
        yield self._bind_search_bind(
            succeed=False,
            password=(self.fred_password
                      + wrong_delim
                      + self.concat_factor),
            bind_requests_expected=2,
        )

    @defer.inlineCallbacks
    def test_concat_delim_config(self):
        delim = '|'
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth(duo_async.API_RESULT_ALLOW,
                              factor=self.concat_factor)
        yield self._bind_search_bind(
            succeed=True,
            delim=delim,
            password=(self.fred_password
                      + delim
                      + self.concat_factor),
        )

    @defer.inlineCallbacks
    def test_concat_false_positive(self):
        """
        Induce a concat false-positive by setting delim to a character
        contained in the primary auth password and not actually doing
        concat. The second primary auth must succeed, leading to Duo
        Auth with the auto-selected factor (*not* whatever the
        incorrectly split factor part of the password was).
        """
        delim = self.fred_password[4]
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.assertNotEqual(self.fred_password[5:],
                            self.default_factor)
        self.enqueue_duo_auth(duo_async.API_RESULT_ALLOW,
                              factor=self.default_factor)
        yield self._bind_search_bind(
            succeed=True,
            delim=delim,
            password=self.fred_password,
            bind_requests_expected=4,
        )

    @defer.inlineCallbacks
    def test_concat_delim_also_in_password(self):
        """
        Concat must still work if the password happens to also contain
        the delim.
        """
        delim = self.fred_password[4]
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth(duo_async.API_RESULT_ALLOW,
                              factor=self.concat_factor)
        yield self._bind_search_bind(
            succeed=True,
            delim=delim,
            password=(self.fred_password
                      + delim
                      + self.concat_factor),
        )

    @defer.inlineCallbacks
    def test_concat_auth_multi_char_delim(self):
        delim = ';;'
        self.enqueue_duo_preauth(duo_async.API_RESULT_ALLOW)
        yield self._bind_search_bind(
            succeed=True,
            delim=delim,
            password=(self.fred_password
                      + delim
                      + self.concat_factor),
        )

    @defer.inlineCallbacks
    def test_concat_rejected_if_disabled(self):
        yield self._bind_search_bind(
            succeed=False,
            allow_concat='0',
            password=(self.fred_password
                      + DEFAULT_DELIM
                      + self.concat_factor),
            bind_requests_expected=2,
        )

    @defer.inlineCallbacks
    def test_concat_disabled_with_non_concat_password(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_ALLOW)
        yield self._bind_search_bind(
            succeed=True,
            allow_concat='0',
            password=self.fred_password,
        )

    @defer.inlineCallbacks
    def test_concat_not_performed_for_exempt_ou(self):
        delim = self.fred_password[4]
        yield self._bind_search_bind(
            succeed=True,
            delim=delim,
            password=self.fred_password,
            # One irrelevant, one matching self.fred_dn
            exempt_ous=['dc=foo', 'DC=test'],
            # Won't even attempt primary auth with split password.
            bind_requests_expected=2,
        )

    @defer.inlineCallbacks
    def test_sasl(self):
        self._make_backing_factory()
        yield self._make_proxy(
            # Ensure that preauth and concat are not attempted because
            # request is SASL by excluding other reasons in the
            # config.
            exempt_primary_bind='0',
            allow_concat='1',
        )

        delim = self.server_module.factory.delim
        sasl_auth = ('i am' + delim + ' a mechanism',
                     'these are' + delim + ' some credentials')
        server_sasl_creds = 'serverSaslCreds go \x03 here'

        def fake_sasl(protocol, request, controls, reply):
            self.assertEqual(self.fred_dn, request.dn)
            self.assertTrue(request.sasl)
            self.assertIs(type(request.auth), tuple)
            protocol.boundUser = "don't look at me!"
            ssc = pureldap.LDAPBindResponse_serverSaslCreds(
                value=server_sasl_creds,
            )
            response = pureldap.LDAPBindResponse(
                resultCode=ldaperrors.LDAPSaslBindInProgress.resultCode,
                matchedDN=str(request.dn),
                serverSaslCreds=ssc,
            )
            return response
        self.backing_factory.handle_LDAPBindRequest = fake_sasl

        self.mox.ReplayAll()
        # Duo Auth not triggered until the subsequent Success.
        request = pureldap.LDAPBindRequest(
            dn=self.fred_dn,
            auth=sasl_auth,
            sasl=True,
        )
        response = yield self.client.send(request)
        self.mox.VerifyAll()
        self.mox.ResetAll()

        # Test ldaptor round-tripping of serverSaslCreds
        self.assertEqual(server_sasl_creds,
                         response.serverSaslCreds.value)

        # No fallback to primary auth with unsplit password.
        self.assertEqual(
            1,
            self.backing_factory.bind_requests_received,
        )

    @defer.inlineCallbacks
    def _test_bind_with_non_canonical_dn(self, username):
        """
        Bind-search-bind and check for the expected Duo auth with a
        "DN" that does not match the object's canonical DN but is
        nonetheless recognized by the backing server instead of an
        actual DN.
        """
        self._make_backing_factory()
        yield self._make_proxy()

        # Primary bind.
        yield self.assert_bind('CN=adminer adminer,CN=Users,DC=bongos,DC=test',
                               'password!')

        # Expect exactly two binds after it:
        expected_usernames = [
            username,  # Second bind will be with the non-canonical DN.
            self.fred_dn,  # Third will be the proxy itself in dn_to_username().
        ]

        # Override bind to find fred's object without the real DN.
        def fake_bind(protocol, request, controls, reply):
            self.assertEqual(expected_usernames.pop(0), request.dn)
            self.assertEqual(self.fred_password, request.auth)
            protocol.boundUser = "don't look at me!"
            response = pureldap.LDAPBindResponse(
                resultCode=ldaperrors.Success.resultCode,
                matchedDN=self.fred_dn,
            )
            return response
        self.backing_factory.handle_LDAPBindRequest = fake_bind

        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth(duo_async.API_RESULT_ALLOW)
        self.mox.ReplayAll()
        # Duo Auth not triggered until the subsequent Success.
        request = pureldap.LDAPBindRequest(
            dn=username,
            auth=self.fred_password,
        )
        response = yield self.client.send(request)
        self.mox.VerifyAll()
        self.mox.ResetAll()

        self.assertEqual(ldaperrors.Success.resultCode,
                         response.resultCode)
        self.assertEqual(self.fred_dn,
                         response.matchedDN)

    def test_netbios_domain_username(self):
        """
        Bind-search-bind with a "DN" of the form "NetBIOS
        domain\username" instead of an actual DN.
        """
        return self._test_bind_with_non_canonical_dn('BONGOS\\fred')

    def test_upn_username(self):
        """
        Bind-search-bind with a "DN" in the UPN username@domain format
        instead of an actual DN.
        """
        return self._test_bind_with_non_canonical_dn('fred@bongos.test')

    def test_duplicate_samaccountname(self):
        """
        Bind-search-bind with a "NetBIOS domain\username" using two accounts
        with the same sAMAccountname but different domains. If the msDS-principalName
        is specified we can select the user with the proper domain
        """
        return self._test_bind_with_non_canonical_dn('bongos\\dupe')

    def assert_create_username_filter(self, username, expected_result):
        client = ldap.client.ADClientProtocol()
        output = client._create_username_filter(username)
        self.assertEqual(expected_result, output)

    def test_create_username_filter_upn(self):
        self.assert_create_username_filter(
            username='genghis@khan.test',
            expected_result={'userPrincipalName': 'genghis@khan.test'}
        )

    def test_create_username_filter_netbios(self):
        self.assert_create_username_filter(
            username='KHAN\genghis',
            expected_result={'sAMAccountName': 'genghis'}
        )

    def test_create_username_filter_bad(self):
        self.assert_create_username_filter(
            username='CN=KhanGenghis',
            expected_result=None
        )

    @defer.inlineCallbacks
    def test_response_with_empty_matchedDN(self):
        self._make_backing_factory()
        yield self._make_proxy(
            exempt_primary_bind='0',
        )

        def handle_LDAPBindRequest(protocol, request, controls, reply):
            self.assertEqual(self.fred_dn, request.dn)
            protocol.boundUser = "don't look at me!"
            response = pureldap.LDAPBindResponse(
                resultCode=ldaperrors.Success.resultCode,
                matchedDN='',
            )
            return response
        self.backing_factory.handle_LDAPBindRequest = handle_LDAPBindRequest

        # Expect Duo auth with requested username even if response had
        # an empty matchedDN.
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth(duo_async.API_RESULT_ALLOW)

        self.mox.ReplayAll()
        request = pureldap.LDAPBindRequest(
            dn=self.fred_dn,
            auth=''.join(('foo', self.server_module.factory.delim, 'push')),
        )
        response = yield self.client.send(request)
        self.mox.VerifyAll()
        self.mox.ResetAll()

        self.assertEqual(ldaperrors.Success.resultCode, response.resultCode)
        self.assertEqual('', response.matchedDN)

    @defer.inlineCallbacks
    def test_anonymous_bind(self):
        self._make_backing_factory()
        yield self._make_proxy(
            exempt_primary_bind='0',
        )

        def handle_LDAPBindRequest(protocol, request, controls, reply):
            self.assertEqual(self.fred_dn, request.dn)
            protocol.boundUser = "don't look at me!"
            response = pureldap.LDAPBindResponse(
                resultCode=ldaperrors.Success.resultCode,
                matchedDN='',
            )
            return response
        self.backing_factory.handle_LDAPBindRequest = handle_LDAPBindRequest

        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth(duo_async.API_RESULT_ALLOW)
        self.mox.ReplayAll()
        request = pureldap.LDAPBindRequest(
            dn=self.fred_dn,
            auth='',
        )
        response = yield self.client.send(request)
        self.mox.VerifyAll()
        self.mox.ResetAll()

        self.assertEqual(ldaperrors.Success.resultCode, response.resultCode)
        self.assertEqual('', response.matchedDN)

    @defer.inlineCallbacks
    def test_concat_does_not_create_anonymous_bind(self):
        """
        The LDAP proxy must never create an anonymous bind from one with a
        password because it speculates about concat mode.
        """
        self._make_backing_factory()
        yield self._make_proxy(
            exempt_primary_bind='0',
        )

        requests = []

        def handle_LDAPBindRequest(protocol, request, controls, reply):
            requests.append((request.dn, request.auth))
            return pureldap.LDAPBindResponse(
                resultCode=ldaperrors.Success.resultCode,
                matchedDN=request.dn,
            )
        self.backing_factory.handle_LDAPBindRequest = handle_LDAPBindRequest

        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth(duo_async.API_RESULT_ALLOW)
        self.mox.ReplayAll()
        request = pureldap.LDAPBindRequest(
            dn='CN=Jane Blackwood,CN=Users,DC=bongos,DC=test',
            auth=self.server_module.factory.delim,
        )
        yield self.client.send(request)
        self.mox.VerifyAll()
        self.mox.ResetAll()

        self.assertEqual([
            (self.jane_dn, self.server_module.factory.delim),
            (self.fred_dn, self.fred_password),
        ], requests)

    @defer.inlineCallbacks
    def test_controls(self):
        # Test critical control codes.  Critical control codes should be
        # proxied through to the backing server.

        self._make_backing_factory()
        yield self._make_proxy(
            exempt_primary_bind='0',
        )

        c = [("1.2.840.113556.1.4.319", 255, '3')]

        def handle_LDAPSearchRequest(request, controls, reply):
            """ Verify that the controls are passed through correctly. """
            self.assertEqual(controls, c,
                             "Expected {0}".format(c))
            return pureldap.LDAPSearchResultDone(resultCode=ldaperrors.Success.resultCode)

        self.backing_factory.handle_LDAPSearchRequest = handle_LDAPSearchRequest

        self.mox.ReplayAll()
        request = pureldap.LDAPSearchRequest()
        response = yield self.client.send(request, controls=c)
        self.mox.VerifyAll()
        self.mox.ResetAll()

        self.assertEqual(
            ldaperrors.Success.resultCode,
            response.resultCode,
            "Expected {0}".format(ldaperrors.Success.resultCode)
        )
        self.assertEqual('', response.matchedDN)

    @defer.inlineCallbacks
    def test_bind_controls_pass(self):
        # Test critical control codes.
        # Control codes should not pass through.

        self._make_backing_factory()
        yield self._make_proxy(
            exempt_primary_bind='0',
        )

        c = [("2.2.840.113556.1.4.319", 0, '3')]

        def handle_LDAPBindRequest(protocol, request, controls, reply):
            self.assertIs(controls, None)
            response = pureldap.LDAPBindResponse(
                resultCode=ldaperrors.Success.resultCode,
                matchedDN=str(request.dn),
            )
            return response
        self.backing_factory.handle_LDAPBindRequest = handle_LDAPBindRequest

        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth(duo_async.API_RESULT_ALLOW)
        self.mox.ReplayAll()
        # Even if there was a DN this BIND shouldn't be interpreted as
        # the empty matchedDN case.
        request = pureldap.LDAPBindRequest(
            dn=self.fred_dn,
            auth=self.fred_password,
        )
        response = yield self.client.send(request, controls=c)
        self.mox.VerifyAll()
        self.mox.ResetAll()

        self.assertEqual(
            ldaperrors.Success.resultCode,
            response.resultCode,
            "Expected {0}".format(ldaperrors.Success.resultCode)
        )
        self.assertEqual(self.fred_dn, response.matchedDN)

    @defer.inlineCallbacks
    def test_bind_controls(self):
        # Test critical control codes.  Critical control codes should cause
        # an error during BIND request processing

        self._make_backing_factory()
        yield self._make_proxy(
            exempt_primary_bind='0',
        )

        c = [("2.2.840.113556.1.4.319", 255, '3')]

        def handle_LDAPBindRequest(protocol, request, controls, reply):
            # This should not be called by the authproxy implementation.
            # If it is called, it should fail the assert, and its
            # success result code should be trapped in the assert at the
            # end of the test function.
            self.assertEqual(controls, [], "Expected {0}".format([]))
            response = pureldap.LDAPBindResponse(
                resultCode=ldaperrors.Success.resultCode,
                matchedDN='',
            )
            return response
        self.backing_factory.handle_LDAPBindRequest = handle_LDAPBindRequest

        self.mox.ReplayAll()
        # Even if there was a DN this BIND shouldn't be interpreted as
        # the empty matchedDN case.
        request = pureldap.LDAPBindRequest(
            dn=self.fred_dn,
            auth='',
        )
        response = yield self.client.send(request, controls=c)
        self.mox.VerifyAll()
        self.mox.ResetAll()

        self.assertEqual(
            ldaperrors.LDAPUnavailableCriticalExtension.resultCode,
            response.resultCode,
            "Expected {0}".format(ldaperrors.LDAPUnavailableCriticalExtension.resultCode)
        )
        self.assertEqual('', response.matchedDN)

    @defer.inlineCallbacks
    def test_controls_to_client_via_proxy(self):
        # Test control codes.  Control codes should be
        # proxied through from the server back to the client.

        self._make_backing_factory()
        yield self._make_proxy(
            exempt_primary_bind='0',
        )

        expected_control = [("1.2.840.113556.1.4.319", 0, "7")]

        def handle_LDAPSearchRequest(request, controls, reply):
            """ Verify that the controls are passed through correctly. """
            reply(
                pureldap.LDAPSearchResultDone(resultCode=ldaperrors.Success.resultCode),
                response_controls=expected_control
            )
            return

        self.backing_factory.handle_LDAPSearchRequest = handle_LDAPSearchRequest

        self.mox.ReplayAll()
        request = pureldap.LDAPSearchRequest()
        response = yield self.client.send(request, handle_msg=True)
        self.mox.VerifyAll()
        self.mox.ResetAll()

        self.assertEqual(
            ldaperrors.Success.resultCode,
            response.value.resultCode,
            "Expected {0}".format(ldaperrors.Success.resultCode)
        )
        self.assertEqual(
            response.controls,
            expected_control,
            "Expected {0}".format(expected_control)
        )

    @defer.inlineCallbacks
    def test_drop_pdu_during_bindrequest_resolution(self):
        """
        TODO See D23705 for more info about this-
        This test as it stands cannot tell the difference between the second bind
        request being dropped and the second bind request simply taking a long time
        to reach the end LDAP server- aka. a false positive. Ideally we'd patch
        ldap_server_auto's handle() here to call an eg. drop_request() method when
        the bad PDU gets dropped so that we can compare the received_dns of patched
        ldap_server_auto:handle() and the received_dns of backing_factory.

        When ldap_server_auto is resolving a client's BindRequest, any other PDUs
        sent by the client before receiving its BindResponse should be dropped.

        Here's the order in which things occur:

        Client sends Jane's bind_request1 and backing_factory receives it
        Client awaits bind_request_recvd_d
        blocking_handle_LDAPBindRequest appends jane_dn to received_dns
        bind_request_recvd_d yields back to the client;
            blocking_handle_LDAPBindRequest awaits bind_response_blocked_d
        Client sends Bob's bind_request2, ldap_server_auto drops it during handle()
        Client unblocks blocking_handle_LDAPBindRequest; blocking_handle_LDAPBindRequest
            returns Jane's/bind_request1's BindResponse
        """
        received_dns = []
        bind_request_recvd_d = defer.Deferred()
        bind_response_blocked_d = defer.Deferred()

        @defer.inlineCallbacks
        def blocking_handle_LDAPBindRequest(protocol, request, controls, reply):
            self.resolving_bind_request = True
            # Keep a history of requests that have been received through their DN's
            received_dns.append(request.dn)
            if not bind_request_recvd_d.called:
                bind_request_recvd_d.callback(None)
            yield bind_response_blocked_d
            defer.returnValue(pureldap.LDAPBindResponse(
                resultCode=ldaperrors.Success.resultCode,
                matchedDN=request.dn,
            ))

        self._make_backing_factory()
        yield self._make_proxy(
            exempt_primary_bind='0'
        )
        self.backing_factory.handle_LDAPBindRequest = blocking_handle_LDAPBindRequest

        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth(duo_async.API_RESULT_ALLOW)
        self.mox.ReplayAll()

        bind_request1 = pureldap.LDAPBindRequest(
            dn=self.jane_dn,
            auth='',
        )
        first_request_d = self.client.send(bind_request1)
        yield bind_request_recvd_d

        bind_request2 = pureldap.LDAPBindRequest(
            dn='CN=Bob,CN=Users,DC=bongos,DC=test',
            auth='',
        )
        second_request_d = self.client.send(bind_request2)

        # Tell backing factory to return a response, which will make
        # blocking_handle_LDAPBindRequest return Jane's BindResponse
        bind_response_blocked_d.callback(None)
        yield first_request_d

        self.mox.VerifyAll()
        self.mox.ResetAll()

        # Since the second request is expected to never come back with a response,
        # we artificially sever the transport's connection and just make sure that
        # the second request's Deferred properly raises a ConnectionDone error.
        self.client.transport.loseConnection()
        with self.assertRaises(error.ConnectionDone):
            yield second_request_d

        # Even though we expect only the first request to have been received
        # by the backing factory, fred_dn is present because all bind requests
        # will trigger a bind as the service account user- in this case,
        # fred_dn is our service account user (see setUp).
        self.assertEqual([self.jane_dn, self.fred_dn], received_dns)


class TestDuoAutoLdapServerLogging(AuthLoggingTest, BaseDuoAutoLdapServer, TestCase):

    @defer.inlineCallbacks
    def test_bind_timeout_failure_logged(self):
        expected = [
            {'msg': 'LDAP Connection error occurred during bind',
             'auth_stage': duo_log.AUTH_PRIMARY,
             'status': duo_log.AUTH_ERROR}
        ]

        self._make_backing_factory()
        self.backing_factory.response_delay = 2
        yield self._make_proxy('safe', network_timeout=1)
        with self.assertRaises(error.ConnectionLost):
            yield self.assert_bind(self.fred_dn,
                                   'password')
        # wait for cleanup of artificially delayed requests
        yield self.backing_factory.delayed

        self.assert_logged(expected)

    @defer.inlineCallbacks
    def test_all_auth_success_logged(self):
        expected = [
            {'msg': 'Primary authentication successful',
             'dn': 'CN=adminer adminer,CN=Users,DC=bongos,DC=test',
             'auth_stage': duo_log.AUTH_PRIMARY,
             'status': duo_log.AUTH_ALLOW},
            {'msg': 'Primary bind exempted from 2FA',
             'dn': 'CN=adminer adminer,CN=Users,DC=bongos,DC=test',
             'auth_stage': duo_log.AUTH_SECONDARY,
             'status': duo_log.AUTH_ALLOW},
            {'msg': 'Primary authentication successful',
             'dn': self.fred_dn,
             'auth_stage': duo_log.AUTH_PRIMARY,
             'status': duo_log.AUTH_ALLOW},
            {'msg': 'Duo authentication succeeded - allow',
             'auth_stage': duo_log.AUTH_SECONDARY,
             'dn': self.fred_dn,
             'username': 'fred',
             'status': duo_log.AUTH_ALLOW}
        ]

        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth(duo_async.API_RESULT_ALLOW)
        yield self._bind_search_bind(succeed=True)

        self.assert_logged(expected)

    @defer.inlineCallbacks
    def test_service_bind_failure_logged(self):
        expected = [
            {'msg': 'Primary authentication rejected',
             'dn': 'CN=adminer adminer,CN=Users,DC=bongos,DC=test',
             'auth_stage': duo_log.AUTH_PRIMARY,
             'status': duo_log.AUTH_REJECT}
        ]

        self._make_backing_factory()
        yield self._make_proxy()
        with self.assertRaises(Exception):
            yield self.assert_bind('CN=adminer adminer,CN=Users,DC=bongos,DC=test', 'wrongpassword!')

        self.assert_logged(expected)

    @defer.inlineCallbacks
    def test_primary_auth_failure_logged(self):
        expected = [
            {'msg': 'Primary authentication successful',
             'dn': 'CN=adminer adminer,CN=Users,DC=bongos,DC=test',
             'auth_stage': duo_log.AUTH_PRIMARY,
             'status': duo_log.AUTH_ALLOW},
            {'msg': 'Primary bind exempted from 2FA',
             'dn': 'CN=adminer adminer,CN=Users,DC=bongos,DC=test',
             'auth_stage': duo_log.AUTH_SECONDARY,
             'status': duo_log.AUTH_ALLOW},
            {'msg': 'Primary authentication rejected',
             'auth_stage': duo_log.AUTH_PRIMARY,
             'status': duo_log.AUTH_REJECT,
             'dn': self.fred_dn}
        ]

        yield self._bind_search_bind(
            succeed=False,
            password=('not' + self.fred_password),
            bind_requests_expected=2)

        self.assert_logged(expected)

    @defer.inlineCallbacks
    def test_no_dn_duo_bypassed_logged(self):
        expected = [
            {'msg': 'Primary authentication successful',
             'dn': '',
             'auth_stage': duo_log.AUTH_PRIMARY,
             'status': duo_log.AUTH_ALLOW},
            {'msg': 'Duo authentication bypassed for anonymous bind',
             'dn': '',
             'auth_stage': duo_log.AUTH_SECONDARY,
             'status': duo_log.AUTH_ALLOW},
        ]

        self._make_backing_factory()
        yield self._make_proxy(exempt_primary_bind='false')
        yield self._bind('', None)

        self.assert_logged(expected)

    @defer.inlineCallbacks
    def test_exempt_ou_success_logged(self):
        expected = [
            {'msg': 'Primary authentication successful',
             'dn': 'CN=adminer adminer,CN=Users,DC=bongos,DC=test',
             'auth_stage': duo_log.AUTH_PRIMARY,
             'status': duo_log.AUTH_ALLOW},
            {'msg': 'Exempt OU: CN=adminer adminer,CN=Users,DC=bongos,DC=test',
             'dn': 'CN=adminer adminer,CN=Users,DC=bongos,DC=test',
             'auth_stage': duo_log.AUTH_SECONDARY,
             'status': duo_log.AUTH_ALLOW},
            {'msg': 'Primary authentication successful',
             'dn': self.fred_dn,
             'auth_stage': duo_log.AUTH_PRIMARY,
             'status': duo_log.AUTH_ALLOW},
            {'msg': 'Exempt OU: {0}'.format(self.fred_dn),
             'dn': self.fred_dn,
             'auth_stage': duo_log.AUTH_SECONDARY,
             'status': duo_log.AUTH_ALLOW}
        ]

        yield self._bind_search_bind(
            succeed=True,
            exempt_primary_bind='false',
            exempt_ous=['dc=foo', 'DC=test'],
            bind_requests_expected=2,
        )

        self.assert_logged(expected)

    @defer.inlineCallbacks
    def test_api_error_fail_open_success_logged(self):
        expected = [
            {'msg': 'Primary authentication successful',
             'dn': 'CN=adminer adminer,CN=Users,DC=bongos,DC=test',
             'auth_stage': duo_log.AUTH_PRIMARY,
             'status': duo_log.AUTH_ALLOW},
            {'msg': 'Primary bind exempted from 2FA',
             'dn': 'CN=adminer adminer,CN=Users,DC=bongos,DC=test',
             'auth_stage': duo_log.AUTH_SECONDARY,
             'status': duo_log.AUTH_ALLOW},
            {'msg': 'Primary authentication successful',
             'dn': self.fred_dn,
             'auth_stage': duo_log.AUTH_PRIMARY,
             'status': duo_log.AUTH_ALLOW},
            {'msg': duo_async.FAILMODE_SAFE_MSG,
             'dn': self.fred_dn,
             'auth_stage': duo_log.AUTH_SECONDARY,
             'status': duo_log.AUTH_ALLOW}
        ]

        self.enqueue_duo_preauth_error_fail_open()
        yield self._bind_search_bind(succeed=True, failmode='safe')

        self.assert_logged(expected)

    @defer.inlineCallbacks
    def test_api_error_fail_closed_failure_logged(self):
        expected = [
            {'msg': 'Primary authentication successful',
             'dn': 'CN=adminer adminer,CN=Users,DC=bongos,DC=test',
             'auth_stage': duo_log.AUTH_PRIMARY,
             'status': duo_log.AUTH_ALLOW},
            {'msg': 'Primary bind exempted from 2FA',
             'dn': 'CN=adminer adminer,CN=Users,DC=bongos,DC=test',
             'auth_stage': duo_log.AUTH_SECONDARY,
             'status': duo_log.AUTH_ALLOW},
            {'msg': 'Primary authentication successful',
             'dn': self.fred_dn,
             'auth_stage': duo_log.AUTH_PRIMARY,
             'status': duo_log.AUTH_ALLOW},
            {'msg': duo_async.FAILMODE_SECURE_MSG,
             'dn': self.fred_dn,
             'auth_stage': duo_log.AUTH_SECONDARY,
             'status': duo_log.AUTH_ERROR}
        ]

        self.enqueue_duo_preauth_error_fail_closed()
        yield self._bind_search_bind(succeed=False, failmode='secure')

        self.assert_logged(expected)

    @defer.inlineCallbacks
    def test_duo_no_username_failure_logged(self):
        expected = [
            {'msg': 'Primary authentication successful',
             'dn': 'CN=Non AD,CN=Users,DC=bongos,DC=test',
             'auth_stage': duo_log.AUTH_PRIMARY,
             'status': duo_log.AUTH_ALLOW},
            {'msg': 'Cannot find username',
             'dn': 'CN=Non AD,CN=Users,DC=bongos,DC=test',
             'username': duo_log.AUTH_UNKNOWN,
             'auth_stage': duo_log.AUTH_SECONDARY,
             'status': duo_log.AUTH_REJECT}
        ]

        self._make_backing_factory()
        yield self._make_proxy(exempt_primary_bind='false')
        with self.assertRaises(Exception):
            yield self._bind('CN=Non AD,CN=Users,DC=bongos,DC=test', self.fred_password)

        self.assert_logged(expected)

    @defer.inlineCallbacks
    def test_duo_preauth_allow_success_logged(self):
        expected = [
            {'msg': 'Primary authentication successful',
             'dn': 'CN=adminer adminer,CN=Users,DC=bongos,DC=test',
             'auth_stage': duo_log.AUTH_PRIMARY,
             'status': duo_log.AUTH_ALLOW},
            {'msg': 'Primary bind exempted from 2FA',
             'dn': 'CN=adminer adminer,CN=Users,DC=bongos,DC=test',
             'auth_stage': duo_log.AUTH_SECONDARY,
             'status': duo_log.AUTH_ALLOW},
            {'msg': 'Primary authentication successful',
             'dn': self.fred_dn,
             'auth_stage': duo_log.AUTH_PRIMARY,
             'status': duo_log.AUTH_ALLOW},
            {'msg': 'Duo preauth result was allow',
             'auth_stage': duo_log.AUTH_SECONDARY,
             'status': duo_log.AUTH_ALLOW,
             'username': 'fred'}
        ]

        self.enqueue_duo_preauth(duo_async.API_RESULT_ALLOW)
        yield self._bind_search_bind(succeed=True)

        self.assert_logged(expected)

    @defer.inlineCallbacks
    def test_duo_preauth_deny_failure_logged(self):
        expected = [
            {'msg': 'Primary authentication successful',
             'dn': 'CN=adminer adminer,CN=Users,DC=bongos,DC=test',
             'auth_stage': duo_log.AUTH_PRIMARY,
             'status': duo_log.AUTH_ALLOW},
            {'msg': 'Primary bind exempted from 2FA',
             'dn': 'CN=adminer adminer,CN=Users,DC=bongos,DC=test',
             'auth_stage': duo_log.AUTH_SECONDARY,
             'status': duo_log.AUTH_ALLOW},
            {'msg': 'Primary authentication successful',
             'dn': self.fred_dn,
             'auth_stage': duo_log.AUTH_PRIMARY,
             'status': duo_log.AUTH_ALLOW},
            {'msg': 'Duo preauth result was deny',
             'username': 'fred',
             'auth_stage': duo_log.AUTH_SECONDARY,
             'status': duo_log.AUTH_REJECT}
        ]

        self.enqueue_duo_preauth(duo_async.API_RESULT_DENY)
        yield self._bind_search_bind(succeed=False)

        self.assert_logged(expected)

    @defer.inlineCallbacks
    def test_duo_preauth_enroll_logged(self):
        expected = [
            {'msg': 'Primary authentication successful',
             'dn': 'CN=adminer adminer,CN=Users,DC=bongos,DC=test',
             'auth_stage': duo_log.AUTH_PRIMARY,
             'status': duo_log.AUTH_ALLOW},
            {'msg': 'Primary bind exempted from 2FA',
             'dn': 'CN=adminer adminer,CN=Users,DC=bongos,DC=test',
             'auth_stage': duo_log.AUTH_SECONDARY,
             'status': duo_log.AUTH_ALLOW},
            {'msg': 'Primary authentication successful',
             'dn': self.fred_dn,
             'auth_stage': duo_log.AUTH_PRIMARY,
             'status': duo_log.AUTH_ALLOW},
            {'msg': duo_log.AUTH_ENROLL_MSG,
             'username': 'fred',
             'auth_stage': duo_log.AUTH_SECONDARY,
             'status': duo_log.AUTH_REJECT}
        ]

        self.enqueue_duo_preauth(duo_async.API_RESULT_ENROLL)
        yield self._bind_search_bind(succeed=False)

        self.assert_logged(expected)

    @defer.inlineCallbacks
    def test_duo_invalid_factors_failure_logged(self):
        expected = [
            {'msg': 'Primary authentication successful',
             'dn': 'CN=adminer adminer,CN=Users,DC=bongos,DC=test',
             'auth_stage': duo_log.AUTH_PRIMARY,
             'status': duo_log.AUTH_ALLOW},
            {'msg': 'Primary bind exempted from 2FA',
             'dn': 'CN=adminer adminer,CN=Users,DC=bongos,DC=test',
             'auth_stage': duo_log.AUTH_SECONDARY,
             'status': duo_log.AUTH_ALLOW},
            {'msg': 'Primary authentication successful',
             'dn': self.fred_dn,
             'auth_stage': duo_log.AUTH_PRIMARY,
             'status': duo_log.AUTH_ALLOW},
            {'msg': 'User has no Duo factors usable with this configuration',
             'auth_stage': duo_log.AUTH_SECONDARY,
             'status': duo_log.AUTH_REJECT,
             'username': 'fred'}
        ]

        self.server_config['factors'] = 'passcode'
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        yield self._bind_search_bind(succeed=False)

        self.assert_logged(expected)

    @defer.inlineCallbacks
    def test_duo_auth_failure_logged(self):
        expected = [
            {'msg': 'Primary authentication successful',
             'dn': 'CN=adminer adminer,CN=Users,DC=bongos,DC=test',
             'auth_stage': duo_log.AUTH_PRIMARY,
             'status': duo_log.AUTH_ALLOW},
            {'msg': 'Primary bind exempted from 2FA',
             'dn': 'CN=adminer adminer,CN=Users,DC=bongos,DC=test',
             'auth_stage': duo_log.AUTH_SECONDARY,
             'status': duo_log.AUTH_ALLOW},
            {'msg': 'Primary authentication successful',
             'dn': self.fred_dn,
             'auth_stage': duo_log.AUTH_PRIMARY,
             'status': duo_log.AUTH_ALLOW},
            {'msg': 'Duo authentication was rejected - deny',
             'username': 'fred',
             'auth_stage': duo_log.AUTH_SECONDARY,
             'status': duo_log.AUTH_REJECT}
        ]

        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth(duo_async.API_RESULT_DENY)
        yield self._bind_search_bind(succeed=False)

        self.assert_logged(expected)


class TestModuleLdapServerAuto(BaseDuoAutoLdapServer, TestCase):
    is_password_logged = False

    def setUp(self):
        super(TestModuleLdapServerAuto, self).setUp()
        self.server_config.update({'debug': 'True'})
        self.original_log_msg = log.msg

    @defer.inlineCallbacks
    def tearDown(self):
        yield super(TestModuleLdapServerAuto, self).tearDown()
        log.msg = self.original_log_msg
        self.mox.ResetAll()

    @defer.inlineCallbacks
    def test_plaintext_password_logged(self):
        self.server_config.update({'is_logging_insecure': 'True'})
        self._make_backing_factory()
        yield self.make_ldap_request()
        self.assertTrue(self.is_password_logged)

    @defer.inlineCallbacks
    def test_plaintext_password_not_logged(self):
        self.server_config.update({'is_logging_insecure': 'False'})
        self._make_backing_factory()
        yield self.make_ldap_request()
        self.assertFalse(self.is_password_logged)

    @defer.inlineCallbacks
    def make_ldap_request(self):
        yield self._make_proxy(
            exempt_primary_bind='0',
        )

        self.backing_factory.handle_LDAPBindRequest = self.handle_LDAPBindRequest

        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth(duo_async.API_RESULT_ALLOW)

        request = pureldap.LDAPBindRequest(
            dn=self.fred_dn,
            auth=''.join((self.fred_password, self.server_module.factory.delim, 'push')),
        )
        log.msg = self.find_password_in_server_log_msg
        self.mox.ReplayAll()
        yield self.client.send(request)

    def handle_LDAPBindRequest(self, protocol, request, controls, reply):
        self.assertEqual(self.fred_dn, request.dn)
        protocol.boundUser = "don't look at me!"
        response = pureldap.LDAPBindResponse(
            resultCode=ldaperrors.Success.resultCode,
            matchedDN='',
        )
        return response

    def find_password_in_server_log_msg(self, msg, **kwargs):
        if 'C->S' not in msg and 'C<-S' not in msg and self.fred_password in msg:
            self.is_password_logged = True


class TestLDAPServerAutoSSLHardeningContext(BaseDuoAutoLdapServer):
    @defer.inlineCallbacks
    def test_default_tls_version_is_tls_1_2(self):
        yield self._make_backing_factory(ssl=True)
        yield self._make_proxy()

        ssl_context_factory = CertificateOptions(
            insecurelyLowerMinimumTo=TLSVersion.TLSv1_2,
            lowerMaximumSecurityTo=TLSVersion.TLSv1_2,
        )

        connection = yield self.client_creator.connectTCP(
            LOCALHOST,
            self.backing_port
        )

        try:
            yield connection.transport.startTLS(ssl_context_factory)
            self.assertEqual(connection.transport.connected, 1)
        finally:
            yield connection.transport.abortConnection()

    @defer.inlineCallbacks
    def test_custom_cipher_list_match(self):
        ciphers = 'AES256-SHA'

        yield self._make_backing_factory(ssl=True, cipher_list=ciphers)
        yield self._make_proxy()

        acceptable_ciphers = AcceptableCiphers.fromOpenSSLCipherString(
            ciphers
        )

        ssl_context_factory = CertificateOptions(
            acceptableCiphers=acceptable_ciphers
        )

        connection = yield self.client_creator.connectTCP(
            LOCALHOST,
            self.backing_port
        )

        try:
            connection.transport.startTLS(ssl_context_factory)
            # Must wait for handshake. startTLS provides no deferreds for this :/
            yield sleep(0.5)
            self.assertEqual(connection.transport.connected, 1)
        finally:
            yield connection.transport.abortConnection()

    @defer.inlineCallbacks
    def test_custom_cipher_list_mismatch(self):
        server_ciphers = 'AES256-SHA'
        client_ciphers = 'DES-CBC3-SHA'

        yield self._make_backing_factory(ssl=True, cipher_list=server_ciphers)
        yield self._make_proxy()

        acceptable_ciphers = AcceptableCiphers.fromOpenSSLCipherString(
            client_ciphers
        )

        ssl_context_factory = CertificateOptions(
            acceptableCiphers=acceptable_ciphers
        )

        connection = yield self.client_creator.connectSSL(
            LOCALHOST,
            self.backing_port,
            ssl_context_factory
        )
        with self.assertRaises(OpenSSL.SSL.Error):
            yield connection.startTLS()

    @defer.inlineCallbacks
    def test_default_minimum_tls_works_with_tls1_2_client(self):
        yield self._make_backing_factory(ssl=True)
        yield self._make_proxy()

        ssl_context_factory = CertificateOptions(
            insecurelyLowerMinimumTo=TLSVersion.TLSv1_2,
            lowerMaximumSecurityTo=TLSVersion.TLSv1_3,  # maximumSecurity is exclusive
        )

        connection = yield self.client_creator.connectTCP(
            LOCALHOST,
            self.backing_port
        )

        try:
            yield connection.transport.startTLS(ssl_context_factory)
            # Must wait for handshake. startTLS provides no deferreds for this :/
            yield sleep(0.5)
            self.assertEqual(connection.transport.connected, 1)
        finally:
            yield connection.transport.abortConnection()

    @defer.inlineCallbacks
    def test_default_minimum_tls_does_not_work_with_sslv3_client(self):
        yield self._make_backing_factory(ssl=True)
        yield self._make_proxy()

        ssl_context_factory = CertificateOptions(
            insecurelyLowerMinimumTo=TLSVersion.SSLv3,
            lowerMaximumSecurityTo=TLSVersion.TLSv1_0,  # maximumSecurity is exclusive
        )

        connection = yield self.client_creator.connectTCP(
            LOCALHOST,
            self.backing_port
        )

        try:
            yield connection.transport.startTLS(ssl_context_factory)
            # Must wait for handshake. startTLS provides no deferreds for this :/
            yield sleep(0.5)
            self.assertEqual(connection.transport.connected, 0)
        finally:
            yield connection.transport.abortConnection()

    @defer.inlineCallbacks
    def test_client_using_tlsv1_unallowed(self):
        yield self._make_backing_factory(ssl=True, minimum_tls_version="tls1.1")
        yield self._make_proxy()

        ssl_context_factory = CertificateOptions(
            lowerMaximumSecurityTo=TLSVersion.TLSv1_1,  # maximumSecurity is exclusive
        )

        connection = yield self.client_creator.connectTCP(
            LOCALHOST,
            self.backing_port
        )

        try:
            connection.transport.startTLS(ssl_context_factory)
            # Must wait for handshake. startTLS provides no deferreds for this :/
            yield sleep(1)
            self.assertEqual(connection.transport.connected, 0)
        finally:
            yield connection.transport.abortConnection()

    @defer.inlineCallbacks
    def test_client_using_tlsv1_1unallowed(self):
        yield self._make_backing_factory(ssl=True, minimum_tls_version="tls1.2")
        yield self._make_proxy()

        ssl_context_factory = CertificateOptions(
            insecurelyLowerMinimumTo=TLSVersion.TLSv1_1,
            lowerMaximumSecurityTo=TLSVersion.TLSv1_2,  # maximumSecurity is exclusive
        )

        connection = yield self.client_creator.connectTCP(
            LOCALHOST,
            self.backing_port
        )

        try:
            connection.transport.startTLS(ssl_context_factory)
            # Must wait for handshake. startTLS provides no deferreds for this :/
            yield sleep(1)
            self.assertEqual(connection.transport.connected, 0)
        finally:
            yield connection.transport.abortConnection()

    @defer.inlineCallbacks
    def test_client_tls_version_and_server_tls_version_are_lower_than_default(self):
        yield self._make_backing_factory(ssl=True, minimum_tls_version="tls1.0")
        yield self._make_proxy()

        ssl_context_factory = CertificateOptions(raiseMinimumTo=TLSVersion.SSLv3, lowerMaximumSecurityTo=TLSVersion.TLSv1_1)

        try:
            connection = yield self.client_creator.connectSSL(
                LOCALHOST,
                self.backing_port,
                ssl_context_factory
            )

            self.assertEqual(connection.transport.connected, 1)
        finally:
            yield connection.transport.abortConnection()


if __name__ == '__main__':
    unittest.main()
