#
# Copyright (c) 2012 Duo Security
# All Rights Reserved
#

import mox
import pyrad
import six
import mock

from twisted.internet import defer

import duoauthproxy.modules.mock_client
import duoauthproxy.modules.radius_server_auto
from duoauthproxy.lib import duo_async, base, protect, log
from duoauthproxy.lib.config_provider import ConfigDict
from duoauthproxy.lib.radius.base import radius_dictionary
from duoauthproxy.lib.primary_only_manager import PrimaryOnlyManager

from base import (
    TestCase,
    ServerModuleTest,
    ServerModuleTestPassThrough,
    ServerModuleTestPassThroughMultipleValues,
    ServerProtectedConfigParsingTestCase,
    AuthLoggingTest,
)


class MockRadiusClientModule(duoauthproxy.modules.mock_client.Module):
    """Helper class to _make_proxying_auth_packet().

    Temporarily mocks out the client Module instance so that all attempts at
    primary authentication passed on to a proxy RADIUS server instead creates
    a challenge for the fake user and stores the challenge state.
    """
    fake_challenge_code = pyrad.packet.AccessChallenge
    fake_challenge_state = 'i challenge you!'

    def radius_proxy(self, request):
        return pyrad.packet.AuthPacket(
            code=self.fake_challenge_code,
            authenticator=pyrad.packet.AuthPacket.CreateAuthenticator(),
            dict=radius_dictionary(),
            secret=six.b(' '),
            State=self.fake_challenge_state,
        )


class ErroringRadiusClientModule(duoauthproxy.modules.mock_client.Module):
    """Helper class to test_proxy_error_failure_logged().

    Temporarily mocks out the client Module instance so that any attempt at
    primary authentication through a proxy RADIUS server errors out.

    Note:
        This separate subclass of the mock_client Module is needed rather than
        just patching radius_proxy() because `self` must point to some instance
        of mock_client.Module rather than eg. RadiusServerAutoAuthLogging.
    """
    def radius_proxy(self, request):
        raise base.AuthError('ERROR')


class RadiusServerAutoTestRunner(ServerModuleTest, TestCase):
    """Tester class inheritable by all other RadiusServerAuto test classes.

    Any test class that wants to inherit from ServerModuleTest, TestCase, or both
    of them can inherit from this class instead.

    This class was created due to multiple inheritance super() incompatibilities
    with TestRadiusServerAutoAuthLogging, though it's since been co-opted
    for use with other RadiusServerAuto test classes as well to cut down on
    extra setUp() redefinitions.
    """
    def setUp(self):
        self.concat_factor = 'phone9'
        super(RadiusServerAutoTestRunner, self).setUp(
            {},
            duoauthproxy.modules.radius_server_auto,
            'radius_server_auto'
        )


class RadiusServerAutoTestCases(object):
    """Test suite covering various DuoAutoRadiusServer auth flows.

    MS-CHAP2 requests are tested separately, as they contain different RADIUS
    attributes and have slightly different auth flows in certain situations,
    such as when a user wants to change their CHAP2 password.

    Note:
        Since there's no setUp() or tearDown() defined in this class, it needs
        to make use of the setUp() defined in RadiusServerAutoTestRunner.
    """

    # coverage for missing username/password:

    @defer.inlineCallbacks
    def test_missing_username(self):
        yield self._test_auth(pyrad.packet.AccessReject,
                              username=None,
                              password=self.password)

    @defer.inlineCallbacks
    def test_chap2_missing_username(self):
        yield self._test_auth(pyrad.packet.AccessReject,
                              username=None,
                              password=None,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_request_attrs)

    @defer.inlineCallbacks
    def test_wrong_primary_auth_password(self):
        yield self._test_auth(pyrad.packet.AccessReject,
                              username=self.username,
                              password=('not ' + self.password))

    @defer.inlineCallbacks
    def test_chap2_wrong_primary_auth_password(self):
        yield self._test_auth(pyrad.packet.AccessReject,
                              username=self.username,
                              password=None,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_bad_request_attrs)

    # coverage for preauth, MS-CHAP2 or not:

    @defer.inlineCallbacks
    def test_preauth_allow(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_ALLOW)
        yield self._test_auth(pyrad.packet.AccessAccept,
                              username=self.username,
                              password=self.password)

    @defer.inlineCallbacks
    def test_chap2_preauth_allow(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_ALLOW)
        yield self._test_auth(pyrad.packet.AccessAccept,
                              username=self.username,
                              password=None,
                              request_attrs=self.chap2_request_attrs)

    @defer.inlineCallbacks
    def test_chap2_expired_preauth_allow(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_ALLOW)
        yield self._test_auth(pyrad.packet.AccessReject,
                              username=self.username,
                              password=None,
                              reject_attrs=self.chap2_ps_expired_attr,
                              request_attrs=self.chap2_expired_request_attrs)

    @defer.inlineCallbacks
    def test_chap2_preauth_deny(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_DENY)
        yield self._test_auth(pyrad.packet.AccessReject,
                              username=self.username,
                              password=None,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_request_attrs)

    @defer.inlineCallbacks
    def test_chap2_expired_preauth_deny(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_DENY)
        yield self._test_auth(pyrad.packet.AccessReject,
                              username=self.username,
                              password=None,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_expired_request_attrs)

    @defer.inlineCallbacks
    def test_preauth_deny(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_DENY)
        yield self._test_auth(pyrad.packet.AccessReject,
                              username=self.username,
                              password=self.password)

    @defer.inlineCallbacks
    def test_preauth_enroll(self):
        # treated as deny by DuoSimpleRadiusServer
        self.enqueue_duo_preauth(duo_async.API_RESULT_ENROLL)
        yield self._test_auth(pyrad.packet.AccessReject,
                              username=self.username,
                              password=self.password)

    @defer.inlineCallbacks
    def test_chap2_preauth_enroll(self):
        # treated as deny by DuoSimpleRadiusServer
        self.enqueue_duo_preauth(duo_async.API_RESULT_ENROLL)
        yield self._test_auth(pyrad.packet.AccessReject,
                              username=self.username,
                              password=None,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_request_attrs)

    @defer.inlineCallbacks
    def test_chap2_expired_preauth_enroll(self):
        # treated as deny by DuoSimpleRadiusServer
        self.enqueue_duo_preauth(duo_async.API_RESULT_ENROLL)
        yield self._test_auth(pyrad.packet.AccessReject,
                              username=self.username,
                              password=None,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_expired_request_attrs)

    @defer.inlineCallbacks
    def test_preauth_fail_open_failmode_secure(self):
        self.enqueue_duo_preauth_error_fail_open()
        yield self._test_auth(pyrad.packet.AccessReject,
                              failmode=duo_async.FAILMODE_SECURE,
                              username=self.username,
                              password=self.password)

    @defer.inlineCallbacks
    def test_preauth_override_failmode_primary_only(self):
        with mock.patch.object(PrimaryOnlyManager, 'is_primary_only_enabled', return_value=True):
            self.enqueue_duo_preauth(duo_async.API_RESULT_ALLOW)
            yield self._test_auth(pyrad.packet.AccessAccept,
                                  failmode=duo_async.FAILMODE_SECURE,
                                  username=self.username,
                                  password=self.password)

    @defer.inlineCallbacks
    def test_chap2_preauth_fail_open_failmode_secure(self):
        self.enqueue_duo_preauth_error_fail_open()
        yield self._test_auth(pyrad.packet.AccessReject,
                              failmode=duo_async.FAILMODE_SECURE,
                              username=self.username,
                              password=None,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_request_attrs)

    @defer.inlineCallbacks
    def test_chap2_expired_preauth_fail_open_failmode_secure(self):
        self.enqueue_duo_preauth_error_fail_open()
        yield self._test_auth(pyrad.packet.AccessReject,
                              failmode=duo_async.FAILMODE_SECURE,
                              username=self.username,
                              password=None,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_expired_request_attrs)

    @defer.inlineCallbacks
    def test_preauth_fail_closed_failmode_secure(self):
        self.enqueue_duo_preauth_error_fail_closed()
        yield self._test_auth(pyrad.packet.AccessReject,
                              failmode=duo_async.FAILMODE_SECURE,
                              username=self.username,
                              password=self.password)

    @defer.inlineCallbacks
    def test_chap2_preauth_fail_closed_failmode_secure(self):
        self.enqueue_duo_preauth_error_fail_closed()
        yield self._test_auth(pyrad.packet.AccessReject,
                              failmode=duo_async.FAILMODE_SECURE,
                              username=self.username,
                              password=None,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_request_attrs)

    @defer.inlineCallbacks
    def test_chap2_expired_preauth_fail_closed_failmode_secure(self):
        self.enqueue_duo_preauth_error_fail_closed()
        yield self._test_auth(pyrad.packet.AccessReject,
                              failmode=duo_async.FAILMODE_SECURE,
                              username=self.username,
                              password=None,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_expired_request_attrs)

    @defer.inlineCallbacks
    def test_preauth_fail_open_failmode_safe(self):
        self.enqueue_duo_preauth_error_fail_open()
        yield self._test_auth(pyrad.packet.AccessAccept,
                              failmode=duo_async.FAILMODE_SAFE,
                              username=self.username,
                              password=self.password)

    @defer.inlineCallbacks
    def test_chap2_preauth_fail_open_failmode_safe(self):
        self.enqueue_duo_preauth_error_fail_open()
        yield self._test_auth(pyrad.packet.AccessAccept,
                              failmode=duo_async.FAILMODE_SAFE,
                              username=self.username,
                              password=None,
                              request_attrs=self.chap2_request_attrs)

    @defer.inlineCallbacks
    def test_chap2_expired_preauth_fail_open_failmode_safe(self):
        self.enqueue_duo_preauth_error_fail_open()
        yield self._test_auth(pyrad.packet.AccessReject,
                              failmode=duo_async.FAILMODE_SAFE,
                              username=self.username,
                              password=None,
                              reject_attrs=self.chap2_ps_expired_attr,
                              request_attrs=self.chap2_expired_request_attrs)

    @defer.inlineCallbacks
    def test_preauth_fail_closed_failmode_safe(self):
        self.enqueue_duo_preauth_error_fail_closed()
        yield self._test_auth(pyrad.packet.AccessReject,
                              failmode=duo_async.FAILMODE_SAFE,
                              username=self.username,
                              password=self.password)

    @defer.inlineCallbacks
    def test_chap2_preauth_fail_closed_failmode_safe(self):
        self.enqueue_duo_preauth_error_fail_closed()
        yield self._test_auth(pyrad.packet.AccessReject,
                              failmode=duo_async.FAILMODE_SAFE,
                              username=self.username,
                              password=None,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_request_attrs)

    # coverage for secondary authentication, MS-CHAP2 or not:

    @defer.inlineCallbacks
    def test_preauth_auth_auth_allow(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth(duo_async.API_RESULT_ALLOW)
        yield self._test_auth(pyrad.packet.AccessAccept,
                              username=self.username,
                              password=self.password)

    @defer.inlineCallbacks
    def test_chap2_preauth_auth_auth_allow(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth(duo_async.API_RESULT_ALLOW)
        yield self._test_auth(pyrad.packet.AccessAccept,
                              username=self.username,
                              password=None,
                              request_attrs=self.chap2_request_attrs)

    @defer.inlineCallbacks
    def test_preauth_auth_auth_fail_open_failmode_safe(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth_error_fail_open()
        yield self._test_auth(pyrad.packet.AccessAccept,
                              failmode=duo_async.FAILMODE_SAFE,
                              username=self.username,
                              password=self.password)

    @defer.inlineCallbacks
    def test_chap2_preauth_auth_auth_fail_open_failmode_safe(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth_error_fail_open()
        yield self._test_auth(pyrad.packet.AccessAccept,
                              failmode=duo_async.FAILMODE_SAFE,
                              username=self.username,
                              password=None,
                              request_attrs=self.chap2_request_attrs)

    @defer.inlineCallbacks
    def test_preauth_auth_auth_fail_closed_failmode_safe(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth_error_fail_closed()
        yield self._test_auth(pyrad.packet.AccessReject,
                              failmode=duo_async.FAILMODE_SAFE,
                              username=self.username,
                              password=self.password)

    @defer.inlineCallbacks
    def test_chap2_preauth_auth_auth_fail_closed_failmode_safe(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth_error_fail_closed()
        yield self._test_auth(pyrad.packet.AccessReject,
                              failmode=duo_async.FAILMODE_SAFE,
                              username=self.username,
                              password=None,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_request_attrs)

    @defer.inlineCallbacks
    def test_preauth_auth_auth_fail_open_failmode_secure(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth_error_fail_open()
        yield self._test_auth(pyrad.packet.AccessReject,
                              failmode=duo_async.FAILMODE_SECURE,
                              username=self.username,
                              password=self.password)

    @defer.inlineCallbacks
    def test_chap2_preauth_auth_auth_fail_open_failmode_secure(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth_error_fail_open()
        yield self._test_auth(pyrad.packet.AccessReject,
                              failmode=duo_async.FAILMODE_SECURE,
                              username=self.username,
                              password=None,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_request_attrs)

    @defer.inlineCallbacks
    def test_preauth_auth_auth_fail_closed_failmode_secure(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth_error_fail_closed()
        yield self._test_auth(pyrad.packet.AccessReject,
                              failmode=duo_async.FAILMODE_SECURE,
                              username=self.username,
                              password=self.password)

    @defer.inlineCallbacks
    def test_chap2_preauth_auth_auth_fail_closed_failmode_secure(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth_error_fail_closed()
        yield self._test_auth(pyrad.packet.AccessReject,
                              failmode=duo_async.FAILMODE_SECURE,
                              username=self.username,
                              password=None,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_request_attrs)

    @defer.inlineCallbacks
    def test_preauth_auth_auth_deny(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth(duo_async.API_RESULT_DENY)
        yield self._test_auth(pyrad.packet.AccessReject,
                              username=self.username,
                              password=self.password)

    @defer.inlineCallbacks
    def test_chap2_preauth_auth_auth_deny(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth(duo_async.API_RESULT_DENY)
        yield self._test_auth(pyrad.packet.AccessReject,
                              username=self.username,
                              password=None,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_request_attrs)

    # coverage for use of passcodes during authentication:

    @defer.inlineCallbacks
    def test_passcode_preauth_enroll(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_ENROLL)
        yield self._test_auth(
            expect_code=pyrad.packet.AccessReject,
            username=self.username,
            challenge_response='2',
            challenge_state={'is_chap2': False},
        )

    @defer.inlineCallbacks
    def test_chap2_passcode_preauth_enroll(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_ENROLL)
        yield self._test_auth(
            expect_code=pyrad.packet.AccessReject,
            username=self.username,
            challenge_response='2',
            challenge_state={'is_chap2': True},
            reject_attrs=self.chap2_error_attr,
        )

    @defer.inlineCallbacks
    def test_passcode_preauth_deny(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_DENY)
        yield self._test_auth(
            expect_code=pyrad.packet.AccessReject,
            username=self.username,
            challenge_response='2',
            challenge_state={'is_chap2': False},
        )

    @defer.inlineCallbacks
    def test_chap2_passcode_preauth_deny(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_DENY)
        yield self._test_auth(
            expect_code=pyrad.packet.AccessReject,
            username=self.username,
            challenge_response='2',
            challenge_state={'is_chap2': True},
            reject_attrs=self.chap2_error_attr,
        )

    @defer.inlineCallbacks
    def test_chap2_cpw_passcode_auth_deny(self):
        request = self.create_cpw_req()
        self.enqueue_duo_auth(duo_async.API_RESULT_DENY)
        yield self._test_auth(
            expect_code=pyrad.packet.AccessReject,
            username=self.username,
            challenge_response='2',
            challenge_state={'is_chap2': True, 'cpw_req': request},
            reject_attrs=self.chap2_error_attr,
        )

    @defer.inlineCallbacks
    def test_chap2_cpw_passcode_auth_allow(self):
        request = self.create_cpw_req()
        self.enqueue_duo_auth(duo_async.API_RESULT_ALLOW)
        yield self._test_auth(
            expect_code=pyrad.packet.AccessAccept,
            username=self.username,
            challenge_response='2',
            challenge_state={'is_chap2': True, 'cpw_req': request},
        )

    @defer.inlineCallbacks
    def test_chap2_cpw_passcode_auth_allow_primary_fail(self):
        request = self.create_cpw_req()
        self.enqueue_duo_auth(duo_async.API_RESULT_ALLOW)
        yield self._test_auth(
            expect_code=pyrad.packet.AccessAccept,
            username=self.username,
            challenge_response='4',
            challenge_state={'is_chap2': True, 'cpw_req': request},
        )

    @defer.inlineCallbacks
    def test_passcode_preauth_fail_closed_failmode_secure(self):
        self.enqueue_duo_preauth_error_fail_closed()
        yield self._test_auth(
            expect_code=pyrad.packet.AccessReject,
            username=self.username,
            challenge_response='2',
            challenge_state={'is_chap2': False},
            failmode=duo_async.FAILMODE_SECURE,
        )

    @defer.inlineCallbacks
    def test_chap2_passcode_preauth_fail_closed_failmode_secure(self):
        self.enqueue_duo_preauth_error_fail_closed()
        yield self._test_auth(
            expect_code=pyrad.packet.AccessReject,
            username=self.username,
            challenge_response='2',
            challenge_state={'is_chap2': True},
            reject_attrs=self.chap2_error_attr,
            failmode=duo_async.FAILMODE_SECURE,
        )

    @defer.inlineCallbacks
    def test_passcode_preauth_fail_closed_failmode_safe(self):
        self.enqueue_duo_preauth_error_fail_closed()
        yield self._test_auth(
            expect_code=pyrad.packet.AccessReject,
            username=self.username,
            challenge_response='2',
            challenge_state={'is_chap2': False},
            failmode=duo_async.FAILMODE_SAFE,
        )

    @defer.inlineCallbacks
    def test_chap2_passcode_preauth_fail_closed_failmode_safe(self):
        self.enqueue_duo_preauth_error_fail_closed()
        yield self._test_auth(
            expect_code=pyrad.packet.AccessReject,
            username=self.username,
            challenge_response='2',
            challenge_state={'is_chap2': True},
            reject_attrs=self.chap2_error_attr,
            failmode=duo_async.FAILMODE_SAFE,
        )

    @defer.inlineCallbacks
    def test_passcode_preauth_fail_open_failmode_secure(self):
        self.enqueue_duo_preauth_error_fail_open()
        yield self._test_auth(
            expect_code=pyrad.packet.AccessReject,
            username=self.username,
            challenge_response='2',
            challenge_state={'is_chap2': False},
            failmode=duo_async.FAILMODE_SECURE,
        )

    @defer.inlineCallbacks
    def test_chap2_passcode_preauth_fail_open_failmode_secure(self):
        self.enqueue_duo_preauth_error_fail_open()
        yield self._test_auth(
            expect_code=pyrad.packet.AccessReject,
            username=self.username,
            challenge_response='2',
            challenge_state={'is_chap2': True},
            reject_attrs=self.chap2_error_attr,
            failmode=duo_async.FAILMODE_SECURE,
        )

    @defer.inlineCallbacks
    def test_passcode_preauth_auth_auth_deny(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth(duo_async.API_RESULT_DENY)
        yield self._test_auth(
            expect_code=pyrad.packet.AccessReject,
            username=self.username,
            challenge_response='2',
            challenge_state={'is_chap2': False},
        )

    @defer.inlineCallbacks
    def test_chap2_passcode_preauth_auth_auth_deny(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth(duo_async.API_RESULT_DENY)
        yield self._test_auth(
            expect_code=pyrad.packet.AccessReject,
            username=self.username,
            challenge_response='2',
            challenge_state={'is_chap2': True},
            reject_attrs=self.chap2_error_attr,
        )

    @defer.inlineCallbacks
    def test_passcode_preauth_auth_auth_allow(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth(duo_async.API_RESULT_ALLOW)
        yield self._test_auth(
            expect_code=pyrad.packet.AccessAccept,
            username=self.username,
            challenge_response='2',
            challenge_state={'is_chap2': False},
        )

    @defer.inlineCallbacks
    def test_chap2_passcode_preauth_auth_auth_allow(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth(duo_async.API_RESULT_ALLOW)
        yield self._test_auth(
            expect_code=pyrad.packet.AccessAccept,
            username=self.username,
            challenge_response='2',
            challenge_state={'is_chap2': True},
        )

    @defer.inlineCallbacks
    def test_passcode_preauth_auth_auth_fail_open_failmode_secure(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth_error_fail_open()
        yield self._test_auth(
            expect_code=pyrad.packet.AccessReject,
            username=self.username,
            challenge_response='2',
            challenge_state={'is_chap2': False},
            failmode=duo_async.FAILMODE_SECURE,
        )

    @defer.inlineCallbacks
    def test_chap2_passcode_preauth_auth_auth_fail_open_failmode_secure(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth_error_fail_open()
        yield self._test_auth(
            expect_code=pyrad.packet.AccessReject,
            username=self.username,
            challenge_response='2',
            challenge_state={'is_chap2': True},
            reject_attrs=self.chap2_error_attr,
            failmode=duo_async.FAILMODE_SECURE,
        )

    @defer.inlineCallbacks
    def test_passcode_preauth_auth_auth_fail_closed_failmode_secure(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth_error_fail_closed()
        yield self._test_auth(
            expect_code=pyrad.packet.AccessReject,
            username=self.username,
            challenge_response='2',
            challenge_state={'is_chap2': False},
            failmode=duo_async.FAILMODE_SECURE,
        )

    @defer.inlineCallbacks
    def test_chap2_passcode_preauth_auth_auth_fail_closed_failmode_secure(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth_error_fail_closed()
        yield self._test_auth(
            expect_code=pyrad.packet.AccessReject,
            username=self.username,
            challenge_response='2',
            challenge_state={'is_chap2': True},
            reject_attrs=self.chap2_error_attr,
            failmode=duo_async.FAILMODE_SECURE,
        )

    @defer.inlineCallbacks
    def test_passcode_preauth_auth_auth_fail_closed_failmode_safe(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth_error_fail_closed()
        yield self._test_auth(
            expect_code=pyrad.packet.AccessReject,
            username=self.username,
            challenge_response='2',
            challenge_state={'is_chap2': False},
            failmode=duo_async.FAILMODE_SAFE,
        )

    @defer.inlineCallbacks
    def test_chap2_passcode_preauth_auth_auth_fail_closed_failmode_safe(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth_error_fail_closed()
        yield self._test_auth(
            expect_code=pyrad.packet.AccessReject,
            username=self.username,
            challenge_response='2',
            challenge_state={'is_chap2': True},
            reject_attrs=self.chap2_error_attr,
            failmode=duo_async.FAILMODE_SAFE,
        )

    # coverage for exempt usernames:

    @defer.inlineCallbacks
    def test_exempt_username(self):
        self.instance.protocol.exempt_usernames = [self.username]
        yield self._test_auth(
            expect_code=pyrad.packet.AccessAccept,
            username=self.username,
            password=self.password
        )

    @defer.inlineCallbacks
    def test_exempt_username_wrong_primary_auth_password(self):
        self.instance.protocol.exempt_usernames = [self.username]
        yield self._test_auth(
            expect_code=pyrad.packet.AccessReject,
            username=self.username,
            password=None
        )

    @defer.inlineCallbacks
    def test_exempt_username_chap2(self):
        self.instance.protocol.exempt_usernames = [self.username]
        yield self._test_auth(
            expect_code=pyrad.packet.AccessAccept,
            username=self.username,
            password=None,
            request_attrs=self.chap2_request_attrs
        )

    @defer.inlineCallbacks
    def test_exempt_username_chap2_cpw(self):
        self.instance.protocol.exempt_usernames = [self.username]
        yield self._test_auth(
            expect_code=pyrad.packet.AccessAccept,
            username=self.username,
            password=None,
            request_attrs=self.chap2_cpw_attrs
        )

    @defer.inlineCallbacks
    def test_exempt_username_chap2_bad_cpw(self):
        self.instance.protocol.exempt_usernames = [self.username]
        yield self._test_auth(
            expect_code=pyrad.packet.AccessReject,
            username=self.username,
            reject_attrs=self.chap2_error_attr,
            request_attrs=self.chap2_bad_cpw_attrs
        )

    # Concat tests! Don't bother testing every variation of preauth,
    # auth, and failmode, only concat detection and functionality.

    @defer.inlineCallbacks
    def test_concat_wrong_password(self):
        delim = self.instance.protocol.delim
        yield self._test_auth(
            expect_code=pyrad.packet.AccessReject,
            username=self.username,
            password=('not ' + self.password + delim + self.concat_factor),
        )

    @defer.inlineCallbacks
    def test_concat_wrong_delim(self):
        delim = self.instance.protocol.delim
        if delim == ',':
            wrong_delim = '|'
        else:
            # Ensure default delim doesn't work when non-default
            # configured.
            wrong_delim = ','
        yield self._test_auth(
            expect_code=pyrad.packet.AccessReject,
            username=self.username,
            password=(self.password + wrong_delim + self.concat_factor),
        )

    @defer.inlineCallbacks
    def test_concat_false_positive(self):
        # Induce a concat false-positive by setting delim to a character
        # contained in the primary auth password and not actually doing
        # concat. The second primary auth must succeed, leading to Duo
        # Auth with the auto-selected factor (*not* whatever the
        # incorrectly split factor part of the password was).
        server = self.instance.protocol
        server.delim = self.password[4]
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.assertNotEqual(self.password[5:], self.default_factor)
        self.enqueue_duo_auth(duo_async.API_RESULT_ALLOW,
                              factor=self.default_factor)
        yield self._test_auth(
            expect_code=pyrad.packet.AccessAccept,
            username=self.username,
            password=self.password,
        )

    @defer.inlineCallbacks
    def test_concat_delim_also_in_password(self):
        # Concat must still work if the password happens to also contain
        # the delim.
        delim = self.password[4]
        self.instance.protocol.delim = delim
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth(duo_async.API_RESULT_ALLOW,
                              factor=self.concat_factor)
        yield self._test_auth(
            expect_code=pyrad.packet.AccessAccept,
            username=self.username,
            password=(self.password + delim + self.concat_factor),
        )

    @defer.inlineCallbacks
    def test_concat_preauth_allow(self):
        delim = self.instance.protocol.delim
        self.enqueue_duo_preauth(duo_async.API_RESULT_ALLOW)
        yield self._test_auth(
            expect_code=pyrad.packet.AccessAccept,
            username=self.username,
            password=(self.password + delim + self.concat_factor),
        )

    @defer.inlineCallbacks
    def test_concat_preauth_deny(self):
        delim = self.instance.protocol.delim
        self.enqueue_duo_preauth(duo_async.API_RESULT_DENY)
        yield self._test_auth(
            expect_code=pyrad.packet.AccessReject,
            username=self.username,
            password=(self.password + delim + self.concat_factor),
        )

    @defer.inlineCallbacks
    def test_concat_auth_uses_specified_factor(self):
        delim = self.instance.protocol.delim
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth(duo_async.API_RESULT_ALLOW,
                              factor=self.concat_factor)
        yield self._test_auth(
            expect_code=pyrad.packet.AccessAccept,
            username=self.username,
            password=(self.password + delim + self.concat_factor),
        )

    @defer.inlineCallbacks
    def test_concat_auth_multi_char_delim(self):
        delim = ';;'
        self.instance.protocol.delim = delim
        self.enqueue_duo_preauth(duo_async.API_RESULT_ALLOW)
        yield self._test_auth(
            expect_code=pyrad.packet.AccessAccept,
            username=self.username,
            password=(self.password + delim + self.concat_factor),
        )

    @defer.inlineCallbacks
    def test_concat_rejected_if_disabled(self):
        self.instance.protocol.allow_concat = False
        delim = self.instance.protocol.delim
        yield self._test_auth(
            expect_code=pyrad.packet.AccessReject,
            username=self.username,
            password=(self.password + delim + self.concat_factor),
        )

    @defer.inlineCallbacks
    def test_concat_disabled_with_non_concat_password(self):
        self.instance.protocol.allow_concat = False
        self.enqueue_duo_preauth(duo_async.API_RESULT_ALLOW)
        yield self._test_auth(
            expect_code=pyrad.packet.AccessAccept,
            username=self.username,
            password=self.password,
        )

    @defer.inlineCallbacks
    def test_concat_delim_only(self):
        self.instance.protocol.allow_concat = True
        delim = self.instance.protocol.delim
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth(duo_async.API_RESULT_ALLOW,
                              factor=self.default_factor)
        yield self._test_auth(
            expect_code=pyrad.packet.AccessAccept,
            username=self.username,
            password=(self.password + delim)
        )

    @defer.inlineCallbacks
    def test_chap2_cpw_preauth_deny(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_DENY)
        yield self._test_auth(pyrad.packet.AccessReject,
                              username=self.username,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_cpw_attrs)

    @defer.inlineCallbacks
    def test_chap2_cpw_preauth_enroll_challenge(self):
        # This test verifies a Rejection.  Should the implementation
        # challenge in this case?  The user should have enrolled
        # in their original login attempt, so either they unenrolled
        # between their initial login and their CPW request, or
        # this is a forged cpw request...
        self.enqueue_duo_preauth(duo_async.API_RESULT_ENROLL)
        yield self._test_auth(pyrad.packet.AccessReject,
                              username=self.username,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_cpw_attrs)

    @defer.inlineCallbacks
    def test_chap2_cpw_preauth_allow(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_ALLOW)
        yield self._test_auth(pyrad.packet.AccessAccept,
                              username=self.username,
                              password=None,
                              request_attrs=self.chap2_cpw_attrs)

    @defer.inlineCallbacks
    def test_chap2_cpw_preauth_apifail_closed_secure(self):
        self.enqueue_duo_preauth_error_fail_closed()
        yield self._test_auth(pyrad.packet.AccessReject,
                              username=self.username,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_cpw_attrs)

    @defer.inlineCallbacks
    def test_chap2_cpw_preauth_apifail_closed_safe(self):
        self.enqueue_duo_preauth_error_fail_closed()
        yield self._test_auth(pyrad.packet.AccessReject,
                              username=self.username,
                              failmode=duo_async.FAILMODE_SAFE,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_cpw_attrs)

    @defer.inlineCallbacks
    def test_chap2_cpw_preauth_apifail_open_safe(self):
        self.enqueue_duo_preauth_error_fail_open()
        yield self._test_auth(pyrad.packet.AccessAccept,
                              failmode=duo_async.FAILMODE_SAFE,
                              username=self.username,
                              request_attrs=self.chap2_cpw_attrs)

    @defer.inlineCallbacks
    def test_chap2_cpw_preauth_apifail_open_secure(self):
        self.enqueue_duo_preauth_error_fail_open()
        yield self._test_auth(pyrad.packet.AccessReject,
                              username=self.username,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_cpw_attrs)

    @defer.inlineCallbacks
    def test_chap2_cpw_preauth_allow_primary_fail(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_ALLOW)
        yield self._test_auth(pyrad.packet.AccessReject,
                              username=self.username,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_bad_cpw_attrs)

    @defer.inlineCallbacks
    def test_chap2_cpw_preauth_auth_duoauth_allow_primary_succeed(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth(duo_async.API_RESULT_ALLOW)
        yield self._test_auth(pyrad.packet.AccessAccept,
                              username=self.username,
                              request_attrs=self.chap2_cpw_attrs)

    @defer.inlineCallbacks
    def test_chap2_cpw_preauth_auth_duoauth_allow_primary_fail(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth(duo_async.API_RESULT_ALLOW)
        yield self._test_auth(pyrad.packet.AccessReject,
                              username=self.username,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_bad_cpw_attrs)

    @defer.inlineCallbacks
    def test_chap2_cpw_preauth_fail_open_failmode_secure(self):
        self.enqueue_duo_preauth_error_fail_open()
        yield self._test_auth(pyrad.packet.AccessReject,
                              failmode=duo_async.FAILMODE_SECURE,
                              username=self.username,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_cpw_attrs)

    @defer.inlineCallbacks
    def test_chap2_cpw_preauth_fail_closed_failmode_secure(self):
        self.enqueue_duo_preauth_error_fail_closed()
        yield self._test_auth(pyrad.packet.AccessReject,
                              failmode=duo_async.FAILMODE_SECURE,
                              username=self.username,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_cpw_attrs)

    @defer.inlineCallbacks
    def test_chap2_cpw_preauth_fail_open_failmode_safe(self):
        self.enqueue_duo_preauth_error_fail_open()
        yield self._test_auth(pyrad.packet.AccessAccept,
                              failmode=duo_async.FAILMODE_SAFE,
                              username=self.username,
                              request_attrs=self.chap2_cpw_attrs)

    @defer.inlineCallbacks
    def test_chap2_cpw_preauth_fail_closed_failmode_safe(self):
        self.enqueue_duo_preauth_error_fail_closed()
        yield self._test_auth(pyrad.packet.AccessReject,
                              failmode=duo_async.FAILMODE_SAFE,
                              username=self.username,
                              reject_attrs=self.chap2_error_attr,
                              request_attrs=self.chap2_cpw_attrs)


class TestRadiusServerAutoConfig(RadiusServerAutoTestRunner):
    """Tests that an instance of DuoAutoRadiusServer can authenticate correctly
    with varying configurations, with a focus on proper second factor selection.
    """

    @defer.inlineCallbacks
    def _test_factor_config(self, config_factors, user_factors, chosen_factor,
                            preauth_has_default_factor=True,
                            auth_result_code=pyrad.packet.AccessAccept):
        """Performs preauth and regular auth given a factor configuration.

        Factors may be specified by an admin's authproxy configuration, results
        from preauth(), and the users themselves while attempting login. In the
        case that no factors are chosen by either 3, or in the case that a factor
        is unavailable by any one of the 3, a proper fallback factor must be
        properly chosen.

        Calls to preauth and auth are essentially mocked out here for when they
        are run during _test_auth().
        """
        server = self.instance.protocol
        duo_client = server.client

        # enqueue a mock preauth
        factors = dict(enumerate(user_factors, start=1))
        if preauth_has_default_factor:
            factors['default'] = self.default_factor
        method_call = duo_client.preauth(mox.IsA(six.string_types),
                                         mox.Or(mox.IsA(type(None)), mox.IsA((six.string_types))))
        method_call.AndReturn({
            'result': duo_async.API_RESULT_AUTH,
            'status': duo_async.API_RESULT_AUTH,
            'factors': factors,
            'prompt': '?',
        })

        # enqueue a mock auth
        if chosen_factor:
            method_call = duo_client.auth(mox.IsA(six.string_types),
                                          chosen_factor,
                                          None)
            method_call.AndReturn({
                'result': duo_async.API_RESULT_ALLOW,
                'status': duo_async.API_RESULT_ALLOW,
            })

        # the actual test
        server.factors = config_factors
        yield self._test_auth(auth_result_code,
                              username=self.username,
                              password=self.password)

    @defer.inlineCallbacks
    def test_factor_config_empty(self):
        server = self.instance.protocol

        # no admin-configured factors => choose whatever preauth said
        # was the best default.
        server.factors = tuple()

        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        yield self._test_auth(pyrad.packet.AccessReject,
                              username=self.username,
                              password=self.password)

    @defer.inlineCallbacks
    def test_factor_config_no_matching_factors(self):
        # No admin-configured factors => choose whatever preauth said
        # was the best default.
        yield self._test_factor_config(
            ['phone'],
            ['push1'],
            chosen_factor=None,
            auth_result_code=pyrad.packet.AccessReject,
        )

    @defer.inlineCallbacks
    def test_factor_config_matching_factor(self):
        # Pick the first admin-configured factor the user can do.
        yield self._test_factor_config(['push', 'phone'],
                                       ['sms1', 'phone2'],
                                       'phone')

    @defer.inlineCallbacks
    def test_factor_config_auto_no_match(self):
        # Prefer push but pick default if it's not present.
        yield self._test_factor_config(['push', 'auto'],
                                       ['phone1', 'phone2'],
                                       self.default_factor)

    @defer.inlineCallbacks
    def test_factor_config_auto_match(self):
        # Prefer push but pick default if it's not present.
        yield self._test_factor_config(['push', 'auto'],
                                       ['push1', 'phone1'],
                                       'push')

    @defer.inlineCallbacks
    def test_factor_config_auto_useless_match(self):
        # Anything after 'auto' is ignored in config. Thus, default is
        # chosen despite 'push' matching.
        yield self._test_factor_config(['auto', 'push'],
                                       ['push1', 'phone1'],
                                       self.default_factor)

    @defer.inlineCallbacks
    def test_factor_config_no_factors_no_passcode(self):
        # If passcode is not an allowed factor and preauth does not
        # return any out-of-band factors for the user, reject.
        yield self._test_factor_config(
            config_factors=['auto', 'push'],
            user_factors=[],
            chosen_factor=None,
            preauth_has_default_factor=False,
            auth_result_code=pyrad.packet.AccessReject,
        )

    @defer.inlineCallbacks
    def test_factor_config_no_default_factor_no_passcode(self):
        # If passcode is not an allowed factor and preauth does not
        # return a default out-of-band factor for the user, reject.

        # (Should never happen because apiserv sends empty factors for
        # this case so it would actually be
        # test_factor_config_no_factors_no_passcode(), but authproxy
        # handles it for robustness.)
        yield self._test_factor_config(
            config_factors=['auto', 'push'],
            user_factors=['phone1'],
            chosen_factor=None,
            preauth_has_default_factor=False,
            auth_result_code=pyrad.packet.AccessReject,
        )

    @defer.inlineCallbacks
    def test_factor_config_passcode_no_factors(self):
        # If passcode is an allowed factor and preauth does not return
        # any out-of-band factors for the user, send passcode
        # challenge.
        yield self._test_factor_config(
            config_factors=['auto', 'push', 'passcode'],
            user_factors=[],
            chosen_factor=None,
            preauth_has_default_factor=False,
            auth_result_code=pyrad.packet.AccessChallenge,
        )

    @defer.inlineCallbacks
    def test_factor_config_passcode_no_default_factor(self):
        # If passcode is an allowed factor and preauth does not return
        # a default out-of-band factor for the user, send passcode
        # challenge.

        # (Should never happen because apiserv sends empty factors for
        # this case so it would actually be
        # test_factor_config_passcode_no_factors(), but authproxy handles it
        # for robustness.)
        yield self._test_factor_config(
            config_factors=['auto', 'push', 'passcode'],
            user_factors=['phone1'],
            chosen_factor=None,
            preauth_has_default_factor=False,
            auth_result_code=pyrad.packet.AccessChallenge,
        )

    @defer.inlineCallbacks
    def test_factor_config_passthrough_no_default_factor(self):
        # If proxying EAP but preauth does not return a default
        # out-of-band factor for the user, reject.
        server = self.instance.protocol
        duo_client = server.client

        auth_packet = yield self._make_proxying_auth_packet()

        # enqueue preauth
        method_call = duo_client.preauth(mox.IsA(six.string_types),
                                         mox.Or(mox.IsA(type(None)), mox.IsA((six.string_types))))
        method_call.AndReturn({
            'result': duo_async.API_RESULT_AUTH,
            'status': duo_async.API_RESULT_AUTH,
            'prompt': '?',
            'factors': dict(),
        })

        # the actual test
        response_packet = yield self._test_process_packet(auth_packet)
        self.assertEqual(response_packet.code,
                         pyrad.packet.AccessReject)

    @defer.inlineCallbacks
    def test_radius_proxy_non_radius_client_module(self):
        # All non-PAP auth requests are rejected if the client module
        # cannot proxy RADIUS packets.
        auth_packet = pyrad.packet.AuthPacket(
            code=pyrad.packet.AccessRequest,
            authenticator=pyrad.packet.AuthPacket.CreateAuthenticator(),
            dict=radius_dictionary(),
            id=self.radius_packet_id.next(),
            secret=self.secret,
        )
        eap_message = 'Oh, whatever!'
        auth_packet['User-Name'] = 'whoever'
        auth_packet['EAP-Message'] = eap_message

        response_packet = yield self._test_process_packet(auth_packet)
        self.assertEqual(response_packet.code,
                         pyrad.packet.AccessReject)
        self.assertEqual([base.RADIUS_PROXY_NOT_IMPLEMENTED_CLIENT_MSG],
                         response_packet['Reply-Message'])

    @defer.inlineCallbacks
    def _make_proxying_auth_packet(self):
        auth_packet = pyrad.packet.AuthPacket(
            code=pyrad.packet.AccessRequest,
            authenticator=pyrad.packet.AuthPacket.CreateAuthenticator(),
            dict=radius_dictionary(),
            id=self.radius_packet_id.next(),
            secret=self.secret,
        )
        eap_message = 'Oh, whatever!'
        auth_packet['User-Name'] = 'whoever'
        auth_packet['EAP-Message'] = eap_message

        self.client_instance.stopService()
        self.client_instance = MockRadiusClientModule(self.client_config)
        self.instance.primary_ator = self.client_instance
        self.instance.protocol.primary_ator = self.client_instance

        # Send an initial request and make the fake backing RADIUS
        # server send an AccessChallenge the authproxy server module
        # will proxy.
        response_packet = yield self._test_process_packet(auth_packet)
        self.assertEqual(response_packet.code,
                         pyrad.packet.AccessChallenge)
        self.assertEqual([MockRadiusClientModule.fake_challenge_state],
                         response_packet['State'])

        # Now send a response to the challenge, test that the
        # authproxy server module looks up the challenge state, and
        # make the fake backing RADIUS server send AccessAccept.
        auth_packet.id = self.radius_packet_id.next()
        auth_packet['State'] = MockRadiusClientModule.fake_challenge_state
        self.client_instance.fake_challenge_code = pyrad.packet.AccessAccept
        defer.returnValue(auth_packet)

    @defer.inlineCallbacks
    def _test_radius_proxy_radius_client_module(self,
                                                duo_auth_passes,
                                                expect_final_code):
        auth_packet = yield self._make_proxying_auth_packet()

        # Because it is done proxying primary auth the authproxy
        # server module must return whatever Duo auth said.
        # Coverage for actual auth part of such a sequence of
        # events is elsewhere so just let this test allow or deny
        # at preauth.
        if duo_auth_passes:
            self.enqueue_duo_preauth(duo_async.API_RESULT_ALLOW)
        else:
            self.enqueue_duo_preauth(duo_async.API_RESULT_DENY)

        response_packet = yield self._test_process_packet(auth_packet)
        self.assertEqual(response_packet.code,
                         expect_final_code)
        if duo_auth_passes:
            self.assertEqual([MockRadiusClientModule.fake_challenge_state],
                             response_packet['State'])

    @defer.inlineCallbacks
    def test_radius_proxy_radius_client_module_duo_allow(self):
        yield self._test_radius_proxy_radius_client_module(
            duo_auth_passes=True,
            expect_final_code=pyrad.packet.AccessAccept,
        )

    @defer.inlineCallbacks
    def test_radius_proxy_radius_client_module_duo_deny(self):
        yield self._test_radius_proxy_radius_client_module(
            duo_auth_passes=False,
            expect_final_code=pyrad.packet.AccessReject,
        )

    def test_concat_config_parsing(self):
        # Only need to check that concat mode config values make it from
        # the config file to the right place. Their actual effects when
        # changed are tested by the cases above.
        delim = 'foo'
        server_config = {
            'radius_ip_1': self.radius_ip,
            'radius_secret_1': self.secret,
            'ikey': 'nope',
            'skey': 'nuh-uh',
            'delimiter': delim,
            'allow_concat': '0',
        }
        self.instance = duoauthproxy.modules.radius_server_auto.Module(
            ConfigDict(server_config),
            None,
            'radius_server_auto',
        )

        server = self.instance.protocol
        self.assertEqual(server.delim, delim)
        self.assertIs(server.allow_concat, False)

    def test_exempt_username_config_parsing(self):
        server_config = {
            'radius_ip_1': self.radius_ip,
            'radius_secret_1': self.secret,
            'ikey': 'nope',
            'skey': 'nuh-uh',
            'allow_concat': '0',
            'exempt_username_1': 'testuser1',
            'exempt_username_2': 'testuser2',
        }
        self.instance = duoauthproxy.modules.radius_server_auto.Module(
            ConfigDict(server_config),
            None,
            'radius_server_auto',
        )

        server = self.instance.protocol
        self.assertEqual(server.exempt_usernames, ['testuser1', 'testuser2'])


class TestRadiusServerAuto(RadiusServerAutoTestCases, RadiusServerAutoTestRunner):
    """Runs all tests in RadiusServerAutoTestCases."""
    pass


class TestRadiusServerAutoPassThrough(RadiusServerAutoTestCases,
                                      ServerModuleTestPassThrough,
                                      RadiusServerAutoTestRunner):
    """Runs all tests in RadiusServerAutoTestCases with an additional
    pass-through attribute to be sent and received by the server."""
    pass


class TestRadiusServerAutoPassThroughMultipleValues(RadiusServerAutoTestCases,
                                                    ServerModuleTestPassThroughMultipleValues,
                                                    RadiusServerAutoTestRunner):
    """Runs all tests in RadiusServerAutoTestCases with multiple additional
    pass-through attributes to be sent and received by the server."""
    pass


class TestRadiusAutoProtectedConfigParsing(ServerProtectedConfigParsingTestCase):
    """Tests that an instance of DuoAutoRadiusServer with a protected
    config properly accesses and/or decrypts certain protected keys defined in
    ProtectedValidatingConfigDict."""

    def test_plain_keys_use_protected_api(self):
        # Both of these new keys should be accessed with get_protected_str()
        # though neither are encrypted
        instance = self._make_instance({
            'radius_secret_1': 'secret!',
            'skey': 'skey!'
        }, duoauthproxy.modules.radius_server_auto, 'radius_server_auto')
        self.assertEqual(instance.protocol.client.duo_creds.get_secret(), 'skey!')
        self.assertEqual(instance.protocol.secrets.values()[0], 'secret!')

    def test_protected_keys(self):
        # radius_secret_protected_1 should decrypt first with get_protected_str()
        if protect.PROTECT_ENABLED:
            protected = protect.protect('secret!')
            instance = self._make_instance(
                {
                    'radius_secret_protected_1': protected,
                    'skey': 'skey!'
                },
                duoauthproxy.modules.radius_server_auto,
                'radius_server_auto'
            )
            self.assertEqual(instance.protocol.client.duo_creds.get_secret(), 'skey!')
            self.assertEqual(instance.protocol.secrets.values()[0], 'secret!')


class TestRadiusServerAutoAuthLogging(AuthLoggingTest, RadiusServerAutoTestRunner):
    """Tests SIEM logging from log.auth_standard() calls during various
    authentication flows in DuoAutoRadiusServer."""

    @defer.inlineCallbacks
    def test_primary_auth_failure_logged(self):
        yield self._test_auth(pyrad.packet.AccessReject,
                              username=self.username,
                              password='wrong_password')

        self.assert_logged(
            [{
                'msg': "Primary credentials rejected - Hello",
                'auth_stage': log.AUTH_PRIMARY,
                'status': log.AUTH_REJECT,
            }]
        )

    @defer.inlineCallbacks
    def test_primary_auth_error_failure_logged(self):
        def bad_authenticator(username, password, client_ip, radius_attrs):
            raise base.AuthError('ERROR')
        self.patch(self.instance.protocol.primary_ator, 'authenticate', bad_authenticator)

        yield self._test_auth(pyrad.packet.AccessReject,
                              username=self.username,
                              password=self.password)

        self.assert_logged(
            [{
                'msg': 'Error performing primary authentication: ERROR',
                'auth_stage': log.AUTH_PRIMARY,
                'status': log.AUTH_ERROR,
            }]
        )

    @defer.inlineCallbacks
    def test_exempt_user_success_logged(self):
        self.instance.protocol.exempt_usernames = [self.username]
        yield self._test_auth(pyrad.packet.AccessAccept,
                              username=self.username,
                              password=self.password)

        self.assert_logged(
            [{
                'msg': 'Primary authentication successful - Hello',
                'auth_stage': log.AUTH_PRIMARY,
                'status': log.AUTH_ALLOW,
                'username': self.username,
            }, {
                'msg': 'User exempted from 2FA',
                'auth_stage': log.AUTH_SECONDARY,
                'status': log.AUTH_ALLOW,
                'username': self.username,
            }]
        )

    @defer.inlineCallbacks
    def test_preauth_allow_success_logged(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_ALLOW)
        yield self._test_auth(pyrad.packet.AccessAccept,
                              username=self.username,
                              password=self.password)

        self.assert_logged(
            [{
                'msg': 'Primary authentication successful - Hello',
                'auth_stage': log.AUTH_PRIMARY,
                'status': log.AUTH_ALLOW,
            }, {
                'msg': 'allow',
                'auth_stage': log.AUTH_SECONDARY,
                'status': log.AUTH_ALLOW,
            }]
        )

    @defer.inlineCallbacks
    def test_preauth_deny_failure_logged(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_DENY)
        yield self._test_auth(pyrad.packet.AccessReject,
                              username=self.username,
                              password=self.password)

        self.assert_logged(
            [{
                'msg': 'Primary authentication successful - Hello',
                'auth_stage': log.AUTH_PRIMARY,
                'status': log.AUTH_ALLOW,
            }, {
                'msg': 'deny',
                'auth_stage': log.AUTH_SECONDARY,
                'status': log.AUTH_REJECT,
            }]
        )

    @defer.inlineCallbacks
    def test_preauth_fail_open_success_logged(self):
        self.enqueue_duo_preauth_error_fail_open()

        yield self._test_auth(pyrad.packet.AccessAccept,
                              failmode=duo_async.FAILMODE_SAFE,
                              username=self.username,
                              password=self.password)

        self.assert_logged(
            [{
                'msg': 'Primary authentication successful - Hello',
                'auth_stage': log.AUTH_PRIMARY,
                'status': log.AUTH_ALLOW,
            }, {
                'msg': duo_async.FAILMODE_SAFE_MSG,
                'auth_stage': log.AUTH_SECONDARY,
                'status': log.AUTH_ALLOW,
            }]
        )

    @defer.inlineCallbacks
    def test_preauth_fail_closed_failure_logged(self):
        self.enqueue_duo_preauth_error_fail_closed()

        yield self._test_auth(pyrad.packet.AccessReject,
                              failmode=duo_async.FAILMODE_SECURE,
                              username=self.username,
                              password=self.password)

        self.assert_logged(
            [{
                'msg': 'Primary authentication successful - Hello',
                'auth_stage': log.AUTH_PRIMARY,
                'status': log.AUTH_ALLOW,
            }, {
                'msg': duo_async.FAILMODE_SECURE_MSG,
                'auth_stage': log.AUTH_SECONDARY,
                'status': log.AUTH_REJECT,
            }]
        )

    @defer.inlineCallbacks
    def test_chap2_cpw_preauth_exempt_username_success_logged(self):
        self.instance.protocol.exempt_usernames = [self.username]
        yield self._test_auth(
            expect_code=pyrad.packet.AccessAccept,
            username=self.username,
            request_attrs=self.chap2_cpw_attrs,
        )

        self.assert_logged(
            [{
                'msg': "Duo preauth returned 'allow': 'User exempted from 2FA'",
                'auth_stage': log.AUTH_SECONDARY,
                'status': log.AUTH_ALLOW,
            }, {
                'auth_stage': log.AUTH_PRIMARY,
                'status': log.AUTH_ALLOW,
            }]
        )

    @defer.inlineCallbacks
    def test_chap2_cpw_preauth_fail_open_success_logged(self):
        self.enqueue_duo_preauth_error_fail_open()
        yield self._test_auth(
            failmode=duo_async.FAILMODE_SAFE,
            expect_code=pyrad.packet.AccessAccept,
            username=self.username,
            request_attrs=self.chap2_cpw_attrs,
        )

        self.assert_logged(
            [{
                'msg': duo_async.FAILMODE_SAFE_MSG,
                'auth_stage': log.AUTH_SECONDARY,
                'status': log.AUTH_ALLOW,
            }, {
                'auth_stage': log.AUTH_PRIMARY,
                'status': log.AUTH_ALLOW,
            }]
        )

    @defer.inlineCallbacks
    def test_chap2_cpw_preauth_fail_closed_failure_logged(self):
        self.enqueue_duo_preauth_error_fail_closed()
        yield self._test_auth(
            failmode=duo_async.FAILMODE_SECURE,
            expect_code=pyrad.packet.AccessReject,
            username=self.username,
            request_attrs=self.chap2_cpw_attrs,
        )

        self.assert_logged(
            [{
                'msg': duo_async.FAILMODE_SECURE_MSG,
                'auth_stage': log.AUTH_SECONDARY,
                'status': log.AUTH_REJECT,
            }]
        )

    @defer.inlineCallbacks
    def test_chap2_cpw_preauth_allow_success_logged(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_ALLOW)
        yield self._test_auth(
            expect_code=pyrad.packet.AccessAccept,
            username=self.username,
            request_attrs=self.chap2_cpw_attrs,
        )

        self.assert_logged(
            [{
                'msg': "Duo preauth returned 'allow': 'allow'",
                'auth_stage': log.AUTH_SECONDARY,
                'status': log.AUTH_ALLOW,
            }, {
                'auth_stage': log.AUTH_PRIMARY,
                'status': log.AUTH_ALLOW,
            }]
        )

    @defer.inlineCallbacks
    def test_chap2_cpw_preauth_deny_failure_logged(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_DENY)
        yield self._test_auth(
            expect_code=pyrad.packet.AccessReject,
            username=self.username,
            request_attrs=self.chap2_cpw_attrs,
        )

        self.assert_logged(
            [{
                'msg': "Duo preauth returned 'deny': 'deny'",
                'auth_stage': log.AUTH_SECONDARY,
                'status': log.AUTH_REJECT,
            }]
        )

    @defer.inlineCallbacks
    def test_chap2_cpw_preauth_enroll_logged(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_ENROLL)
        yield self._test_auth(
            expect_code=pyrad.packet.AccessReject,
            username=self.username,
            request_attrs=self.chap2_cpw_attrs,
        )

        self.assert_logged(
            [{
                'msg': log.AUTH_ENROLL_MSG,
                'auth_stage': log.AUTH_SECONDARY,
                'status': log.AUTH_REJECT,
            }]
        )

    @defer.inlineCallbacks
    def test_duo_auth_allow_success_logged(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth(duo_async.API_RESULT_ALLOW)

        yield self._test_auth(pyrad.packet.AccessAccept,
                              username=self.username,
                              password=self.password)

        self.assert_logged(
            [{
                'auth_stage': log.AUTH_PRIMARY,
                'status': log.AUTH_ALLOW,
            }, {
                'msg': 'allow',
                'auth_stage': log.AUTH_SECONDARY,
                'status': log.AUTH_ALLOW,
            }]
        )

    @defer.inlineCallbacks
    def test_duo_auth_failed_logged(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth(duo_async.API_RESULT_DENY)

        yield self._test_auth(pyrad.packet.AccessReject,
                              username=self.username,
                              password=self.password)

        self.assert_logged(
            [{
                'auth_stage': log.AUTH_PRIMARY,
                'status': log.AUTH_ALLOW,
            }, {
                'msg': 'deny',
                'auth_stage': log.AUTH_SECONDARY,
                'status': log.AUTH_REJECT,
            }]
        )

    @defer.inlineCallbacks
    def test_duo_auth_fail_open_success_logged(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth_error_fail_open()
        yield self._test_auth(pyrad.packet.AccessAccept,
                              failmode=duo_async.FAILMODE_SAFE,
                              username=self.username,
                              password=self.password)

        self.assert_logged(
            [{
                'auth_stage': log.AUTH_PRIMARY,
                'status': log.AUTH_ALLOW,
            }, {
                'msg': duo_async.FAILMODE_SAFE_MSG,
                'auth_stage': log.AUTH_SECONDARY,
                'status': log.AUTH_ALLOW,
            }]
        )

    @defer.inlineCallbacks
    def test_duo_auth_fail_closed_failure_logged(self):
        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        self.enqueue_duo_auth_error_fail_closed()
        yield self._test_auth(pyrad.packet.AccessReject,
                              failmode=duo_async.FAILMODE_SECURE,
                              username=self.username,
                              password=self.password)

        self.assert_logged(
            [{
                'auth_stage': log.AUTH_PRIMARY,
                'status': log.AUTH_ALLOW,
            }, {
                'msg': duo_async.FAILMODE_SECURE_MSG,
                'auth_stage': log.AUTH_SECONDARY,
                'status': log.AUTH_ERROR,
            }]
        )

    @defer.inlineCallbacks
    def test_missing_username_failure_logged(self):
        yield self._test_auth(pyrad.packet.AccessReject,
                              username=None,
                              password=self.password)
        self.assert_logged(
            [{
                'msg': 'No username provided',
                'auth_stage': log.AUTH_PRIMARY,
                'status': log.AUTH_ERROR,
            }]
        )

    @defer.inlineCallbacks
    def test_no_duo_factor_in_cpw_failure_logged(self):
        self.instance.protocol.factors = tuple()

        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        yield self._test_auth(
            expect_code=pyrad.packet.AccessReject,
            username=self.username,
            request_attrs=self.chap2_cpw_attrs,
        )

        self.assert_logged(
            [{
                'msg': 'User has no Duo factors usable with this configuration',
                'auth_stage': log.AUTH_SECONDARY,
                'status': log.AUTH_REJECT,
            }]
        )

    @defer.inlineCallbacks
    def test_missing_password_failure_logged(self):
        yield self._test_auth(pyrad.packet.AccessReject,
                              username=self.username,
                              password=None)
        self.assert_logged(
            [{
                'msg': 'Missing or improperly-formatted password',
                'auth_stage': log.AUTH_PRIMARY,
                'status': log.AUTH_ERROR,
            }]
        )

    @defer.inlineCallbacks
    def test_unimplemented_proxy_failure_logged(self):
        auth_packet = pyrad.packet.AuthPacket(
            code=pyrad.packet.AccessRequest,
            authenticator=pyrad.packet.AuthPacket.CreateAuthenticator(),
            dict=radius_dictionary(),
            id=self.radius_packet_id.next(),
            secret=self.secret,
        )
        eap_message = 'Oh, whatever!'
        auth_packet['User-Name'] = 'whoever'
        auth_packet['EAP-Message'] = eap_message

        yield self._test_process_packet(auth_packet)
        self.assert_logged(
            [{
                'msg': base.RADIUS_PROXY_NOT_IMPLEMENTED_CLIENT_MSG,
                'auth_stage': log.AUTH_PRIMARY,
                'status': log.AUTH_ERROR,
            }]
        )

    @defer.inlineCallbacks
    def test_proxy_error_failure_logged(self):
        auth_packet = pyrad.packet.AuthPacket(
            code=pyrad.packet.AccessRequest,
            authenticator=pyrad.packet.AuthPacket.CreateAuthenticator(),
            dict=radius_dictionary(),
            id=self.radius_packet_id.next(),
            secret=self.secret,
        )
        eap_message = 'Oh, whatever!'
        auth_packet['User-Name'] = 'whoever'
        auth_packet['EAP-Message'] = eap_message
        self.client_instance.stopService()
        self.client_instance = ErroringRadiusClientModule(self.client_config)
        self.instance.primary_ator = self.client_instance
        self.instance.protocol.primary_ator = self.client_instance

        yield self._test_process_packet(auth_packet)

        self.assert_logged(
            [{
                'msg': 'Error performing primary authentication: ERROR',
                'auth_stage': log.AUTH_PRIMARY,
                'status': log.AUTH_ERROR,
            }]
        )

    @defer.inlineCallbacks
    def test_proxying_primary_auth_reject_logged(self):
        auth_packet = pyrad.packet.AuthPacket(
            code=pyrad.packet.AccessRequest,
            authenticator=pyrad.packet.AuthPacket.CreateAuthenticator(),
            dict=radius_dictionary(),
            id=self.radius_packet_id.next(),
            secret=self.secret,
        )
        eap_message = 'Oh, whatever!'
        auth_packet['User-Name'] = 'whoever'
        auth_packet['EAP-Message'] = eap_message

        self.client_instance.stopService()
        self.client_instance = MockRadiusClientModule(self.client_config)
        # Cause client to send a Reject code
        self.client_instance.fake_challenge_code = pyrad.packet.AccessReject
        self.instance.primary_ator = self.client_instance
        self.instance.protocol.primary_ator = self.client_instance

        yield self._test_process_packet(auth_packet)

        self.assert_logged(
            [{
                'msg': 'Primary credentials rejected - {0}'.format(base.NO_REPLY_MESSAGE),
                'auth_stage': log.AUTH_PRIMARY,
                'status': log.AUTH_REJECT,
            }]
        )

    @defer.inlineCallbacks
    def test_proxying_primary_auth_success_logged(self):
        auth_packet = pyrad.packet.AuthPacket(
            code=pyrad.packet.AccessRequest,
            authenticator=pyrad.packet.AuthPacket.CreateAuthenticator(),
            dict=radius_dictionary(),
            id=self.radius_packet_id.next(),
            secret=self.secret,
        )
        eap_message = 'Oh, whatever!'
        auth_packet['User-Name'] = 'whoever'
        auth_packet['EAP-Message'] = eap_message

        self.client_instance.stopService()
        self.client_instance = MockRadiusClientModule(self.client_config)
        self.instance.primary_ator = self.client_instance
        self.instance.protocol.primary_ator = self.client_instance

        yield self._test_process_packet(auth_packet)

        auth_packet.id = self.radius_packet_id.next()
        auth_packet['State'] = MockRadiusClientModule.fake_challenge_state
        self.client_instance.fake_challenge_code = pyrad.packet.AccessAccept
        self.enqueue_duo_preauth(duo_async.API_RESULT_ALLOW)
        yield self._test_process_packet(auth_packet)

        self.assert_logged(
            [{
                'msg': 'Primary authentication successful - {0}'.format(base.NO_REPLY_MESSAGE),
                'auth_stage': log.AUTH_PRIMARY,
                'status': log.AUTH_ALLOW,
            }, {
                'auth_stage': log.AUTH_SECONDARY,
                'status': log.AUTH_ALLOW,
            }]
        )

    @defer.inlineCallbacks
    def test_proxying_unexpected_packet_failure_logged(self):
        auth_packet = pyrad.packet.AuthPacket(
            code=pyrad.packet.AccessRequest,
            authenticator=pyrad.packet.AuthPacket.CreateAuthenticator(),
            dict=radius_dictionary(),
            id=self.radius_packet_id.next(),
            secret=self.secret,
        )
        eap_message = 'Oh, whatever!'
        auth_packet['User-Name'] = 'whoever'
        auth_packet['EAP-Message'] = eap_message

        self.client_instance.stopService()
        self.client_instance = MockRadiusClientModule(self.client_config)
        # Cause client to send a nonsense packet code
        self.client_instance.fake_challenge_code = 8675309
        self.instance.primary_ator = self.client_instance
        self.instance.protocol.primary_ator = self.client_instance

        yield self._test_process_packet(auth_packet)

        self.assert_logged(
            [{
                'msg': 'Unexpected packet in primary authentication',
                'auth_stage': log.AUTH_PRIMARY,
                'status': log.AUTH_ERROR,
            }]
        )

    @defer.inlineCallbacks
    def test_no_duo_factor_failure_logged(self):
        self.instance.protocol.factors = tuple()

        self.enqueue_duo_preauth(duo_async.API_RESULT_AUTH)
        yield self._test_auth(pyrad.packet.AccessReject,
                              username=self.username,
                              password=self.password)

        self.assert_logged(
            [{
                'msg': 'Primary authentication successful - Hello',
                'auth_stage': log.AUTH_PRIMARY,
                'status': log.AUTH_ALLOW,
            }, {
                'msg': 'User has no Duo factors usable with this configuration',
                'auth_stage': log.AUTH_SECONDARY,
                'status': log.AUTH_REJECT,
            }]
        )

    @defer.inlineCallbacks
    def test_no_passcode_response_failure_logged(self):
        yield self._test_auth(
            expect_code=pyrad.packet.AccessReject,
            username=self.username,
            challenge_response='',
            challenge_state={'is_chap2': False},
        )

        self.assert_logged(
            [{
                'msg': 'Missing or improperly-formatted password',
                'auth_stage': log.AUTH_SECONDARY,
                'status': log.AUTH_ERROR,
            }]
        )


if __name__ == '__main__':
    import unittest
    unittest.main()
