#
# Copyright (c) 2012 Duo Security
# All Rights Reserved
#

from pyrad import packet

from twisted.internet import defer, reactor
from twisted.internet.protocol import Protocol, Factory
from twisted.internet.task import Clock

import duoauthproxy.modules.radius_server_iframe
from duoauthproxy.lib import protect, log, duo_async
from duoauthproxy.lib.config_provider import ConfigDict
from duoauthproxy.lib.radius.challenge import BadStateError

from base import (
    ServerModuleTest,
    ServerModuleTestPassThrough,
    ServerModuleTestPassThroughMultipleValues,
    ServerProtectedConfigParsingTestCase,
    AuthLoggingTest,
    TestCase,
)


class RadiusServerIFrameTestRunner(ServerModuleTest, TestCase):
    """Tester class inheritable by all other RadiusServerIFrame test classes.

    Any test class that wants to inherit from ServerModuleTest, TestCase, or both
    of them can inherit from this class instead.

    This class was created due to multiple inheritance super() incompatibilities
    with TestRadiusServerIFrameAuthLogging, though it's since been co-opted
    for use with other RadiusServerIFrame test classes as well to cut down on
    extra setUp() redefinitions.
    """
    def setUp(self):
        super(RadiusServerIFrameTestRunner, self).setUp(
            {
                'type': 'array',    
            },
            duoauthproxy.modules.radius_server_iframe,
            'radius_server_iframe',
        )


class RadiusServerIFrameTestCases(object):
    """Test suite covering various DuoIFrameRadiusServer auth flows.

    Since DuoIFrameRadiusServer works via RADIUS Challenge-Response, tests have
    been divided into those dealing with the user's initial request pre-challenge
    as well as those dealing with the user's response to the iframe challenge.

    Note:
        Since there's no setUp() or tearDown() defined in this class, it needs
        to make use of the setUp() defined in RadiusServerIFrameTestRunner.
    """

    # coverage for response to initial request:

    @defer.inlineCallbacks
    def test_missing_username(self):
        yield self._test_auth(packet.AccessReject,
                              username=None,
                              password=self.password)

    @defer.inlineCallbacks
    def test_wrong_primary_auth_password(self):
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              password='foo')

    @defer.inlineCallbacks
    def test_proxy_init_fail_open_failmode_secure(self):
        self.enqueue_duo_proxy_init_error_fail_open()
        yield self._test_auth(packet.AccessReject,
                              failmode=duo_async.FAILMODE_SECURE,
                              username=self.username,
                              password=self.password)

    @defer.inlineCallbacks
    def test_proxy_init_fail_closed_failmode_secure(self):
        self.enqueue_duo_proxy_init_error_fail_closed()
        yield self._test_auth(packet.AccessReject,
                              failmode=duo_async.FAILMODE_SECURE,
                              username=self.username,
                              password=self.password)

    @defer.inlineCallbacks
    def test_proxy_init_fail_open_failmode_safe(self):
        self.enqueue_duo_proxy_init_error_fail_open()
        yield self._test_auth(packet.AccessAccept,
                              failmode=duo_async.FAILMODE_SAFE,
                              username=self.username,
                              password=self.password)

    @defer.inlineCallbacks
    def test_proxy_init_fail_closed_failmode_safe(self):
        self.enqueue_duo_proxy_init_error_fail_closed()
        yield self._test_auth(packet.AccessReject,
                              failmode=duo_async.FAILMODE_SAFE,
                              username=self.username,
                              password=self.password)

    @defer.inlineCallbacks
    def test_proxy_init_challenge(self):
        # primary auth and duo proxy_init pass -> challenge.
        self.enqueue_duo_proxy_init()
        yield self._test_auth(packet.AccessChallenge,
                              username=self.username,
                              password=self.password)

    @defer.inlineCallbacks
    def test_exempt_username(self):
        self.instance.protocol.exempt_usernames = [self.username]
        yield self._test_auth(packet.AccessAccept,
                              username=self.username,
                              password=self.password,
                              expect_msg_contains='User exempted from 2FA')

    # coverage for response to challenge:

    @defer.inlineCallbacks
    def test_proxy_finish_valid_cookie_false(self):
        self.enqueue_duo_proxy_finish(valid_cookie=False,
                                      username=self.username)
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              challenge_response='push')

    @defer.inlineCallbacks
    def test_proxy_finish_mismatched_username(self):
        self.enqueue_duo_proxy_finish(valid_cookie=True,
                                      username=('not ' + self.username))
        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              challenge_response='push')

    @defer.inlineCallbacks
    def test_proxy_finish_success(self):
        self.enqueue_duo_proxy_finish(valid_cookie=True,
                                      username=self.username)
        yield self._test_auth(packet.AccessAccept,
                              username=self.username,
                              challenge_response='push')

    @defer.inlineCallbacks
    def test_proxy_finish_fail_open_failmode_safe(self):
        self.enqueue_duo_proxy_finish_error_fail_open()
        yield self._test_auth(packet.AccessAccept,
                              failmode=duo_async.FAILMODE_SAFE,
                              username=self.username,
                              challenge_response='push')

    @defer.inlineCallbacks
    def test_proxy_finish_fail_closed_failmode_safe(self):
        self.enqueue_duo_proxy_finish_error_fail_closed()
        yield self._test_auth(packet.AccessReject,
                              failmode=duo_async.FAILMODE_SAFE,
                              username=self.username,
                              challenge_response='push')

    @defer.inlineCallbacks
    def test_proxy_finish_fail_open_failmode_secure(self):
        self.enqueue_duo_proxy_finish_error_fail_open()
        yield self._test_auth(packet.AccessReject,
                              failmode=duo_async.FAILMODE_SECURE,
                              username=self.username,
                              challenge_response='push')

    @defer.inlineCallbacks
    def test_proxy_finish_fail_closed_failmode_secure(self):
        self.enqueue_duo_proxy_finish_error_fail_closed()
        yield self._test_auth(packet.AccessReject,
                              failmode=duo_async.FAILMODE_SECURE,
                              username=self.username,
                              challenge_response='push')


class TestRadiusServerIFrame(RadiusServerIFrameTestCases,
                             RadiusServerIFrameTestRunner):
    """Runs all tests in RadiusServerIFrameTestCases.

    Note:
        test_exempt_username_config_parsing is kept here and not in
        RadiusServerIFrameTestCases since we don't need to run it more than once.
    """

    def test_exempt_username_config_parsing(self):
        server_config = {
            'radius_ip_1': self.radius_ip,
            'radius_secret_1': self.secret,
            'ikey': 'nope',
            'skey': 'nuh-uh',
            'type': 'array',
            'exempt_username_1': 'testuser1',
            'exempt_username_2': 'testuser2',
        }
        self.instance = duoauthproxy.modules.radius_server_iframe.Module(
            ConfigDict(server_config),
            None,
            'test_radius_iframe',
        )
        server = self.instance.protocol
        self.assertEqual(server.exempt_usernames, ['testuser1', 'testuser2'])


# Note that the time out logic is the same for all the radius server sections
# so we don't need a test like this in all of the test_module_radius_*.py files
class TestRadiusServerIFrameTimeout(TestCase):

    @defer.inlineCallbacks
    def test_api_timeout(self):
        api_timeout = 20
        api_port = 1234
        server_config = {
            'radius_ip_1': '127.0.0.1',
            'radius_secret_1': 'secret',
            'ikey': 'nope',
            'skey': 'nuh-uh',
            'type': 'array',
            'exempt_username_1': 'testuser1',
            'exempt_username_2': 'testuser2',
            'api_timeout': '%s' % api_timeout,
            'api_host': '127.0.0.1',
            'api_port': '%s' % api_port,
        }

        instance = duoauthproxy.modules.radius_server_iframe.Module(
            ConfigDict(server_config),
            None,
            'test_radius_iframe',
        )

        class HangingProtocol(Protocol):
            def dataReceived(self, data):
                return

        def fake_success(_a, _b):
            class result(object):
                success = True
                radius_attrs = 'nope'

            return result()

        class Response(object):
            username = 'user'
            password = 'pass'
            source = 'test'
            id = '0'
            client_ip = '127.0.0.1'
            packet = 'fake-packet'

        def fake_error_handler(_a, _b, error, _c):
            return error

        # jump over the part where the TCP connection times out
        clock = Clock()
        self.wrap_do_connect(clock, 20)
        self.patch(instance.protocol, 'primary_auth', fake_success)
        self.patch(instance.protocol, 'response_for_api_error', fake_error_handler)

        f = Factory()
        f.protocol = HangingProtocol

        try:
            fake_server = reactor.listenTCP(api_port, f)
            fake_response = Response()
            result = yield instance.protocol.get_initial_response(fake_response)
        finally:
            fake_server.stopListening()

        self.assertTrue('TimeoutError' in result[0] and 'took longer than' in result[0])


class TestRadiusServerIFramePassThrough(RadiusServerIFrameTestCases,
                                        ServerModuleTestPassThrough,
                                        RadiusServerIFrameTestRunner):
    """Runs all tests in RadiusServerChallengeTestCases with an additional
    pass-through attribute to be sent and received by the server."""
    pass


class TestRadiusServerIFramePassThroughMultipleValues(
        RadiusServerIFrameTestCases,
        ServerModuleTestPassThroughMultipleValues,
        RadiusServerIFrameTestRunner):
    """Runs all tests in RadiusServerChallengeTestCases with multiple additional
    pass-through attributes to be send and received by the server."""
    pass


class TestRadiusIframeProtectedConfigParsing(ServerProtectedConfigParsingTestCase):
    """Tests that an instance of DuoIFrameRadiusServer with a protected
    config properly accesses and/or decrypts certain protected keys defined in
    ProtectedValidatingConfigDict."""

    def test_plain_keys_use_protected_api(self):
        # Both of these new keys should be accessed with get_protected_str()
        # though neither are encrypted
        instance = self._make_instance(
            {
                'type': 'array',
                'radius_secret_1': 'secret!',
                'skey': 'skey!'
            },
            duoauthproxy.modules.radius_server_iframe,
            'radius_server_iframe',
        )
        self.assertEqual(instance.protocol.client.duo_creds.get_secret(), 'skey!')
        self.assertEqual(instance.protocol.secrets.values()[0], 'secret!')

    def test_protected_keys(self):
        # radius_secret_protected_1 should decrypt first with get_protected_str()
        if protect.PROTECT_ENABLED:
            protected = protect.protect('secret!')
            instance = self._make_instance(
                {
                    'type': 'array',
                    'radius_secret_protected_1': protected,
                    'skey': 'skey!'
                },
                duoauthproxy.modules.radius_server_iframe,
                'radius_server_iframe',
            )
            self.assertEqual(instance.protocol.client.duo_creds.get_secret(), 'skey!')
            self.assertEqual(instance.protocol.secrets.values()[0], 'secret!')


class TestRadiusServerIFrameAuthLogging(AuthLoggingTest, RadiusServerIFrameTestRunner):
    """Tests SIEM logging from log.auth_standard() calls during various
    authentication flows in DuoIFrameRadiusServer."""

    @defer.inlineCallbacks
    def test_missing_username(self):
        expected = [{
            'auth_stage': 'Unknown',
            'status': log.AUTH_ERROR,
        }]

        yield self._test_auth(packet.AccessReject,
                              username=None,
                              password=self.password)

        self.assert_logged(expected)

    @defer.inlineCallbacks
    def test_missing_password(self):
        expected = [{
            'auth_stage': log.AUTH_PRIMARY,
            'status': log.AUTH_ERROR,
        }]

        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              password=None)

        self.assert_logged(expected)

    @defer.inlineCallbacks
    def test_exempt_username(self):
        expected = [{
            'auth_stage': log.AUTH_PRIMARY,
            'status': log.AUTH_ALLOW,
        }, {
            'auth_stage': log.AUTH_SECONDARY,
            'status': log.AUTH_ALLOW,
            'msg': 'User exempted from 2FA',
        }]

        self.instance.protocol.exempt_usernames = [self.username]
        yield self._test_auth(packet.AccessAccept,
                              username=self.username,
                              password=self.password)

        self.assert_logged(expected)

    @defer.inlineCallbacks
    def test_wrong_primary_auth_password(self):
        expected = [{
            'auth_stage': log.AUTH_PRIMARY,
            'status': log.AUTH_REJECT,
            'msg': 'Primary credentials rejected - Hello',
        }]

        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              password='foo')

        self.assert_logged(expected)

    @defer.inlineCallbacks
    def test_proxy_finish_success(self):
        expected = [{
            'auth_stage': log.AUTH_SECONDARY,
            'status': log.AUTH_ALLOW,
            'msg': 'Valid login from iframe',
        }]
        self.enqueue_duo_proxy_finish(valid_cookie=True,
                                      username=self.username)

        yield self._test_auth(packet.AccessAccept,
                              username=self.username,
                              challenge_response='push')

        self.assert_logged(expected)

    @defer.inlineCallbacks
    def test_proxy_finish_valid_cookie_false(self):
        expected = [{
            'auth_stage': log.AUTH_SECONDARY,
            'status': log.AUTH_REJECT,
            'msg': 'Invalid login from iframe',
        }]
        self.enqueue_duo_proxy_finish(valid_cookie=False,
                                      username=self.username)

        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              challenge_response='push')

        self.assert_logged(expected)

    @defer.inlineCallbacks
    def test_proxy_init_fail_open(self):
        expected = [{
            'auth_stage': log.AUTH_PRIMARY,
            'status': log.AUTH_ALLOW,
        }, {
            'auth_stage': log.AUTH_SECONDARY,
            'status': log.AUTH_ALLOW,
            'msg': duo_async.FAILMODE_SAFE_MSG,
        }]
        self.enqueue_duo_proxy_init_error_fail_open()

        yield self._test_auth(packet.AccessAccept,
                              failmode=duo_async.FAILMODE_SAFE,
                              username=self.username,
                              password=self.password)

        self.assert_logged(expected)

    @defer.inlineCallbacks
    def test_proxy_init_fail_closed(self):
        expected = [{
            'auth_stage': log.AUTH_PRIMARY,
            'status': log.AUTH_ALLOW,
        }, {
            'auth_stage': log.AUTH_SECONDARY,
            'status': log.AUTH_REJECT,
            'msg': duo_async.FAILMODE_SECURE_MSG,
        }]
        self.enqueue_duo_proxy_init_error_fail_closed()

        yield self._test_auth(packet.AccessReject,
                              failmode=duo_async.FAILMODE_SECURE,
                              username=self.username,
                              password=self.password)

        self.assert_logged(expected)

    @defer.inlineCallbacks
    def test_bad_state_error(self):
        error = "Bad State Error"
        expected = [{
            'auth_stage': log.AUTH_PRIMARY,
            'status': log.AUTH_ALLOW,
        }, {
            'auth_stage': 'Unknown',
            'status': log.AUTH_REJECT,
            'msg': error,
        }]
        e = BadStateError(error)
        self.enqueue_duo_proxy_init().AndRaise(e)

        yield self._test_auth(packet.AccessReject,
                              username=self.username,
                              password=self.password)

        self.assert_logged(expected)


if __name__ == '__main__':
    import unittest
    unittest.main()
