#
# Copyright (c) 2013 Duo Security
# All Rights Reserved
#
import mock
from pyrad import packet

from base import TestCase
import duoauthproxy.lib.mppe as mppe
from duoauthproxy.lib.radius.base import radius_dictionary

OPENSSL_VERSION_110 = int('0x10100000', 16)
OPENSSL_VERSION_100 = int('0x1000000f', 16)


class MockSSLConnection(object):
    def export_keying_material(self, label, olen):
        recv_key = 'a' * (olen/2)
        send_key = 'b' * (olen/2)
        return recv_key + send_key


class TestMPPE(TestCase):
    def setUp(self):
        salt1 = 35331
        salt2 = 64405
        iter = [salt1, salt2].__iter__()
        self.oldgen = mppe._generator.randint
        mppe._generator.randint = lambda x, y: iter.next()

    def tearDown(self):
        mppe._generator.randint = self.oldgen

    def test_mppe(self):
        """Test MPPE key generation"""

        test_packet = packet.Packet(
            authenticator='foo',
            secret='bar',
            dict=radius_dictionary())

        keys = ['a00001793f603eb8658516d4e8e493a93d93378b5fa368295adcdaae8a231bf6'.decode('hex'),
                '8919904f39cb2933e8ecd109c97d94674199c998f842e506d566fbb3b14d0e98'.decode('hex')]
        secret = '74686973697361726164697573736563726574'.decode('hex')
        auth = 'bfaaf628407c91472ae414d6671400bc'.decode('hex')

        mppe.add_mppe(test_packet, keys, secret, auth)

        new_packet = packet.Packet(secret='bar',
                                   dict=radius_dictionary())
        new_packet.DecodePacket(test_packet.ReplyPacket())

        expected1 = '8a03922d77123ac130a29ee3d284861236adbb611d14ec33294ae9e9290332397f911da30872498f1ccc954bdf1406064f53'.decode('hex')
        expected2 = 'fb959f88b3399c6c25678f5285ec7d978818371f9b0c2c8d8bf38022e616e36bef9c6c75138a19981d803671f7e17c04e0fd'.decode('hex')

        self.assertEqual(new_packet['MS-MPPE-Recv-Key'][0], expected1)
        self.assertEqual(new_packet['MS-MPPE-Send-Key'][0], expected2)

    @mock.patch('OpenSSL.SSL.OPENSSL_VERSION_NUMBER', OPENSSL_VERSION_110)
    def test_get_mppe_keys_openssl_above_101(self):
        olen = 64
        expected = ['a' * (olen / 2), 'b' * (olen / 2)]
        mock_ssl_connection = MockSSLConnection()

        result = mppe.get_mppe_keys('master_key', 'client_random', 'server_random', 'label', mock_ssl_connection)
        self.assertItemsEqual(expected, result)

    @mock.patch('duoauthproxy.lib.mppe.PRF')
    @mock.patch('OpenSSL.SSL.OPENSSL_VERSION_NUMBER', OPENSSL_VERSION_100)
    def test_get_mppe_keys_openssl_below_101(self, mock_prf):
        olen = 64
        recv_key = 'a' * (olen/2)
        send_key = 'b' * (olen/2)
        expected = [recv_key, send_key]
        mock_prf.return_value = recv_key + send_key

        result = mppe.get_mppe_keys('master_key', 'client_random', 'server_random', 'label', 'ssl_con')
        self.assertItemsEqual(expected, result)


if __name__ == '__main__':
    import unittest
    unittest.main()
