from base import TestCase
from twisted.internet.task import Clock
from twisted.internet import defer
import mock
from duoauthproxy.lib import duo_async
from duoauthproxy.lib.duo_creds import DuoCreds
import duoauthproxy.lib.primary_only_manager


class TempReactor(object):
    def __init__(self, target):
        self.clock = Clock()
        self.initial_reactor = getattr(target, 'reactor')
        self.target = target

    def __enter__(self):
        self.target.reactor = self.clock
        return self.clock

    def __getattr__(self, name):
        return getattr(self.clock, name)

    def __exit__(self, t, value, traceback):
        self.target.reactor = self.initial_reactor


class MockPrimaryOnlyManager(object):
    def __init__(self, time=0):
        self.wait_time = time
        self.reactor = TempReactor(duoauthproxy.lib.primary_only_manager)

    def __enter__(self):
        self.reactor.__enter__()
        self.reactor.clock.stop = mock.MagicMock(return_value=None)
        duoauthproxy.lib.primary_only_manager.PrimaryOnlyManager.enable_primary_only(self.wait_time)
        return self

    def __exit__(self, t, value, traceback):
        duoauthproxy.lib.primary_only_manager.PrimaryOnlyManager._disable_primary_only()
        self.reactor.__exit__(t, value, traceback)


class TestPrimaryOnly(TestCase):
    def setUp(self):
        self.client = duo_async.AsyncDuoClient('non-host', DuoCreds('noikey', 'noskey'))

    def test_enable_primary_only(self):
        for wait_time in [0, 10, 100, 240]:
            with MockPrimaryOnlyManager(wait_time) as mock_manager:
                mock_manager.reactor.advance(wait_time * 60)
                self.assertTrue(mock_manager.reactor.stop.called)

    def test_enable_primary_only_too_long(self):
        wait_time = 600
        with MockPrimaryOnlyManager(wait_time) as mock_manager:
            mock_manager.reactor.advance(240 * 60)
            self.assertTrue(mock_manager.reactor.stop.called)

    @defer.inlineCallbacks
    def test_auth_primary_only(self):
        with MockPrimaryOnlyManager():
            with self.assertRaisesRegexp(duo_async.DuoAPIFailOpenError, 'operating in primary-only'):
                yield self.client.auth('some_user_1', '123456', '1.2.3.4')

    @defer.inlineCallbacks
    def test_preauth_primary_only(self):
        with MockPrimaryOnlyManager():
            with self.assertRaisesRegexp(duo_async.DuoAPIFailOpenError, 'operating in primary-only'):
                yield self.client.preauth('some_user_1', '1.2.3.4')


if __name__ == '__main__':
    import unittest
    unittest.main()
