#
# Copyright (c) 2011 Duo Security
# All Rights Reserved
#
import socket
import unittest
import hashlib

from twisted.internet import defer, reactor

from pyrad import packet

from duoauthproxy.lib.base import AuthError
from duoauthproxy.lib.radius.base import radius_dictionary
from duoauthproxy.lib.radius.client import RadiusClient
from duoauthproxy.lib.md5_wrapper import Depends

from mock_objects import DeferredMockTransport, BasicMockTransport
from base import TestCase

md5_constructor = hashlib.md5


class TestRadiusClient(TestCase):
    def setUp(self):
        self.srv_ip = '1.2.3.4'
        self.srv_port = 1812
        self.secret = 'jawokenmahtubercon'
        self.nas_ip = '5.6.7.8'
        self.addrs = [
            (self.srv_ip, self.srv_port),
            ('1.2.3.5', 11812),
        ]
        self.client = RadiusClient(self.addrs,
                                   self.nas_ip,
                                   self.secret, 3, 0.5, False, 'utf-8')
        self.client.transport = DeferredMockTransport()

    def tearDown(self):
        self.client.cleanup_all()

    @defer.inlineCallbacks
    def test_basic_response(self):
        pass_through_attrs = {
            'NAS-Identifier': ['localhost', '127.0.0.1'],
            'Login-LAT-Node': ['cheese'],
            'Message-Authenticator': ['\1' * 16]
        }
        d = self.client.authenticate('someuser', 'somePassword!', '12.34.56.78', pass_through_attrs)

        queue = self.client.transport.queue
        (req_data, dst) = yield queue.get()
        self.assertEqual(dst, (self.srv_ip, self.srv_port))

        # parse request packet
        request = packet.AuthPacket(packet=req_data,
                                    dict=radius_dictionary())
        request.secret = self.secret
        self.assertEqual(request.code, packet.AccessRequest)
        self.assertEqual(request[1], ['someuser'])
        self.assertEqual(request.PwDecrypt(request[2][0]), 'somePassword!')
        self.assertEqual(request[4], [socket.inet_aton(self.nas_ip)])
        self.assertEqual(request[31], ['12.34.56.78'])
        for attr_name, value in pass_through_attrs.iteritems():
            if attr_name != "Message-Authenticator":
                self.assertEqual(request[attr_name], value)

        # create a positive response with no reply message
        # (note: we expect no reply-message to be the common case. in fact
        # probably nobody uses Reply-Message in practice, except in
        # AccessChallenge packets...)
        response_packet = request.CreateReply()
        response_packet.code = packet.AccessAccept
        self.client.datagramReceived(response_packet.ReplyPacket(),
                                     (self.srv_ip, self.srv_port))
        actual_response_packet = yield d
        self.assertNotIn('Reply-Message',
                         actual_response_packet)
        self.assertEqual(actual_response_packet.code,
                         packet.AccessAccept)

    @defer.inlineCallbacks
    def test_pw_codec(self):
        self.client.pw_codec = 'latin-1'
        d = self.client.authenticate('someuser', u'\xa3password', '12.34.56.78')

        queue = self.client.transport.queue
        (req_data, dst) = yield queue.get()
        self.assertEqual(dst, (self.srv_ip, self.srv_port))

        # parse request packet
        request = packet.AuthPacket(packet=req_data,
                                    dict=radius_dictionary())
        request.secret = self.secret
        self.assertEqual(request.code, packet.AccessRequest)
        self.assertEqual(request[1], ['someuser'])
        self.assertEqual(request.PwDecrypt(request[2][0], pw_codec='latin-1'), u'\xa3password')
        with self.assertRaises(UnicodeDecodeError):
            # With a latin-1 encoding, a utf-8 decoding should fail
            request.PwDecrypt(request[2][0], pw_codec='utf-8')
        self.assertEqual(request[4], [socket.inet_aton(self.nas_ip)])
        self.assertEqual(request[31], ['12.34.56.78'])

        # create a positive response to clean the reactor
        response_packet = request.CreateReply()
        response_packet.code = packet.AccessAccept
        self.client.datagramReceived(response_packet.ReplyPacket(),
                                     (self.srv_ip, self.srv_port))
        yield d

    @defer.inlineCallbacks
    def test_message_response(self):
        d = self.client.authenticate('someuser', 'somePassword!', '12.34.56.78')

        queue = self.client.transport.queue
        (req_data, dst) = yield queue.get()
        self.assertEqual(dst, (self.srv_ip, self.srv_port))

        # parse request packet
        request = packet.AuthPacket(packet=req_data,
                                    dict=radius_dictionary())
        request.secret = self.secret
        self.assertEqual(request.code, packet.AccessRequest)
        self.assertEqual(request[1], ['someuser'])
        self.assertEqual(request.PwDecrypt(request[2][0]), 'somePassword!')
        self.assertEqual(request[4], [socket.inet_aton(self.nas_ip)])
        self.assertEqual(request[31], ['12.34.56.78'])

        # create a negative response with a reply message
        response_packet = request.CreateReply()
        response_packet.code = packet.AccessReject
        msg = 'Hello, World!'
        response_packet['Reply-Message'] = msg
        self.client.datagramReceived(response_packet.ReplyPacket(),
                                     (self.srv_ip, self.srv_port))
        actual_response_packet = yield d
        self.assertEqual([msg],
                         actual_response_packet['Reply-Message'])
        self.assertEqual(actual_response_packet.code,
                         packet.AccessReject)

    @defer.inlineCallbacks
    def test_response_from_wrong_addr(self):
        d = self.client.authenticate('someuser',
                                     'somePassword!',
                                     '12.34.56.78')

        queue = self.client.transport.queue
        (req_data, dst) = yield queue.get()
        self.assertEqual(dst, (self.srv_ip, self.srv_port))

        # parse request packet
        request = packet.AuthPacket(packet=req_data,
                                    dict=radius_dictionary())
        request.secret = self.secret

        # create a negative response with a reply message
        response_packet = request.CreateReply()
        response_packet.code = packet.AccessAccept
        response_packet['Reply-Message'] = 'Hello, World!'

        # wrong IP
        with self.assertRaisesRegexp(packet.PacketError,
                                     'response packet from unknown address:'):
            self.client._handle_response(
                response_packet.ReplyPacket(),
                ('9.9.9.9', self.srv_port),
            )

        # wrong port
        with self.assertRaisesRegexp(packet.PacketError,
                                     'response packet from unknown address:'):
            self.client._handle_response(
                response_packet.ReplyPacket(),
                (self.srv_ip, 9999),
            )

        with self.assertRaises(AuthError):
            yield d

    @defer.inlineCallbacks
    def test_response_from_wrong_id(self):
        d = self.client.authenticate('someuser',
                                     'somePassword!',
                                     '12.34.56.78')

        queue = self.client.transport.queue
        (req_data, dst) = yield queue.get()
        self.assertEqual(dst, (self.srv_ip, self.srv_port))

        # parse request packet
        request = packet.AuthPacket(packet=req_data,
                                    dict=radius_dictionary())
        request.secret = self.secret

        # create a negative response with a reply message
        response_packet = request.CreateReply()
        response_packet.code = packet.AccessAccept
        response_packet['Reply-Message'] = 'Hello, World!'
        if response_packet.id == 255:
            # If ID was max then we need to loop back to 1 so we don't go over the max value
            response_packet.id = 1
        else:
            response_packet.id = response_packet.id + 1

        with self.assertRaisesRegexp(packet.PacketError,
                                     'unrecognized id in response packet:'):
            self.client._handle_response(
                response_packet.ReplyPacket(),
                (self.srv_ip, self.srv_port),
            )

        with self.assertRaises(AuthError):
            yield d

    @defer.inlineCallbacks
    def test_good_message_authenticator(self):
        d = self.client.authenticate('someuser', 'somePassword!', '12.34.56.78')

        queue = self.client.transport.queue
        (req_data, dst) = yield queue.get()
        self.assertEqual(dst, (self.srv_ip, self.srv_port))

        # parse request packet
        request = packet.AuthPacket(packet=req_data,
                                    dict=radius_dictionary())
        request.secret = self.secret

        response_packet = request.CreateReply()
        response_packet.code = packet.AccessAccept
        response_packet['Message-Authenticator'] = b'\x01' * 16

        self.client.datagramReceived(response_packet.ReplyPacket(),
                                     (self.srv_ip, self.srv_port))
        actual_response_packet = yield d
        self.assertEqual(actual_response_packet.code,
                         packet.AccessAccept)

    @defer.inlineCallbacks
    def test_bad_message_authenticator(self):
        d = self.client.authenticate('someuser', 'somePassword!', '12.34.56.78')

        queue = self.client.transport.queue
        (req_data, dst) = yield queue.get()
        self.assertEqual(dst, (self.srv_ip, self.srv_port))

        # parse request packet
        request = packet.AuthPacket(packet=req_data,
                                    dict=radius_dictionary())
        request.secret = self.secret

        response_packet = request.CreateReply()
        response_packet.code = packet.AccessAccept
        response_packet['Message-Authenticator'] = b'\x01' * 16
        res_data = response_packet.ReplyPacket()
        # corrupt the Message-Authenticator with one bad char
        res_data_attrs = res_data[20:37] + 'b'
        res_authenticator = md5_constructor(
            res_data[0:4] +
            response_packet.authenticator +
            res_data_attrs +
            response_packet.secret,
            used_for_security=Depends,
        ).digest()
        res_data = res_data[0:4] + res_authenticator + res_data_attrs

        with self.assertRaisesRegexp(packet.PacketError,
                                     'Invalid Message-Authenticator from'):
            self.client._handle_response(
                res_data,
                (self.srv_ip, self.srv_port),
            )

        with self.assertRaises(AuthError):
            yield d

    @defer.inlineCallbacks
    def test_bad_message_authenticator_excepts_then_good_message_authenticator_passes(self):
        d = self.client.authenticate('someuser', 'somePassword!', '12.34.56.78')

        queue = self.client.transport.queue
        (req_data, dst) = yield queue.get()
        self.assertEqual(dst, (self.srv_ip, self.srv_port))

        # parse request packet
        request = packet.AuthPacket(packet=req_data,
                                    dict=radius_dictionary())
        request.secret = self.secret

        response_packet = request.CreateReply()
        response_packet.code = packet.AccessAccept
        response_packet['Message-Authenticator'] = b'\x01' * 16
        res_data = response_packet.ReplyPacket()

        # corrupt the Message-Authenticator with one bad char
        if res_data[37] == 'b':
            bad_char = 'c'
        else:
            bad_char = 'b'
        res_data_attrs = res_data[20:37] + bad_char

        res_authenticator = md5_constructor(
            res_data[0:4] +
            response_packet.authenticator +
            res_data_attrs +
            response_packet.secret,
            used_for_security=Depends,
        ).digest()
        res_data = res_data[0:4] + res_authenticator + res_data_attrs

        with self.assertRaisesRegexp(packet.PacketError,
                                     'Invalid Message-Authenticator from'):
            self.client._handle_response(
                res_data,
                (self.srv_ip, self.srv_port),
            )

        # now send the good packet and confirm it
        self.client.datagramReceived(response_packet.ReplyPacket(),
                                     (self.srv_ip, self.srv_port))
        actual_response_packet = yield d
        self.assertEqual(actual_response_packet.code,
                         packet.AccessAccept)

    @defer.inlineCallbacks
    def test_retry(self):
        self.client.transport = BasicMockTransport()

        queue = self.client.transport.queue
        d = self.client.authenticate('someuser', 'somePassword!', '12.34.56.78')

        d2 = defer.Deferred()
        reactor.callLater(0.75, d2.callback, None)
        yield d2

        self.assertEqual(len(queue), 2)
        # nth retry uses nth addr
        self.assertEqual(queue[0][1], self.addrs[0])
        self.assertEqual(queue[1][1], self.addrs[1])

        d3 = defer.Deferred()
        reactor.callLater(1, d3.callback, None)
        yield d3
        # cycles back around to 0th addr
        self.assertEqual(queue[2][1], self.addrs[0])

        self.assertEqual(len(queue), 4)

        try:
            yield d
            self.fail('Did not raise a timeout error')
        except AuthError:
            pass


if __name__ == '__main__':
    unittest.main()
