#
# Copyright (c) 2011 Duo Security
# All Rights Reserved
#
import functools
import itertools

from pyrad import packet
from twisted.internet import defer
import netaddr

from duoauthproxy.lib import forward_serv, log
from duoauthproxy.lib.base import ServerModule
from duoauthproxy.lib.config_provider import ConfigDict
from duoauthproxy.lib.config_error import ConfigError
from duoauthproxy.lib.radius.base import radius_dictionary
from duoauthproxy.lib.radius.server import SimpleRadiusServer
from duoauthproxy.lib.radius.server import parse_radius_secrets
from duoauthproxy.lib.radius.server import parse_exempt_usernames
from duoauthproxy.lib.radius.server import parse_client_ip_attribute
from duoauthproxy.lib.radius.challenge import ChallengeResponseRadiusServer

import mock_objects
from base import TestCase, AuthLoggingTest


class SimpleMockServer(SimpleRadiusServer):
    def __init__(self, secrets, pw_codec='utf-8'):
        SimpleRadiusServer.__init__(self, secrets, None, '', pw_codec=pw_codec)
        self.action = self.create_accept_packet

    def get_response(self, request):
        return defer.succeed(self.action(request))


class MockServerModule(ServerModule):
    def __init__(self, protocol):
        self.protocol = protocol


class MockConfigDictProtectedNone(ConfigDict):
    def get_protected_str(self, key, protected_key, default=None):
        return None


class TestForwardRadiusServer(TestCase):
    def setUp(self):
        self.port = 9563
        self.secret = 'throughthreecheesetreesthreefreefleasflew'

        networks = [
            '0.0.0.0/0',
            '10.0.0.0/8',
            '192.0.0.0/8',
            '192.168.0.0/16',
            '192.168.1.0/24',
            '192.168.2.3/32']
        self.srv_insts = []
        self.srv_mods = []
        self.srv_networks = {}
        self.srv_names = {}

        i = itertools.count()
        for net in networks:
            inst = SimpleMockServer({
                netaddr.IPNetwork(net).cidr: self.secret
            })
            # make the instance send a reply-message containing the
            # appropriate netblock
            inst.action = functools.partial(
                inst.create_accept_packet, msg=net)
            self.srv_insts.append(inst)

            mod = MockServerModule(inst)
            self.srv_mods.append(mod)

            self.srv_names[mod] = 'mock_server_%i' % i.next()
            self.srv_networks[mod] = inst.secrets.keys()

        server_port = 9563
        self.srv = forward_serv.Module(
            server_port, self.srv_networks, self.srv_names, '', False)

        self.srv.protocol.transport = mock_objects.BasicMockTransport()

    def tearDown(self):
        self.srv.stopService()
        for s in self.srv_insts:
            s.cleanup_all()

    @defer.inlineCallbacks
    def test_forwarding(self):
        """Make sure that requests from various source IPs get route
        to the correct server modules."""

        test_addrs = {
            '1.2.3.4': '0.0.0.0/0',
            '10.1.2.3': '10.0.0.0/8',
            '192.1.2.3': '192.0.0.0/8',
            '192.168.244.1': '192.168.0.0/16',
            '192.168.1.1': '192.168.1.0/24',
            '192.168.2.3': '192.168.2.3/32'
        }

        for addr, net in test_addrs.iteritems():
            # create access request
            request = packet.AuthPacket(code=packet.AccessRequest,
                                        id=1,
                                        secret=self.secret,
                                        dict=radius_dictionary())
            request[1] = ['some_username']
            request[2] = [request.PwCrypt('THISisAPassword!')]
            req_data = request.RequestPacket()

            # send into mock server
            yield self.srv.protocol.datagramReceived(req_data, (addr, self.port))

            # read response out
            self.assertEqual(len(self.srv.protocol.transport.queue), 1)
            rsp_data, rsp_addr = self.srv.protocol.transport.queue.pop(0)
            self.assertEqual(rsp_addr, (addr, self.port))

            # verify response is correct
            response = packet.AuthPacket(packet=rsp_data,
                                         dict=radius_dictionary())
            response.secret = self.secret
            self.assertEqual(response.code, packet.AccessAccept)
            self.assertTrue(request.VerifyReply(response, rsp_data))

            # verify the reply-message contains the expected netblock string!
            self.assertEqual(response[18], [net])


class SimpleRadiusServerTestRunner(TestCase):
    def setUp(self):
        self.ip = '98.76.54.32'
        self.port = 9563
        self.secret = 'throughthreecheesetreesthreefreefleasflew'
        self.srv_inst = SimpleMockServer({
            netaddr.IPNetwork(self.ip).cidr: self.secret,
        })
        self.srv_mod = MockServerModule(self.srv_inst)

        server_port = 9563
        server_networks = {self.srv_mod: [netaddr.IPNetwork(self.ip).cidr]}
        server_names = {self.srv_mod: 'test-server'}
        self.srv = forward_serv.Module(server_port, server_networks, server_names, '', False)
        self.srv.protocol.transport = mock_objects.BasicMockTransport()

    def tearDown(self):
        self.srv.stopService()
        self.srv_inst.cleanup_all()


class TestSimpleRadiusServer(SimpleRadiusServerTestRunner):
    @defer.inlineCallbacks
    def test_basic_request(self):
        # create access request
        request = packet.AuthPacket(code=packet.AccessRequest,
                                    id=1,
                                    secret=self.secret,
                                    dict=radius_dictionary())
        request[1] = ['some_username']
        request[2] = [request.PwCrypt('THISisAPassword!')]
        req_data = request.RequestPacket()

        # send into mock server
        def action(request):
            self.assertEqual(request.username, 'some_username')
            self.assertEqual(request.password, 'THISisAPassword!')
            return self.srv_inst.create_accept_packet(request)
        self.srv_inst.action = action
        yield self.srv.protocol.datagramReceived(req_data, (self.ip, self.port))

        # read response out
        self.assertEqual(len(self.srv.protocol.transport.queue), 1)
        rsp_data, rsp_addr = self.srv.protocol.transport.queue.pop(0)
        self.assertEqual(rsp_addr, (self.ip, self.port))

        # verify response is correct
        response = packet.AuthPacket(packet=rsp_data,
                                     dict=radius_dictionary())
        response.secret = self.secret
        self.assertEqual(response.code, packet.AccessAccept)
        self.assertTrue(request.VerifyReply(response, rsp_data))

        # test retransmission
        # basically, change what the response would be to a new incoming request
        # but we should still get exactly the same response back
        self.srv_inst.action = self.srv_inst.create_reject_packet
        yield self.srv.protocol.datagramReceived(req_data, (self.ip, self.port))

        self.assertEqual(len(self.srv.protocol.transport.queue), 1)
        rsp2_data, rsp2_addr = self.srv.protocol.transport.queue.pop(0)
        self.assertEqual(rsp_data, rsp2_data)
        self.assertEqual(rsp_addr, rsp2_addr)

    @defer.inlineCallbacks
    def test_pw_codec_latin1_succeed(self):

        # Default is utf-8
        self.assertEqual(self.srv_inst.pw_codec, 'utf-8')
        # Make the server decode latin-1 instead of utf-8
        self.srv_inst.pw_codec = 'latin-1'

        # create access request
        request = packet.AuthPacket(code=packet.AccessRequest,
                                    id=1,
                                    secret=self.secret,
                                    dict=radius_dictionary())
        # Encode the pw in latin-1
        request[1] = ['some_username']
        request[2] = [request.PwCrypt(u'\xa3password', pw_codec='latin-1')]
        req_data = request.RequestPacket()

        # send into mock server
        def action(request):
            self.assertEqual(request.username, 'some_username')
            self.assertEqual(request.password, u'\xa3password')
            return self.srv_inst.create_accept_packet(request)
        self.srv_inst.action = action
        yield self.srv.protocol.datagramReceived(req_data, (self.ip, self.port))

        # read response out
        self.assertEqual(len(self.srv.protocol.transport.queue), 1)
        rsp_data, rsp_addr = self.srv.protocol.transport.queue.pop(0)
        self.assertEqual(rsp_addr, (self.ip, self.port))

        # verify response is correct
        response = packet.AuthPacket(packet=rsp_data,
                                     dict=radius_dictionary())
        response.secret = self.secret
        self.assertEqual(response.code, packet.AccessAccept)
        self.assertTrue(request.VerifyReply(response, rsp_data))

    @defer.inlineCallbacks
    def test_pw_codec_latin1_fail(self):
        # Default is utf-8
        self.assertEqual(self.srv_inst.pw_codec, 'utf-8')

        # create access request
        request = packet.AuthPacket(code=packet.AccessRequest,
                                    id=1,
                                    secret=self.secret,
                                    dict=radius_dictionary())
        # Encode the pw in latin-1
        request[1] = ['some_username']
        request[2] = [request.PwCrypt(u'\xa3password', pw_codec='latin-1')]
        req_data = request.RequestPacket()

        yield self.srv.protocol.datagramReceived(req_data, (self.ip, self.port))

        # read response out
        self.assertEqual(len(self.srv.protocol.transport.queue), 1)
        rsp_data, rsp_addr = self.srv.protocol.transport.queue.pop(0)
        self.assertEqual(rsp_addr, (self.ip, self.port))

        # verify response is correct
        response = packet.AuthPacket(packet=rsp_data,
                                     dict=radius_dictionary())
        response.secret = self.secret
        self.assertEqual(response.code, packet.AccessReject)
        self.assertTrue(request.VerifyReply(response, rsp_data))

    @defer.inlineCallbacks
    def test_good_auth_request(self):
        # create access request
        request = packet.AuthPacket(code=packet.AccessRequest,
                                    id=1,
                                    secret=self.secret,
                                    dict=radius_dictionary())
        request[1] = ['some_username']
        request[2] = [request.PwCrypt('THISisAPassword!')]
        req_data = request.RequestPacket()

        # send into mock server
        self.srv_inst.action = self.srv_inst.create_accept_packet
        yield self.srv.protocol.datagramReceived(req_data, (self.ip, self.port))

        # verify response
        self.assertEqual(
            len(self.srv.protocol.transport.queue),
            1,
            "Expected 1 message on the queue"
        )

    @defer.inlineCallbacks
    def test_bad_message_authenticator_request(self):
        # create access request
        request = packet.AuthPacket(code=packet.AccessRequest,
                                    id=1,
                                    secret=self.secret,
                                    dict=radius_dictionary())
        request[1] = ['some_username']
        request['Message-Authenticator'] = b'\xFF' * 16
        req_data = request.RequestPacket()

        # Misconfigure the shared secret, to force a MessageAuth failure.
        request.secret = 'incorrect_secret'
        req_data_attrs = request.SetMessageAuthenticatorAttrs(req_data[0:4], req_data[4:20], req_data[20:])
        req_data = req_data[0:20] + req_data_attrs

        # send into mock server
        self.srv_inst.action = self.srv_inst.create_accept_packet
        yield self.srv.protocol.datagramReceived(req_data, (self.ip, self.port))

        # verify no response was returned (due to bad Message-Authenticator)
        self.assertEqual(
            len(self.srv.protocol.transport.queue),
            0,
            "Expected 0 messages on the queue"
        )

    @defer.inlineCallbacks
    def test_good_message_authenticator_request(self):
        # create access request
        request = packet.AuthPacket(code=packet.AccessRequest,
                                    id=1,
                                    secret=self.secret,
                                    dict=radius_dictionary())
        request[1] = ['some_username']
        request[2] = [request.PwCrypt('THISisAPassword!')]
        request['Message-Authenticator'] = b'\xFF' * 16
        req_data = request.RequestPacket()

        # send into mock server
        self.srv_inst.action = self.srv_inst.create_accept_packet
        yield self.srv.protocol.datagramReceived(req_data, (self.ip, self.port))

        # verify no response was returned (due to bad Message-Authenticator)
        self.assertEqual(
            len(self.srv.protocol.transport.queue),
            1,
            "Expected 1 messages on the queue"
        )

    @defer.inlineCallbacks
    def test_bad_request(self):
        # create access request
        request = packet.AuthPacket(code=packet.AccessRequest,
                                    id=1,
                                    secret=self.secret,
                                    dict=radius_dictionary())
        request[1] = ['some_username']
        request[2] = ['THISisAPassword!']
        req_data = request.RequestPacket()

        # send into mock server
        # It shoud be dropped by the server due to its src IP address
        self.srv_inst.action = self.srv_inst.create_accept_packet
        yield self.srv.protocol.datagramReceived(req_data, ('1.2.3.4', self.port))

        # verify no response was returned
        self.assertEqual(len(self.srv.protocol.transport.queue), 0)

        # create packet with wrong code
        request = packet.AuthPacket(code=packet.AccessAccept,
                                    id=1,
                                    secret=self.secret,
                                    dict=radius_dictionary())
        req_data = request.RequestPacket()

        # send into mock server
        self.srv_inst.action = self.srv_inst.create_accept_packet
        yield self.srv.protocol.datagramReceived(req_data, ('1.2.3.4', self.port))

        # verify no response was returned
        self.assertEqual(len(self.srv.protocol.transport.queue), 0)

    def test_parse_secrets_None(self):
        config = MockConfigDictProtectedNone(
            {'radius_ip_1': '1.1.1.1', 'radius_secret_protected_1': None}
        )

        with self.assertRaises(ConfigError):
            parse_radius_secrets(config)

    def test_parse_secrets(self):
        config = ConfigDict({'radius_ip_1': '1.1.1.1',
                             'radius_secret_1': 'secret1',
                             'radius_ip_2': '2.2.2.2',
                             'radius_secret_2': 'secret2',
                             'radius_ip_3': '3.3.3.3',
                             'radius_secret_3': 'secret3'})
        secrets = {
            netaddr.IPNetwork('1.1.1.1/32'): 'secret1',
            netaddr.IPNetwork('2.2.2.2/32'): 'secret2',
            netaddr.IPNetwork('3.3.3.3/32'): 'secret3',
        }
        self.assertEqual(parse_radius_secrets(config), secrets)

    def test_parse_secrets_bad_ip(self):
        good = '1.1.1.1'
        for bad in ('', '/', '2.2/', '.', 'z', '2.2/z',
                    '2.2.2.2.2', '2.2.2.2/2.2.2.2.2'):
            # test with bad IP for radius_ip_1 and radius_ip_N where N
            # > 1 because the code is different.
            config = ConfigDict({
                'radius_ip_1': bad,
                'radius_secret_1': 'secret1',
                'radius_ip_2': good,
                'radius_secret_2': 'secret2',
            })
            try:
                parse_radius_secrets(config)
            except ConfigError:
                pass
            else:
                self.fail('must ConfigError with bad radius_ip_1=%s'
                          % bad)
            config = ConfigDict({
                'radius_ip_1': good,
                'radius_secret_1': 'secret1',
                'radius_ip_2': bad,
                'radius_secret_2': 'secret2',
            })
            try:
                parse_radius_secrets(config)
            except ConfigError:
                pass
            else:
                self.fail('must ConfigError with bad radius_ip_2=%s')

    def test_parse_secrets_empty_input(self):
        config = ConfigDict({})
        try:
            parse_radius_secrets(config)
        except ConfigError:
            pass
        else:
            self.fail('did not ConfigError with no radius_ip_1')
        config = ConfigDict({
            'radius_ip_1': '1.1.1.1',
        })
        try:
            parse_radius_secrets(config)
        except ConfigError:
            pass
        else:
            self.fail('did not ConfigError with no radius_secret_1')

    def test_parse_secrets_ip_range(self):
        config = ConfigDict({
            'radius_ip_1': '1.1.1.1-1.1.1.3',
            'radius_secret_1': 'secret1',
            'radius_ip_2': '2.2.2.2-2.2.2.2',
            'radius_secret_2': 'secret2',
            'radius_ip_3': ' 3.3.3.3 - 3.3.3.9 ',
            'radius_secret_3': 'secret3',
        })
        secrets = {
            netaddr.IPNetwork('1.1.1.1/32'): 'secret1',
            netaddr.IPNetwork('1.1.1.2/31'): 'secret1',
            netaddr.IPNetwork('2.2.2.2/32'): 'secret2',
            netaddr.IPNetwork('3.3.3.3/32'): 'secret3',
            netaddr.IPNetwork('3.3.3.4/30'): 'secret3',
            netaddr.IPNetwork('3.3.3.8/31'): 'secret3',
        }
        self.assertEqual(parse_radius_secrets(config), secrets)

    def test_parse_secrets_bad_ip_range(self):
        for bad in ('1.1.1.1-',
                    '-1.1.1.1',
                    '-',
                    '1.1.1.1-1.z',
                    '1.1.1.1-1.1.1.2-1.1.1.3'):
            config = ConfigDict({
                'radius_ip_1': bad,
                'radius_secret_1': 'secret1',
            })
            try:
                parse_radius_secrets(config)
            except ConfigError:
                pass
            else:
                self.fail(
                    'no ConfigError with bad range "%s" in radius_ip_1'
                    % (bad,)
                )

    def test_parse_secrets_nonconsecutive(self):
        config = ConfigDict({'radius_ip_1': '1.1.1.1',
                             'radius_secret_1': 'secret1',
                             'radius_ip_2': '2.2.2.2',
                             'radius_secret_2': 'secret2',
                             'radius_ip_44': '4.4.4.4',
                             'radius_secret_44': 'secret44'})
        secrets = {
            netaddr.IPNetwork('1.1.1.1/32'): 'secret1',
            netaddr.IPNetwork('2.2.2.2/32'): 'secret2',
            netaddr.IPNetwork('4.4.4.4/32'): 'secret44',
        }
        self.assertEqual(parse_radius_secrets(config), secrets)

    def test_parse_secrets_bad_keys(self):
        config = ConfigDict({'radius_ip_1': '1.1.1.1',
                             'radius_secret_1': 'secret1',
                             'radius_ip_NOTANUMBER': '2.2.2.2',
                             'radius_secret_2': 'secret2'})
        with self.assertRaises(ConfigError):
            parse_radius_secrets(config)

    def test_parse_exempt_usernames(self):
        config = ConfigDict({'exempt_username_1': 'testuser1',
                             'exempt_username_2': 'testuser2',
                             'exempt_username_4': 'testuser4'})
        exempt_usernames = ['testuser1', 'testuser2']
        self.assertEqual(parse_exempt_usernames(config), exempt_usernames)

    def test_paloalto_parse_client_ip_attribute(self):
        config = ConfigDict({'client_ip_attr': 'paloalto'})
        self.assertEqual(parse_client_ip_attribute(config), 'PaloAlto-Client-Source-IP')

    def test_direct_parse_client_ip_attribute(self):
        config = ConfigDict({'client_ip_attr': 'PaloAlto-Client-Source-IP'})
        self.assertEqual(parse_client_ip_attribute(config), 'PaloAlto-Client-Source-IP')

    def test_fail_parse_client_ip_attribute(self):
        config = ConfigDict({'client_ip_attr': 'palo_alto'})
        with self.assertRaisesRegexp(ConfigError, 'Invalid configuration value for client_ip_attr: RADIUS attribute'):
            parse_client_ip_attribute(config)

    def test_secret_for_host(self):
        ip1 = '1.1.1.1'
        secret1 = 'secret1'
        ip2 = '2.2.2.2'
        secret2 = 'secret2'
        srv = SimpleMockServer({
            netaddr.IPNetwork(ip1).cidr: secret1,
            netaddr.IPNetwork(ip2).cidr: secret2,
        })
        self.assertEqual(srv.secret_for_host(ip1),
                         secret1)
        self.assertEqual(srv.secret_for_host(ip2),
                         secret2)
        self.assertIs(srv.secret_for_host('3.3.3.3'),
                      None)

    def test_secret_for_host_multiple_subnets_match(self):
        # Returns secret for smallest matching subnet.
        secret1 = 'secret1'
        secret2 = 'secret2'
        srv = SimpleMockServer({
            netaddr.IPNetwork('1.1.0.0/16'): secret1,
            netaddr.IPNetwork('1.1.1.0/24'): secret2,
        })
        self.assertEqual(srv.secret_for_host('1.1.1.1'),
                         secret2)

    @defer.inlineCallbacks
    def test_proxy_state_return(self):
        # create access request
        request = packet.AuthPacket(code=packet.AccessRequest,
                                    id=1,
                                    secret=self.secret,
                                    dict=radius_dictionary())
        request[1] = ['some_username']
        request[2] = [request.PwCrypt('THISisAPassword!')]
        request['Proxy-State'] = 'ProxyMePlease!'
        req_data = request.RequestPacket()

        # send into mock server

        self.srv_inst.action = self.srv_inst.create_accept_packet
        yield self.srv.protocol.datagramReceived(req_data, (self.ip, self.port))

        # read response out
        self.assertEqual(len(self.srv.protocol.transport.queue), 1)
        rsp_data, rsp_addr = self.srv.protocol.transport.queue.pop(0)
        self.assertEqual(rsp_addr, (self.ip, self.port))

        # verify response is correct
        response = packet.AuthPacket(packet=rsp_data,
                                     dict=radius_dictionary())
        response.secret = self.secret
        self.assertEqual(
            response['Proxy-State'],
            ['ProxyMePlease!']
        )


class ChallengeMockServer(ChallengeResponseRadiusServer):
    def __init__(self, secrets):
        super(ChallengeMockServer, self).__init__(
            primary_ator=None,
            secrets=secrets,
            pass_through_attr_names=''
        )

    def get_initial_response(self, request):
        return self.create_challenge(request, 'This is very challenging',
                                     'blarg')

    def get_challenge_response(self, request, state):
        return self.action(request, state)


class TestChallengeRadiusServer(TestCase):
    def setUp(self):
        self.ip = '98.76.54.32'
        self.wrong_ip = '98.76.54.31'
        self.port = 9563
        self.secret = 'throughthreecheesetreesthreefreefleasflew'
        self.srv_inst = ChallengeMockServer({
            netaddr.IPNetwork(self.ip).cidr: self.secret,
            netaddr.IPNetwork(self.wrong_ip).cidr: self.secret,
        })
        self.srv_mod = MockServerModule(self.srv_inst)

        server_port = 9563
        server_networks = {
            self.srv_mod: [
                netaddr.IPNetwork(self.ip).cidr,
                netaddr.IPNetwork(self.wrong_ip).cidr
            ]
        }
        server_names = {self.srv_mod: 'test-server'}
        self.srv = forward_serv.Module(server_port, server_networks, server_names, '', False)
        self.srv.protocol.transport = mock_objects.BasicMockTransport()

    def tearDown(self):
        self.srv.stopService()
        self.srv_inst.cleanup_all()

    @defer.inlineCallbacks
    def test_challenge_request(self):
        # create access request
        request = packet.AuthPacket(code=packet.AccessRequest,
                                    id=1,
                                    secret=self.secret,
                                    dict=radius_dictionary())
        request[1] = ['some_username']
        request[2] = [request.PwCrypt('THISisAPassword!')]
        req_data = request.RequestPacket()

        # send into mock server
        def init_rsp(request):
            self.assertEqual(request.username, 'some_username')
            self.assertEqual(request.password, 'THISisAPassword!')
            return self.srv_inst.create_challenge(
                request,
                'I challenge you!',
                'blarg'
            )
        self.srv_inst.get_initial_response = init_rsp
        yield self.srv.protocol.datagramReceived(req_data, (self.ip, self.port))

        # read response out
        self.assertEqual(len(self.srv.protocol.transport.queue), 1)
        rsp_data, rsp_addr = self.srv.protocol.transport.queue.pop(0)
        self.assertEqual(rsp_addr, (self.ip, self.port))

        # verify response is correct
        response = packet.AuthPacket(packet=rsp_data,
                                     dict=radius_dictionary())
        response.secret = self.secret
        self.assertEqual(response.code, packet.AccessChallenge)
        self.assertTrue(request.VerifyReply(response, rsp_data))
        self.assertTrue('State' in response)

        # create response for challenge
        request = packet.AuthPacket(code=packet.AccessRequest,
                                    id=2,
                                    secret=self.secret,
                                    dict=radius_dictionary())
        request['User-Name'] = 'some_username'
        request['User-Password'] = request.PwCrypt('Swords at dawn')
        request[24] = response[24]
        req_data = request.RequestPacket()

        # send into mock server
        def challenge_rsp(request, state):
            self.assertEqual(state, 'blarg')
            self.assertEqual(request.username, 'some_username')
            self.assertEqual(request.password, 'Swords at dawn')
            return self.srv_inst.create_accept_packet(request)
        self.srv_inst.get_challenge_response = challenge_rsp
        yield self.srv.protocol.datagramReceived(req_data, (self.ip, self.port))

        # read response out
        self.assertEqual(len(self.srv.protocol.transport.queue), 1)
        rsp_data, rsp_addr = self.srv.protocol.transport.queue.pop(0)
        self.assertEqual(rsp_addr, (self.ip, self.port))

        # verify response is correct
        response = packet.AuthPacket(packet=rsp_data,
                                     dict=radius_dictionary())
        response.secret = self.secret
        self.assertEqual(response.code, packet.AccessAccept)
        self.assertTrue(request.VerifyReply(response, rsp_data))

    @defer.inlineCallbacks
    def test_bad_request(self):
        # create access request
        request = packet.AuthPacket(code=packet.AccessRequest,
                                    id=1,
                                    secret=self.secret,
                                    dict=radius_dictionary())
        request[1] = ['some_username']
        request[2] = [request.PwCrypt('THISisAPassword!')]
        req_data = request.RequestPacket()

        # send into mock server
        def init_rsp(request):
            return self.srv_inst.create_challenge(
                request,
                'I challenge you!',
                'blarg'
            )
        self.srv_inst.get_initial_response = init_rsp
        yield self.srv.protocol.datagramReceived(req_data, (self.ip, self.port))

        # read response out
        self.assertEqual(len(self.srv.protocol.transport.queue), 1)
        rsp_data, rsp_addr = self.srv.protocol.transport.queue.pop(0)
        self.assertEqual(rsp_addr, (self.ip, self.port))

        # verify response is correct
        init_response = packet.AuthPacket(packet=rsp_data,
                                          dict=radius_dictionary())
        init_response.secret = self.secret

        #
        # create response for challenge with wrong state
        #
        request = packet.AuthPacket(code=packet.AccessRequest,
                                    id=2,
                                    secret=self.secret,
                                    dict=radius_dictionary())
        request['User-Name'] = 'some_username'
        request['User-Password'] = request.PwCrypt('Swords at dawn')
        request['State'] = init_response['State'][0] + 'a'
        req_data = request.RequestPacket()

        # send into mock server
        def challenge_rsp(request, state):
            self.fail('Should not try to handle bad request!')
        self.srv_inst.get_challenge_response = challenge_rsp
        yield self.srv.protocol.datagramReceived(req_data, (self.ip, self.port))

        # read response out
        self.assertEqual(len(self.srv.protocol.transport.queue), 1)
        rsp_data, rsp_addr = self.srv.protocol.transport.queue.pop(0)
        self.assertEqual(rsp_addr, (self.ip, self.port))

        # verify response is an accessreject
        response = packet.AuthPacket(packet=rsp_data,
                                     dict=radius_dictionary())
        response.secret = self.secret
        self.assertEqual(response.code, packet.AccessReject)

        #
        # now try correct state, but wrong src ip
        #
        request[24] = init_response[24]
        request.id = 3
        req_data = request.RequestPacket()
        yield self.srv.protocol.datagramReceived(req_data, (self.wrong_ip, self.port))

        # read response out
        self.assertEqual(len(self.srv.protocol.transport.queue), 1)
        rsp_data, rsp_addr = self.srv.protocol.transport.queue.pop(0)
        self.assertEqual(rsp_addr, (self.wrong_ip, self.port))

        # verify response is an accessreject
        response = packet.AuthPacket(packet=rsp_data,
                                     dict=radius_dictionary())
        response.secret = self.secret
        self.assertEqual(response.code, packet.AccessReject)

        #
        # now try correct ip, but wrong username
        #
        request['User-Name'] = 'some_username2'
        request.id = 4
        req_data = request.RequestPacket()
        yield self.srv.protocol.datagramReceived(req_data, (self.ip, self.port))

        # read response out
        self.assertEqual(len(self.srv.protocol.transport.queue), 1)
        rsp_data, rsp_addr = self.srv.protocol.transport.queue.pop(0)
        self.assertEqual(rsp_addr, (self.ip, self.port))

        # verify response is correct
        response = packet.AuthPacket(packet=rsp_data,
                                     dict=radius_dictionary())
        response.secret = self.secret
        self.assertEqual(response.code, packet.AccessReject)


class TestSimpleRadiusServerAuthLogging(AuthLoggingTest, SimpleRadiusServerTestRunner):
    @defer.inlineCallbacks
    def test_bad_radius_secret(self):
        # create access request with mismatched secret
        request = packet.AuthPacket(code=packet.AccessRequest,
                                    id=1,
                                    secret='badbadsecret',
                                    dict=radius_dictionary())
        request[1] = ['some_username']
        request[2] = [request.PwCrypt('THISisAPassword!')]
        req_data = request.RequestPacket()

        # send into mock server
        self.srv_inst.action = self.srv_inst.create_reject_packet
        yield self.srv.protocol.datagramReceived(req_data, (self.ip, self.port))

        expected = [{
            "msg": "Cannot decode password",
            "auth_stage": log.AUTH_UNKNOWN,
            "status": log.AUTH_ERROR,
        }]
        self.assert_logged(expected)


if __name__ == '__main__':
    import unittest
    unittest.main()
