from mock_ldap_server import server_key, server_cert  # pylint: disable=W0403
import unittest
from twisted.internet import _sslverify as sslverify
import mock

from twisted.internet._sslverify import TLSVersion

from duoauthproxy.lib import fips_manager
from duoauthproxy.lib.config_error import ConfigError
from duoauthproxy.modules.ssl_server import ChainingOpenSSLContextFactory


class TestSslServer(unittest.TestCase):
    @unittest.skipIf(fips_manager.status(), "TLS locked to v1.2 in FIPS mode")
    def test_bad_minimum_tls_value_raises_an_error(self):
        with self.assertRaisesRegexp(ConfigError, "incorrect minimum_tls_version"):
            ChainingOpenSSLContextFactory(
                minimum_tls_version="ssl2",
                privatekey_data=server_key,
                certificate_data=server_cert,
            )

    def test_bad_cert_file_raises_an_error(self):
        with self.assertRaisesRegexp(ConfigError, "not open certificate file at badcert.pem"):
            ChainingOpenSSLContextFactory(
                privatekey_data=server_key,
                certificate_filename="badcert.pem",
            )

    def test_bad_private_key_file_raises_an_error(self):
        with self.assertRaisesRegexp(ConfigError, "not open private key file at badkey.pem"):
            ChainingOpenSSLContextFactory(
                privatekey_filename="badkey.pem",
                certificate_data=server_cert,
            )

    def test_empty_cipher_list(self):
        try:
            ChainingOpenSSLContextFactory(
                privatekey_data=server_key,
                certificate_data=server_cert,
            )
        except ConfigError:
            self.fail("valid custom cipher list yielded no usable ciphers")

    def test_invalid_cipher_list(self):
        cipher_list = 'BAD:BAD:BAD'

        with self.assertRaisesRegexp(ConfigError, "No usable ciphers specified in cipher_list"):
            ChainingOpenSSLContextFactory(
                privatekey_data=server_key,
                certificate_data=server_cert,
                cipher_list=cipher_list
            )

    def test_valid_custom_cipher_list(self):
        cipher_list = 'ALL'

        try:
            ChainingOpenSSLContextFactory(
                privatekey_data=server_key,
                certificate_data=server_cert,
                cipher_list=cipher_list,
            )
        except ConfigError:
            self.fail("valid custom cipher list yielded no usable ciphers")

    def test_tls_default(self):
        result = ChainingOpenSSLContextFactory(
            privatekey_data=server_key,
            certificate_data=server_cert,
        )

        self.assertEqual(result.minimum_tls_version, TLSVersion.TLSv1_2)

    def test_select_sslv3(self):
        result = ChainingOpenSSLContextFactory(
            privatekey_data=server_key,
            minimum_tls_version="ssl3",
            certificate_data=server_cert,
        )

        if fips_manager.status():
            self.assertEqual(result.minimum_tls_version, TLSVersion.TLSv1_2)
        else:
            self.assertEqual(result.minimum_tls_version, TLSVersion.SSLv3)

    def test_select_tls1_0(self):
        result = ChainingOpenSSLContextFactory(
            privatekey_data=server_key,
            minimum_tls_version="tls1.0",
            certificate_data=server_cert,
        )

        if fips_manager.status():
            self.assertEqual(result.minimum_tls_version, TLSVersion.TLSv1_2)
        else:
            self.assertEqual(result.minimum_tls_version, TLSVersion.TLSv1_0)

    def test_select_tls1_1(self):
        result = ChainingOpenSSLContextFactory(
            privatekey_data=server_key,
            minimum_tls_version="tls1.1",
            certificate_data=server_cert,
        )

        if fips_manager.status():
            self.assertEqual(result.minimum_tls_version, TLSVersion.TLSv1_2)
        else:
            self.assertEqual(result.minimum_tls_version, TLSVersion.TLSv1_1)

    def test_select_tls1_2(self):
        result = ChainingOpenSSLContextFactory(
            privatekey_data=server_key,
            minimum_tls_version="tls1.2",
            certificate_data=server_cert,
        )

        self.assertEqual(result.minimum_tls_version, TLSVersion.TLSv1_2)

    @mock.patch('duoauthproxy.lib.fips_manager.status')
    def test_select_tls12_in_fips_mode_regardless_of_value(self, mock_fips):
        mock_fips.return_value = True
        result = ChainingOpenSSLContextFactory(
            privatekey_data=server_key,
            minimum_tls_version="tls1.0",
            certificate_data=server_cert,
        )

        self.assertEqual(result.minimum_tls_version, TLSVersion.TLSv1_2)

    @mock.patch('duoauthproxy.lib.fips_manager.status')
    def test_select_fips_cipers_in_fips_mode(self, mock_fips):
        mock_fips.return_value = True
        result = ChainingOpenSSLContextFactory(
            privatekey_data=server_key,
            certificate_data=server_cert,
        )

        self.assertEqual([
            sslverify.OpenSSLCipher(u'ECDHE-RSA-AES256-GCM-SHA384'),
            sslverify.OpenSSLCipher(u'ECDHE-RSA-AES256-SHA384'),
            sslverify.OpenSSLCipher(u'ECDHE-RSA-AES128-GCM-SHA256'),
            sslverify.OpenSSLCipher(u'ECDHE-RSA-AES128-SHA256'),
            sslverify.OpenSSLCipher(u'DES-CBC3-SHA')
        ], result.cipher_list.selectCiphers([
            sslverify.OpenSSLCipher(u'ECDHE-RSA-AES128-SHA256'),
            sslverify.OpenSSLCipher(u'ECDHE-RSA-AES128-GCM-SHA256'),
            sslverify.OpenSSLCipher(u'ECDHE-RSA-AES256-SHA384'),
            sslverify.OpenSSLCipher(u'ECDHE-RSA-AES256-GCM-SHA384'),
            sslverify.OpenSSLCipher(u'ECDHE-RSA-3DES-EDE-SHA'),
            sslverify.OpenSSLCipher(u'ECDHE-RSA-AES128-SHA'),
            sslverify.OpenSSLCipher(u'ECDHE-RSA-AES256-SHA'),
            sslverify.OpenSSLCipher(u'DES-CBC3-SHA')
        ]))


if __name__ == '__main__':
    unittest.main()
