import StringIO

from twisted.internet import ssl
from OpenSSL import crypto

from duoauthproxy.lib import ssl_verify

from base import TestCase

CERT1 = """\
-----BEGIN CERTIFICATE-----
MIIE2DCCBEGgAwIBAgIEN0rSQzANBgkqhkiG9w0BAQUFADCBwzELMAkGA1UE
BhMCVVMxFDASBgNVBAoTC0VudHJ1c3QubmV0MTswOQYDVQQLEzJ3d3cuZW50
cnVzdC5uZXQvQ1BTIGluY29ycC4gYnkgcmVmLiAobGltaXRzIGxpYWIuKTEl
MCMGA1UECxMcKGMpIDE5OTkgRW50cnVzdC5uZXQgTGltaXRlZDE6MDgGA1UE
AxMxRW50cnVzdC5uZXQgU2VjdXJlIFNlcnZlciBDZXJ0aWZpY2F0aW9uIEF1
dGhvcml0eTAeFw05OTA1MjUxNjA5NDBaFw0xOTA1MjUxNjM5NDBaMIHDMQsw
CQYDVQQGEwJVUzEUMBIGA1UEChMLRW50cnVzdC5uZXQxOzA5BgNVBAsTMnd3
dy5lbnRydXN0Lm5ldC9DUFMgaW5jb3JwLiBieSByZWYuIChsaW1pdHMgbGlh
Yi4pMSUwIwYDVQQLExwoYykgMTk5OSBFbnRydXN0Lm5ldCBMaW1pdGVkMTow
OAYDVQQDEzFFbnRydXN0Lm5ldCBTZWN1cmUgU2VydmVyIENlcnRpZmljYXRp
b24gQXV0aG9yaXR5MIGdMA0GCSqGSIb3DQEBAQUAA4GLADCBhwKBgQDNKIM0
VBuJ8w+vN5Ex/68xYMmo6LIQaO2f55M28Qpku0f1BBc/I0dNxScZgSYMVHIN
iC3ZH5oSn7yzcdOAGT9HZnuMNSjSuQrfJNqc1lB5gXpa0zf3wkrYKZImZNHk
mGw6AIr1NJtl+O3jEP/9uElY3KDegjlrgbEWGWG5VLbmQwIBA6OCAdcwggHT
MBEGCWCGSAGG+EIBAQQEAwIABzCCARkGA1UdHwSCARAwggEMMIHeoIHboIHY
pIHVMIHSMQswCQYDVQQGEwJVUzEUMBIGA1UEChMLRW50cnVzdC5uZXQxOzA5
BgNVBAsTMnd3dy5lbnRydXN0Lm5ldC9DUFMgaW5jb3JwLiBieSByZWYuIChs
aW1pdHMgbGlhYi4pMSUwIwYDVQQLExwoYykgMTk5OSBFbnRydXN0Lm5ldCBM
aW1pdGVkMTowOAYDVQQDEzFFbnRydXN0Lm5ldCBTZWN1cmUgU2VydmVyIENl
cnRpZmljYXRpb24gQXV0aG9yaXR5MQ0wCwYDVQQDEwRDUkwxMCmgJ6AlhiNo
dHRwOi8vd3d3LmVudHJ1c3QubmV0L0NSTC9uZXQxLmNybDArBgNVHRAEJDAi
gA8xOTk5MDUyNTE2MDk0MFqBDzIwMTkwNTI1MTYwOTQwWjALBgNVHQ8EBAMC
AQYwHwYDVR0jBBgwFoAU8BdiE1U9s/8KAGv7UISX8+1i0BowHQYDVR0OBBYE
FPAXYhNVPbP/CgBr+1CEl/PtYtAaMAwGA1UdEwQFMAMBAf8wGQYJKoZIhvZ9
B0EABAwwChsEVjQuMAMCBJAwDQYJKoZIhvcNAQEFBQADgYEAkNwwAvpkdMKn
CqV8IY00F6j7Rw7/JXyNEwr75Ji174z4xRAN95K+8cPV1ZVqBLssziY2Zcgx
xufuP+NXdYR6Ee9GTxj005i7qIcyunL2POI9n9cd2cNgQ4xYDiKWL2KjLB+6
rQXvqzJ4h6BUcxm1XAX5Uj5tLUUL9wqT6u0G+bI=
-----END CERTIFICATE-----
"""

CERT1_NAME = [('C', 'US'), ('O', 'Entrust.net'), ('OU', 'www.entrust.net/CPS incorp. by ref. (limits liab.)'), ('OU', '(c) 1999 Entrust.net Limited'), ('CN', 'Entrust.net Secure Server Certification Authority')]

CERT2 = """\
-----BEGIN CERTIFICATE-----
MIIDtzCCAp+gAwIBAgIQDOfg5RfYRv6P5WD8G/AwOTANBgkqhkiG9w0BAQUFADBl
MQswCQYDVQQGEwJVUzEVMBMGA1UEChMMRGlnaUNlcnQgSW5jMRkwFwYDVQQLExB3
d3cuZGlnaWNlcnQuY29tMSQwIgYDVQQDExtEaWdpQ2VydCBBc3N1cmVkIElEIFJv
b3QgQ0EwHhcNMDYxMTEwMDAwMDAwWhcNMzExMTEwMDAwMDAwWjBlMQswCQYDVQQG
EwJVUzEVMBMGA1UEChMMRGlnaUNlcnQgSW5jMRkwFwYDVQQLExB3d3cuZGlnaWNl
cnQuY29tMSQwIgYDVQQDExtEaWdpQ2VydCBBc3N1cmVkIElEIFJvb3QgQ0EwggEi
MA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQCtDhXO5EOAXLGH87dg+XESpa7c
JpSIqvTO9SA5KFhgDPiA2qkVlTJhPLWxKISKityfCgyDF3qPkKyK53lTXDGEKvYP
mDI2dsze3Tyoou9q+yHyUmHfnyDXH+Kx2f4YZNISW1/5WBg1vEfNoTb5a3/UsDg+
wRvDjDPZ2C8Y/igPs6eD1sNuRMBhNZYW/lmci3Zt1/GiSw0r/wty2p5g0I6QNcZ4
VYcgoc/lbQrISXwxmDNsIumH0DJaoroTghHtORedmTpyoeb6pNnVFzF1roV9Iq4/
AUaG9ih5yLHa5FcXxH4cDrC0kqZWs72yl+2qp/C3xag/lRbQ/6GW6whfGHdPAgMB
AAGjYzBhMA4GA1UdDwEB/wQEAwIBhjAPBgNVHRMBAf8EBTADAQH/MB0GA1UdDgQW
BBRF66Kv9JLLgjEtUYunpyGd823IDzAfBgNVHSMEGDAWgBRF66Kv9JLLgjEtUYun
pyGd823IDzANBgkqhkiG9w0BAQUFAAOCAQEAog683+Lt8ONyc3pklL/3cmbYMuRC
dWKuh+vy1dneVrOfzM4UKLkNl2BcEkxY5NM9g0lFWJc1aRqoR+pWxnmrEthngYTf
fwk8lOa4JiwgvT2zKIn3X/8i4peEH+ll74fg38FnSbNd67IJKusm7Xi+fT8r87cm
NW1fiQG2SVufAQWbqz0lwcy2f8Lxb4bG+mRo64EtlOtCt/qMHt1i8b5QZ7dsvfPx
H2sMNgcWfzd8qVttevESRmCD1ycEvkvOl77DZypoEd+A5wwzZr8TDRRu838fYxAe
+o0bJW1sj6W3YQGx0qMmoRBxna3iw/nDmVG3KwcIzi7mULKn+gpFL6Lw8g==
-----END CERTIFICATE-----
"""

CERT2_NAME = [('C', 'US'), ('O', 'DigiCert Inc'), ('OU', 'www.digicert.com'), ('CN', 'DigiCert Assured ID Root CA')]

# Subject: C=US, ST=MI, O=Duo Security, OU=HTTP UNIT TESTS, CN=127.0.0.1/emailAddress=akgood@duosecurity.com
# (no subjectAltName)
NORMAL_SERVER_CERT = """-----BEGIN CERTIFICATE-----
MIIE3DCCA8SgAwIBAgIBATANBgkqhkiG9w0BAQUFADCBnTELMAkGA1UEBhMCVVMx
CzAJBgNVBAgMAk1JMRIwEAYDVQQHDAlBbm4gQXJib3IxFTATBgNVBAoMDER1byBT
ZWN1cml0eTEYMBYGA1UECwwPSFRUUCBVTklUIFRFU1RTMRUwEwYDVQQDDAxUZXN0
IFJvb3QgQ0ExJTAjBgkqhkiG9w0BCQEWFmFrZ29vZEBkdW9zZWN1cml0eS5jb20w
HhcNMTQwMTIyMjIyNTAzWhcNMjQwMTIwMjIyNTAzWjCBhjELMAkGA1UEBhMCVVMx
CzAJBgNVBAgMAk1JMRUwEwYDVQQKDAxEdW8gU2VjdXJpdHkxGDAWBgNVBAsMD0hU
VFAgVU5JVCBURVNUUzESMBAGA1UEAwwJMTI3LjAuMC4xMSUwIwYJKoZIhvcNAQkB
FhZha2dvb2RAZHVvc2VjdXJpdHkuY29tMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8A
MIIBCgKCAQEAu+QLLybUhhRfYTYT4HkLQDR2TpeUp3sTi1UvIfGio8oGuAWA/LPq
gpN2oBr/x9LjCG8kRAjfrVerGi43N0V3Fl6SmGy3RqdiN/iXSFUn9af+o9HZEX2i
zOcWoLs3yjvHyrzDN6xSgXi1HJtW6FsDQNW58pfLwdA3yb4a7W1sloj1yEUsmjzI
uhKgMf5ZIQ/vd5cYnmzdEmyPGwyoY01GFM9iHWHsPyR+QKBoNJy9LLpJZGZ6UypV
0YwLkXOfL6R2SEDHm4ft5TB1UmRZCGfoBq5nuBwZKFUhdrTLh9Rb5BDCuzUlv8kS
VLawaoTyuH9VWw3obi0AVh7gtx7XQehqoQIDAQABo4IBOjCCATYwCQYDVR0TBAIw
ADAdBgNVHQ4EFgQUI8l5gqaG8yjz6QctWMbojB4Ve0QwgdIGA1UdIwSByjCBx4AU
UZtbhwIJaFP8SIHNZdX6Ikfv9G6hgaOkgaAwgZ0xCzAJBgNVBAYTAlVTMQswCQYD
VQQIDAJNSTESMBAGA1UEBwwJQW5uIEFyYm9yMRUwEwYDVQQKDAxEdW8gU2VjdXJp
dHkxGDAWBgNVBAsMD0hUVFAgVU5JVCBURVNUUzEVMBMGA1UEAwwMVGVzdCBSb290
IENBMSUwIwYJKoZIhvcNAQkBFhZha2dvb2RAZHVvc2VjdXJpdHkuY29tggkAt5Mu
43VAJZ8wNQYJYIZIAYb4QgENBCgWJk9wZW5TU0wgQ2VydGlmaWNhdGUgZm9yIFNT
TCBXZWIgU2VydmVyMA0GCSqGSIb3DQEBBQUAA4IBAQABisV1k/UeatAAUtNSfUwx
KDf2tx7JO3IMr06dyGcCjwm9hUTaY/JcZi7xztgjrbp/AtrZYwdtJX4GN0lfGQ8m
8oaR7pMSfvRMEMaDtrQcR05FmuxH8lIIZZft8sgjOoZjtBbVrDfMTjMtsORZdt8q
EuJj0S1W8umkjr5mzCgwILdlX4832NbIgmPT+hWVnEooMgiOthMVnSoyy6IEjwLj
8WyDBQn2xpuasVSQPw5GzwaMShzjrsdAAVUD1z6K+xZAJ/ZrTZ2/Y5ckOQdwEEoh
982ziP6tHtzfUKn8CkuOoF3JvInjMb35grLt36ycT7ljk7hpfEK7lmENlGaOJ8ED
-----END CERTIFICATE-----
"""


# Subject: C=US, ST=MI, O=Duo Security, OU=HTTP UNIT TESTS, CN=127.0.0.1/emailAddress=akgood@duosecurity.com
# X509v3 Subject Alternative Name: critical
#     DNS:localhost, DNS:127.0.0.1
TWO_SANS_SERVER_CERT = """-----BEGIN CERTIFICATE-----
MIIFADCCA+igAwIBAgIBAzANBgkqhkiG9w0BAQUFADCBnTELMAkGA1UEBhMCVVMx
CzAJBgNVBAgMAk1JMRIwEAYDVQQHDAlBbm4gQXJib3IxFTATBgNVBAoMDER1byBT
ZWN1cml0eTEYMBYGA1UECwwPSFRUUCBVTklUIFRFU1RTMRUwEwYDVQQDDAxUZXN0
IFJvb3QgQ0ExJTAjBgkqhkiG9w0BCQEWFmFrZ29vZEBkdW9zZWN1cml0eS5jb20w
HhcNMTQwMTIyMjIyNTAzWhcNMjQwMTIwMjIyNTAzWjCBhjELMAkGA1UEBhMCVVMx
CzAJBgNVBAgMAk1JMRUwEwYDVQQKDAxEdW8gU2VjdXJpdHkxGDAWBgNVBAsMD0hU
VFAgVU5JVCBURVNUUzESMBAGA1UEAwwJMTI3LjAuMC4xMSUwIwYJKoZIhvcNAQkB
FhZha2dvb2RAZHVvc2VjdXJpdHkuY29tMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8A
MIIBCgKCAQEAx7ns8aPV+oC0bP9y5rmKrzOeySKtZNP4ZXh7oCLLmNPVdzwEFIJE
pdBdxC23Rypy8O1KLmdErMuYW14WUBBW/AmUgfHDVHg/gdkfKaKxhCDENbKK7vcA
PGpGJkeWYRJAcf1LYfxV1XnLJ0bWGMebnhOSd1f61K4tXCymUJ2wLEWgrQ3LFV7x
WTw/UbX2dAuEaRhOXYihQjfK7AZ0mtB1FZsM+bCPk4JgyHA/vxPF7UZ/X0wsx5Ir
bDx6tcMOCUD5UxDlsRuPNs+xvwNwVjxE+PwEK9IHheretnAbyFFg6TB8KHo06EOZ
8EPVePNn13nFmekgAXKBWdW8gtr7wmJxmwIDAQABo4IBXjCCAVowCQYDVR0TBAIw
ADAdBgNVHQ4EFgQUgjiAYWXC0t1JvdcJTBffLOa08VowgdIGA1UdIwSByjCBx4AU
UZtbhwIJaFP8SIHNZdX6Ikfv9G6hgaOkgaAwgZ0xCzAJBgNVBAYTAlVTMQswCQYD
VQQIDAJNSTESMBAGA1UEBwwJQW5uIEFyYm9yMRUwEwYDVQQKDAxEdW8gU2VjdXJp
dHkxGDAWBgNVBAsMD0hUVFAgVU5JVCBURVNUUzEVMBMGA1UEAwwMVGVzdCBSb290
IENBMSUwIwYJKoZIhvcNAQkBFhZha2dvb2RAZHVvc2VjdXJpdHkuY29tggkAt5Mu
43VAJZ8wNQYJYIZIAYb4QgENBCgWJk9wZW5TU0wgQ2VydGlmaWNhdGUgZm9yIFNT
TCBXZWIgU2VydmVyMCIGA1UdEQEB/wQYMBaCCWxvY2FsaG9zdIIJMTI3LjAuMC4x
MA0GCSqGSIb3DQEBBQUAA4IBAQCkdoje4gsr/y0mjkr2VewFYlBWdxwZ64MF0mTI
GRBWoZQywKUm7RQDuW8r8/dP4V3/IkqH5DwkJE8cnBQdseJdHepIpTZtX2kirfwW
WNiReLHAbm1MIjAXqmGV3QL8d9/K7loozV8JxgmZOCLNE5WJ/Wm+xgFyPC/TqLqc
g1kxTHcjpvBQWJNBLhCCrw4czIOVmI/VC71FEkn2avEpC5RLv0VhqWxNhaTrbwvj
xiqgOZUbxQHNxltZzbWEnpOSSFRfXckCoZvjgB9OHu1u2+re6+TAWjUJxzueS9nT
CkOU4NzanBShJDIuhOYHhDVjAti+5At/zRbAfrR/h2I4mT7A
-----END CERTIFICATE-----
"""

# Subject: C=US, ST=MI, O=Duo Security, OU=HTTP UNIT TESTS, CN=127.0.0.1/emailAddress=akgood@duosecurity.com
# X509v3 Subject Alternative Name:
#     DNS:localhost\x00127.0.0.1
NULLBYTE_SAN_SERVER_CERT = """-----BEGIN CERTIFICATE-----
MIIE/DCCA+SgAwIBAgIBAjANBgkqhkiG9w0BAQUFADCBnTELMAkGA1UEBhMCVVMx
CzAJBgNVBAgMAk1JMRIwEAYDVQQHDAlBbm4gQXJib3IxFTATBgNVBAoMDER1byBT
ZWN1cml0eTEYMBYGA1UECwwPSFRUUCBVTklUIFRFU1RTMRUwEwYDVQQDDAxUZXN0
IFJvb3QgQ0ExJTAjBgkqhkiG9w0BCQEWFmFrZ29vZEBkdW9zZWN1cml0eS5jb20w
HhcNMTQwMTIzMDAyNTQ0WhcNMjQwMTIxMDAyNTQ0WjCBhjELMAkGA1UEBhMCVVMx
CzAJBgNVBAgMAk1JMRUwEwYDVQQKDAxEdW8gU2VjdXJpdHkxGDAWBgNVBAsMD0hU
VFAgVU5JVCBURVNUUzESMBAGA1UEAwwJMTI3LjAuMC4xMSUwIwYJKoZIhvcNAQkB
FhZha2dvb2RAZHVvc2VjdXJpdHkuY29tMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8A
MIIBCgKCAQEAlwNfrgkTrCQgB3WCp7gHvEXdIVUxZZC9pWp0tYiq4N9IvWSGilqs
RgGO/kk5EDBg2wGJdZUALNP8ISrNl/MSVloJQvJNN769WmdnlmmwqbULLuZDvU3j
VI62K9//ZDiTJy0OKM9A2S7ILj30YmX3hrK76J7keXC3hqVchjlwU1wM/U1fPGRq
0lD7x6+bDs6POVLWMmzWNtXSFZFwXTOXUCwB/KVwx41YTyY3Hx7GMBAkGwr1tEKs
LPwl76eeR5Tuz2dMRgwU+D3ERlmsJE3lO8UMrYg7thNtFkAe3ni47rzpJfw4KpA5
Ji5AqTFHHFd2PmGAoYJ4N5ndvvzG9gGaIQIDAQABo4IBWjCCAVYwCQYDVR0TBAIw
ADAdBgNVHQ4EFgQUubXz6rXZXKU6A38ehpkC7eRSQgQwgdIGA1UdIwSByjCBx4AU
DMGKjxEW/8tRGRu4DcPUd8s0SkqhgaOkgaAwgZ0xCzAJBgNVBAYTAlVTMQswCQYD
VQQIDAJNSTESMBAGA1UEBwwJQW5uIEFyYm9yMRUwEwYDVQQKDAxEdW8gU2VjdXJp
dHkxGDAWBgNVBAsMD0hUVFAgVU5JVCBURVNUUzEVMBMGA1UEAwwMVGVzdCBSb290
IENBMSUwIwYJKoZIhvcNAQkBFhZha2dvb2RAZHVvc2VjdXJpdHkuY29tggkA75d+
mYhRo/EwNQYJYIZIAYb4QgENBCgWJk9wZW5TU0wgQ2VydGlmaWNhdGUgZm9yIFNT
TCBXZWIgU2VydmVyMB4GA1UdEQQXMBWCE2xvY2FsaG9zdAAxMjcuMC4wLjEwDQYJ
KoZIhvcNAQEFBQADggEBAD9lfPH6bUvMxhvs4+lkfvQnuPHysNWjytepS4/HuPwX
gIoFHncpjW1fBHJBD4UCWMOg5amsTuJvfzVEOCPIYOjtpHysrpN3Kp0hYAL6XsvJ
0a6OIXate2puRNBI+rbcck8nQcUg/SoYsJzlg+tvqs+rS7O2SbZfdYYpHWpFywQp
hMjC4EJJS91LsGL71fhE4wLoNOYq0Ieoj7ZF6sK+t1UAeXLGMUxblFYOISOkSUSR
ene2DDsaUAgGk9FYs4suvEQOfM+/DfqxnLtp6aCEA+Sh+FPQb/6nZGv3ceWmJvy4
2W7jkh/nIRALhw+xoZJaZLH2Ad+QqaYF/75QnKlUsKE=
-----END CERTIFICATE-----
"""

# Subject: C=US, ST=MI, O=Duo Security, OU=HTTP UNIT TESTS, CN=127.0.0.1/emailAddress=akgood@duosecurity.com
# X509v3 Subject Alternative Name:
#    DNS:localhost, DNS:localhost,127.0.0.1, email:127.0.0.1, IP Address:127.0.0.1, IP Address:192.168.1.1, URI:http://127.0.0.1
MANY_SANS_SERVER_CERT = """-----BEGIN CERTIFICATE-----
MIIFMDCCBBigAwIBAgIBBDANBgkqhkiG9w0BAQUFADCBnTELMAkGA1UEBhMCVVMx
CzAJBgNVBAgMAk1JMRIwEAYDVQQHDAlBbm4gQXJib3IxFTATBgNVBAoMDER1byBT
ZWN1cml0eTEYMBYGA1UECwwPSFRUUCBVTklUIFRFU1RTMRUwEwYDVQQDDAxUZXN0
IFJvb3QgQ0ExJTAjBgkqhkiG9w0BCQEWFmFrZ29vZEBkdW9zZWN1cml0eS5jb20w
HhcNMTQwMTIyMjIyNTAzWhcNMjQwMTIwMjIyNTAzWjCBhjELMAkGA1UEBhMCVVMx
CzAJBgNVBAgMAk1JMRUwEwYDVQQKDAxEdW8gU2VjdXJpdHkxGDAWBgNVBAsMD0hU
VFAgVU5JVCBURVNUUzESMBAGA1UEAwwJMTI3LjAuMC4xMSUwIwYJKoZIhvcNAQkB
FhZha2dvb2RAZHVvc2VjdXJpdHkuY29tMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8A
MIIBCgKCAQEAxC6AnaG3geQYX1LQA7pwjNSg8qUW9P/dQQuinVGs9F92jU2Xdfmk
0TjfPAGRn5rGYpIFP4GD1Y0avzemiXBVC/7r9UBzbF6ikXtnMFl0/sze3S95r/A5
th1XX2lmfIj9qFSW0daxpk4KiWet7D95GpUNBwzFtzodtMceq7ALJMATjBV0azXf
7W4DdaOO4UspHwjWNpHCI/4MVxSsnBCN7o4SX/M+bQajqbLBQKNHCnJtteGMyCeI
5KyL36rzPidr8yNV6AClt1fCrLFmPs42S16FTe1/ozvgaG7x5mIQvsVrlUSGtJAK
JZmTBL4ZMQOiXZiI2RSRnQQBJ4cfssgOYwIDAQABo4IBjjCCAYowCQYDVR0TBAIw
ADAdBgNVHQ4EFgQUk+CQQf20SDa6v/oT7jYRh69xW2wwgdIGA1UdIwSByjCBx4AU
UZtbhwIJaFP8SIHNZdX6Ikfv9G6hgaOkgaAwgZ0xCzAJBgNVBAYTAlVTMQswCQYD
VQQIDAJNSTESMBAGA1UEBwwJQW5uIEFyYm9yMRUwEwYDVQQKDAxEdW8gU2VjdXJp
dHkxGDAWBgNVBAsMD0hUVFAgVU5JVCBURVNUUzEVMBMGA1UEAwwMVGVzdCBSb290
IENBMSUwIwYJKoZIhvcNAQkBFhZha2dvb2RAZHVvc2VjdXJpdHkuY29tggkAt5Mu
43VAJZ8wNQYJYIZIAYb4QgENBCgWJk9wZW5TU0wgQ2VydGlmaWNhdGUgZm9yIFNT
TCBXZWIgU2VydmVyMFIGA1UdEQRLMEmCCWxvY2FsaG9zdIITbG9jYWxob3N0LDEy
Ny4wLjAuMYEJMTI3LjAuMC4xhwR/AAABhwTAqAEBhhBodHRwOi8vMTI3LjAuMC4x
MA0GCSqGSIb3DQEBBQUAA4IBAQAkZoTF19MFWD1VPAfWayECUqVnW1ItIwethtx1
Qp3jvq5KAWc1ndUih16+jPEjyZUOFbvBnXi5Vg1EEN/zcx3YbokNCdOX2xcJiPQS
lm7fX+D8Y2ClW63b6EU5COPIF793vOgLGwIr6uVrwxCPv/c6BxBM2x+zgq5BvP76
jkKruV7gclGmMWYq/eIbPmtn6J68n/KrO9Hf9sBR5sZwK+QRRxVvN8k2f2bK/n1O
RFruv3gBLF0VMGQp0fbvek1YqmBXZn4NtvafljnLjC/N9/lvKIDe6UMfHGbjmIZ+
52XDDXjz5WUhs2qr9D45I2zF28XO/jj3Ri6gGAX+7tg+mWfl
-----END CERTIFICATE-----
"""


class Test_load_ca_bundle(TestCase):
    def assert_load_ca_bundle(self, bundle_string, expected_names):
        # Test with both file and string input:
        for bundle_file in (bundle_string, StringIO.StringIO(bundle_string)):
            certs = ssl_verify.load_ca_bundle(bundle_string)
            names = [c.get_subject().get_components() for c in certs]
            self.assertEqual(sorted(expected_names), sorted(names))
            ctx_factory = ssl.CertificateOptions(caCerts=certs, verify=True)
            # Must not raise:
            ctx_factory.getContext()

    def test_one_cert(self):
        """
        Load one cert.
        """
        self.assert_load_ca_bundle(CERT1, [CERT1_NAME])

    def test_two_certs(self):
        self.assert_load_ca_bundle(CERT1 + CERT2, [CERT1_NAME, CERT2_NAME])

    def test_de_dupe_if_duplicate_cert_in_input(self):
        self.assert_load_ca_bundle(CERT1 + CERT1, [CERT1_NAME])


class Test_match_hostnames(TestCase):
    def assert_match_hostnames(self, cert_string, hostname):
        cert = crypto.load_certificate(crypto.FILETYPE_PEM, cert_string)
        # Must not raise:
        ssl_verify._match_hostnames(cert, hostname)

    def assert_mismatch_hostnames(self, cert_string, hostname):
        with self.assertRaises(ssl_verify._CertificateError):
            self.assert_match_hostnames(cert_string, hostname)

    def test_normal_cert(self):
        """CN-only cert should match the CN, not other things"""
        self.assert_match_hostnames(NORMAL_SERVER_CERT, '127.0.0.1')
        self.assert_mismatch_hostnames(NORMAL_SERVER_CERT, 'localhost')

    def test_two_sans_cert(self):
        """SAN cert with both names should match both names"""
        self.assert_match_hostnames(TWO_SANS_SERVER_CERT, '127.0.0.1')
        self.assert_match_hostnames(TWO_SANS_SERVER_CERT, 'localhost')

    def test_nullbyte_san_cert(self):
        """SAN cert with 'localhost\x00127.0.0.1' as its (only) SAN should
        match *nothing*

        This is true in spite of the CN, because the presence of any DNS
        SAN means we should ignore the CN completely!"""
        self.assert_mismatch_hostnames(NULLBYTE_SAN_SERVER_CERT, '127.0.0.1')
        self.assert_mismatch_hostnames(NULLBYTE_SAN_SERVER_CERT, 'localhost')

    def test_many_sans_cert(self):
        """SAN cert with lots of SANs - but only 'localhost' as a valid
        DNS SAN - should match only 'localhost'

        This is true in spite of the 'IP Address' SAN, because we only
        check DNS SANs (this is true upstream in the Python SSL
        module). Also, once again, the CN is ignored."""
        self.assert_mismatch_hostnames(MANY_SANS_SERVER_CERT, '127.0.0.1')
        self.assert_match_hostnames(MANY_SANS_SERVER_CERT, 'localhost')


if __name__ == '__main__':
    import unittest
    unittest.main()
