#
# Copyright (c) 2018 Duo Security
# All Rights Reserved
#
import mock
import os

from duoauthproxy.lib import util
from duoauthproxy.lib import const
from duoauthproxy.lib.config_provider import ConfigDict
from duoauthproxy.lib.config_error import ConfigError
from duoauthproxy.lib.ldap.client import AD_TRANSPORT_CLEAR, AD_TRANSPORT_LDAPS, AD_TRANSPORTS
from duoauthproxy.lib.ldap.client import AD_AUTH_TYPE_PLAIN, AD_AUTH_TYPE_NTLM_V1, AD_AUTH_TYPE_SSPI, AD_AUTH_TYPE_NTLM_V2
from duoauthproxy.lib.ldap.client import DEFAULT_SSL_VERIFY_DEPTH
from duoauthproxy.lib.const import DEFAULT_LDAP_PORT, DEFAULT_LDAPS_PORT
from base import TestCase


class TestParseFilter(TestCase):
    def test_good_filter(self):
        input_filter = (
            '(|'
            '(&(objectClass=user)(objectCategory=person))'
            '(objectClass=inetOrgPerson)'
            ')')
        result = util.parse_ldap_filter(input_filter)
        self.assertIsNotNone(result)

    def test_bad_filter(self):
        input_filter = 'asdf'
        result = util.parse_ldap_filter(input_filter)
        self.assertIsNone(result)


class TestParseDelimitedSet(TestCase):
    def test_normal(self):
        expected = ['push', 'phone', 'sms']
        delimited_input = 'push,phone,sms'

        actual = util.parse_delimited_set(delimited_input)
        self.assertItemsEqual(expected, actual)

    def test_duplicates(self):
        expected = ['a', 'b']
        delimited_input = 'a,b,a'

        actual = util.parse_delimited_set(delimited_input)
        self.assertItemsEqual(expected, actual)

    def test_non_default_deliminator(self):
        expected = ['a,b', 'b', 'c', 'd']
        delimited_input = 'a,b#b#c#d'

        actual = util.parse_delimited_set(delimited_input, delimiter='#')
        self.assertItemsEqual(expected, actual)

    def test_whitespace(self):
        expected = ['white', 'space', 'be', 'gone', '!']
        delimited_input = '{0}white{1},space,be{2},gone,!{3}'.format('\t', '\n', '\r', '    ')

        actual = util.parse_delimited_set(delimited_input, delimiter=',')
        self.assertItemsEqual(expected, actual)

    def test_trailing_comma(self):
        expected = ['commas', 'rule', 'they', 'do']
        delimited_input = 'commas,rule,they,do,'

        actual = util.parse_delimited_set(delimited_input, delimiter=',')
        self.assertItemsEqual(expected, actual)


class TestParseFactorList(TestCase):
    def test_empty(self):
        with self.assertRaisesRegexp(ConfigError,
                                     'cannot be empty'):
            util.parse_factor_list('')

    def test_only_delimiters(self):
        with self.assertRaisesRegexp(ConfigError,
                                     'cannot be empty'):
            util.parse_factor_list(',')

    def test_one_bad_factor(self):
        with self.assertRaisesRegexp(ConfigError,
                                     'is not one of'):
            util.parse_factor_list('foo')

    def test_bad_factors(self):
        with self.assertRaisesRegexp(ConfigError,
                                     'is not one of'):
            util.parse_factor_list('foo,bar,baz')

    def test_bad_good_factors(self):
        with self.assertRaisesRegexp(ConfigError,
                                     'is not one of'):
            util.parse_factor_list('foo,bar,push')

    def test_good_bad_factors(self):
        with self.assertRaisesRegexp(ConfigError,
                                     'is not one of'):
            util.parse_factor_list('push,foo,bar')

    def test_good_bad_good_factors(self):
        with self.assertRaisesRegexp(ConfigError,
                                     'is not one of'):
            util.parse_factor_list('push,foo,phone')

    def test_good_factor(self):
        for factor in ('auto', 'phone', 'push', 'passcode'):
            self.assertItemsEqual([factor],
                                  util.parse_factor_list(factor))

    def test_good_factors(self):
        self.assertItemsEqual(['push', 'phone', 'auto'],
                              util.parse_factor_list('push,phone,auto'))

    def test_whitespace_good_factors(self):
        self.assertItemsEqual(
            ['push', 'phone', 'auto'],
            util.parse_factor_list(' push, phone, auto '),
        )

    def test_trailing_comma_good_factor(self):
        self.assertItemsEqual(['push'],
                              util.parse_factor_list(' push,'))

    def test_trailing_leading_double_commas_whitespace_good_factors(self):
        self.assertItemsEqual(
            ['push', 'phone'],
            util.parse_factor_list(' , push ,,phone , '),
        )


class TestIsFactor(TestCase):
    def test_empty(self):
        factor_input = ''

        actual = util.is_factor(factor_input)
        self.assertFalse(actual)

    def test_None(self):
        factor_input = None

        actual = util.is_factor(factor_input)
        self.assertFalse(actual)

    def test_bad_factor(self):
        factor_input = 'bad_factor!'

        actual = util.is_factor(factor_input)
        self.assertFalse(actual)

    def test_good_factor(self):
        factor_input = 'push'

        actual = util.is_factor(factor_input)
        self.assertTrue(actual)

    def test_good_factor_with_number(self):
        factor_input = 'phone9'

        actual = util.is_factor(factor_input)
        self.assertTrue(actual)

    def test_good_factor_with_long_number(self):
        factor_input = 'phone19'

        actual = util.is_factor(factor_input)
        self.assertTrue(actual)

    def test_bad_factor_with_number(self):
        factor_input = 'bad_factor!9'

        actual = util.is_factor(factor_input)
        self.assertFalse(actual)

    def test_good_factor_with_whitespace(self):
        factor_input = ' push '

        actual = util.is_factor(factor_input)
        self.assertTrue(actual)

    def test_good_factor_with_number_and_whitespace(self):
        factor_input = ' phone9 '

        actual = util.is_factor(factor_input)
        self.assertTrue(actual)

    def test_bad_factor_with_whitespace(self):
        factor_input = ' badfactor! '

        actual = util.is_factor(factor_input)
        self.assertFalse(actual)


class TestFactorForRequest(TestCase):
    def setUp(self):
        self.mock_preauth_res = {
            'factors': {
                'default': 'push1',
                '1': 'push1',
                '2': 'phone1',
                '3': 'sms1',
            }
        }

    def test_factor_not_auto(self):
        preferred_factors = ['push', 'phone', 'passcode']

        factor = util.factor_for_request(preferred_factors, self.mock_preauth_res)
        self.assertEqual(factor, 'push')

    def test_default_auto(self):
        """If no factors are given in the config auto is passed in to this func."""
        preferred_factors = ['auto']
        self.mock_preauth_res['factors']['default'] = 'sms1'

        factor = util.factor_for_request(preferred_factors, self.mock_preauth_res)
        self.assertEqual(factor, 'sms1')

    def test_no_default(self):
        preferred_factors = ['auto']
        del self.mock_preauth_res['factors']['default']

        factor = util.factor_for_request(preferred_factors, self.mock_preauth_res)
        self.assertIsNone(factor)

    def test_no_default_passcode_backup(self):
        preferred_factors = ['auto', 'passcode']
        del self.mock_preauth_res['factors']['default']

        factor = util.factor_for_request(preferred_factors, self.mock_preauth_res)
        self.assertEqual(factor, 'passcode')

    def test_no_factor_overlap(self):
        preferred_factors = ['sms']
        del self.mock_preauth_res['factors']['3']

        factor = util.factor_for_request(preferred_factors, self.mock_preauth_res)
        self.assertIsNone(factor)


class TestGetAddrPortPairs(TestCase):
    def test_one_host_one_port(self):
        expected_result = [
            ('1.1.1.1', 1000),
        ]
        config = ConfigDict({'host': '1.1.1.1', 'port': '1000'})

        actual_result = util.get_addr_port_pairs(config)
        self.assertEqual(expected_result, actual_result)

    def test_start_with_zero(self):
        expected_result = [
            ('1.1.1.1', 1000),
            ('2.2.2.2', 2000)
        ]
        config = ConfigDict({
            'host': '1.1.1.1',
            'port': '1000',
            'host_2': '2.2.2.2',
            'port_2': '2000',
        })

        actual_result = util.get_addr_port_pairs(config)
        self.assertEqual(expected_result, actual_result)

    def test_start_with_one(self):
        expected_result = [
            ('1.1.1.1', 1000),
            ('2.2.2.2', 2000),
            ('3.3.3.3', 3000)
        ]
        config = ConfigDict({
            'host_1': '1.1.1.1',
            'port_1': '1000',
            'host_2': '2.2.2.2',
            'port_2': '2000',
            'host_3': '3.3.3.3',
            'port_3': '3000',
        })

        actual_result = util.get_addr_port_pairs(config)
        self.assertEqual(expected_result, actual_result)

    def test_use_default_port(self):
        expected_result = [
            ('1.1.1.1', 1000),
            ('2.2.2.2', const.DEFAULT_RADIUS_PORT),
            ('3.3.3.3', const.DEFAULT_RADIUS_PORT)
        ]
        config = ConfigDict({
            'host': '1.1.1.1',
            'port': '1000',
            'host_2': '2.2.2.2',
            'host_3': '3.3.3.3',
        })

        actual_result = util.get_addr_port_pairs(config)
        self.assertEqual(expected_result, actual_result)


class TestGetLDAPPort(TestCase):
    def test_port_already_defined(self):
        config = ConfigDict({
            'port': '2000',
        })

        for transport_type in AD_TRANSPORTS:
            port = util.get_ldap_port(config, transport_type)
            self.assertEqual(port, 2000)

    def test_port_not_given_clear(self):
        config = ConfigDict({})
        port = util.get_ldap_port(config, AD_TRANSPORT_CLEAR)
        self.assertEqual(port, DEFAULT_LDAP_PORT)

    def test_port_not_given_ldaps(self):
        config = ConfigDict({})
        port = util.get_ldap_port(config, AD_TRANSPORT_LDAPS)
        self.assertEqual(port, DEFAULT_LDAPS_PORT)


class TestWarnInsecureSettings(TestCase):
    def test_plain_clear(self):
        msg = util.warn_insecure_settings(AD_AUTH_TYPE_PLAIN, AD_TRANSPORT_CLEAR)
        self.assertIn('WARNING', msg)

    def test_ntlmv1_clear(self):
        msg = util.warn_insecure_settings(AD_AUTH_TYPE_NTLM_V1, AD_TRANSPORT_CLEAR)
        self.assertIn('WARNING', msg)

    def test_no_warnings(self):
        msg = util.warn_insecure_settings(AD_AUTH_TYPE_PLAIN, AD_TRANSPORT_LDAPS)
        self.assertEqual('', msg)


class TestGetSSLCACerts(TestCase):
    def test_no_file_given(self):
        config = ConfigDict({})

        result = util.get_ssl_ca_certs(config)
        self.assertIsNone(result)

    @mock.patch("__builtin__.open")
    def test_cant_load_file(self, mock_open):
        mock_open.side_effect = Exception("Can't open file")
        config = ConfigDict({
            'ssl_ca_certs_file': 'doesntexist.txt',
        })

        with self.assertRaises(ConfigError):
            util.get_ssl_ca_certs(config)


class TestResolveFilePath(TestCase):
    def setUp(self):
        self.old_home_dir = util.get_home_dir()
        util._home_dir = os.path.join(os.sep, 'tmp')

    def tearDown(self):
        util._home_dir = self.old_home_dir

    def test_resolve_file_path_relative(self):
        path = 'myfile'
        resolved_path = util.resolve_file_path(path)
        self.assertEqual(
            resolved_path,
            os.path.join(os.path.join(util._home_dir, 'conf'), path),
        )

    def test_resolve_file_path_absolute(self):
        path = os.path.join(os.sep, 'myfile')
        resolved_path = util.resolve_file_path(path)
        self.assertEqual(
            resolved_path,
            path,
        )


class TestParseADClient(TestCase):
    def setUp(self):
        self.config = ConfigDict({
            'host': '10.0.0.1',
            'service_account_username': 'admin',
            'service_account_password': 'password',
            'search_dn': 'DC=Example,DC=com',
            'auth_type': AD_AUTH_TYPE_PLAIN,
            'transport': AD_TRANSPORT_CLEAR,
            'bind_dn': 'cn=admin,cn=users,dc=example,dc=com',
        })

        self.expected_result = {
            'debug': False,
            'is_logging_insecure': False,
            'bind_dn': 'cn=admin,cn=users,dc=example,dc=com',
            'service_account_username': 'admin',
            'service_account_password': 'password',
            'search_dn': 'DC=Example,DC=com',
            'security_group': '',
            'username_attribute': 'sAMAccountName',
            'at_attribute': 'userPrincipalName',
            'auth_type': AD_AUTH_TYPE_PLAIN,
            'ntlm_domain': '',
            'domain_discovery': False,
            'ntlm_workstation': '',
            'timeout': 10,
            'transport_type': AD_TRANSPORT_CLEAR,
            'ldap_filter': None,
            'ssl_ca_certs': None,
            'ssl_verify_depth': DEFAULT_SSL_VERIFY_DEPTH,
            'ssl_verify_hostname': True,
        }

    def test_parse_basic_config(self):
        actual_result = util.parse_ad_client(self.config)
        self.assertDictEqual(self.expected_result, actual_result)

    def test_parse_sspi_config(self):
        self.config['auth_type'] = AD_AUTH_TYPE_SSPI
        del self.config['service_account_username']
        del self.config['service_account_password']

        self.expected_result.update({
            'auth_type': AD_AUTH_TYPE_SSPI,
            'service_account_username': '',
            'service_account_password': '',
        })

        actual_result = util.parse_ad_client(self.config)
        self.assertDictEqual(self.expected_result, actual_result)

    def test_bind_dn_root(self):
        self.config['auth_type'] = AD_AUTH_TYPE_NTLM_V2
        del self.config['bind_dn']

        self.expected_result.update({
            'auth_type': AD_AUTH_TYPE_NTLM_V2,
            'bind_dn': '<ROOT>',
        })

        actual_result = util.parse_ad_client(self.config)
        self.assertDictEqual(self.expected_result, actual_result)


class TestGetHostList(TestCase):
    def test_one_host_no_number(self):
        expected_result = [
            '1.1.1.1',
        ]
        config = ConfigDict({'host': '1.1.1.1'})

        actual_result = util.get_host_list(config)
        self.assertEqual(expected_result, actual_result)

    def test_one_host_with_number(self):
        expected_result = [
            'hostname',
        ]
        config = ConfigDict({'host_1': 'hostname'})

        actual_result = util.get_host_list(config)
        self.assertEqual(expected_result, actual_result)

    def test_no_host(self):
        config = ConfigDict({})

        with self.assertRaises(ConfigError):
            util.get_host_list(config)

    def test_missing_any_hosts(self):
        config = ConfigDict({})

        with self.assertRaises(ConfigError):
            util.get_host_list(config)

    def test_multiple_host_all_numbered(self):
        expected_result = [
            'hostname',
            'hostname2',
            '1.1.1.1',
        ]
        config = ConfigDict({
            'host_1': 'hostname',
            'host_2': 'hostname2',
            'host_3': '1.1.1.1',
        })

        actual_result = util.get_host_list(config)
        self.assertEqual(expected_result, actual_result)

    def test_multiple_host_first_unnumbered(self):
        expected_result = [
            'hostname',
            'hostname2',
            '1.1.1.1',
        ]
        config = ConfigDict({
            'host': 'hostname',
            'host_2': 'hostname2',
            'host_3': '1.1.1.1',
        })

        actual_result = util.get_host_list(config)
        self.assertEqual(expected_result, actual_result)

    def test_host_and_host1(self):
        expected_result = [
            'hostname',
            '1.1.1.1',
            '2.2.2.2',
        ]
        config = ConfigDict({
            'host': 'hostname',
            'host_1': '1.1.1.1',
            'host_2': '2.2.2.2',
        })

        actual_result = util.get_host_list(config)
        self.assertEqual(expected_result, actual_result)

    def test_unsequential_hosts(self):
        expected_result = [
            'hostname',
            '1.1.1.1',
            '2.2.2.2',
            '4.4.4.4',
        ]
        config = ConfigDict({
            'host': 'hostname',
            'host_1': '1.1.1.1',
            'host_2': '2.2.2.2',
            'host_4': '4.4.4.4',
        })

        actual_result = util.get_host_list(config)
        self.assertEqual(expected_result, actual_result)

    def test_missing_first_host(self):
        expected_result = [
            'hostname2',
        ]
        config = ConfigDict({'host_2': 'hostname2'})

        actual_result = util.get_host_list(config)
        self.assertEqual(expected_result, actual_result)

    def test_unordered_host_in_config(self):
        expected_result = [
            'hostname1',
            'hostname2',
        ]
        config = ConfigDict({
            'host_2': 'hostname2',
            'host': 'hostname1'
        })

        actual_result = util.get_host_list(config)
        self.assertEqual(expected_result, actual_result)


class TestGetDynamicKeys(TestCase):
    def test_double_protected(self):
        """This method is intentionally ignorant and will just return both protected and unprotected"""
        expected_result = [
            'base_1',
            'base_2',
            'base_protected_2',
        ]
        config = ConfigDict({
            'base_1': 'val1',
            'base_2': 'val2',
            'base_protected_2': 'val2',
        })

        actual_result = util.get_dynamic_keys(config, 'base', protectable=True)
        self.assertEqual(expected_result, actual_result)

    def test_no_match_success(self):
        expected_result = []

        config = ConfigDict({
            'a': 'b',
            'c': 'd',
        })

        actual_result = util.get_dynamic_keys(config, 'none')
        self.assertItemsEqual(expected_result, actual_result)

    def test_protected_base_only(self):
        """This method should pull out the protected stem if the key is protectable """
        expected_result = [
            'base_protected',
        ]
        config = ConfigDict({
            'base_protected': 'val2',
        })

        actual_result = util.get_dynamic_keys(config, 'base', protectable=True)
        self.assertEqual(expected_result, actual_result)


if __name__ == '__main__':
    import unittest
    unittest.main()
