#
# Copyright (c) 2018 Duo Security
# All Rights Reserved
#
import __builtin__
import textwrap
from collections import OrderedDict

from twisted.internet import defer
from twisted.logger import Logger, FileLogObserver, LogLevel
import mock
import colorama

from duoauthproxy.lib import util
from duoauthproxy.lib.ldap.client import AD_TRANSPORT_CLEAR
from duoauthproxy.lib.config_provider import ConfigProvider
from duoauthproxy.lib.validation.config.config_results import ConfigResultType

from duoauthproxy.lib.validation.connectivity import connectivity_results
from duoauthproxy.lib.validation.connectivity.check import ad_client
from base import (
    ModuleConnectivityTest,
    MockConnectivityTestToolbox,
    MockTesterResolver,
    MockProtocol,
    MockConfigTestToolbox,
)
from duoauthproxy.lib.validation.validator import (
    ConfigAndConnectivityValidator, test_connectivity,
    SUCCESSFUL_VALIDATION_MESSAGE, test_config_and_connectivity,
    CONNECTIVITY_TESTS_SKIPPED_MESSAGE, test_config
)


class TestValidator(ModuleConnectivityTest):

    def setUp(self):
        super(TestValidator, self).setUp()
        mock_print = mock.Mock()
        self.patch(__builtin__, 'print', mock_print)
        self.maxDiff = None

        self.mock_logger = mock.Mock(spec=Logger)
        self.mock_logger.attach_mock(FileLogObserver, 'observer')

        self.BIG_CONFIG = ConfigProvider(textwrap.dedent("""
            [main]
            debug=True
            interface=127.0.0.1
            http_ca_certs_file=conf/ca-bundle-2.crt
            log_auth_events=True

            [cloud]
            ikey=DITHISISANIKEYFFFFFF
            skey=skeythatislongenoughtopassvalidationffff
            api_host=some.api.host

            [http_proxy]
            api_host=some.other.api.host
            port=8765
            client_ip=1.2.3.4,10.20.30.40

            [ad_client1]
            host=ad.domain.controller.one
            service_account_username=user
            service_account_password=password
            search_dn=dc=example,dc=com

            [ad_client2]
            host=ad.domain.controller.two
            host_2=ad.domain.controller.two.by.two
            service_account_username=user
            service_account_password=password
            search_dn=dc=example,dc=com

            [radius_client]
            host=11.22.33.44
            port=1813
            secret=radiussecretone
            host_2=44.33.22.11
            port_2=1814

            [duo_only_client]

            [radius_server_auto]
            ikey=DITHISISANIKEYFFFFFF
            skey=skeythatislongenoughtopassvalidationffff
            api_host=api.host
            port=1818
            radius_ip_1=111.222.111.222
            radius_secret_1=radiussecretone
            radius_ip_2=222.222.222.222
            radius_secret_2=radiussecrettwo
            client=radius_client

            [radius_server_iframe]
            ikey=DITHISISANIKEYFFFFFF
            skey=skeythatislongenoughtopassvalidationffff
            api_host=api.host
            port=1819
            radius_ip_1=111.222.111.222
            radius_secret_protected_1=radiussecretone
            radius_ip_2=222.222.222.222
            radius_secret_protected_2=radiussecrettwo
            type=array
            client=radius_client

            [radius_server_challenge]
            ikey=DITHISISANIKEYFFFFFF
            skey=skeythatislongenoughtopassvalidationffff
            api_host=api.host
            radius_ip_1=111.222.111.222
            radius_secret_1=radiussecretone
            enroll_challenge=false
            prompt_format=html
            client=radius_client

            [radius_server_concat]
            ikey=DITHISISANIKEYFFFFFF
            skey=skeythatislongenoughtopassvalidationffff
            api_host=api.host
            radius_ip_1=111.222.111.222
            radius_secret_1=radiussecretone
            client=radius_client

            [radius_server_duo_only]
            ikey=DITHISISANIKEYFFFFFF
            skey=skeythatislongenoughtopassvalidationffff
            api_host=api.host
            radius_ip_1=111.222.111.222
            radius_secret_1=radiussecretone

            [radius_server_eap]
            ikey=DITHISISANIKEYFFFFFF
            skey=skeythatislongenoughtopassvalidationffff
            api_host=api.host
            radius_ip_1=111.222.111.222
            radius_secret_1=radiussecretone
            certs=this/is/a/cert.file
            pkey=this/is/a/pkey.file
            client=radius_client

            [ldap_server_auto]
            ikey=DITHISISANIKEYFFFFFF
            skey=skeythatislongenoughtopassvalidationffff
            api_host=api.host
            client=ad_client1
        """))

        # Config with at least one invalid or missing value in each section
        self.BIG_BAD_CONFIG = ConfigProvider(textwrap.dedent("""
            [main]
            interface=127.0.0.1
            http_ca_certs_file=conf/ca-bundle-2.crt
            unexpected_key=unexpectedvalue
            log_auth_events=no

            [cloud]
            ikey=
            skey=someskey
            api_host=some.api.host

            [http_proxy]
            api_host=some.other.api.host
            port=port
            client_ip=1.2.3.4,10.20.30.40

            [ad_client1]
            host=ad.domain.controller.one
            service_account_username=user
            service_account_password=password
            # search_dn=dc=example,dc=com

            [ad_client2]
            # host=ad.domain.controller.two
            # host_2=ad.domain.controller.two.by.two
            service_account_username=user
            service_account_password=password
            search_dn=dc=example,dc=com

            [radius_client]
            host=11.22.33.44
            port=1813
            secret=
            host_2=44.33.22.11
            port_2=1814

            [duo_only_client]
            host=thisshouldntbehere

            [radius_server_auto]
            ikey=
            skey=
            api_host=api.host
            port=1818
            radius_ip_1=111.222.111.222
            radius_secret_1=radiussecretone
            radius_ip_2=222.222.222.222
            radius_secret_2=radiussecrettwo
            client=radius_client

            [radius_server_iframe]
            # ikey=ikey
            # skey=skey
            api_host=api.host
            port=1819
            radius_ip_1=111.222.111.222
            radius_secret_protected_1=radiussecretone
            radius_ip_2=222.222.222.222
            radius_secret_protected_2=radiussecrettwo

            [radius_server_challenge]
            ikey=ikey
            skey=skey
            api_host=
            radius_ip_1=111.222.111.222
            radius_secret_1=radiussecretone
            enroll_challenge=notabool
            prompt_format=html

            [radius_server_concat]
            ikey=ikey
            skey=skey
            # api_host=api.host
            radius_ip_1=111.222.111.222
            radius_secret_1=radiussecretone
            client=radius_client

            [radius_server_duo_only]
            ikey=ikey
            skey=skey
            api_host=api.host
            radius_ip_1=111.222.111.222
            # radius_secret_1=radiussecretone

            [radius_server_eap]
            ikey=ikey
            skey=skey
            api_host=api.host
            radius_ip_1=notanip
            radius_secret_1=radiussecretone

            [ldap_server_auto]
            ikey=ikey
            skey=
            api_host=api.host
        """))

        self.RADIUS_SERVER_RESULT = connectivity_results.RadiusServerResult(
            self.GOOD_CONFIG_CHECK_RESULT,
            self.GOOD_PING_RESULT,
            self.GOOD_TIME_DRIFT_RESULT,
            self.GOOD_VALIDATE_RESULT,
            self.GOOD_LISTEN_UDP_RESULT
        )

        self.CONFIG_RESULT = OrderedDict([
            ('main', connectivity_results.MainSectionResult(
                self.GOOD_CERTS_RESULT,
                connectivity_results.NOT_APPLICABLE_TEST_RESULT)),
            ('cloud', connectivity_results.CloudResult(
                self.GOOD_PING_RESULT,
                self.GOOD_TIME_DRIFT_RESULT,
            )),
            ('http_proxy', connectivity_results.HttpProxyResult(
                self.GOOD_CONFIG_CHECK_RESULT,
                self.GOOD_LISTEN_TCP_RESULT,
                self.GOOD_PING_RESULT,
                self.GOOD_TIME_DRIFT_RESULT)),
            ('ad_client1', connectivity_results.AdClientConnectivityResult(
                [
                    connectivity_results.LdapHostResult(
                        'ad.domain.controller.one',
                        389,
                        AD_TRANSPORT_CLEAR,
                        self.GOOD_CONNECT_TCP_RESULT,
                        connectivity_results.LdapConnectResult(True, 'ad.domain.controller.one', 389),
                        self.GOOD_BIND_SERVICE_ACCOUNT,
                        self.GOOD_SEARCH_HAS_RESULTS
                    )
                ]
            )),
            ('ad_client2', connectivity_results.AdClientConnectivityResult(
                [
                    connectivity_results.LdapHostResult(
                        'ad.domain.controller.two',
                        389,
                        AD_TRANSPORT_CLEAR,
                        self.GOOD_CONNECT_TCP_RESULT,
                        connectivity_results.LdapConnectResult(True, 'ad.domain.controller.two', 389),
                        self.GOOD_BIND_SERVICE_ACCOUNT,
                        self.GOOD_SEARCH_HAS_RESULTS
                    ),
                    connectivity_results.LdapHostResult(
                        'ad.domain.controller.two.by.two',
                        389,
                        AD_TRANSPORT_CLEAR,
                        self.GOOD_CONNECT_TCP_RESULT,
                        connectivity_results.LdapConnectResult(True, 'ad.domain.controller.two.by.two', 389),
                        self.GOOD_BIND_SERVICE_ACCOUNT,
                        self.GOOD_SEARCH_HAS_RESULTS
                    )
                ]
            )),
            ('radius_client', connectivity_results.RadiusClientResult(
                [self.GOOD_CONNECT_RADIUS_RESULT, self.GOOD_CONNECT_RADIUS_RESULT]
            )),
            ('duo_only_client', connectivity_results.SKIPPED_SECTION_RESULT),
            ('radius_server_auto', self.RADIUS_SERVER_RESULT),
            ('radius_server_iframe', self.RADIUS_SERVER_RESULT),
            ('radius_server_challenge', self.RADIUS_SERVER_RESULT),
            ('radius_server_concat', self.RADIUS_SERVER_RESULT),
            ('radius_server_duo_only', self.RADIUS_SERVER_RESULT),
            ('radius_server_eap', self.RADIUS_SERVER_RESULT),
            ('ldap_server_auto', connectivity_results.LdapServerResult(
                self.GOOD_PING_RESULT,
                self.GOOD_TIME_DRIFT_RESULT,
                self.GOOD_VALIDATE_RESULT,
                connectivity_results.NOT_APPLICABLE_TEST_RESULT,
                self.GOOD_LISTEN_TCP_RESULT
            )),
        ])

        self.SUCCESS_TOOLBOX = MockConnectivityTestToolbox()
        self.RESOLVER = MockTesterResolver()

    @defer.inlineCallbacks
    def test_connectivity_tester_with_all_possible_sections(self):
        expected_result = {
            'main': self.RESOLVER.MOCK_RESULT,
            'cloud': self.RESOLVER.MOCK_RESULT,
            'http_proxy': self.RESOLVER.MOCK_RESULT,
            'ad_client1': self.RESOLVER.MOCK_RESULT,
            'ad_client2': self.RESOLVER.MOCK_RESULT,
            'radius_client': self.RESOLVER.MOCK_RESULT,
            'duo_only_client': self.RESOLVER.MOCK_RESULT,
            'radius_server_auto': self.RESOLVER.MOCK_RESULT,
            'radius_server_iframe': self.RESOLVER.MOCK_RESULT,
            'radius_server_challenge': self.RESOLVER.MOCK_RESULT,
            'radius_server_concat': self.RESOLVER.MOCK_RESULT,
            'radius_server_duo_only': self.RESOLVER.MOCK_RESULT,
            'radius_server_eap': self.RESOLVER.MOCK_RESULT,
            'ldap_server_auto': self.RESOLVER.MOCK_RESULT,
        }

        actual_result = yield test_connectivity(self.BIG_CONFIG,
                                                self.mock_logger,
                                                toolbox=self.SUCCESS_TOOLBOX,
                                                resolver=self.RESOLVER)
        self.mock_logger.info.assert_called()
        self.assertDictEqual(expected_result, actual_result)

    @defer.inlineCallbacks
    def test_connectivity_tester_with_real_resolver(self):
        """An integration test of the connectivity tester + real resolver"""
        mock_connect_host = mock.Mock(return_value=MockProtocol(bind_success=True))
        self.patch(ad_client, '_connect_host', mock_connect_host)
        expected_result = self.CONFIG_RESULT

        actual_result = yield test_connectivity(self.BIG_CONFIG,
                                                self.mock_logger,
                                                toolbox=self.SUCCESS_TOOLBOX)

        # Main Section
        self.mock_logger.info.assert_any_call(connectivity_results.MainSectionResult.SUCCESS_MESSAGE)
        # http proxy
        self.mock_logger.info.assert_any_call(connectivity_results.HttpProxyResult.SUCCESS_MESSAGE)
        # ad client 1
        self.mock_logger.info.assert_any_call(connectivity_results.AdClientConnectivityResult.SUCCESS_MESSAGE)
        # ad client 2
        self.mock_logger.info.assert_any_call(connectivity_results.AdClientConnectivityResult.SUCCESS_MESSAGE)
        # radius client
        self.mock_logger.info.assert_any_call(connectivity_results.RadiusClientResult.SUCCESS_MESSAGE)
        # radius auto
        self.mock_logger.info.assert_any_call(connectivity_results.RadiusServerResult.SUCCESS_MESSAGE)
        # radius iframe
        self.mock_logger.info.assert_any_call(connectivity_results.RadiusServerResult.SUCCESS_MESSAGE)
        # radius challenge
        self.mock_logger.info.assert_any_call(connectivity_results.RadiusServerResult.SUCCESS_MESSAGE)
        # radius concat
        self.mock_logger.info.assert_any_call(connectivity_results.RadiusServerResult.SUCCESS_MESSAGE)
        # radius eap
        self.mock_logger.info.assert_any_call(connectivity_results.RadiusServerResult.SUCCESS_MESSAGE)
        # ldap auto
        self.mock_logger.info.assert_any_call(connectivity_results.LdapServerResult.SUCCESS_MESSAGE)
        # cloud
        self.mock_logger.info.assert_any_call(connectivity_results.CloudResult.SUCCESS_MESSAGE)

        self.assertDictEqual(expected_result, actual_result)

    @defer.inlineCallbacks
    def test_connectivity_tester_with_real_resolver_and_failed_ad_client(self):
        """An integration test where we mock ad client to fail the bind"""
        mock_connect_host = mock.Mock(return_value=MockProtocol(bind_success=True))
        self.patch(ad_client, '_connect_host', mock_connect_host)

        skipped_search_result = connectivity_results.UnmetPrerequisiteSkippedTestResult('search', 'bind')
        for res in self.CONFIG_RESULT['ad_client1'].ldap_host_results:
            res.bind_result = self.BAD_BIND_SERVICE_ACCOUNT
            res.search_result = skipped_search_result

        for res in self.CONFIG_RESULT['ad_client2'].ldap_host_results:
            res.bind_result = self.BAD_BIND_SERVICE_ACCOUNT
            res.search_result = skipped_search_result
        expected_result = self.CONFIG_RESULT

        toolbox = MockConnectivityTestToolbox(fail_bind_service_account=True)
        actual_result = yield test_connectivity(self.BIG_CONFIG,
                                                self.mock_logger,
                                                toolbox=toolbox)

        # Main Section
        self.mock_logger.info.assert_any_call(connectivity_results.MainSectionResult.SUCCESS_MESSAGE)
        # http proxy
        self.mock_logger.info.assert_any_call(connectivity_results.HttpProxyResult.SUCCESS_MESSAGE)
        # ad client 1
        self.mock_logger.warn.assert_any_call(connectivity_results.AdClientConnectivityResult.FAILURE_MESSAGE)
        self.mock_logger.warn.assert_any_call(
            connectivity_results.LdapHostResult.FAILURE_MESSAGE,
            host='ad.domain.controller.one',
            port=389,
            transport=AD_TRANSPORT_CLEAR
        )
        self.mock_logger.error.assert_any_call(
            connectivity_results.LdapBindResult.FAILURE_MESSAGE,
            bind_username='bad_username'
        )
        # ad client 2
        self.mock_logger.warn.assert_any_call(connectivity_results.AdClientConnectivityResult.FAILURE_MESSAGE)
        self.mock_logger.warn.assert_any_call(
            connectivity_results.LdapHostResult.FAILURE_MESSAGE,
            host='ad.domain.controller.two',
            port=389,
            transport=AD_TRANSPORT_CLEAR
        )
        self.mock_logger.warn.assert_any_call(
            connectivity_results.LdapHostResult.FAILURE_MESSAGE,
            host='ad.domain.controller.two.by.two',
            port=389,
            transport=AD_TRANSPORT_CLEAR
        )
        # radius client
        self.mock_logger.info.assert_any_call(connectivity_results.RadiusClientResult.SUCCESS_MESSAGE)
        # radius auto
        self.mock_logger.info.assert_any_call(connectivity_results.RadiusServerResult.SUCCESS_MESSAGE)
        # radius iframe
        self.mock_logger.info.assert_any_call(connectivity_results.RadiusServerResult.SUCCESS_MESSAGE)
        # radius challenge
        self.mock_logger.info.assert_any_call(connectivity_results.RadiusServerResult.SUCCESS_MESSAGE)
        # radius concat
        self.mock_logger.info.assert_any_call(connectivity_results.RadiusServerResult.SUCCESS_MESSAGE)
        # radius eap
        self.mock_logger.info.assert_any_call(connectivity_results.RadiusServerResult.SUCCESS_MESSAGE)
        # ldap auto
        self.mock_logger.info.assert_any_call(connectivity_results.LdapServerResult.SUCCESS_MESSAGE)
        # cloud
        self.mock_logger.info.assert_any_call(connectivity_results.CloudResult.SUCCESS_MESSAGE)

        self.assertDictEqual(expected_result, actual_result)

    @defer.inlineCallbacks
    def test_connectivity_tester_catches_high_level_exceptions(self):
        """
        An integration test where we catch any exceptions at a high
        level to prevent the tool from completely crashing
        """

        with mock.patch('duoauthproxy.lib.validation.validator.ConfigAndConnectivityValidator._test_all_sections',
                        mock.Mock(side_effect=Exception)):
            yield test_connectivity(self.BIG_CONFIG,
                                    self.mock_logger,
                                    toolbox=self.SUCCESS_TOOLBOX,
                                    resolver=self.RESOLVER)

            # Assert that the error was printed and that the summary section
            # was not printed since the validation was not able to complete
            self.mock_logger.error.assert_called()
            self.assert_call_count_with_args(self.mock_logger.info, 0, 'SUMMARY')

    @defer.inlineCallbacks
    def test_connectivity_tester_with_skipped_sections(self):
        skipping_config = ConfigProvider(textwrap.dedent("""
            [duo_only_client]

            [misspelled_section_name]
            ikey=DIAAA
            skey=AAA
            api_host=api-host.com
        """))

        yield test_connectivity(skipping_config,
                                self.mock_logger,
                                toolbox=self.SUCCESS_TOOLBOX)

        # Duo Only Client skip
        self.mock_logger.info.assert_any_call(connectivity_results.SkippedSectionResult.MESSAGE)
        # Bad section name
        self.mock_logger.error.assert_any_call(connectivity_results.InvalidSectionResult.MESSAGE)

    @defer.inlineCallbacks
    def test_successful_validation(self):
        config = ConfigProvider(textwrap.dedent("""
            [radius_client]
            host=127.0.0.1
            secret=wowsosecret
        """))

        validator = ConfigAndConnectivityValidator(connectivity_toolbox=self.SUCCESS_TOOLBOX)
        yield validator.validate(config, self.mock_logger)
        self.mock_logger.info.assert_any_call(SUCCESSFUL_VALIDATION_MESSAGE)

    @defer.inlineCallbacks
    def test_no_duplicate_sections_in_summary(self):
        config = ConfigProvider(textwrap.dedent("""
            [ldap_server_auto]
            ikey=DI000000000000000000
            skey=thisisntarealskeybutitcanplaysoneonthetv
            api_host=api-12345678.test.duosecurity.com
            client=ad_client
            ssl_port=6636
        """))
        # Test both config and connectivity since there will be results in the
        # summary section for both (a warning for config and error for connectivity)
        validator = ConfigAndConnectivityValidator(connectivity_toolbox=MockConnectivityTestToolbox(fail_ping=True))
        yield validator.validate(config, self.mock_logger)
        self.assert_call_count_with_args(self.mock_logger.info, 1,
                                         'Section [{section}]'.format(section='ldap_server_auto'))

    @defer.inlineCallbacks
    def test_summary_printed_for_warnings(self):
        config = ConfigProvider(textwrap.dedent("""
            [ldap_server_auto]
            ikey=DI000000000000000000
            skey=thisisntarealskeybutitcanplaysoneonthetv
            api_host=api-12345678.test.duosecurity.com
            client=ad_client
            ssl_port=6636
        """))
        # Provide config that results in warnings for missing ssl_key_path and
        # ssl_cert_path, which should be the only result output in the summary
        # section
        validator = ConfigAndConnectivityValidator(connectivity_toolbox=self.SUCCESS_TOOLBOX)
        yield validator.validate(config, self.mock_logger)

        # Assert that the summary was printed with a warning for the
        # [ldap_server_auto] section and that no errors were printed.
        self.mock_logger.info.assert_any_call("Section [{section}]".format(section='ldap_server_auto'))
        self.mock_logger.emit.assert_any_call(LogLevel.warn, ConfigResultType.MissingKey.value, key='ssl_key_path')
        self.mock_logger.emit.assert_any_call(LogLevel.warn, ConfigResultType.MissingKey.value, key='ssl_cert_path')
        self.mock_logger.error.assert_not_called()

    @defer.inlineCallbacks
    def test_section_without_errors_or_warnings_not_in_summary(self):
        config = ConfigProvider(textwrap.dedent("""
            [ad_client]
            host=127.0.0.1
            service_account_username=blah
            service_account_password=password
            search_dn=DC=example,DC=com

            [ldap_server_auto]
            ikey=DI000000000000000000
            skey=thisisntarealskeybutitcanplaysoneonthetv
            api_host=api-12345678.test.duosecurity.com
            client=ad_client
            ssl_port=6636
        """))
        mock_connect_host = mock.Mock(return_value=MockProtocol(bind_success=True))
        self.patch(ad_client, '_connect_host', mock_connect_host)
        # Provide config that results in warnings for missing ssl_key_path and
        # ssl_cert_path, which should be the only result output in the summary
        # section
        validator = ConfigAndConnectivityValidator(connectivity_toolbox=self.SUCCESS_TOOLBOX)
        yield validator.validate(config, self.mock_logger)

        # Assert that the summary was printed with a warning for the
        # [ldap_server_auto] section and the [ad_client] section is not included
        # in the summary.
        self.mock_logger.info.assert_any_call("Section [{section}]".format(section='ldap_server_auto'))
        self.assert_call_count_with_args(self.mock_logger.info, 0, 'Section [{section}]', section='ad_client')
        self.mock_logger.emit.assert_any_call(LogLevel.warn, ConfigResultType.MissingKey.value, key='ssl_key_path')
        self.mock_logger.emit.assert_any_call(LogLevel.warn, ConfigResultType.MissingKey.value, key='ssl_cert_path')
        self.mock_logger.error.assert_not_called()

    @defer.inlineCallbacks
    def test_no_connectivity_tests_if_config_errors(self):
        config = ConfigProvider(textwrap.dedent("""
            [ad_client]
            host=
        """))

        # _validate_connectivity shouldn't get called in this case since there
        # are config errors. If it does get called, raise an exception.
        exception = Exception("Connectivity tests shouldn't be run if there are configuration errors")
        with mock.patch('duoauthproxy.lib.validation.validator.ConfigAndConnectivityValidator._validate_connectivity',
                        mock.Mock(side_effect=exception)):
            yield test_config_and_connectivity(config, self.mock_logger)

    @defer.inlineCallbacks
    def test_no_connectivity_skipped_message_when_testing_config_only(self):
        yield test_config(self.BIG_CONFIG, self.mock_logger, toolbox=MockConfigTestToolbox())
        self.assert_call_count_with_args(self.mock_logger.warn, 0, CONNECTIVITY_TESTS_SKIPPED_MESSAGE)

    @defer.inlineCallbacks
    def test_no_connectivity_tests_if_cross_section_errors(self):
        config = ConfigProvider(textwrap.dedent("""
            [main]
            http_proxy_host=somehost

            [http_proxy]
            api_host=some.api.host
        """))

        # Assert that the message for skipping connectivity tests gets printed
        # and that connectivity was not validated since there are cross-section
        # config errors
        exception = Exception("Connectivity tests shouldn't be run if there are configuration errors")
        with mock.patch('duoauthproxy.lib.validation.validator.ConfigAndConnectivityValidator._validate_connectivity',
                        mock.Mock(side_effect=exception)):
            yield test_config_and_connectivity(config, self.mock_logger)
            self.assert_call_count_with_args(self.mock_logger.warn, 1, CONNECTIVITY_TESTS_SKIPPED_MESSAGE)

    @defer.inlineCallbacks
    def test_no_cross_section_summary_if_no_cross_section_errors(self):
        config = ConfigProvider(textwrap.dedent("""
            [http_proxy]
        """))

        yield test_config(config, self.mock_logger)
        self.assert_call_count_with_args(self.mock_logger.info, 0, 'Cross-section results')

    @defer.inlineCallbacks
    def test_config_tester_with_good_full_config(self):
        results = yield test_config(self.BIG_CONFIG, self.mock_logger, toolbox=MockConfigTestToolbox())

        self.assertTrue(results['ad_client1'].is_successful())
        self.assertTrue(results['ad_client2'].is_successful())
        self.assertTrue(results['ldap_server_auto'].is_successful())
        self.assertTrue(results['radius_client'].is_successful())
        self.assertTrue(results['radius_server_auto'].is_successful())
        self.assertTrue(results['radius_server_concat'].is_successful())
        self.assertTrue(results['radius_server_challenge'].is_successful())
        self.assertTrue(results['radius_server_iframe'].is_successful())
        self.assertTrue(results['radius_server_duo_only'].is_successful())
        self.assertTrue(results['radius_server_eap'].is_successful())
        self.assertTrue(results['http_proxy'].is_successful())
        self.assertTrue(results['cloud'].is_successful())
        self.assertTrue(results['main'].is_successful())
        self.assertTrue(results['duo_only_client'].is_successful())
        self.assertTrue(results[ConfigAndConnectivityValidator.CROSS_CONFIG_KEY].is_successful())

    @defer.inlineCallbacks
    def test_config_tester_with_bad_full_config(self):
        results = yield test_config(self.BIG_BAD_CONFIG, self.mock_logger, toolbox=MockConfigTestToolbox())

        self.assertFalse(results['ad_client1'].is_successful())
        self.assertFalse(results['ad_client2'].is_successful())
        self.assertFalse(results['ldap_server_auto'].is_successful())
        self.assertFalse(results['radius_client'].is_successful())
        self.assertFalse(results['radius_server_auto'].is_successful())
        self.assertFalse(results['radius_server_concat'].is_successful())
        self.assertFalse(results['radius_server_challenge'].is_successful())
        self.assertFalse(results['radius_server_iframe'].is_successful())
        self.assertFalse(results['radius_server_duo_only'].is_successful())
        self.assertFalse(results['radius_server_eap'].is_successful())
        self.assertFalse(results['http_proxy'].is_successful())
        self.assertFalse(results['cloud'].is_successful())
        self.assertFalse(results['main'].is_successful())
        self.assertFalse(results['duo_only_client'].is_successful())

    @defer.inlineCallbacks
    def test_config_using_main_overrides(self):
        config = ConfigProvider(textwrap.dedent("""
            [main]
            debug=True
            http_ca_certs_file=/tmp/certs
            http_proxy_host=my_proxy_host.com
            interface=127.0.0.1
            client=ad_client1
            server=radius_server_auto

            [ad_client1]
            host=ad.domain.controller.one
            service_account_username=user
            service_account_password=password
            search_dn=dc=example,dc=com

            [ldap_server_auto]
            ikey=DITHISISANIKEYFFFFFF
            skey=skeythatislongenoughtopassvalidationffff
            api_host=api.host
            client=ad_client1

            [radius_server_auto]
            ikey=DITHISISANIKEYFFFFFF
            skey=skeythatislongenoughtopassvalidationffff
            api_host=api.host
            radius_ip_1=111.222.111.222
            radius_secret_1=radiussecretone

            [radius_server_eap]
            ikey=DITHISISANIKEYFFFFFF
            skey=skeythatislongenoughtopassvalidationffff
            api_host=api.host
            radius_ip_1=111.222.111.222
            radius_secret_1=radiussecretone
            certs=this/is/a/cert.file
            pkey=this/is/a/pkey.file
            client=ad_client1

            [cloud]
            ikey=DITHISISANIKEYFFFFFF
            skey=skeythatislongenoughtopassvalidationffff
            api_host=api.host
        """))
        results = yield test_config(config, self.mock_logger, toolbox=MockConfigTestToolbox())
        self.assertTrue(all([res.is_successful() for res in results.values()]))


class TestValidatorColorized(ModuleConnectivityTest):

    def setUp(self):
        super(TestValidatorColorized, self).setUp()
        self.maxDiff = None

        self.mock_logger = mock.Mock(spec=Logger)
        self.mock_logger.attach_mock(FileLogObserver, 'observer')

        self.SUCCESS_TOOLBOX = MockConnectivityTestToolbox()

        windows = util.is_windows_os()
        self.RED = colorama.Fore.LIGHTRED_EX if windows else colorama.Fore.RED
        self.GREEN = colorama.Fore.LIGHTGREEN_EX if windows else colorama.Fore.GREEN

    @mock.patch('__builtin__.print')
    @defer.inlineCallbacks
    def test_section_without_errors_has_green_printing(self, mock_print):
        config = ConfigProvider(textwrap.dedent("""
            [ad_client]
            host=127.0.0.1
            service_account_username=blah
            service_account_password=password
            search_dn=DC=example,DC=com
        """))
        mock_connect_host = mock.Mock(return_value=MockProtocol(bind_success=True))
        self.patch(ad_client, '_connect_host', mock_connect_host)

        validator = ConfigAndConnectivityValidator(connectivity_toolbox=self.SUCCESS_TOOLBOX)
        yield validator.validate(config, self.mock_logger)

        # 1 green for config
        # 2 green for connectivity
        # 3 green for summary
        self.assert_call_count_with_args(mock_print, 3, self.GREEN, end='')

    @mock.patch('__builtin__.print')
    @defer.inlineCallbacks
    def test_section_with_errors_has_red_printing(self, mock_print):
        config = ConfigProvider(textwrap.dedent("""
            [ad_client]
            host=
            service_account_username=blah
            service_account_password=password
            search_dn=DC=example,DC=com
        """))
        mock_connect_host = mock.Mock(return_value=MockProtocol(bind_success=True))
        self.patch(ad_client, '_connect_host', mock_connect_host)

        validator = ConfigAndConnectivityValidator(connectivity_toolbox=self.SUCCESS_TOOLBOX)
        yield validator.validate(config, self.mock_logger)

        # 1 red for config
        # 2 red for summary
        self.assert_call_count_with_args(mock_print, 2, self.RED, end='')

    @mock.patch('__builtin__.print')
    @defer.inlineCallbacks
    def test_mix_section_success_mix_coloring(self, mock_print):
        config = ConfigProvider(textwrap.dedent("""
            [ad_client]
            host=
            service_account_username=blah
            service_account_password=password
            search_dn=DC=example,DC=com

            [ldap_server_auto]
            ikey=DI000000000000000000
            skey=thisisntarealskeybutitcanplaysoneonthetv
            api_host=api-12345678.test.duosecurity.com
            client=ad_client
        """))
        mock_connect_host = mock.Mock(return_value=MockProtocol(bind_success=True))
        self.patch(ad_client, '_connect_host', mock_connect_host)

        validator = ConfigAndConnectivityValidator(connectivity_toolbox=self.SUCCESS_TOOLBOX)
        yield validator.validate(config, self.mock_logger)

        # 1 red for config ad client
        # 2 red for summary
        self.assert_call_count_with_args(mock_print, 2, self.RED, end='')

        # 1 green for config ldap_server_auto
        self.assert_call_count_with_args(mock_print, 1, self.GREEN, end='')


if __name__ == '__main__':
    import unittest
    unittest.main()
