from duoauthproxy.lib import ldap
from duoauthproxy.lib.validation.config.check import ad_client
from duoauthproxy.lib.config_provider import ConfigDict
from duoauthproxy.lib.validation.config.config_results import (
    MissingKey, UnexpectedKey, InvalidValue, SkippedTest, UnmetDependency)
from duoauthproxy.lib.validation.connectivity.connectivity_results import (
    ConfigCheckResult
)

from ....base import TestCase, MockConfigTestToolbox


class TestCheckAdClientConfig(TestCase):
    """
    Tests the [ad_client] config section tester module
    """

    def setUp(self):
        super(TestCheckAdClientConfig, self).setUp()
        self.BASIC_CONFIG_INPUT = ConfigDict({
            'host': '127.0.0.1',
            'service_account_username': 'user',
            'service_account_password': 'pass',
            'search_dn': 'dc=example,dc=com',
        })

    def test_basic_config(self):
        result = ad_client.check_ad_client(self.BASIC_CONFIG_INPUT)
        self.assertEqual(result, ConfigCheckResult([]))

    def test_full_good_config(self):
        config = ConfigDict({
            'host': '127.0.0.1',
            'host_2': '127.0.0.2',
            'service_account_username': 'user',
            'service_account_password_protected': 'pass',
            'search_dn': 'dc=example,dc=com',
            'security_group_dn': 'CN=DuoVPNUsers,OU=Groups,DC=example,DC=com',
            'ldap_filter': '(&(username=*foo*)(primaryGroupID=513))',
            'timeout': 4,
            'ssl_ca_certs_file': 'certspath',
            'ssl_verify_hostname': 'true',
            'bind_dn': 'UID=jsmith,DC=example,DC=net',
            'ntlm_domain': 'testdomain',
            'ntlm_workstation': 'test_workstation',
            'port': 8080,
            'transport': 'ldaps',
            'auth_type': 'ntlm2',
            'username_attribute': 'sAMAccountName',
            'at_attribute': 'userPrincipalName',
            'domain_discovery': 'False',
        })

        expected_result = ConfigCheckResult([])
        actual_result = ad_client.check_ad_client(config,
                                                  MockConfigTestToolbox())
        self.assert_results_equal(expected_result, actual_result)

    def test_no_host(self):
        expected_result = ConfigCheckResult([
            MissingKey(key='host')
        ])

        del self.BASIC_CONFIG_INPUT['host']
        actual_result = ad_client.check_ad_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_unsequential_hosts(self):
        """We don't want to flag a config that is missing a host number as invalid"""
        expected_result = ConfigCheckResult([])

        self.BASIC_CONFIG_INPUT['host_3'] = '127.0.0.3'
        actual_result = ad_client.check_ad_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_no_search_dn(self):
        expected_result = ConfigCheckResult([
            MissingKey(key='search_dn')
        ])

        del self.BASIC_CONFIG_INPUT['search_dn']
        actual_result = ad_client.check_ad_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_no_service_account_username(self):
        expected_result = ConfigCheckResult([
            MissingKey(key='service_account_username')
        ])

        del self.BASIC_CONFIG_INPUT['service_account_username']
        actual_result = ad_client.check_ad_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_no_service_account_password(self):
        expected_result = ConfigCheckResult([
            MissingKey(key='service_account_password/service_account_password_protected')
        ])

        del self.BASIC_CONFIG_INPUT['service_account_password']
        actual_result = ad_client.check_ad_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_invalid_host(self):
        invalid_host = ''
        expected_result = ConfigCheckResult([
            InvalidValue(key='host', value=invalid_host)
        ])

        self.BASIC_CONFIG_INPUT['host'] = invalid_host
        actual_result = ad_client.check_ad_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_out_of_sequence_host(self):
        config = self.BASIC_CONFIG_INPUT
        expected_result = ConfigCheckResult([
        ])

        del config['host']
        config['host_2'] = "the.real.host"
        actual_result = ad_client.check_ad_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_host_1_and_host(self):
        config = self.BASIC_CONFIG_INPUT
        expected_result = ConfigCheckResult([
        ])

        config['host_1'] = config['host']
        actual_result = ad_client.check_ad_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_invalid_search_dn(self):
        invalid_search_dn = 'DC=example,com'
        expected_result = ConfigCheckResult([
            InvalidValue(key='search_dn', value=invalid_search_dn)
        ])

        self.BASIC_CONFIG_INPUT['search_dn'] = invalid_search_dn
        actual_result = ad_client.check_ad_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_invalid_service_account_username(self):
        invalid_username = ''
        expected_result = ConfigCheckResult([
            InvalidValue(key='service_account_username', value=invalid_username)
        ])

        self.BASIC_CONFIG_INPUT['service_account_username'] = invalid_username
        actual_result = ad_client.check_ad_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_invalid_service_account_password(self):
        invalid_password = ''
        expected_result = ConfigCheckResult([
            InvalidValue(key='service_account_password', value=invalid_password)
        ])

        self.BASIC_CONFIG_INPUT['service_account_password'] = invalid_password
        actual_result = ad_client.check_ad_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_invalid_service_account_password_protected(self):
        invalid_protected_password = ''
        expected_result = ConfigCheckResult([
            InvalidValue(key='service_account_password_protected',
                         value=invalid_protected_password)
        ])

        del self.BASIC_CONFIG_INPUT['service_account_password']
        self.BASIC_CONFIG_INPUT['service_account_password_protected'] = invalid_protected_password
        actual_result = ad_client.check_ad_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_invalid_ldap_filter(self):
        invalid_ldap_filter = '(cn=andy*)(cn=steve*)(cn=margaret*))'
        expected_result = ConfigCheckResult([
            InvalidValue(key='ldap_filter', value=invalid_ldap_filter)
        ])

        self.BASIC_CONFIG_INPUT['ldap_filter'] = invalid_ldap_filter
        actual_result = ad_client.check_ad_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_invalid_transport(self):
        invalid_transport = 'bad_transport'
        expected_result = ConfigCheckResult([
            InvalidValue(key='transport', value=invalid_transport),
            SkippedTest(test='check_valid_cert_for_transport', key='transport')
        ])

        self.BASIC_CONFIG_INPUT['transport'] = invalid_transport
        actual_result = ad_client.check_ad_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_mixed_case_transport(self):
        """ Transports are lowered before validating """
        mixed_case_transport = 'cLEAr'
        expected_result = ConfigCheckResult([])

        self.BASIC_CONFIG_INPUT['transport'] = mixed_case_transport
        actual_result = ad_client.check_ad_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_invalid_timeout(self):
        invalid_timeout = 'not_a_number'
        expected_result = ConfigCheckResult([
            InvalidValue(key='timeout', value=invalid_timeout)
        ])

        self.BASIC_CONFIG_INPUT['timeout'] = invalid_timeout
        actual_result = ad_client.check_ad_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_invalid_ssl_ca_certs_file(self):
        invalid_certs_file = 'bad_file_path'
        expected_result = ConfigCheckResult([
            InvalidValue(key='ssl_ca_certs_file', value=invalid_certs_file)
        ])

        self.BASIC_CONFIG_INPUT['ssl_ca_certs_file'] = invalid_certs_file
        actual_result = ad_client.check_ad_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_invalid_ssl_verify_hostname(self):
        invalid_certs_file = 'bad_file_path'
        expected_result = ConfigCheckResult([
            InvalidValue(key='ssl_ca_certs_file', value=invalid_certs_file)
        ])

        self.BASIC_CONFIG_INPUT['ssl_ca_certs_file'] = invalid_certs_file
        actual_result = ad_client.check_ad_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_invalid_auth_type(self):
        invalid_auth_type = 'bad_auth_type'
        expected_result = ConfigCheckResult([
            InvalidValue(key='auth_type', value=invalid_auth_type),
            SkippedTest(test='check_valid_bind_dn_for_auth_type',
                        key='auth_type')
        ])

        self.BASIC_CONFIG_INPUT['auth_type'] = invalid_auth_type
        actual_result = ad_client.check_ad_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_mixed_case_auth_type(self):
        mixed_case_auth_type = 'PlAIn'
        expected_result = ConfigCheckResult([
            UnmetDependency(message='bind_dn is required for auth_type %s'
                                    % mixed_case_auth_type.lower())
        ])

        self.BASIC_CONFIG_INPUT['auth_type'] = mixed_case_auth_type
        actual_result = ad_client.check_ad_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_invalid_bind_dn(self):
        invalid_bind_dn = 'UID=jsmith,example,DC=net'
        expected_result = ConfigCheckResult([
            InvalidValue(key='bind_dn', value=invalid_bind_dn)
        ])

        self.BASIC_CONFIG_INPUT['bind_dn'] = invalid_bind_dn
        actual_result = ad_client.check_ad_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_invalid_ntlm_domain(self):
        invalid_ntlm_domain = None
        expected_result = ConfigCheckResult([
            InvalidValue(key='ntlm_domain', value=invalid_ntlm_domain)
        ])

        self.BASIC_CONFIG_INPUT['ntlm_domain'] = invalid_ntlm_domain
        actual_result = ad_client.check_ad_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_invalid_ntlm_workstation(self):
        invalid_ntlm_workstation = None
        expected_result = ConfigCheckResult([
            InvalidValue(key='ntlm_workstation', value=invalid_ntlm_workstation)
        ])

        self.BASIC_CONFIG_INPUT['ntlm_workstation'] = invalid_ntlm_workstation
        actual_result = ad_client.check_ad_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_invalid_port(self):
        invalid_port = 'not_a_port'
        expected_result = ConfigCheckResult([
            InvalidValue(key='port', value=invalid_port)
        ])

        self.BASIC_CONFIG_INPUT['port'] = invalid_port
        actual_result = ad_client.check_ad_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_invalid_username_attribute(self):
        invalid_username_attribute = ''
        expected_result = ConfigCheckResult([
            InvalidValue(key='username_attribute',
                         value=invalid_username_attribute)
        ])

        self.BASIC_CONFIG_INPUT['username_attribute'] = \
            invalid_username_attribute
        actual_result = ad_client.check_ad_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_invalid_at_attribute(self):
        invalid_at_attribute = ''
        expected_result = ConfigCheckResult([
            InvalidValue(key='at_attribute', value=invalid_at_attribute)
        ])

        self.BASIC_CONFIG_INPUT['at_attribute'] = invalid_at_attribute
        actual_result = ad_client.check_ad_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_no_service_account_required_for_sspi(self):
        expected_result = ConfigCheckResult([])
        del self.BASIC_CONFIG_INPUT['service_account_username']
        del self.BASIC_CONFIG_INPUT['service_account_password']

        self.BASIC_CONFIG_INPUT['auth_type'] = ldap.client.AD_AUTH_TYPE_SSPI
        actual_result = ad_client.check_ad_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_certs_provided_for_secure_transport(self):
        transport_type = ldap.client.AD_TRANSPORT_LDAPS
        expected_result = ConfigCheckResult([
            UnmetDependency(message='ssl_ca_certs_file is required for '
                                    'transport type %s' % transport_type)
        ])

        self.BASIC_CONFIG_INPUT['transport'] = transport_type
        actual_result = ad_client.check_ad_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_certs_optional_if_verify_hostname_disabled(self):
        transport_type = ldap.client.AD_TRANSPORT_LDAPS
        expected_result = ConfigCheckResult([])

        self.BASIC_CONFIG_INPUT['transport'] = transport_type
        self.BASIC_CONFIG_INPUT['ssl_verify_hostname'] = 'false'
        actual_result = ad_client.check_ad_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_bind_dn_required_for_plain_auth_type(self):
        auth_type = ldap.client.AD_AUTH_TYPE_PLAIN
        expected_result = ConfigCheckResult([
            UnmetDependency(message='bind_dn is required for auth_type %s'
                                    % auth_type)
        ])

        self.BASIC_CONFIG_INPUT['auth_type'] = auth_type
        actual_result = ad_client.check_ad_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_unexpected_key(self):
        unexpected_key = 'unexpected_key'
        expected_result = ConfigCheckResult([
            UnexpectedKey(key=unexpected_key)
        ])

        self.BASIC_CONFIG_INPUT[unexpected_key] = unexpected_key
        actual_result = ad_client.check_ad_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def assert_results_equal(self, expected_result, actual_result):
        self.assertEqual(expected_result, actual_result,
                         "{0} and {1} should be equal".format(
                             str(expected_result), str(actual_result)))


if __name__ == '__main__':
    import unittest
    unittest.main()
