import copy

from duoauthproxy.lib.validation.config.config_toolbox import STANDARD_CONFIG_TOOLBOX
from duoauthproxy.lib.validation.config.check import base
from duoauthproxy.lib.config_provider import ConfigDict
from duoauthproxy.lib.validation.connectivity.connectivity_results import ConfigCheckResult
from duoauthproxy.lib.validation.config.config_results import (
    UnpairedKey,
    UnexpectedKey,
    IncompatibleKeys,
    InvalidValue,
    MissingKey
)
from ....base import TestCase


class TestConfigCheckBase(TestCase):
    """
    Tests the base methods used by the various config validation modules
    """

    def setUp(self):
        super(TestConfigCheckBase, self).setUp()
        self.BASIC_RADIUS_SERVER_CONFIG = ConfigDict({
            'ikey': 'DI000000000000000000',
            'skey': 'thisisntarealskeybutitcanplaysoneonthetv',
            'api_host': 'api-12345678.test.duosecurity.com',
            'radius_ip_1': '127.0.0.1',
            'client': 'radius_client',
            'radius_secret_1': 'radius_secret',
        })

    def test_mismatched_radius_config(self):
        config = self.BASIC_RADIUS_SERVER_CONFIG
        config['radius_ip_2'] = config['radius_ip_1']
        del config['radius_ip_1']

        expected_result = ConfigCheckResult([
            UnpairedKey(key1='radius_secret_2', key2='radius_ip_2'),
            UnpairedKey(key1='radius_ip_1', key2='radius_secret_1')
        ])

        problems = base.check_radius_secret_radius_ip_balance(
            self.BASIC_RADIUS_SERVER_CONFIG)
        actual_result = ConfigCheckResult(problems)
        self.assert_results_equal(expected_result, actual_result)

    def test_duplicate_protected_radius_config(self):
        config = self.BASIC_RADIUS_SERVER_CONFIG
        config['radius_secret_protected_1'] = config['radius_secret_1']
        expected_result = ConfigCheckResult([
            IncompatibleKeys(key1='radius_secret_1',
                             key2='radius_secret_protected_1')
        ])

        problems = base.check_radius_secret_radius_ip_balance(config)
        actual_result = ConfigCheckResult(problems)
        self.assert_results_equal(expected_result, actual_result)

    def test_unexpected_keys(self):
        toolbox = STANDARD_CONFIG_TOOLBOX
        config = self.BASIC_RADIUS_SERVER_CONFIG
        config['bad_key'] = 'test'
        config['type'] = 'citrix_netscaler'
        expected_result = ConfigCheckResult([
            UnexpectedKey(key='bad_key'),
            UnexpectedKey(key='type')
        ])

        resolver = base.get_basic_radius_server_resolver(config, toolbox)
        problems = base.check_for_unexpected_keys(config, toolbox, resolver)
        actual_result = ConfigCheckResult(problems)
        self.assert_results_equal(expected_result, actual_result)

    def test_skey_protected(self):
        config = self.BASIC_RADIUS_SERVER_CONFIG
        del config['skey']
        config['skey_protected'] = 'A' * 80

        expected_result = ConfigCheckResult([])
        problems = base.check_not_both_skey_and_skey_protected(config, STANDARD_CONFIG_TOOLBOX)
        actual_result = ConfigCheckResult(problems)
        self.assert_results_equal(expected_result, actual_result)

    def test_skey_and_skey_protected(self):
        config = self.BASIC_RADIUS_SERVER_CONFIG
        config['skey_protected'] = 'A' * 80
        expected_result = ConfigCheckResult([
            IncompatibleKeys(key1='skey', key2='skey_protected')
        ])
        problems = base.check_not_both_skey_and_skey_protected(config, STANDARD_CONFIG_TOOLBOX)
        actual_result = ConfigCheckResult(problems)
        self.assert_results_equal(expected_result, actual_result)

    def test_mixed_protected_and_unprotected_secrets(self):
        config = ConfigDict({
            'ikey': 'DI000000000000000000',
            'skey': 'thisisntarealskeybutitcanplaysoneonthetv',
            'api_host': 'api-12345678.test.duosecurity.com',
            'client': 'radius_client',
            'radius_ip_1': '127.0.0.1',
            'radius_secret_1': 'radius_secret',
            'radius_ip_2': '127.0.1.0/24',
            'radius_secret_2': 'radius_secret',
            'radius_ip_3': '127.0.2.0-127.0.2.255',
            'radius_secret_protected_3': 'radius_secret',
            'radius_ip_4': '127.0.3.0/255.255.0.0',
            'radius_secret_4': 'radius_secret',
            'radius_ip_5': '127.0.0.1',
            'radius_secret_protected_5': 'radius_secret',
        })

        expected_result = ConfigCheckResult([])
        problems = base.check_radius_secret_radius_ip_balance(config)
        actual_result = ConfigCheckResult(problems)
        self.assert_results_equal(expected_result, actual_result)

    def test_common_radius_required_keys(self):
        config = ConfigDict({})
        expected_result = ConfigCheckResult([
            MissingKey(key='ikey'),
            MissingKey(key='skey'),
            MissingKey(key='api_host'),
            MissingKey(key='radius_ip_1'),
            MissingKey(key='radius_secret_1'),
        ])

        problems = base.check_common_required_radius_keys(config, STANDARD_CONFIG_TOOLBOX)
        actual_result = ConfigCheckResult(problems)
        self.assert_results_equal(expected_result, actual_result)

    def test_common_radius_server_config(self):
        bad_values = {
            'ikey': ['IDnotanikey',
                     50,
                     'IDButjusttoolongijom',
                     'IDTHISISANIKEYBUTITSTOOLONGTOPASS',
                     'IDBUTALLCPASAC19IOAJ'],
            'skey': ['tooshorttobeanskey',
                     0,
                     '_punctuationisnotallowedinthesekeys.....'],
            'api_host': ['', 1, False, []],
            'radius_ip_1': ['notanip', '127.0.0.0.0', '2001:0db8:85a3:0000:0000:8a2e:0370:'],
            'radius_secret_1': ['', 1, False, []],
            'client': [0, False, []],
            'api_timeout': ['-1', '1.2'],
            'failmode': ['notsafe', 'open', '', False, []],
            'port': ['0', '65536', '1.1', '127.0.0.1', 'http'],
            'interface': ['eth1', 0, '127.0.0.0.0', 'iface', True],
            'pass_through_attr_names': ['', False, []],
            'exempt_username_1': ['', False, []],
            'pw_codec': ['latin-0', 'utf8080', 'notacodec', 1, False, []],
        }

        for key in bad_values:
            config = copy.deepcopy(self.BASIC_RADIUS_SERVER_CONFIG)
            for value in bad_values[key]:
                config[key] = value
                resolver = base.get_basic_radius_server_resolver(config, STANDARD_CONFIG_TOOLBOX)
                expected_result = ConfigCheckResult([
                    InvalidValue(key=key, value=config[key])
                ])
                problems = base.run_config_value_checks(config, resolver)
                actual_result = ConfigCheckResult(problems)
                self.assert_results_equal(expected_result, actual_result)

    def assert_results_equal(self, expected_result, actual_result):
        self.assertEqual(expected_result, actual_result,
                         "{0} and {1} should be equal".format(
                             expected_result, actual_result))


if __name__ == '__main__':
    import unittest
    unittest.main()
