import unittest

from duoauthproxy.lib.validation.config.check import cloud
from duoauthproxy.lib.config_provider import ConfigDict
from duoauthproxy.lib.validation.connectivity.connectivity_results import ConfigCheckResult
from duoauthproxy.lib.validation.config import config_results

from ....base import MockConfigTestToolbox


class TestCloudConfig(unittest.TestCase):

    def test_valid_config_succeeds(self):
        expected_result = ConfigCheckResult([])

        config = ConfigDict({
            'ikey': 'DI000000000000000000',
            'skey': 'Fm4PkSiPn2iQnsZWQ5v3qUrmO8K86yZeGgZVwfSa',
            'api_host': 'api-12345678.test.duosecurity.com',
        })
        result = cloud.check_cloud(config, toolbox=MockConfigTestToolbox())
        self.assertEqual(expected_result, result)

    def test_valid_config_skey_protected_succeeds(self):
        expected_result = ConfigCheckResult([])

        config = ConfigDict({
            'ikey': 'DI000000000000000000',
            'skey_protected': 'thisissomeprotectedskey',
            'api_host': 'api-12345678.test.duosecurity.com',
        })
        result = cloud.check_cloud(config, toolbox=MockConfigTestToolbox())
        self.assertEqual(expected_result, result)

    def test_valid_config_with_optional_keys_succeeds(self):
        expected_result = ConfigCheckResult([])

        config = ConfigDict({
            'ikey': 'DI000000000000000000',
            'skey_protected': 'thisissomeprotectedskey',
            'api_host': 'api-12345678.test.duosecurity.com',
            'service_account_username': 'user_foo',
            'service_account_password': 'password_foo',
        })
        result = cloud.check_cloud(config, toolbox=MockConfigTestToolbox())
        self.assertEqual(expected_result, result)

    def test_missing_required_keys_fails(self):
        expected_result = ConfigCheckResult([
            config_results.MissingKey(key=key)
            for key in ['ikey', 'api_host', 'skey']
        ])
        config = ConfigDict({})
        result = cloud.check_cloud(config, toolbox=MockConfigTestToolbox())
        self.assertEqual(expected_result, result)

    def test_invalid_optional_keys_fails(self):
        '''service_account_username and service_account_password must exist together'''
        expected_result = ConfigCheckResult([
            config_results.MissingKey(key='service_account_username'),
        ])
        config = ConfigDict({
            'ikey': 'DI000000000000000000',
            'skey': 'Fm4PkSiPn2iQnsZWQ5v3qUrmO8K86yZeGgZVwfSa',
            'api_host': 'api-12345678.test.duosecurity.com',
            'service_account_password': 'password_foo',
        })
        result = cloud.check_cloud(config, toolbox=MockConfigTestToolbox())
        self.assertEqual(expected_result, result)

        expected_result = ConfigCheckResult([
            config_results.MissingKey(key='service_account_password'),
        ])
        config = ConfigDict({
            'ikey': 'DI000000000000000000',
            'skey': 'Fm4PkSiPn2iQnsZWQ5v3qUrmO8K86yZeGgZVwfSa',
            'api_host': 'api-12345678.test.duosecurity.com',
            'service_account_username': 'user_foo',
        })
        result = cloud.check_cloud(config, toolbox=MockConfigTestToolbox())
        self.assertEqual(expected_result, result)

    def test_bad_values_fails(self):
        config = ConfigDict({
            'ikey': 'thisisntarealskeybutitcanplaysoneonthetv',
            'skey': 'DI000000000000000000',
            'api_host': '',
            'service_account_username': '',
            'service_account_password': '',
        })

        expected_result = ConfigCheckResult(
            [config_results.InvalidValue(key=key, value=value)
                for key, value in config.items()],
        )

        result = cloud.check_cloud(config, toolbox=MockConfigTestToolbox())
        self.assertEqual(expected_result, result)

    def test_unexpected_keys_fails(self):
        expected_result = ConfigCheckResult([
            config_results.UnexpectedKey(key='lolwut'),
        ])

        config = ConfigDict({
            'ikey': 'DI000000000000000000',
            'skey': 'Fm4PkSiPn2iQnsZWQ5v3qUrmO8K86yZeGgZVwfSa',
            'api_host': 'api-12345678.test.duosecurity.com',
            'service_account_username': 'some username',
            'service_account_password': 'you think this is a pw',
            'lolwut': 'wow such unexpected',
        })

        result = cloud.check_cloud(config, toolbox=MockConfigTestToolbox())
        self.assertEqual(expected_result, result)


if __name__ == '__main__':
    unittest.main()
