import unittest
import textwrap

from duoauthproxy.lib.validation.config.check import cross_config
from duoauthproxy.lib.validation.connectivity.connectivity_results import ConfigCheckResult
from duoauthproxy.lib.config_provider import ConfigProvider
from duoauthproxy.lib.validation.config.config_results import (
    KeyIncompatibleWithSection,
    NetworkContention,
    InvalidClientServerMapping,
    PortCollision,
    KeyInvalidWithClient,
    IneffectiveConfig
)

from ....base import TestCase


class TestConfigBase(unittest.TestCase):
    def assert_results_equal(self, expected_result, actual_result):
        self.assertEqual(expected_result, actual_result,
                         "{0} and {1} should be equal".format(
                             unicode(expected_result), unicode(actual_result)))

    def _check_network_contention(self, config):
        '''Wrapped version of check_network_contention that returns ConfigCheckResult'''
        return ConfigCheckResult(cross_config.check_network_contention(config))


class TestHTTPProxy(TestConfigBase):

    def test_good_http_proxy_section(self):
        config = ConfigProvider(textwrap.dedent("""
            [main]
            debug=True
            interface=127.0.0.1
            http_ca_certs_file=conf/ca-bundle-2.crt
            log_auth_events=True

            [http_proxy]
            api_host=some.other.api.host
            port=8765
            client_ip=1.2.3.4,10.20.30.40

        """))
        result = cross_config.check_cross_config(config)
        self.assert_results_equal(result, ConfigCheckResult([]))

    def test_good_http_proxy2_section(self):
        config = ConfigProvider(textwrap.dedent("""
            [main]
            debug=True
            interface=127.0.0.1
            http_ca_certs_file=conf/ca-bundle-2.crt
            log_auth_events=True

            [http_proxy2]
            api_host=some.other.api.host
            port=8765
            client_ip=1.2.3.4,10.20.30.40

        """))
        result = cross_config.check_cross_config(config)
        self.assert_results_equal(result, ConfigCheckResult([]))

    def test_both(self):
        config = ConfigProvider(textwrap.dedent("""
            [main]
            debug=True
            interface=127.0.0.1
            http_ca_certs_file=conf/ca-bundle-2.crt
            log_auth_events=True
            http_proxy_host=127.0.0.1

            [http_proxy]
            api_host=some.other.api.host
            port=8765
            client_ip=1.2.3.4,10.20.30.40

        """))
        expected_result = ConfigCheckResult([
            KeyIncompatibleWithSection(key1='http_proxy_host', key2='http_proxy')
        ])

        actual_result = cross_config.check_cross_config(config)
        self.assert_results_equal(expected_result, actual_result)

    def test_both_with_different_section_name(self):
        config = ConfigProvider(textwrap.dedent("""
            [main]
            debug=True
            interface=127.0.0.1
            http_ca_certs_file=conf/ca-bundle-2.crt
            log_auth_events=True
            http_proxy_host=127.0.0.1

            [http_proxy2]
            api_host=some.other.api.host
            port=8765
            client_ip=1.2.3.4,10.20.30.40

        """))
        expected_result = ConfigCheckResult([
            KeyIncompatibleWithSection(key1='http_proxy_host', key2='http_proxy2')
        ])

        actual_result = cross_config.check_cross_config(config)
        self.assert_results_equal(expected_result, actual_result)

    def test_good_http_proxy_keys(self):
        config = ConfigProvider(textwrap.dedent("""
            [main]
            debug=True
            interface=127.0.0.1
            http_ca_certs_file=conf/ca-bundle-2.crt
            log_auth_events=True
            http_proxy_host=127.0.0.1
            http_proxy_port=8080

        """))
        result = cross_config.check_cross_config(config)
        self.assert_results_equal(result, ConfigCheckResult([]))


class TestCheckPortContentionRadius(TestConfigBase):
    def test_no_ports(self):
        config = ConfigProvider(textwrap.dedent("""
            [cloud]
            ikey=DITHISISANIKEYFFFFFF
            skey=skeythatislongenoughtopassvalidationffff
            api_host=some.api.host
        """))
        expected = ConfigCheckResult([
        ])
        actual = self._check_network_contention(config)
        self.assert_results_equal(expected, actual)

    def test_radius_no_collision(self):
        config = ConfigProvider(textwrap.dedent("""
            [radius_server_auto1]
            radius_ip_1=0.0.0.0

            [radius_server_auto2]
            radius_ip_1=0.0.0.0
            port=1813
        """))
        expected = ConfigCheckResult([
        ])
        actual = self._check_network_contention(config)
        self.assert_results_equal(expected, actual)

    def test_multiple_radius(self):
        """ Radius auto1 is good. But 2 3 and 4 collide with each other"""
        config = ConfigProvider(textwrap.dedent("""
            [radius_server_auto1]
            ikey=DITHISISANIKEYFFFFFF
            radius_ip_1=127.0.0.1

            [radius_server_auto2]
            ikey=DITHISISANIKEYFFFFFF
            radius_ip_1=127.0.0.1
            port=1813

            [radius_server_auto3]
            ikey=DITHISISANIKEYFFFFFF
            radius_ip_1=127.0.0.1
            port=1813

            [radius_server_auto4]
            ikey=DITHISISANIKEYFFFFFF
            radius_ip_1=127.0.0.1
            port=1813
        """))
        expected = ConfigCheckResult([
            NetworkContention(section1='radius_server_auto4', section2='radius_server_auto2', interface='*', port=1813, ip1='127.0.0.1', ip2='127.0.0.1'),
            NetworkContention(section1='radius_server_auto4', section2='radius_server_auto3', interface='*', port=1813, ip1='127.0.0.1', ip2='127.0.0.1'),
            NetworkContention(section1='radius_server_auto3', section2='radius_server_auto2', interface='*', port=1813, ip1='127.0.0.1', ip2='127.0.0.1')
        ])
        actual = self._check_network_contention(config)
        self.assert_results_equal(expected, actual)

    def test_mutliple_unique(self):
        """ 1 and 2 collide. 3 and 4 collide"""
        config = ConfigProvider(textwrap.dedent("""
            [radius_server_auto1]
            radius_ip_1=127.0.0.1
            ikey=DITHISISANIKEYFFFFFF
            port=1813

            [radius_server_auto2]
            radius_ip_1=127.0.0.1
            ikey=DITHISISANIKEYFFFFFF
            port=1813

            [radius_server_auto3]
            radius_ip_1=1.2.3.4
            ikey=DITHISISANIKEYFFFFFF
            port=1813

            [radius_server_auto4]
            radius_ip_1=1.2.3.4
            ikey=DITHISISANIKEYFFFFFF
            port=1813
        """))
        expected = ConfigCheckResult([
            NetworkContention(section1='radius_server_auto4', section2='radius_server_auto3', interface='*', port=1813, ip1='1.2.3.4', ip2='1.2.3.4'),
            NetworkContention(section1='radius_server_auto2', section2='radius_server_auto1', interface='*', port=1813, ip1='127.0.0.1', ip2='127.0.0.1'),
        ])
        actual = self._check_network_contention(config)
        self.assert_results_equal(expected, actual)


class TestInterfaceContention(TestConfigBase):

    def test_only_interface_is_different(self):
        """ No conflict because interface differs """
        config = ConfigProvider(textwrap.dedent("""
            [radius_server_auto1]
            radius_ip_1=127.0.0.1
            interface=127.0.0.1
            ikey=DITHISISANIKEYFFFFFF
            port=1813

            [radius_server_auto2]
            radius_ip_1=127.0.0.1
            interface=127.0.0.2
            ikey=DITHISISANIKEYFFFFFF
            port=1813
        """))
        expected = ConfigCheckResult([
        ])
        actual = self._check_network_contention(config)
        self.assert_results_equal(expected, actual)

    def test_interface_all_default_conflict(self):
        """ Default interface (*) conflicts with all others """
        config = ConfigProvider(textwrap.dedent("""
            [radius_server_auto1]
            radius_ip_1=127.0.0.1
            ikey=DITHISISANIKEYFFFFFF

            [radius_server_auto2]
            radius_ip_1=127.0.0.1
            interface=127.0.0.2
            ikey=DITHISISANIKEYFFFFFF
        """))
        expected = ConfigCheckResult([
            NetworkContention(section1='radius_server_auto2', section2='radius_server_auto1', interface='127.0.0.2', port=1812, ip1='127.0.0.1', ip2='127.0.0.1'),
        ])
        actual = self._check_network_contention(config)
        self.assert_results_equal(expected, actual)


class TestIPContention(TestConfigBase):

    def test_ip_subnet_conflict(self):
        """ IP range contains a single IP specified in a different section """
        config = ConfigProvider(textwrap.dedent("""
            [radius_server_auto1]
            radius_ip_1=127.0.0.1
            ikey=DITHISISANIKEYFFFFFF

            [radius_server_auto2]
            radius_ip_1=127.0.0.0/24
            interface=127.0.0.2
            ikey=DITHISISANIKEYFFFFFF
        """))
        expected = ConfigCheckResult([
            NetworkContention(section1='radius_server_auto2', section2='radius_server_auto1', interface='127.0.0.2', port=1812, ip1='127.0.0.0/24', ip2='127.0.0.1'),
        ])
        actual = self._check_network_contention(config)
        self.assert_results_equal(expected, actual)

    def test_ip_range_subset(self):
        """ Test two ip ranges where one is the subset of the other """
        config = ConfigProvider(textwrap.dedent("""
            [radius_server_auto1]
            radius_ip_1=127.0.0.1-127.0.0.255
            ikey=DITHISISANIKEYFFFFFF

            [radius_server_auto2]
            radius_ip_1=127.0.0.100-127.0.0.105
            interface=127.0.0.2
            ikey=DITHISISANIKEYFFFFFF
        """))
        expected = ConfigCheckResult([
            NetworkContention(section1='radius_server_auto2', section2='radius_server_auto1', interface='127.0.0.2', port=1812, ip1='127.0.0.100-127.0.0.105', ip2='127.0.0.1-127.0.0.255'),
        ])
        actual = self._check_network_contention(config)
        self.assert_results_equal(expected, actual)

    def test_ip_range_shared_boundary(self):
        """ test ip ranges which overlap only on the start/end ip """
        config = ConfigProvider(textwrap.dedent("""
            [radius_server_auto1]
            radius_ip_1=127.0.0.1-127.0.0.100
            ikey=DITHISISANIKEYFFFFFF

            [radius_server_auto2]
            radius_ip_1=127.0.0.100-127.0.0.200
            ikey=DITHISISANIKEYFFFFFF
        """))
        expected = ConfigCheckResult([
            NetworkContention(section1='radius_server_auto2', section2='radius_server_auto1', interface='*', port=1812, ip1='127.0.0.100-127.0.0.200', ip2='127.0.0.1-127.0.0.100'),
        ])
        actual = self._check_network_contention(config)
        self.assert_results_equal(expected, actual)

    def test_ip_range_overlap(self):
        """ test ip ranges where the intersection of the ranges is non-empty """
        config = ConfigProvider(textwrap.dedent("""
            [radius_server_auto1]
            radius_ip_1=127.0.0.1-127.0.0.050
            ikey=DITHISISANIKEYFFFFFF

            [radius_server_auto2]
            radius_ip_1=127.0.0.010-127.0.0.200
            ikey=DITHISISANIKEYFFFFFF
        """))
        expected = ConfigCheckResult([
            NetworkContention(section1='radius_server_auto2', section2='radius_server_auto1', interface='*', port=1812, ip1='127.0.0.010-127.0.0.200', ip2='127.0.0.1-127.0.0.050'),
        ])
        actual = self._check_network_contention(config)
        self.assert_results_equal(expected, actual)

    def test_ip_range_overlap_with_different_interfaces(self):
        """ test ip ranges where the intersection of the ranges is non-empty """
        config = ConfigProvider(textwrap.dedent("""
            [radius_server_auto1]
            radius_ip_1=127.0.0.1-127.0.0.050
            interface=127.0.0.1
            ikey=DITHISISANIKEYFFFFFF

            [radius_server_auto2]
            radius_ip_1=127.0.0.010-127.0.0.200
            interface=127.0.0.2
            ikey=DITHISISANIKEYFFFFFF
        """))
        expected = ConfigCheckResult([
        ])
        actual = self._check_network_contention(config)
        self.assert_results_equal(expected, actual)

    def test_multiple_ips_one_range_conflict(self):
        """ Test shared ip in range when multipe radius_ips are specified """
        config = ConfigProvider(textwrap.dedent("""
            [radius_server_auto1]
            radius_ip_1=127.0.0.1-127.0.0.050
            radius_ip_2=127.0.0.51-127.0.0.100
            interface=127.0.0.1
            ikey=DITHISISANIKEYFFFFFF

            [radius_server_auto2]
            radius_ip_1=127.0.1.010-127.0.1.200
            radius_ip_2=127.0.0.100-127.0.0.200
            ikey=DITHISISANIKEYFFFFFF
        """))
        expected = ConfigCheckResult([
            NetworkContention(section1='radius_server_auto2', section2='radius_server_auto1', interface='127.0.0.1', port=1812, ip1='127.0.0.100-127.0.0.200', ip2='127.0.0.51-127.0.0.100'),
        ])
        actual = self._check_network_contention(config)
        self.assert_results_equal(expected, actual)

    def test_multiple_ips_one_subset_conflict(self):
        """ Test multipe ips in subset """
        config = ConfigProvider(textwrap.dedent("""
            [radius_server_auto1]
            radius_ip_1=127.0.0.1-127.0.0.050
            radius_ip_2=127.0.0.51-127.0.0.100
            radius_ip_3=127.0.1.0/24
            ikey=DITHISISANIKEYFFFFFF

            [radius_server_auto2]
            radius_ip_1=127.0.1.010-127.0.1.200
            radius_ip_2=127.0.1.100-127.0.1.200
            ikey=DITHISISANIKEYFFFFFF
        """))
        expected = ConfigCheckResult([
            NetworkContention(section1='radius_server_auto2', section2='radius_server_auto1', interface='*', port=1812, ip1='127.0.1.010-127.0.1.200', ip2='127.0.1.0/24'),
            NetworkContention(section1='radius_server_auto2', section2='radius_server_auto1', interface='*', port=1812, ip1='127.0.1.100-127.0.1.200', ip2='127.0.1.0/24'),
        ])
        actual = self._check_network_contention(config)
        self.assert_results_equal(expected, actual)

    def test_multiple_ips_one_single_conflict(self):
        """ Test shared ip in range when multipe radius_ips are specified """
        config = ConfigProvider(textwrap.dedent("""
            [radius_server_auto1]
            radius_ip_1=127.0.0.1-127.0.0.050
            radius_ip_2=127.0.0.51-127.0.0.100
            radius_ip_3=127.0.1.0/24
            ikey=DITHISISANIKEYFFFFFF

            [radius_server_auto2]
            radius_ip_1=127.0.2.010-127.0.2.200
            radius_ip_2=127.0.2.100-127.0.2.200
            radius_ip_3=127.0.1.100
            ikey=DITHISISANIKEYFFFFFF
        """))
        expected = ConfigCheckResult([
            NetworkContention(section1='radius_server_auto2', section2='radius_server_auto1', interface='*', port=1812, ip1='127.0.1.100', ip2='127.0.1.0/24')
        ])
        actual = self._check_network_contention(config)
        self.assert_results_equal(expected, actual)


class TestPortContentionLDAP(TestCase):
    """LDAP Sections cannot share ports like radius so we must make sure that interface:port
    pairs are unique across all the sections and even between the same section because LDAP server
    sections can listen on two ports"""
    def test_no_ldap(self):
        config = ConfigProvider(textwrap.dedent("""
            [ad_client]
            fake=fake

            [radius_server_auto]
            fake=fake
        """))
        expected = []
        actual = cross_config.check_port_contention_for_ldap(config)

        self.assertUnorderedUnhashableListsEqual(expected, actual)

    def test_one_ldap(self):
        config = ConfigProvider(textwrap.dedent("""
            [ldap_server_auto]
            port=389
            ssl_port=633
            ssl_key_path=/tmp/fake.key
            ssl_cert_path=/tmp/fake.crt
        """))
        expected = []
        actual = cross_config.check_port_contention_for_ldap(config)

        self.assertUnorderedUnhashableListsEqual(expected, actual)

    def test_ldap_specify_ports_no_collission(self):
        config = ConfigProvider(textwrap.dedent("""
            [ldap_server_auto1]
            port=1234
            ssl_port=5678
            ssl_key_path=/tmp/fake.key
            ssl_cert_path=/tmp/fake.crt

            [ldap_server_auto2]
            port=1111
            ssl_port=2222
            ssl_key_path=/tmp/fake.key
            ssl_cert_path=/tmp/fake.crt
        """))
        expected = []
        actual = cross_config.check_port_contention_for_ldap(config)

        self.assertUnorderedUnhashableListsEqual(expected, actual)

    def test_collide_default_ssl(self):
        config = ConfigProvider(textwrap.dedent("""
            [ldap_server_auto1]
            port=1234
            ssl_key_path=/tmp/fake.key
            ssl_cert_path=/tmp/fake.crt

            [ldap_server_auto2]
            port=1111
            ssl_key_path=/tmp/fake.key
            ssl_cert_path=/tmp/fake.crt
        """))
        expected = [
            PortCollision(sections=['ldap_server_auto1', 'ldap_server_auto2'], port=636, interface='*')
        ]
        actual = cross_config.check_port_contention_for_ldap(config)

        self.assertUnorderedUnhashableListsEqual(expected, actual)

    def test_collide_default_clear(self):
        config = ConfigProvider(textwrap.dedent("""
            [ldap_server_auto1]
            port=3890

            [ldap_server_auto2]
            port=3890
        """))
        expected = [
            PortCollision(sections=['ldap_server_auto1', 'ldap_server_auto2'], port=3890, interface='*')
        ]
        actual = cross_config.check_port_contention_for_ldap(config)

        self.assertUnorderedUnhashableListsEqual(expected, actual)

    def test_three_sections_collide(self):
        config = ConfigProvider(textwrap.dedent("""
            [ldap_server_auto1]
            port=3890

            [ldap_server_auto2]
            port=3890

            [ldap_server_auto3]
            port=3890
        """))
        expected = [
            PortCollision(sections=['ldap_server_auto1', 'ldap_server_auto2'], port=3890, interface='*'),
            PortCollision(sections=['ldap_server_auto1', 'ldap_server_auto3'], port=3890, interface='*'),
            PortCollision(sections=['ldap_server_auto2', 'ldap_server_auto3'], port=3890, interface='*'),
        ]
        actual = cross_config.check_port_contention_for_ldap(config)

        self.assertUnorderedUnhashableListsEqual(expected, actual)

    def test_two_sets_of_collisions(self):
        """ 1 and 2 collide. 3 and 4 collide"""
        config = ConfigProvider(textwrap.dedent("""
            [ldap_server_auto1]
            port=1234

            [ldap_server_auto2]
            port=1234

            [ldap_server_auto3]
            port=3333

            [ldap_server_auto4]
            port=4444
            ssl_port=3333
            ssl_key_path=/tmp/fake.key
            ssl_cert_path=/tmp/fake.crt
        """))
        expected = [
            PortCollision(sections=['ldap_server_auto1', 'ldap_server_auto2'], port=1234, interface='*'),
            PortCollision(sections=['ldap_server_auto3', 'ldap_server_auto4'], port=3333, interface='*'),
        ]
        actual = cross_config.check_port_contention_for_ldap(config)

        self.assertUnorderedUnhashableListsEqual(expected, actual)

    def test_section_collides_with_self(self):
        config = ConfigProvider(textwrap.dedent("""
            [ldap_server_auto1]
            port=1234
            ssl_port=1234
            ssl_key_path=/tmp/fake.key
            ssl_cert_path=/tmp/fake.crt
        """))
        expected = [
            PortCollision(sections=['ldap_server_auto1', 'ldap_server_auto1'], port=1234, interface='*'),
        ]
        actual = cross_config.check_port_contention_for_ldap(config)

        self.assertUnorderedUnhashableListsEqual(expected, actual)

    def test_interface_prevents_collision(self):
        config = ConfigProvider(textwrap.dedent("""
            [ldap_server_auto1]
            interface=127.0.0.1
            port=1234
            ssl_port=5678
            ssl_key_path=/tmp/fake.key
            ssl_cert_path=/tmp/fake.crt

            [ldap_server_auto2]
            interface=127.0.0.2
            port=1234
            ssl_port=5678
            ssl_key_path=/tmp/fake.key
            ssl_cert_path=/tmp/fake.crt
        """))
        expected = []
        actual = cross_config.check_port_contention_for_ldap(config)

        self.assertUnorderedUnhashableListsEqual(expected, actual)

    def test_overlapping_interface_and_port(self):
        config = ConfigProvider(textwrap.dedent("""
            [ldap_server_auto1]
            port=1234

            [ldap_server_auto2]
            interface=127.0.0.1
            port=1234
        """))
        expected = [
            PortCollision(sections=['ldap_server_auto1', 'ldap_server_auto2'], port=1234, interface='127.0.0.1'),
        ]
        actual = cross_config.check_port_contention_for_ldap(config)

        self.assertUnorderedUnhashableListsEqual(expected, actual)


class TestClientServerMapping(TestConfigBase):

    def _assert_success(self, config):
        result = cross_config.check_cross_config(config)
        self.assert_results_equal(result, ConfigCheckResult([]))

    def _assert_fail(self, config, client_name, section):
        expected_result = ConfigCheckResult([
            InvalidClientServerMapping(
                client_name=client_name, section=section)
        ])
        result = cross_config.check_cross_config(config)
        self.assert_results_equal(result, expected_result)

    def test_main_section_client_invalid(self):
        '''Invalid client in the main section fails'''
        config = ConfigProvider(textwrap.dedent("""
            [main]
            client=some_client_setting

            [radius_server_auto]
            ikey=there_is_no_client_config
        """))
        self._assert_fail(config, 'some_client_setting', 'radius_server_auto')

    def test_main_section_client_success(self):
        '''Client config in main section succeeds'''
        config = ConfigProvider(textwrap.dedent("""
            [main]
            client=radius_client

            [radius_server_auto]
            ikey=this_is_successful

            [radius_client]
            host=blah
        """))
        self._assert_success(config)

    def test_no_client_config_with_server_config_fail(self):
        '''A server config with no valid client config fails'''
        config = ConfigProvider(textwrap.dedent("""
            [radius_server_auto]
            ikey=there_is_no_client_config
        """))
        self._assert_fail(config, None, 'radius_server_auto')

    def test_no_client_config_with_no_server_config_success(self):
        '''No server config and no client config succeeds'''
        config = ConfigProvider(textwrap.dedent("""
            [main]
            debug=True
        """))
        self._assert_success(config)

    def test_missing_client_config_with_one_client_section_success(self):
        '''No client config in server section with only one client section suceeeds'''
        config = ConfigProvider(textwrap.dedent("""
            [radius_server_auto]
            ikey=foo

            [ad_client]
            host=blah
        """))
        self._assert_success(config)

    def test_valid_client_config_in_section_success(self):
        '''Client config in server section with matching client section succeeds'''
        config = ConfigProvider(textwrap.dedent("""
            [radius_server_auto]
            ikey=foo
            client=ad_client2

            [ad_client]
            host=blah

            [ad_client2]
            host=bar
        """))
        self._assert_success(config)

    def test_invalid_client_config_in_section_fail(self):
        '''Client config in server section without matching client section fails'''
        config = ConfigProvider(textwrap.dedent("""
            [radius_server_auto]
            ikey=foo
            client=ad_client
        """))
        self._assert_fail(config, 'ad_client', 'radius_server_auto')

    def test_optional_duo_only_config(self):
        '''radius_server_duo_only does not require a client config'''
        config = ConfigProvider(textwrap.dedent("""
            [radius_server_duo_only]
            ikey=foo
        """))
        self._assert_success(config)

    def test_ldap_server_with_single_ad_client_sucess(self):
        '''Maps successfully when only one ad_client section'''
        config = ConfigProvider(textwrap.dedent("""
            [ldap_server_auto]
            ikey=foo

            [ad_client]
            host=blah
        """))
        self._assert_success(config)

    def test_ldap_server_with_multiple_ad_client_fails(self):
        '''Mapping fails when multiple ad_client sections without a specific client selected'''
        config = ConfigProvider(textwrap.dedent("""
            [ldap_server_auto]
            ikey=foo

            [ad_client]
            host=blah

            [ad_client2]
            host=bar
        """))
        self._assert_fail(config, None, 'ldap_server_auto')

    def test_ldap_server_valid_client_config_success(self):
        '''Valid config of ldap_server succeeds'''
        config = ConfigProvider(textwrap.dedent("""
            [ldap_server_auto]
            ikey=foo
            client=ad_client2

            [ad_client]
            host=blah

            [ad_client2]
            host=bar
        """))
        self._assert_success(config)

    def test_ldap_server_non_ad_client_fail(self):
        '''ldap server configured with non-ad client fails'''
        config = ConfigProvider(textwrap.dedent("""
            [ldap_server_auto]
            ikey=foo
            client=radius_client

            [ad_client]
            host=blah

            [radius_client]
            host=bar
        """))
        self._assert_fail(config, 'radius_client', 'ldap_server_auto')


class TestPassThroughAttrs(TestConfigBase):
    def test_ad_client_with_pass_through(self):
        '''radius server configured with ad client and pass_through_attr_names fails'''
        config = ConfigProvider(textwrap.dedent("""
            [radius_server_auto]
            ikey=foo
            client=ad_client
            pass_through_attr_names=name,birthday

            [ad_client]
            host=blah

            [radius_client]
            host=bar
        """))
        expected = ConfigCheckResult([
            KeyInvalidWithClient(key='pass_through_attr_names', client='ad_client')
        ])
        actual = cross_config.check_cross_config(config)
        self.assert_results_equal(actual, expected)

    def test_ad_client_without_pass_through(self):
        '''radius server configured with ad client but without pass_through_attr_names succeeds'''
        config = ConfigProvider(textwrap.dedent("""
            [radius_server_auto]
            ikey=foo
            client=ad_client

            [ad_client]
            host=blah

            [radius_client]
            host=bar
        """))
        expected = ConfigCheckResult([])
        actual = cross_config.check_cross_config(config)
        self.assert_results_equal(actual, expected)

    def test_radius_client_with_pass_through(self):
        '''radius server configured with radius client and with pass_through_attr_names succeeds'''
        config = ConfigProvider(textwrap.dedent("""
            [radius_server_auto]
            ikey=foo
            client=radius_client
            pass_through_attr_names=name,birthday

            [ad_client]
            host=blah

            [radius_client]
            host=bar
        """))
        expected = ConfigCheckResult([])
        actual = cross_config.check_cross_config(config)
        self.assert_results_equal(actual, expected)

    def test_radius_client_without_pass_through(self):
        '''radius server configured with radius client but without pass_through_attr_names succeeds'''
        config = ConfigProvider(textwrap.dedent("""
            [radius_server_auto]
            ikey=foo
            client=radius_client

            [ad_client]
            host=blah

            [radius_client]
            host=bar
        """))
        expected = ConfigCheckResult([])
        actual = cross_config.check_cross_config(config)
        self.assert_results_equal(actual, expected)

    def test_no_client_with_pass_through(self):
        ''' make sure we don't fail with an empty client'''
        config = ConfigProvider(textwrap.dedent("""
            [radius_server_auto]
            ikey=foo
            pass_through_attr_names=name,birthday

            [ad_client]
            host=blah

            [radius_client]
            host=bar
        """))
        expected = ConfigCheckResult([
            InvalidClientServerMapping(client_name=None, section='radius_server_auto')
        ])
        actual = cross_config.check_cross_config(config)
        self.assert_results_equal(actual, expected)


class TestMinimumTlsVersionWithFipsMode(TestConfigBase):
    def test_minimum_tls_version_with_fips_mode_enabled(self):
        config = ConfigProvider(textwrap.dedent("""
            [main]
            fips_mode=true

            [ad_client]
            host=blah

            [ldap_server_auto]
            ikey=DITHISISANIKEYFFFFFF
            skey=skeythatislongenoughtopassvalidationffff
            api_host=some.api.host
            port=8080

            [ldap_server_auto2]
            ikey=DITHISISANIKEYFFFFFF
            skey=skeythatislongenoughtopassvalidationffff
            api_host=some.api.host
            minimum_tls_version=tls1.0
            port=9090
        """))

        expected = ConfigCheckResult([
            IneffectiveConfig(key='minimum_tls_version', section='ldap_server_auto2',
                              condition='fips_mode is enabled')
        ])
        actual = cross_config.check_cross_config(config)
        self.assert_results_equal(actual, expected)

    def test_minimum_tls_version_with_fips_mode_disabled(self):
        config = ConfigProvider(textwrap.dedent("""
            [main]
            fips_mode=false

            [ad_client]
            host=blah

            [ldap_server_auto]
            ikey=DITHISISANIKEYFFFFFF
            skey=skeythatislongenoughtopassvalidationffff
            api_host=some.api.host
            port=8080

            [ldap_server_auto2]
            ikey=DITHISISANIKEYFFFFFF
            skey=skeythatislongenoughtopassvalidationffff
            api_host=some.api.host
            minimum_tls_version=tls1.0
            port=9090
        """))

        expected = ConfigCheckResult([])
        actual = cross_config.check_cross_config(config)
        self.assert_results_equal(actual, expected)

    def test_no_main_section(self):
        config = ConfigProvider(textwrap.dedent("""
            [ad_client]
            host=blah

            [ldap_server_auto]
            ikey=DITHISISANIKEYFFFFFF
            skey=skeythatislongenoughtopassvalidationffff
            api_host=some.api.host
            port=8080

            [ldap_server_auto2]
            ikey=DITHISISANIKEYFFFFFF
            skey=skeythatislongenoughtopassvalidationffff
            api_host=some.api.host
            minimum_tls_version=tls1.0
            port=9090
        """))

        expected = ConfigCheckResult([])
        actual = cross_config.check_cross_config(config)
        self.assert_results_equal(actual, expected)

    def test_no_minimum_tls_version(self):
        config = ConfigProvider(textwrap.dedent("""
            [main]
            fips_mode=true

            [ad_client]
            host=blah

            [ldap_server_auto]
            ikey=DITHISISANIKEYFFFFFF
            skey=skeythatislongenoughtopassvalidationffff
            api_host=some.api.host
        """))

        expected = ConfigCheckResult([])
        actual = cross_config.check_cross_config(config)
        self.assert_results_equal(actual, expected)

    def test_radius_server_duo_only_client(self):
        config = ConfigProvider(textwrap.dedent("""
            [radius_server_duo_only]
            ikey=DITHISISANIKEYFFFFFF
            skey=skeythatislongenoughtopassvalidationffff
            api_host=some.api.host
            radius_ip_1=127.0.0.1
            radius_secret_1=testest
            client=something
        """))

        expected = ConfigCheckResult([
            IneffectiveConfig(key='client', section='radius_server_duo_only',
                              condition='in this type of section')
        ])
        actual = cross_config.check_cross_config(config)
        self.assert_results_equal(actual, expected)


if __name__ == '__main__':
    unittest.main()
