#
# Copyright (c) 2018 Duo Security
# All Rights Reserved
#


import unittest

from duoauthproxy.lib.validation.config.check import ldap_server_auto
from duoauthproxy.lib.config_provider import ConfigDict
from duoauthproxy.lib.validation.connectivity.connectivity_results import ConfigCheckResult
from duoauthproxy.lib.validation.config import config_results

from ....base import MockConfigTestToolbox


class TestCheckLdapServerAutoConfig(unittest.TestCase):

    def test_basic_config_succeeds(self):
        expected_result = ConfigCheckResult([])

        BASIC_CONFIG = ConfigDict({
            'ikey': 'DI000000000000000000',
            'skey_protected': 'totallyaprotectedskey',
            'api_host': 'api-12345678.test.duosecurity.com',
            'client': 'ad_client2',
        })
        result = ldap_server_auto.check_ldap_server_auto(BASIC_CONFIG, toolbox=MockConfigTestToolbox())
        self.assertEqual(expected_result, result, 'Expected {0} should equal actual {1}'.format(expected_result, result))

    def test_big_config_succeeds(self):
        expected_result = ConfigCheckResult([])

        BIG_CONFIG = ConfigDict({
            'ikey': 'DI000000000000000000',
            'skey': 'thisisntarealskeybutitcanplaysoneonthetv',
            'api_host': 'api-12345678.test.duosecurity.com',
            'api_timeout': '17',
            'client': 'ad_client',
            'factors': 'auto,phone',
            'failmode': 'safe',
            'port': '3389',
            'interface': '127.0.0.10',
            'ssl_port': '6636',
            'ssl_key_path': 'this/is/a/good.path',
            'ssl_cert_path': 'another/good.path',
            'exempt_primary_bind': 'true',
            'exempt_ou_1': 'ou=exempt,dc=example,dc=com',
            'exempt_ou_2': 'ou=bypassed,dc=example,dc=com',
            'delimiter': 'q',
            'allow_concat': 'false',
            'allow_searches_after_bind': 'true',
            'allow_unlimited_binds': 'false',
            'cipher_list': 'DEFAULT',
            'network_timeout': 60,
            'idle_timeout': 60,
        })
        result = ldap_server_auto.check_ldap_server_auto(BIG_CONFIG, toolbox=MockConfigTestToolbox())
        self.assertEqual(expected_result, result, 'Expected {0} should equal actual {1}'.format(expected_result, result))

    def test_missing_required_keys_failure(self):
        expected_result = ConfigCheckResult([
            config_results.MissingKey(key='ikey'),
            config_results.MissingKey(key='skey'),
        ])

        MISSING_KEYS_CONFIG = ConfigDict({
            'api_host': 'api-12345678.test.duosecurity.com',
            'client': 'ad_client2',
        })
        result = ldap_server_auto.check_ldap_server_auto(MISSING_KEYS_CONFIG, toolbox=MockConfigTestToolbox())
        self.assertEqual(expected_result, result, 'Expected {0} should equal actual {1}'.format(expected_result, result))

    def test_bad_values_failure(self):
        BAD_VALUES_CONFIG = ConfigDict({
            'ikey': 'thisisntarealskeybutitcanplaysoneonthetv',
            'skey': 'DI000000000000000000',
            'api_host': '',
            'api_timeout': '',
            'client': '',
            'factors': 'auto,carrierpigeon',
            'failmode': 'open',
            'port': '333389',
            'interface': '327.0.0.10',
            'ssl_port': '66636',
            'ssl_key_path': 'this/is/a/bad.path',
            'ssl_cert_path': 'another/bad.path',
            'exempt_primary_bind': 'yes',
            'exempt_ou_1': 'this_is_not_a_dn_format',
            'exempt_ou_2': 'wrong,wrong,wrong',
            'delimiter': '',
            'allow_concat': 'maybe',
            'allow_searches_after_bind': 'allow',
            'allow_unlimited_binds': 'fudgiecakes',
        })

        expected_result = ConfigCheckResult(
            [config_results.InvalidValue(key=key, value=value) for key, value in BAD_VALUES_CONFIG.iteritems()],
        )

        result = ldap_server_auto.check_ldap_server_auto(BAD_VALUES_CONFIG, toolbox=MockConfigTestToolbox(
            fail_test_file=True))
        self.assertEqual(expected_result, result, 'Expected {0} should equal actual {1}'.format(expected_result, result))

    def test_unexpected_keys_failure(self):
        expected_result = ConfigCheckResult([
            config_results.UnexpectedKey(key='exempt_ou1'),
            config_results.UnexpectedKey(key='fail_mode'),
            config_results.UnexpectedKey(key='concat'),
        ])

        UNEXPECTED_KEYS_CONFIG = ConfigDict({
            'ikey': 'DI000000000000000000',
            'skey_protected': 'totallyaprotectedskey',
            'api_host': 'api-12345678.test.duosecurity.com',
            'client': 'ad_client2',
            'exempt_ou1': 'ou=exempt,dc=example,dc=com',
            'fail_mode': 'secure',
            'concat': 'true',
        })
        result = ldap_server_auto.check_ldap_server_auto(UNEXPECTED_KEYS_CONFIG, toolbox=MockConfigTestToolbox())
        self.assertEqual(expected_result, result, 'Expected {0} should equal actual {1}'.format(expected_result, result))

    def test_ssl_port_requires_key_and_cert_failure(self):
        expected_result = ConfigCheckResult([
            config_results.MissingKey(level=config_results.ConfigResultLevel.Warning, key='ssl_key_path'),
            config_results.MissingKey(level=config_results.ConfigResultLevel.Warning, key='ssl_cert_path'),
        ])

        BAD_SSL_CONFIG = ConfigDict({
            'ikey': 'DI000000000000000000',
            'skey': 'thisisntarealskeybutitcanplaysoneonthetv',
            'api_host': 'api-12345678.test.duosecurity.com',
            'client': 'ad_client',
            'ssl_port': '6636',
        })
        result = ldap_server_auto.check_ldap_server_auto(BAD_SSL_CONFIG, toolbox=MockConfigTestToolbox())
        self.assertEqual(expected_result, result, 'Expected {0} should equal actual {1}'.format(expected_result, result))

    def test_ssl_key_without_cert_failure(self):
        expected_result = ConfigCheckResult([
            config_results.MissingKey(key='ssl_cert_path'),
        ])

        BAD_SSL_CONFIG = ConfigDict({
            'ikey': 'DI000000000000000000',
            'skey': 'thisisntarealskeybutitcanplaysoneonthetv',
            'api_host': 'api-12345678.test.duosecurity.com',
            'client': 'ad_client',
            'ssl_key_path': 'totally/a/key.file',
        })
        result = ldap_server_auto.check_ldap_server_auto(BAD_SSL_CONFIG, toolbox=MockConfigTestToolbox())
        self.assertEqual(expected_result, result, 'Expected {0} should equal actual {1}'.format(expected_result, result))

    def test_ssl_cert_without_key_failure(self):
        expected_result = ConfigCheckResult([
            config_results.MissingKey(key='ssl_key_path'),
        ])

        BAD_SSL_CONFIG = ConfigDict({
            'ikey': 'DI000000000000000000',
            'skey': 'thisisntarealskeybutitcanplaysoneonthetv',
            'api_host': 'api-12345678.test.duosecurity.com',
            'client': 'ad_client',
            'ssl_cert_path': 'totally/a/cert/file',
        })
        result = ldap_server_auto.check_ldap_server_auto(BAD_SSL_CONFIG, toolbox=MockConfigTestToolbox())
        self.assertEqual(expected_result, result, 'Expected {0} should equal actual {1}'.format(expected_result, result))

    def test_both_unlimited_binds_and_exempt_primary_failure(self):
        expected_result = ConfigCheckResult([
            config_results.IncompatibleKeys(key1='allow_unlimited_binds', key2='exempt_primary_bind')
        ])

        BAD_BIND_FLAGS_CONFIG = ConfigDict({
            'ikey': 'DI000000000000000000',
            'skey': 'thisisntarealskeybutitcanplaysoneonthetv',
            'api_host': 'api-12345678.test.duosecurity.com',
            'client': 'ad_client',
            'exempt_primary_bind': 'true',
            'allow_unlimited_binds': 'true',
        })
        result = ldap_server_auto.check_ldap_server_auto(BAD_BIND_FLAGS_CONFIG, toolbox=MockConfigTestToolbox())
        self.assertEqual(expected_result, result, 'Expected {0} should equal actual {1}'.format(expected_result, result))

    def test_bad_cipher_list(self):
        config = ConfigDict({
            'ikey': 'DI000000000000000000',
            'skey_protected': 'totallyaprotectedskey',
            'api_host': 'api-12345678.test.duosecurity.com',
            'client': 'ad_client2',
            'cipher_list': 'AES-bad',
        })

        result = ldap_server_auto.check_ldap_server_auto(config, toolbox=MockConfigTestToolbox())
        expected_result = ConfigCheckResult([
            config_results.InvalidValue(key='cipher_list', value='AES-bad'),
        ])
        self.assertEqual(expected_result, result, 'Expected {0} should equal actual {1}'.format(expected_result, result))

    def test_good_cipher_list(self):
        config = ConfigDict({
            'ikey': 'DI000000000000000000',
            'skey_protected': 'totallyaprotectedskey',
            'api_host': 'api-12345678.test.duosecurity.com',
            'client': 'ad_client2',
            'cipher_list': 'TLSv1.2:kRSA:!eNULL:!aNULL'
        })

        result = ldap_server_auto.check_ldap_server_auto(config, toolbox=MockConfigTestToolbox())
        expected_result = ConfigCheckResult([
        ])
        self.assertEqual(expected_result, result, 'Expected {0} should equal actual {1}'.format(expected_result, result))

    def test_good_cipher_list2(self):
        config = ConfigDict({
            'ikey': 'DI000000000000000000',
            'skey_protected': 'totallyaprotectedskey',
            'api_host': 'api-12345678.test.duosecurity.com',
            'client': 'ad_client2',
            'cipher_list': 'RSA+FIPS:!TLSv1.2',
        })

        result = ldap_server_auto.check_ldap_server_auto(config, toolbox=MockConfigTestToolbox())
        expected_result = ConfigCheckResult([
        ])
        self.assertEqual(expected_result, result, 'Expected {0} should equal actual {1}'.format(expected_result, result))

    def test_good_minimum_tls_version(self):
        config = ConfigDict({
            'ikey': 'DI000000000000000000',
            'skey_protected': 'totallyaprotectedskey',
            'api_host': 'api-12345678.test.duosecurity.com',
            'client': 'ad_client2',
            'minimum_tls_version': 'tls1.0',
        })

        result = ldap_server_auto.check_ldap_server_auto(config, toolbox=MockConfigTestToolbox())
        expected_result = ConfigCheckResult([
        ])
        self.assertEqual(expected_result, result, 'Expected {0} should equal actual {1}'.format(expected_result, result))

    def test_bad_minimum_tls_version(self):
        config = ConfigDict({
            'ikey': 'DI000000000000000000',
            'skey_protected': 'totallyaprotectedskey',
            'api_host': 'api-12345678.test.duosecurity.com',
            'client': 'ad_client2',
            'minimum_tls_version': 'ssl2',
        })

        result = ldap_server_auto.check_ldap_server_auto(config, toolbox=MockConfigTestToolbox())
        expected_result = ConfigCheckResult([
            config_results.InvalidValue(key='minimum_tls_version', value='ssl2'),
        ])
        self.assertEqual(expected_result, result, 'Expected {0} should equal actual {1}'.format(expected_result, result))

    def test_bad_minimum_tls_version2(self):
        config = ConfigDict({
            'ikey': 'DI000000000000000000',
            'skey_protected': 'totallyaprotectedskey',
            'api_host': 'api-12345678.test.duosecurity.com',
            'client': 'ad_client2',
            'minimum_tls_version': 'ssl',
        })

        result = ldap_server_auto.check_ldap_server_auto(config, toolbox=MockConfigTestToolbox())
        expected_result = ConfigCheckResult([
            config_results.InvalidValue(key='minimum_tls_version', value='ssl'),
        ])
        self.assertEqual(expected_result, result, 'Expected {0} should equal actual {1}'.format(expected_result, result))


if __name__ == '__main__':
    unittest.main()
