from duoauthproxy.lib.validation.config.check import main
from duoauthproxy.lib.config_provider import ConfigDict
from duoauthproxy.lib.validation.config.config_results import (
    MissingKey, UnexpectedKey, InvalidValue)
from duoauthproxy.lib.validation.connectivity.connectivity_results import (
    ConfigCheckResult
)

from ....base import TestCase, MockConfigTestToolbox


class TestCheckMainConfig(TestCase):
    """
    Tests the [main] config section tester module
    """

    def setUp(self):
        super(TestCheckMainConfig, self).setUp()
        self.BASIC_CONFIG_INPUT = ConfigDict({})

    def test_basic_config(self):
        result = main.check_main(self.BASIC_CONFIG_INPUT)
        self.assertEqual(result, ConfigCheckResult([]))

    def test_full_good_config(self):
        config = ConfigDict({
            'debug': 'true',
            'log_dir': '/tmp',
            'log_auth_events': 'true',
            'log_max_files': '100',
            'log_max_size': '200000',
            'log_file': 'true',
            'log_stdout': 'true',
            'log_syslog': 'false',
            'syslog_facility': 'AUTHPROXY_SVC',
            'http_ca_certs_file': '/tmp/other.crts',
            'interface': '10.0.0.1',
            'http_proxy_host': 'mylocalproxy.com',
            'http_proxy_port': '8080',
            'fips_mode': 'true',
            'client': 'ad_client',
            'server': 'ldap_server_auto',
        })

        expected_result = ConfigCheckResult([])
        actual_result = main.check_main(config, MockConfigTestToolbox())
        self.assert_results_equal(expected_result, actual_result)

    def test_debug(self):
        invalid_debug = 'yes'
        config = ConfigDict({
            'debug': invalid_debug,
        })
        expected_result = ConfigCheckResult([
            InvalidValue(key='debug', value=invalid_debug)
        ])

        actual_result = main.check_main(config, MockConfigTestToolbox())
        self.assert_results_equal(expected_result, actual_result)

    def test_log_dir(self):
        invalid_log_dir = ''
        config = ConfigDict({
            'log_dir': invalid_log_dir,
        })
        expected_result = ConfigCheckResult([
            InvalidValue(key='log_dir', value=invalid_log_dir)
        ])

        actual_result = main.check_main(config, MockConfigTestToolbox(fail_test_directory=True))
        self.assert_results_equal(expected_result, actual_result)

    def test_log_auth_events(self):
        invalid_log_auth_events = 'no'
        config = ConfigDict({
            'log_auth_events': invalid_log_auth_events,
        })
        expected_result = ConfigCheckResult([
            InvalidValue(key='log_auth_events', value=invalid_log_auth_events)
        ])

        actual_result = main.check_main(config, MockConfigTestToolbox())
        self.assert_results_equal(expected_result, actual_result)

    def test_log_max_files(self):
        invalid_log_max_files = 'two'
        config = ConfigDict({
            'log_max_files': invalid_log_max_files,
        })
        expected_result = ConfigCheckResult([
            InvalidValue(key='log_max_files', value=invalid_log_max_files)
        ])

        actual_result = main.check_main(config, MockConfigTestToolbox())
        self.assert_results_equal(expected_result, actual_result)

    def test_log_max_size(self):
        invalid_log_max_size = '-1'
        config = ConfigDict({
            'log_max_size': invalid_log_max_size,
        })
        expected_result = ConfigCheckResult([
            InvalidValue(key='log_max_size', value=invalid_log_max_size)
        ])

        actual_result = main.check_main(config, MockConfigTestToolbox())
        self.assert_results_equal(expected_result, actual_result)

    def test_log_output_locations(self):
        invalid_log_file = 'no'
        invalid_log_stdout = 'yes'
        invalid_syslog = ''
        config = ConfigDict({
            'log_file': invalid_log_file,
            'log_stdout': invalid_log_stdout,
            'log_syslog': invalid_syslog,
        })
        expected_result = ConfigCheckResult([
            InvalidValue(key='log_file', value=invalid_log_file),
            InvalidValue(key='log_stdout', value=invalid_log_stdout),
            InvalidValue(key='log_syslog', value=invalid_syslog),
        ])

        actual_result = main.check_main(config, MockConfigTestToolbox())
        self.assert_results_equal(expected_result, actual_result)

    def test_syslog_facility(self):
        invalid_syslog_facility = ''
        config = ConfigDict({
            'syslog_facility': invalid_syslog_facility,
        })
        expected_result = ConfigCheckResult([
            InvalidValue(key='syslog_facility', value=invalid_syslog_facility)
        ])

        actual_result = main.check_main(config, MockConfigTestToolbox())
        self.assert_results_equal(expected_result, actual_result)

    def test_ca_certs_file(self):
        invalid_http_ca_certs_file = ''
        config = ConfigDict({
            'http_ca_certs_file': invalid_http_ca_certs_file,
        })
        expected_result = ConfigCheckResult([
            InvalidValue(key='http_ca_certs_file', value=invalid_http_ca_certs_file)
        ])

        actual_result = main.check_main(config, MockConfigTestToolbox(fail_test_file=True))
        self.assert_results_equal(expected_result, actual_result)

    def test_interface(self):
        invalid_interface = '300.0.0.1'
        config = ConfigDict({
            'interface': invalid_interface,
        })
        expected_result = ConfigCheckResult([
            InvalidValue(key='interface', value=invalid_interface)
        ])

        actual_result = main.check_main(config, MockConfigTestToolbox())
        self.assert_results_equal(expected_result, actual_result)

    def test_just_proxy_host(self):
        config = ConfigDict({
            'http_proxy_host': 'mylocalproxy.com',
        })
        expected_result = ConfigCheckResult([])

        actual_result = main.check_main(config, MockConfigTestToolbox())
        self.assert_results_equal(expected_result, actual_result)

    def test_just_proxy_port(self):
        config = ConfigDict({
            'http_proxy_port': '8081',
        })
        expected_result = ConfigCheckResult([
            MissingKey(key='http_proxy_host')
        ])

        actual_result = main.check_main(config, MockConfigTestToolbox())
        self.assert_results_equal(expected_result, actual_result)

    def test_bad_proxy_host_and_port(self):
        invalid_host = ''
        invalid_port = '-1'
        config = ConfigDict({
            'http_proxy_host': invalid_host,
            'http_proxy_port': invalid_port,
        })
        expected_result = ConfigCheckResult([
            InvalidValue(key='http_proxy_host', value=invalid_host),
            InvalidValue(key='http_proxy_port', value=invalid_port),
        ])

        actual_result = main.check_main(config, MockConfigTestToolbox())
        self.assert_results_equal(expected_result, actual_result)

    def test_bad_client(self):
        invalid_client = ''
        config = ConfigDict({
            'client': invalid_client,
        })
        expected_result = ConfigCheckResult([
            InvalidValue(key='client', value=invalid_client),
        ])

        actual_result = main.check_main(config, MockConfigTestToolbox())
        self.assert_results_equal(expected_result, actual_result)

    def test_bad_fips_mode(self):
        invalid_fips_mode = 'on'
        config = ConfigDict({
            'fips_mode': invalid_fips_mode,
        })
        expected_result = ConfigCheckResult([
            InvalidValue(key='fips_mode', value=invalid_fips_mode),
        ])

        actual_result = main.check_main(config, MockConfigTestToolbox())
        self.assert_results_equal(expected_result, actual_result)

    def test_unexpected_keys(self):
        config = ConfigDict({
            'debugg': 'true',
            'notakey': 'yes',
        })
        expected_result = ConfigCheckResult([
            UnexpectedKey(key='notakey'),
            UnexpectedKey(key='debugg'),
        ])

        actual_result = main.check_main(config, MockConfigTestToolbox())
        self.assert_results_equal(expected_result, actual_result)

    def assert_results_equal(self, expected_result, actual_result):
        self.assertEqual(expected_result, actual_result,
                         "{0} and {1} should be equal".format(
                             str(expected_result), str(actual_result)))


if __name__ == '__main__':
    import unittest
    unittest.main()
