from duoauthproxy.lib.validation.config.check import radius_client
from duoauthproxy.lib.config_provider import ConfigDict
from duoauthproxy.lib.validation.config.config_results import (
    MissingKey, UnexpectedKey, InvalidValue)
from duoauthproxy.lib.validation.connectivity.connectivity_results import ConfigCheckResult

from ....base import TestCase, MockConfigTestToolbox


class TestCheckRadiusClientConfig(TestCase):
    """
    Tests the [ad_client] config section tester module
    """

    def setUp(self):
        super(TestCheckRadiusClientConfig, self).setUp()
        self.BASIC_CONFIG_INPUT = ConfigDict({
            'host': '127.0.0.1',
            'secret': 's3cr3t',
        })

    def test_basic_config(self):
        result = radius_client.check_radius_client(self.BASIC_CONFIG_INPUT)
        self.assertEqual(result, ConfigCheckResult([]))

    def test_full_good_config(self):
        config = ConfigDict({
            'host': '127.0.0.1',
            'host_2': '127.0.0.2',
            'secret_protected': 'lookatmeallprotectedandstuff',
            'port': '1812',
            'port_2': '1813',
            'retries': '15',
            'retry_wait': '2',
            'nas_ip': '127.0.0.3',
            'pass_through_attr_names': 'XXX,XXX,XXX',
            'pass_through_all': 'true',
            'pw_codec': 'latin-1',
        })

        expected_result = ConfigCheckResult([])
        actual_result = radius_client.check_radius_client(config, MockConfigTestToolbox())
        self.assert_results_equal(expected_result, actual_result)

    def test_no_host(self):
        expected_result = ConfigCheckResult([
            MissingKey(key='host')
        ])

        del self.BASIC_CONFIG_INPUT['host']
        actual_result = radius_client.check_radius_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_out_of_sequenc_host(self):
        config = self.BASIC_CONFIG_INPUT
        expected_result = ConfigCheckResult([
        ])

        del config['host']
        config['host_2'] = "127.0.2.1"
        actual_result = radius_client.check_radius_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_weird_port_naming(self):
        config = self.BASIC_CONFIG_INPUT
        expected_result = ConfigCheckResult([
            MissingKey(key='host_1')
        ])

        config['port_1'] = '1234'
        actual_result = radius_client.check_radius_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_host_1_and_host(self):
        config = self.BASIC_CONFIG_INPUT
        expected_result = ConfigCheckResult([
        ])

        config['host_1'] = config['host']
        actual_result = radius_client.check_radius_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_no_secret(self):
        expected_result = ConfigCheckResult([
            MissingKey(key='secret/secret_protected')
        ])

        del self.BASIC_CONFIG_INPUT['secret']
        actual_result = radius_client.check_radius_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_invalid_port(self):
        invalid_port = '9999999999999'
        expected_result = ConfigCheckResult([
            InvalidValue(key='port', value=invalid_port)
        ])

        self.BASIC_CONFIG_INPUT['port'] = invalid_port
        actual_result = radius_client.check_radius_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_one_bad_ip(self):
        invalid_ip = '127.0.0.1.0'
        expected_result = ConfigCheckResult([
            InvalidValue(key='host', value=invalid_ip)
        ])

        self.BASIC_CONFIG_INPUT['host'] = invalid_ip
        actual_result = radius_client.check_radius_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_multiple_bad_ip(self):
        invalid_ip1 = '127.0.0.1.0'
        invalid_ip2 = '100'
        invalid_ip3 = '100.10.1.10/24'
        expected_result = ConfigCheckResult([
            InvalidValue(key='host', value=invalid_ip1),
            InvalidValue(key='host_2', value=invalid_ip2),
            InvalidValue(key='host_3', value=invalid_ip3),
        ])

        self.BASIC_CONFIG_INPUT['host'] = invalid_ip1
        self.BASIC_CONFIG_INPUT['host_2'] = invalid_ip2
        self.BASIC_CONFIG_INPUT['host_3'] = invalid_ip3
        actual_result = radius_client.check_radius_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_bad_ip_no_secret(self):
        bad_ip = 'hostname_not_ip'
        expected_result = ConfigCheckResult([
            MissingKey(key='secret/secret_protected'),
            InvalidValue(key='host', value=bad_ip)
        ])

        self.BASIC_CONFIG_INPUT['host'] = bad_ip
        del self.BASIC_CONFIG_INPUT['secret']
        actual_result = radius_client.check_radius_client(self.BASIC_CONFIG_INPUT)
        self.assertEqual(expected_result,
                         actual_result,
                         "{0} and {1} should be equal".format(str(expected_result), str(actual_result)))

    def test_invalid_retries(self):
        invalid_retries = '-1'
        expected_result = ConfigCheckResult([
            InvalidValue(key='retries', value=invalid_retries)
        ])

        self.BASIC_CONFIG_INPUT['retries'] = invalid_retries
        actual_result = radius_client.check_radius_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_invalid_retry_wait(self):
        invalid_retry_wait = '-1'
        expected_result = ConfigCheckResult([
            InvalidValue(key='retry_wait', value=invalid_retry_wait)
        ])

        self.BASIC_CONFIG_INPUT['retry_wait'] = invalid_retry_wait
        actual_result = radius_client.check_radius_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_invalid_nas_ip(self):
        invalid_nas_ip = '300.300.300.300'
        expected_result = ConfigCheckResult([
            InvalidValue(key='nas_ip', value=invalid_nas_ip),
        ])

        self.BASIC_CONFIG_INPUT['nas_ip'] = invalid_nas_ip
        actual_result = radius_client.check_radius_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_invalid_pass_through_attrs(self):
        invalid_pass_through_attrs = ''
        expected_result = ConfigCheckResult([
            InvalidValue(key='pass_through_attr_names',
                         value=invalid_pass_through_attrs)
        ])

        self.BASIC_CONFIG_INPUT['pass_through_attr_names'] = invalid_pass_through_attrs
        actual_result = radius_client.check_radius_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_invalid_pass_through_all(self):
        invalid_pass_through_all = 'yes'
        expected_result = ConfigCheckResult([
            InvalidValue(key='pass_through_all', value=invalid_pass_through_all)
        ])

        self.BASIC_CONFIG_INPUT['pass_through_all'] = invalid_pass_through_all
        actual_result = radius_client.check_radius_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_invalid_password_codec(self):
        invalid_password_codec = 'ImNotaRealCodec'
        expected_result = ConfigCheckResult([
            InvalidValue(key='pass_through_all', value=invalid_password_codec)
        ])

        self.BASIC_CONFIG_INPUT['pass_through_all'] = invalid_password_codec
        actual_result = radius_client.check_radius_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_port_without_a_host(self):
        config = ConfigDict({
            'host': '127.0.0.1',
            'secret': 's3cr3t',
            'port': '1812',
            'port_2': '1813',
        })

        expected_result = ConfigCheckResult([
            MissingKey(key='host_2'),
        ])
        actual_result = radius_client.check_radius_client(config)
        self.assert_results_equal(expected_result, actual_result)

    def test_unexpected_key(self):
        unexpected_key = 'unexpected_key'
        expected_result = ConfigCheckResult([
            UnexpectedKey(key=unexpected_key)
        ])

        self.BASIC_CONFIG_INPUT[unexpected_key] = unexpected_key
        actual_result = radius_client.check_radius_client(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def assert_results_equal(self, expected_result, actual_result):
        self.assertEqual(expected_result, actual_result,
                         "{0} and {1} should be equal".format(
                             str(expected_result), str(actual_result)))


if __name__ == '__main__':
    import unittest
    unittest.main()
