from duoauthproxy.lib.validation.config.check import radius_server_auto
from duoauthproxy.lib.config_provider import ConfigDict
from duoauthproxy.lib.validation.connectivity.connectivity_results import ConfigCheckResult
from duoauthproxy.lib.validation.config.config_results import InvalidValue, UnexpectedKey
from ....base import TestCase, MockConfigTestToolbox


class TestCheckRadiusServerAutoConfig(TestCase):
    """
    Tests the [radius_server_auto] config section tester module
    """

    def setUp(self):
        super(TestCheckRadiusServerAutoConfig, self).setUp()
        self.BASIC_CONFIG_INPUT = ConfigDict({
            'ikey': 'DI000000000000000000',
            'skey': 'thisisntarealskeybutitcanplaysoneonthetv',
            'api_host': 'api-12345678.test.duosecurity.com',
            'radius_ip_1': '127.0.0.1',
            'client': 'radius_client',
            'radius_secret_1': 'radius_secret',
        })

    def test_basic_config(self):
        result = radius_server_auto.check_radius_server_auto(self.BASIC_CONFIG_INPUT)
        self.assertEqual(result, ConfigCheckResult([]))

    def test_full_good_config(self):
        config = ConfigDict({
            'ikey': 'DI000000000000000000',
            'skey': 'thisisntarealskeybutitcanplaysoneonthetv',
            'api_host': 'api-12345678.test.duosecurity.com',
            'radius_ip_1': '127.0.0.1',
            'client': 'radius_client',
            'radius_secret_protected_1': 'aimcoeami923u892',
            'radius_ip_2': '127.0.0.2',
            'radius_secret_2': 'non_protected_secret',
            'factors': 'push,auto',
            'delimiter': '!',
            'failmode': 'secure',
            'pw_codec': 'latin-1',
            'allow_concat': 'True',
            'pass_through_attr_names': 'NAS-Identifier,Calling-Station-Id',
            'exempt_username_1': 'alice',
            'exempt_username_2': 'bob',
            'client_ip_attr': 'some-random-id',
            'api_port': '443',
        })

        expected_result = ConfigCheckResult([])
        actual_result = radius_server_auto.check_radius_server_auto(config, MockConfigTestToolbox())
        self.assert_results_equal(expected_result, actual_result)

    def test_invalid_delimiter(self):
        invalid_delimiter = ''
        expected_result = ConfigCheckResult([
            InvalidValue(key='delimiter', value=invalid_delimiter)
        ])

        self.BASIC_CONFIG_INPUT['delimiter'] = invalid_delimiter
        actual_result = radius_server_auto.check_radius_server_auto(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_invalid_factors(self):
        invalid_factors = 'not_a_factor,xfactor,factorial'
        expected_result = ConfigCheckResult([
            InvalidValue(key='factors', value=invalid_factors)
        ])

        self.BASIC_CONFIG_INPUT['factors'] = invalid_factors
        actual_result = radius_server_auto.check_radius_server_auto(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_invalid_allow_concat(self):
        invalid_allow_concat = 'notabool'
        expected_result = ConfigCheckResult([
            InvalidValue(key='allow_concat', value=invalid_allow_concat)
        ])

        self.BASIC_CONFIG_INPUT['allow_concat'] = invalid_allow_concat
        actual_result = radius_server_auto.check_radius_server_auto(self.BASIC_CONFIG_INPUT)
        self.assert_results_equal(expected_result, actual_result)

    def test_radius_ip_not_starting_with_one(self):
        config = ConfigDict({
            'ikey': 'DI000000000000000000',
            'skey': 'thisisntarealskeybutitcanplaysoneonthetv',
            'api_host': 'api-12345678.test.duosecurity.com',
            'client': 'radius_client',
            'radius_ip_2': '127.0.0.2',
            'radius_secret_2': 'non_protected_secret',
            'radius_ip_3': '127.0.0.2',
            'radius_secret_protected_3': 'protected_secret',
        })
        expected_result = ConfigCheckResult([])
        actual_result = radius_server_auto.check_radius_server_auto(config, MockConfigTestToolbox())
        self.assert_results_equal(expected_result, actual_result)

    def test_no_ip_secret_suffix(self):
        config = ConfigDict({
            'ikey': 'DI000000000000000000',
            'skey': 'thisisntarealskeybutitcanplaysoneonthetv',
            'api_host': 'api-12345678.test.duosecurity.com',
            'client': 'radius_client',
            'radius_ip': '127.0.0.2',
            'radius_secret': 'non_protected_secret',
        })
        expected_result = ConfigCheckResult([
            UnexpectedKey(key='radius_ip'),
            UnexpectedKey(key='radius_secret / radius_secret_protected'),
        ])
        actual_result = radius_server_auto.check_radius_server_auto(config, MockConfigTestToolbox())
        self.assert_results_equal(expected_result, actual_result)

    def test_no_ip_secret_protected_suffix(self):
        config = ConfigDict({
            'ikey': 'DI000000000000000000',
            'skey': 'thisisntarealskeybutitcanplaysoneonthetv',
            'api_host': 'api-12345678.test.duosecurity.com',
            'client': 'radius_client',
            'radius_ip': '127.0.0.2',
            'radius_secret_protected': 'protected_secret',
        })
        expected_result = ConfigCheckResult([
            UnexpectedKey(key='radius_ip'),
            UnexpectedKey(key='radius_secret / radius_secret_protected'),
        ])
        actual_result = radius_server_auto.check_radius_server_auto(config, MockConfigTestToolbox())
        self.assert_results_equal(expected_result, actual_result)

    def assert_results_equal(self, expected_result, actual_result):
        self.assertEqual(expected_result, actual_result,
                         "{0} and {1} should be equal".format(
                             str(expected_result), str(actual_result)))


if __name__ == '__main__':
    import unittest
    unittest.main()
