#
# Copyright (c) 2018 Duo Security
# All Rights Reserved
#

import copy
import unittest

from duoauthproxy.lib.validation.config import config_key_tools
from duoauthproxy.lib.config_provider import ConfigDict


BASE_CONFIG = ConfigDict({
    'some_key': 'some_value',
    'another_key': 'another_value',
    'one_more_key': 'one_more_value',
    'dynamic_key_1': 'value_1',
    'dynamic_key_2': 'value_2',
    'dynamic_key_3': 'value_3',
    'pair_member_one': 'need_member_two',
    'pair_member_two': 'need_member_one'
})


class TestConfigKeyToolsKeyUsable(unittest.TestCase):
    def test_success(self):
        test_key = 'some_key'
        self.assertTrue(config_key_tools.is_key_usable(BASE_CONFIG, test_key),
                        'Expected to succeed test for {0} presence'.format(test_key))

    def test_blank_value_failure(self):
        test_key = 'blank_key'

        TEST_CONFIG = copy.deepcopy(BASE_CONFIG)
        TEST_CONFIG.update({test_key: ''})

        self.assertFalse(config_key_tools.is_key_usable(TEST_CONFIG, test_key),
                         'Expected failure for blank key {0}'.format(test_key))

    def test_no_key_failure(self):
        test_key = 'not_a_key'
        self.assertFalse(config_key_tools.is_key_usable(BASE_CONFIG, test_key),
                         'Expected to fail test for {0} presence'.format(test_key))


class TestConfigKeyToolsUnexpectedKeys(unittest.TestCase):
    VALID_FIXED_KEYS = ['some_key', 'another_key', 'one_more_key', 'pair_member_one', 'pair_member_two']
    VALID_DYNAMIC_KEYS = ['dynamic_key']

    def test_no_unexpected_keys(self):
        unexpected_keys = config_key_tools.get_unexpected_keys(BASE_CONFIG, self.VALID_FIXED_KEYS, self.VALID_DYNAMIC_KEYS)
        self.assertItemsEqual([], unexpected_keys)

    def test_mix_of_keys(self):
        TEST_CONFIG = copy.deepcopy(BASE_CONFIG)
        TEST_CONFIG.update({'unexpected_key': 'unexpected_value', 'another_unexpected_key': 'another_unexpected_value'})
        unexpected_keys = config_key_tools.get_unexpected_keys(TEST_CONFIG, self.VALID_FIXED_KEYS, self.VALID_DYNAMIC_KEYS)
        expected_result = ['unexpected_key', 'another_unexpected_key']

        self.assertItemsEqual(expected_result, unexpected_keys)

    def test_with_dynamic_keys(self):
        TEST_CONFIG = copy.deepcopy(BASE_CONFIG)
        TEST_CONFIG.update({'dynamic_key_4': 'value_4', 'dynamic_key5': 'value5'})
        unexpected_keys = config_key_tools.get_unexpected_keys(TEST_CONFIG, self.VALID_FIXED_KEYS, self.VALID_DYNAMIC_KEYS)
        expected_result = ['dynamic_key5']

        self.assertItemsEqual(expected_result, unexpected_keys)

    def test_all_unexpected_keys(self):
        TEST_CONFIG = ConfigDict({
            'what': 'the',
            'heck': 'this',
            'config': 'is',
            'total': 'garbage',
        })
        unexpected_keys = config_key_tools.get_unexpected_keys(TEST_CONFIG, self.VALID_FIXED_KEYS, self.VALID_DYNAMIC_KEYS)
        expected_result = ['what', 'heck', 'config', 'total']

        self.assertItemsEqual(expected_result, unexpected_keys)


class TestConfigKeyToolsRequiredUnpairedKeys(unittest.TestCase):
    def test_both_present_failure(self):
        key_one = 'some_key'
        key_two = 'another_key'
        self.assertFalse(config_key_tools.are_keys_unpaired(BASE_CONFIG, key_one, key_two),
                         'Expected success for key pair {0}, {1}'.format(key_one, key_two))

    def test_one_present_success(self):
        key_one = 'some_key'
        key_two = 'absent_key'
        self.assertTrue(config_key_tools.are_keys_unpaired(BASE_CONFIG, key_one, key_two),
                        'Expected success for key pair {0}, {1}'.format(key_one, key_two))

    def test_neither_present_failure(self):
        key_one = 'absent_key'
        key_two = 'also_absent'
        self.assertTrue(config_key_tools.are_keys_unpaired(BASE_CONFIG, key_one, key_two),
                        'Expected success for absent key pair {0}, {1}'.format(key_one, key_two))

    def test_both_blank_failure(self):
        key_one = 'blank_key_one'
        key_two = 'blank_key_two'

        TEST_CONFIG = copy.deepcopy(BASE_CONFIG)
        TEST_CONFIG.update({key_one: '', key_two: ''})

        self.assertFalse(config_key_tools.are_keys_unpaired(TEST_CONFIG, key_one, key_two),
                         'Expected failure for blank keys {0} and {1}'.format(key_one, key_two))

    def test_one_blank_one_usable_failure(self):
        key_one = 'some_key'
        key_two = 'blank_key'

        TEST_CONFIG = copy.deepcopy(BASE_CONFIG)
        TEST_CONFIG.update({key_two: ''})

        self.assertFalse(config_key_tools.are_keys_unpaired(TEST_CONFIG, key_one, key_two),
                         'Expected failure for key {0} and blank key {1}'.format(key_one, key_two))


if __name__ == '__main__':
    unittest.main()
