import mock
import platform
import duoauthproxy
from duoauthproxy.lib.validation.connectivity.check import base_duo_check
from ....base import ModuleConnectivityTest, MockConnectivityTestToolbox
from duoauthproxy.lib.validation.connectivity.connectivity_results import (
    ConfigProblemSkippedTestResult,
    UnmetPrerequisiteSkippedTestResult,
    BAD_PROTECTED_SKEY_RESULT
)


class TestBaseDuoCheck(ModuleConnectivityTest):
    def test_perform_duo_ping_no_api_host(self):
        toolbox = MockConnectivityTestToolbox()

        result = base_duo_check.perform_duo_ping(toolbox, api_host="")

        self.assertEqual(result, ConfigProblemSkippedTestResult('ping', 'api_host'))

    def test_perform_duo_ping_failure(self):
        toolbox = MockConnectivityTestToolbox(fail_ping=True)

        result = base_duo_check.perform_duo_ping(toolbox, api_host='badhost')

        self.assertEqual(result, self.BAD_PING_RESULT)

    def test_perform_duo_ping_success(self):
        toolbox = MockConnectivityTestToolbox()

        result = base_duo_check.perform_duo_ping(toolbox, api_host='goodhost')

        self.assertEqual(result, self.GOOD_PING_RESULT)

    def test_perform_time_drift_success(self):
        toolbox = MockConnectivityTestToolbox()

        mocked_ping_result = mock.Mock()
        mocked_ping_result.is_successful.return_value = True

        result = base_duo_check.perform_time_drift(toolbox, mocked_ping_result, api_host='doesntmatter')

        self.assertEqual(result, self.GOOD_TIME_DRIFT_RESULT)

    def test_perform_time_drift_failure(self):
        toolbox = MockConnectivityTestToolbox(fail_time_drift=True)

        mocked_ping_result = mock.Mock()
        mocked_ping_result.is_successful.return_value = True

        result = base_duo_check.perform_time_drift(toolbox, mocked_ping_result, api_host='doesntmatter')

        self.assertEqual(result, self.BAD_TIME_DRIFT_RESULT)

    def test_perform_time_drift_skipped(self):
        toolbox = MockConnectivityTestToolbox()

        mocked_ping_result = mock.Mock()
        mocked_ping_result.is_successful.return_value = False

        result = base_duo_check.perform_time_drift(toolbox, mocked_ping_result, api_host='doesntmatter')

        self.assertEqual(result, UnmetPrerequisiteSkippedTestResult('time drift', 'ping'))

    def test_perform_credentials_check_success(self):
        toolbox = MockConnectivityTestToolbox()

        result = base_duo_check.perform_credentials_check(toolbox, 'anyhost', skey="a", ikey="b")

        self.assertEqual(result, self.GOOD_VALIDATE_RESULT)

    def test_perform_credentials_check_failure(self):
        toolbox = MockConnectivityTestToolbox(fail_validate_api_creds=True)

        result = base_duo_check.perform_credentials_check(toolbox, 'anyhost', skey="a", ikey="b")

        self.assertEqual(result, self.BAD_VALIDATE_RESULT)

    def test_perform_credentials_check_missing_skey(self):
        toolbox = MockConnectivityTestToolbox()

        result = base_duo_check.perform_credentials_check(toolbox, 'anyhost', skey=None)

        self.assertEqual(result, BAD_PROTECTED_SKEY_RESULT)

    def test_perform_credentials_check_missing_ikey(self):
        toolbox = MockConnectivityTestToolbox()

        result = base_duo_check.perform_credentials_check(toolbox, 'anyhost', ikey=None)

        self.assertEqual(result, ConfigProblemSkippedTestResult('validate api credentials', 'api credentials'))

    def test_building_duo_client(self):
        result = base_duo_check._get_duo_api_client('a.hostname', 'skeyvalue', 'ikeyvalue', '/capath', 12.34)

        expected_user_agent = "duoauthproxy connectivity tool/{0}; ({1})".format(
            duoauthproxy.__version__,
            platform.platform())

        self.assertEqual(result.host, 'a.hostname')
        self.assertEqual(result.skey, 'skeyvalue')
        self.assertEqual(result.ikey, 'ikeyvalue')
        self.assertEqual(result.ca_certs, '/capath')
        self.assertEqual(result.user_agent, expected_user_agent)
        self.assertEqual(result.timeout, 12.34)

    def test_duo_client_with_http_proxy(self):
        proxy_host = 'a.proxy_host'
        proxy_port = 8000

        result = base_duo_check._get_duo_api_client('a.hostname', http_proxy=(proxy_host, proxy_port))

        self.assertEqual(result.host, 'a.hostname')
        self.assertEqual(result.proxy_host, proxy_host)
        self.assertEqual(result.proxy_port, proxy_port)


if __name__ == '__main__':
    import unittest
    unittest.main()
